package anonvpn.anon_next.core;

import java.net.ConnectException;

import HTTPClient.HTTPClientSocketFactory;
import anon.AnonChannel;
import anon.AnonServerDescription;
import anon.AnonServiceEventListener;
import anon.ErrorCodes;
import anon.client.AnonClient;
import anon.client.DummyTrafficControlChannel;
import anon.error.AnonServiceException;
import anon.infoservice.MixCascade;
import anon.infoservice.SimpleMixCascadeContainer;
import anonvpn.anon_next.core.networking.*;
import anonvpn.anon_next.core.notification.Event;
import anonvpn.anon_next.core.notification.NotificationDispatcher;
import anonvpn.anon_next.core.notification.Observer;
import anonvpn.anon_next.core.persistence.IConfig;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;
import pcap.PCAPTUNDevice;

public class CascadeConnectionManager implements Observer, AnonServiceEventListener
	{

		public final static int MAX_FAILED_CONNECT_ATTEMPTS = 5;
		public final static int RECONNECT_INTERVAL_IN_S = 3;

		protected class ReconnectTask extends Thread
			{

				private boolean m_bIsFinished = false;
				private int m_iFailedConnectionAttempts = 0;

				public boolean isFinished()
					{
						return m_bIsFinished;
					}

				@Override
				public void run()
					{

						m_iFailedConnectionAttempts += 1;
						boolean connected = false;

						try
							{
								while (!Thread.interrupted() && connected == false
										&& m_iFailedConnectionAttempts < MAX_FAILED_CONNECT_ATTEMPTS
										&& mConnectivityMonitor.deviceIsConnected())
									{
										Thread.sleep(RECONNECT_INTERVAL_IN_S * 1000);

										LogHolder.log(LogLevel.INFO, LogType.NET,
												"CASCADECONNNECTIONMANAGER: Reconnection Attempt " + m_iFailedConnectionAttempts + "...");
										try
											{
												setupMixCascade();
												connected = true;
											}
										catch (AnonServiceException e)
											{
												LogHolder.log(LogLevel.EXCEPTION, LogType.NET,
														"CASCADECONNNECTIONMANAGER: Reconnection Attempt " + m_iFailedConnectionAttempts
																+ "failed ...",
														e);
												m_iFailedConnectionAttempts += 1;
											}
									}

								if (m_iFailedConnectionAttempts == MAX_FAILED_CONNECT_ATTEMPTS)
									{
										LogHolder.log(LogLevel.ERR, LogType.NET, "CASCADECONNNECTIONMANAGER: Failed to connect after "
												+ m_iFailedConnectionAttempts + " attempts...");
										mNotification.notify(Event.CASCADE_LOST);
									}

								if (!mConnectivityMonitor.deviceIsConnected())
									{
										LogHolder.log(LogLevel.INFO, LogType.NET,
												"CASCADECONNNECTIONMANAGER: Waiting for network to reconnect...");
										mNotification.subscribe(Event.NETWORK_CONNECTED, CascadeConnectionManager.this);
										mNotification.notify(Event.CASCADE_DISCONNECTED);
									}
							}
						catch (InterruptedException e)
							{
								e.printStackTrace();
							}
						finally
							{
								m_bIsFinished = true;
								m_iFailedConnectionAttempts = 0;
							}

						LogHolder.log(LogLevel.INFO, LogType.NET, "CASCADECONNNECTIONMANAGER: Reconnect thread done...");
					}
			}

		private final NotificationDispatcher mNotification;
		private final INetworkFactory mNetworkFactory;
		private final IConfig mConfig;
		private final HTTPClientSocketFactory mMixSocketFactory;
		private final IConnectivityMonitor mConnectivityMonitor;
		private SenderThread mSender = null;
		private ReceiverThread mReceiver = null;
		private AnonClient mAnonClient = null;
		private AnonChannel mAnonChannel = null;
		private boolean mCascadeConnected = false;

		protected ReconnectTask mReconnectTask = null;

		private ITunneling mTunDevice;

		public static boolean USE_FAKE_DNS = false;
		public static boolean USE_TCP_SOCKS = true;

		///if true --> use the ANONVPNServer UDP protocol for handling UDP (otherwise UDP packets a treated as IP packets)
		public static boolean USE_UDP_SOCKS = true;

		//Which UDP channel type to use for transmitting UDP packets
		public static int UDP_ANON_CHANNEL_TYPE = AnonChannel.UDP;

		// for debugging with local SOCKS server
		public static boolean USE_SOCKS_WITH_CASCADE = true;

		public static boolean USE_PCAP_WRITER = false;
		public static String PCAP_FILENAME = "vpn.pcap";
		public static boolean USE_DEBUG_LOG = false;

		/**
		 * Assumes that tunneling has already been enabled.
		 * 
		 * @param notification
		 * @param networkFactory
		 * @param config
		 * @param mixSocketFactory
		 */
		public CascadeConnectionManager(NotificationDispatcher notification, INetworkFactory networkFactory, IConfig config,
				IConnectivityMonitor connMonitor, HTTPClientSocketFactory mixSocketFactory)
			{
				this.mNotification = notification;
				this.mNetworkFactory = networkFactory;
				this.mConfig = config;
				this.mMixSocketFactory = mixSocketFactory;
				this.mConnectivityMonitor = connMonitor;

				mNotification.subscribe(Event.CASCADE_CHOSEN, this);
				mNotification.subscribe(Event.VPN_DISABLED, this);

				try
					{
						mNotification.notify(Event.CASCADE_CONNECTING);
						setupMixCascade();
					}
				catch (AnonServiceException e)
					{
						LogHolder.log(LogLevel.EXCEPTION, LogType.NET, "CASCADECONNNECTIONMANAGER: START 1", e);
						startReconnectTask();
					}
			}

		@Override
		public void notify(Event event)
			{
				switch (event)
					{
						case CASCADE_CHOSEN:

							if (mCascadeConnected)
								{
									mAnonClient.shutdown(true);
								}

							if (reconnectTaskIsRunning())
								{
									stopReconnectTask();
								}

							try
								{
									mNotification.notify(Event.CASCADE_CONNECTING);
									setupMixCascade();
								}
							catch (Exception e)
								{
									LogHolder.log(LogLevel.EXCEPTION, LogType.NET, "CASCADECONNNECTIONMANAGER: START 2", e);
									startReconnectTask();
								}

							break;

						case VPN_DISABLED:
							if (reconnectTaskIsRunning())
								{
									LogHolder.log(LogLevel.INFO, LogType.NET, "Stopping reconnect task..");
									stopReconnectTask();
									LogHolder.log(LogLevel.INFO, LogType.NET, "Stopped reconnect task.");
								}

							mNotification.unsubscribe(Event.NETWORK_CONNECTED, this);

							if (mAnonClient != null)
								{
									if (mAnonClient.isConnected())
										{
											if (mAnonChannel != null)
												{
													try
														{
															mAnonChannel.close();
														}
													catch (Throwable t)
														{
														}
													LogHolder.log(LogLevel.INFO, LogType.NET, "TERMINATING SENDER");
													stopSenderThread();
													LogHolder.log(LogLevel.INFO, LogType.NET, "TERMINATING RECEIVER");
													stopReceiverThread();
												}
										}
									mAnonClient.shutdown(true);
								}
							break;

						case NETWORK_CONNECTED:
							mNotification.unsubscribe(Event.NETWORK_CONNECTED, this);
							LogHolder.log(LogLevel.INFO, LogType.NET,
									"CASCADECONNNECTIONMANAGER: Network is connected, contacting cascade...");
							mNotification.notify(Event.CASCADE_CONNECTING);
							try
								{
									setupMixCascade();
								}
							catch (AnonServiceException e)
								{
									LogHolder.log(LogLevel.EXCEPTION, LogType.NET, "CASCADECONNNECTIONMANAGER: START 3", e);
									startReconnectTask();
								}
							break;
					}

			}

		/**
		 * Sets up the connection to the mix cascade after the user has chosen one
		 * cascade.
		 * <p>
		 * TODO: Make anonlib accept prepared sockets as argument for initialize or
		 * so.
		 */
		protected void setupMixCascade() throws AnonServiceException
			{
				MixCascade mixCascade = null;
				try
					{
						LogHolder.log(LogLevel.INFO, LogType.NET, "SETTING UP CASCADE!!!");

						// TODO: Remove in the future
						String credential = mConfig.getChosenCascadeCredential();

						// TODO: Check if it is safe to reuse an existing anon client
						if (mAnonClient == null)
							{
								mAnonClient = getNewAnonClient();
								mAnonClient.setSocketFactory(mMixSocketFactory);
								mAnonClient.setDummyTraffic(DummyTrafficControlChannel.DT_DISABLE);
								mAnonClient.addEventListener(this);
							}

						//initialize the client --> establish the connection
						mixCascade = mConfig.getChosenMixCascade();
						LogHolder.log(LogLevel.INFO, LogType.NET, "CASCADE NAME: " + mixCascade.getName());
						mixCascade.setAccessControlPreSharedSecret(credential);
						LogHolder.log(LogLevel.INFO, LogType.NET, "CASCADE CREDENTIAL: " + credential);
						SimpleMixCascadeContainer mixcascadeContainer = new SimpleMixCascadeContainer(mixCascade);
						mConfig.setSentBytes(0);
						mConfig.setReceivedBytes(0);
						mAnonClient.initialize(mixCascade, mixcascadeContainer, null, true);
						mAnonClient.setDummyTraffic(DummyTrafficControlChannel.DT_DISABLE);
						LogHolder.log(LogLevel.INFO, LogType.NET, "VPNSERVICE - CONNECTION ESTABLISHED!");
					}
				catch (Throwable t)
					{
						throw new AnonServiceException(mixCascade, t.getMessage(), ErrorCodes.E_UNKNOWN);
					}
			}

		protected synchronized void startReconnectTask()
			{
				if (mReconnectTask == null || mReconnectTask.isFinished())
					{
						mReconnectTask = new ReconnectTask();
						mReconnectTask.start();
					}
			}

		protected synchronized void stopReconnectTask()
			{
				if (!mReconnectTask.isFinished())
					{
						mReconnectTask.interrupt();
						try
							{
								mReconnectTask.join();
							}
						catch (InterruptedException e)
							{
								e.printStackTrace();
							}
					}
			}

		protected boolean reconnectTaskIsRunning()
			{
				return mReconnectTask != null && !mReconnectTask.isFinished();
			}

		protected AnonClient getNewAnonClient()
			{
				return new AnonClient(null);
			}

		protected synchronized ITunneling getTunnelingInterface()
			{
				if (mTunDevice == null)
					{
						mTunDevice = mNetworkFactory.getTunnelingInterface();
						if (USE_PCAP_WRITER)
							mTunDevice = new PCAPTUNDevice(mTunDevice, PCAP_FILENAME);
					}
				return mTunDevice;
			}

		protected void startReceiverThread()
			{
				mReceiver = new ReceiverThread(getTunnelingInterface(), mAnonChannel.getInputStream());
				mReceiver.start();
			}

		protected void startSenderThread()
			{
				mSender = new SenderThread(getTunnelingInterface(), mAnonChannel.getOutputStream(), mAnonClient,
						mMixSocketFactory);
				mSender.start();
			}

		protected void stopReceiverThread()
			{
				mReceiver.terminate();
			}

		protected void stopSenderThread()
			{
				mSender.terminate();
			}

		@Override
		public void connectionError(AnonServiceException e)
			{
				stopSenderThread();
				stopReceiverThread();

				mCascadeConnected = false;

				LogHolder.log(LogLevel.INFO, LogType.NET, "CASCADECONNNECTIONMANAGER: START 4");
				e.printStackTrace();
				mNotification.notify(Event.CASCADE_CONNECTING);
				startReconnectTask();
			}

		@Override
		public void currentServiceChanged(AnonServerDescription anonServerDescription)
			{
				LogHolder.log(LogLevel.INFO, LogType.NET, "ANONMGR - SERVICE CHANGED");
			}

		@Override
		public void disconnected()
			{
				LogHolder.log(LogLevel.INFO, LogType.NET, "ANONMGR - DISCONNECTED");
			}

		@Override
		public void connecting(AnonServerDescription anonServerDescription, boolean b)
			{
				LogHolder.log(LogLevel.INFO, LogType.NET, "ANONMGR - CONNECTING");
			}

		@Override
		public void connectionEstablished(AnonServerDescription anonServerDescription)
			{
				LogHolder.log(LogLevel.INFO, LogType.NET, "ANONMGR - CONNECTED");

				try
					{
						mAnonChannel = mAnonClient.createChannel(AnonChannel.VPN);

						startSenderThread();
						startReceiverThread();
						mCascadeConnected = true;
					}
				catch (ConnectException e)
					{
						LogHolder.log(LogLevel.EXCEPTION, LogType.NET, "ANONMGR - FAILED TO CREATE CHANNEL: " + e.getMessage(), e);
					}
				LogHolder.log(LogLevel.INFO, LogType.NET, "VPNSERVICE - STARTED SENDER AND RECEIVER THREAD");
				mNotification.notify(Event.CASCADE_CONNECTED);
				// TODO: Might be wrong point to notify this
			}

		@Override
		public void packetMixed(long l)
			{

			}

		@Override
		public void dataChainErrorSignaled(AnonServiceException e)
			{
				LogHolder.log(LogLevel.ERR, LogType.NET, "ANONMGR - DATA CHAIN ERROR " + e.getMessage());
			}

		@Override
		public void integrityErrorSignaled(AnonServiceException e)
			{
				LogHolder.log(LogLevel.ERR, LogType.NET, "ANONMGR - INTEGRITY ERROR " + e.getMessage());
			}
	}
