package anonvpn.anon_next.core.networking.TcpSocks;

import anonvpn.anon_next.core.entities.IPv4Packet;
import anonvpn.anon_next.core.networking.ITunneling;

import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.Set;

/**
 * Created by nod on 7/14/17.
 */

final public class DNSResolver implements Runnable
	{
		private final static int DNS_ENTRY_TTL_SECONDS=60;
		private final static int DNS_ENTRY_REMOVE_TTL_MILLISECONDS=DNS_ENTRY_TTL_SECONDS*1000+10000;
		
		
		
		final private class DNSCacheEntry
			{
				final String domainname;
				final long ttl;

				DNSCacheEntry(String dn)
					{
						domainname = dn;
						ttl = System.currentTimeMillis() + DNS_ENTRY_REMOVE_TTL_MILLISECONDS;
					}
			}

		private final byte[] dnsIP = { (byte) 0x0A, (byte) 0x00, (byte) 0x00, (byte) 0x01 };
		private int m_nextIP = 0x000A000002;
		private final ITunneling m_TunDevice;
		private HashMap<Integer, DNSCacheEntry> m_dnsCache; //domain name, ip
		private byte[] m_Response = new byte[200]; // 200 should be enough for all cases
		private byte[] m_Domainname = new byte[500];
		private byte ownIPexclusion;

		private Thread m_threadDNSCacheCleaner;
		private volatile boolean m_bRun;

		public DNSResolver(ITunneling a_TunDevice)
			{
				// TODO: exclude own IP as dns response
				m_dnsCache = new HashMap<Integer, DNSCacheEntry>();
				m_TunDevice = a_TunDevice;
				ownIPexclusion = (byte) (m_TunDevice.getIP() >>> 16);
				m_threadDNSCacheCleaner = new Thread(this, "DNS Cache Cleaner");
				m_bRun = true;
				m_threadDNSCacheCleaner.setDaemon(true);
				m_threadDNSCacheCleaner.start();
			}

		public void stop()
			{
				m_bRun = false;
				m_threadDNSCacheCleaner.interrupt();
				try
					{
						m_threadDNSCacheCleaner.join();
					}
				catch (Exception e)
					{
					}
			}

		private void incAndGetIpBytes(byte[] to, int ito)
			{
				m_nextIP++;
				BitUtils.intToBytes(m_nextIP, to, ito);

			}

		//Set all fixed fields of the response
		private void prepareResponse()
			{
				//set answer rr
				m_Response[35] = (byte) 0x01;
				// set flags (response = 1, recursion desired = 1, recursion available = 1)
				m_Response[30] = (byte) 0x81;
				m_Response[31] = (byte) 0x80;
				// disable udp checksum
				m_Response[26] = m_Response[27] = 0;

			}

		public synchronized void sendResponse(IPv4Packet pkt)
			{
				try
					{
						int len = pkt.getLength();

						byte[] rawPacket = pkt.getRawPacket();

						// only for IPv4 (A) queries
						if (rawPacket[len - 3] != 0x01)
							{
								return;
							}

						//get domain name
						//            ByteBuffer domainname = ByteBuffer.allocate(64);//41
						int pos = 41;
						byte zero = 0;
						int dnLen = rawPacket[40];

						for (int i = 0; i <= dnLen; ++i)
							{
								if (i == dnLen)
									{
										if (rawPacket[pos + i] == zero)
											{
												break;
											}
										else
											{
												dnLen += rawPacket[pos + i] + 1;
												m_Domainname[i] = (byte) 0x2E;
											}
									}
								else
									{
										m_Domainname[i] = rawPacket[pos + i];
									}
							}

						String dn = new String(m_Domainname, 0, dnLen);
						// protection agains chromes random dns querries. more at https://unix.stackexchange.com/questions/363512/dns-requests-with-random-dns-names-malware
						if (!dn.contains("."))
							{
								return;
							}
						//            byte[] ip = null;

						//            if(dnsRecords.containsValue(dn)) { // TODO: could return the saved ip, but iterating Map too costly?
						//                return;
						//            }

						System.arraycopy(rawPacket, 0, m_Response, 0, len);
						prepareResponse();

						m_Response[3] = (byte) (len + 16);

						incAndGetIpBytes(m_Response, len + 12);

						// get type and class
						int type = (rawPacket[len - 4] << 8) | (0x00FF & rawPacket[len - 3]);
						int clas = (rawPacket[len - 2] << 8) | (0x00FF & rawPacket[len - 1]);

						//            byte[] response = new byte[16];

						// answer
						m_Response[len] = (byte) 0xC0;
						m_Response[len + 1] = (byte) 0x0C;
						m_Response[len + 2] = (byte) (type >> 8);
						m_Response[len + 3] = (byte) type;
						m_Response[len + 4] = (byte) (clas >> 8);
						m_Response[len + 5] = (byte) clas;
						// ttl = DNS_ENTRY_TTL_SECONDS
						m_Response[len + 6] = m_Response[len + 7] = m_Response[len + 8] = (byte) 0x00;
						m_Response[len + 9] = DNS_ENTRY_TTL_SECONDS;
						// data length = 4
						m_Response[len + 10] = 0;
						m_Response[len + 11] = (byte) 0x04;
						//ip

						// swap IPs
						System.arraycopy(rawPacket, 12, m_Response, 16, 4);
						System.arraycopy(rawPacket, 16, m_Response, 12, 4);
						// swap ports
						System.arraycopy(rawPacket, 20, m_Response, 22, 2);
						System.arraycopy(rawPacket, 22, m_Response, 20, 2);

						// UDP length
						int udpLen = len - 20 + 16;
						m_Response[24] = (byte) (udpLen >>> 8);
						m_Response[25] = (byte) udpLen;

						// IP checksum
						m_Response[IPv4Packet.IP_CHECKSUM_FIELD_POSITION] = 0;
						m_Response[IPv4Packet.IP_CHECKSUM_FIELD_POSITION + 1] = 0;
						BitUtils.calcIPHeaderChecksum(m_Response);

						synchronized (m_dnsCache)
							{
								m_dnsCache.put(m_nextIP, new DNSCacheEntry(dn));
							}
						m_TunDevice.write(m_Response, 0, len + 16);

					}
				catch (IOException e)
					{
						e.printStackTrace();
					}
			}

		public String getDomainName(int i)
			{
				DNSCacheEntry entry = m_dnsCache.get(i);
				if (entry != null)
					return entry.domainname;
				return null;
			}

		@Override
		public void run()
			{
				while (m_bRun)
					{
						long timeOut = System.currentTimeMillis();
						synchronized (m_dnsCache)
							{
								LinkedList<Integer> obsoleteEntries = new LinkedList<Integer>();
								Set<Entry<Integer, DNSCacheEntry>> s = m_dnsCache.entrySet();
								for (Entry<Integer, DNSCacheEntry> entry : s)
									{
										if (entry.getValue().ttl < timeOut)
											obsoleteEntries.add(entry.getKey());
									}
								for (Integer key : obsoleteEntries) {
									m_dnsCache.remove(key);
								}
							}
						try
							{
								Thread.sleep(DNS_ENTRY_REMOVE_TTL_MILLISECONDS);
							}
						catch (Exception e)
							{
							}
					}
			}
	}