package anonvpn.anon_next.core.networking;

import anonvpn.anon_next.core.CascadeConnectionManager;
import anonvpn.anon_next.core.TestAnonClient;
import anonvpn.anon_next.core.networking.mocks.*;
import anonvpn.anon_next.core.notification.Event;
import anonvpn.anon_next.core.notification.NotificationDispatcher;
import anonvpn.anon_next.core.persistence.IConfig;
import anonvpn.anon_next.core.persistence.mocks.TestConfig;
import org.junit.Test;
import org.w3c.dom.Document;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.util.LinkedList;
import java.util.List;

import HTTPClient.HTTPClientSocketFactory;
import anon.client.AnonClient;
import anon.crypto.SignatureVerifier;
import anon.error.AnonServiceException;
import anon.infoservice.MixCascade;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertNotNull;
import static junit.framework.Assert.assertTrue;
import static junit.framework.TestCase.assertFalse;

/**
 * Created by martin on 8/14/17.
 */

public class CascadeConnectionManagerTest {

    class TestCascadeConnectionManager extends CascadeConnectionManager {

        public TestAnonClient mTestClient;
        public boolean mSenderStarted = false;
        public boolean mReceiverStarted = false;

        public TestCascadeConnectionManager(NotificationDispatcher notification,
                                            INetworkFactory networkFactory,
                                            IConnectivityMonitor connMonitor,
                                            IConfig config,
                                            HTTPClientSocketFactory mixSocketFactory) {
            super(notification, networkFactory, config, connMonitor, mixSocketFactory);
        }

        @Override
        protected AnonClient getNewAnonClient() {
            if (mTestClient == null) {
                mTestClient = new TestAnonClient();
            }
            return mTestClient;
        }

        @Override
        protected void startReceiverThread() {
            mReceiverStarted = true;
        }

        @Override
        protected void startSenderThread() {
            mSenderStarted = true;
        }

        @Override
        protected void stopReceiverThread() {
            mReceiverStarted = false;
        }

        @Override
        protected void stopSenderThread() {
            mSenderStarted = false;
        }

        public boolean pReconnectTaskIsRunning () {
            return reconnectTaskIsRunning();
        }

        public void waitForReconnectTask() {
            if (reconnectTaskIsRunning()) {
                try {
                    mReconnectTask.join();
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    class TestNotificationDispatcher extends NotificationDispatcher {

        public List<Event> mNotifications = new LinkedList<Event>();

        @Override
        public void notify(Event event) {
            mNotifications.add(event);
            super.notify(event);
        }

    }

    @Test
    public void testInitialSetup() {

        TestNotificationDispatcher notification = new TestNotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();
        TestConnectivityMonitor testConnMonitor = new TestConnectivityMonitor();

        try {

            /**
             * The user has chosen a cascade and enabled tunneling. We now
             * expect the CascadeConnectionManager to instantiate an AnonClient
             * and call its initialize method.
             */
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    testConnMonitor,
                    config,
                    testHTTPFactory);

            // Check if our test client has been initialized with the correct cascade
            assertNotNull(connectionManager.mTestClient);
            assertNotNull(connectionManager.mTestClient.mUsedCascade);
            assertTrue(connectionManager.mTestClient.initialized);
            assertTrue(connectionManager.mTestClient.mUsedCascade.compareMixIDs(testCascade));
            assertFalse(connectionManager.mReceiverStarted);
            assertFalse(connectionManager.mSenderStarted);
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            notification.mNotifications.clear();

            /**
             * After the connection to the chosen mix cascade has been established, the CascadeConnectionManager-instance
             * should set up the sender and receiver thread.
             */

            connectionManager.connectionEstablished(testCascade);

            assertTrue(connectionManager.mReceiverStarted);
            assertTrue(connectionManager.mSenderStarted);
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTED);
            notification.mNotifications.clear();

        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        }
    }

    /**
     * The following test checks if the CascadeConnectionManager class
     * properly shuts down the VPN functionality if the user disables
     * the anonymization while the app is connected to the anonymization service.
     */
    @Test
    public void testShutdownWhenConnected() {

        TestNotificationDispatcher notification = new TestNotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();
        TestConnectivityMonitor testConnMonitor = new TestConnectivityMonitor();

        try {
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    testConnMonitor,
                    config,
                    testHTTPFactory);
            notification.mNotifications.clear();
            connectionManager.connectionEstablished(testCascade);
            notification.mNotifications.clear();

            /**
             * After the user disabled the tunneling, we expect the CascadeConnectionManager
             * to shut down the sender and receiver thread.
             */
            assertTrue(connectionManager.mReceiverStarted);
            assertTrue(connectionManager.mSenderStarted);
            assertFalse(connectionManager.pReconnectTaskIsRunning());
            notification.notify(Event.VPN_DISABLED);
            assertFalse(connectionManager.mReceiverStarted);
            assertFalse(connectionManager.mSenderStarted);
            assertFalse(connectionManager.mTestClient.initialized);

        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        }

    }

    @Test
    public void testSuccessfulReconnectAfterFailedInitialize() {

        TestNotificationDispatcher notification = new TestNotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();
        TestConnectivityMonitor testConnMonitor = new TestConnectivityMonitor();
        testConnMonitor.setConnected(true);

        try {
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestAnonClient.numOfInitializationsToFail = 2;
            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    testConnMonitor,
                    config,
                    testHTTPFactory);

            connectionManager.waitForReconnectTask();
            assertEquals(0, TestAnonClient.numOfInitializationsToFail = 0);
            connectionManager.connectionEstablished(testCascade);

            assertEquals(2, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            assertTrue(notification.mNotifications.get(1) == Event.CASCADE_CONNECTED);
            notification.mNotifications.clear();

            assertTrue(connectionManager.mTestClient.initialized);

        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        } finally {
            TestAnonClient.numOfInitializationsToFail = 0;
        }

    }

    @Test
    public void testSuccessfulReconnectAfterSuccessfulInitialize() {

        TestNotificationDispatcher notification = new TestNotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();
        TestConnectivityMonitor testConnMonitor = new TestConnectivityMonitor();
        testConnMonitor.setConnected(true);

        try {
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    testConnMonitor,
                    config,
                    testHTTPFactory);

            assertEquals(1, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            notification.mNotifications.clear();

            connectionManager.connectionEstablished(testCascade);
            assertEquals(1, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTED);
            notification.mNotifications.clear();

            /**
             * We now simulate the case that the connection still fails.
             */
            TestAnonClient.numOfInitializationsToFail = 1;

            /**
             * The connection gets interrupted while the device is still connected to the
             * Wifi or mobile network.
             */
            connectionManager.connectionError(new AnonServiceException(testCascade, "Test service exception", 0));

            connectionManager.waitForReconnectTask();
            assertEquals(0, TestAnonClient.numOfInitializationsToFail = 0);
            connectionManager.connectionEstablished(testCascade);

            assertEquals(2, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            assertTrue(notification.mNotifications.get(1) == Event.CASCADE_CONNECTED);
            notification.mNotifications.clear();

            assertTrue(connectionManager.mTestClient.initialized);

        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        } finally {
            TestAnonClient.numOfInitializationsToFail = 0;
        }

    }

    @Test
    public void testFailedReconnectAfterSuccessfulInitialize() {

        TestNotificationDispatcher notification = new TestNotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();
        TestConnectivityMonitor testConnMonitor = new TestConnectivityMonitor();
        testConnMonitor.setConnected(true);

        try {
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    testConnMonitor,
                    config,
                    testHTTPFactory);

            assertEquals(1, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            notification.mNotifications.clear();

            connectionManager.connectionEstablished(testCascade);
            assertEquals(1, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTED);
            notification.mNotifications.clear();

            /**
             * The connection gets interrupted while the device is still connected to the
             * Wifi or mobile network.
             */
            TestAnonClient.numOfInitializationsToFail = CascadeConnectionManager.MAX_FAILED_CONNECT_ATTEMPTS + 1;
            connectionManager.connectionError(new AnonServiceException(testCascade, "Test service exception", 0));

            connectionManager.waitForReconnectTask();

            assertEquals(2, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            assertTrue(notification.mNotifications.get(1) == Event.CASCADE_LOST);
            notification.mNotifications.clear();

        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        } finally {
            TestAnonClient.numOfInitializationsToFail = 0;
        }

    }

    @Test
    public void testFailedReconnectAfterFailedInitialize() {

        TestNotificationDispatcher notification = new TestNotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();
        TestConnectivityMonitor testConnMonitor = new TestConnectivityMonitor();
        testConnMonitor.setConnected(true);

        try {
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestAnonClient.numOfInitializationsToFail = CascadeConnectionManager.MAX_FAILED_CONNECT_ATTEMPTS;

            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    testConnMonitor,
                    config,
                    testHTTPFactory);

            connectionManager.waitForReconnectTask();
            assertEquals(0, TestAnonClient.numOfInitializationsToFail);
            assertEquals(2, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            assertTrue(notification.mNotifications.get(1) == Event.CASCADE_LOST);
            notification.mNotifications.clear();

        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        } finally {
            TestAnonClient.numOfInitializationsToFail = 0;
        }

    }

    @Test
    public void testAutomaticReconnectAfterFailedInitialize() {

        TestNotificationDispatcher notification = new TestNotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();
        TestConnectivityMonitor testConnMonitor = new TestConnectivityMonitor();
        testConnMonitor.setConnected(false);

        try {
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestAnonClient.numOfInitializationsToFail = 1;
            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    testConnMonitor,
                    config,
                    testHTTPFactory);

            connectionManager.waitForReconnectTask();

            assertEquals(2, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            assertTrue(notification.mNotifications.get(1) == Event.CASCADE_DISCONNECTED);
            notification.mNotifications.clear();

            assertFalse(connectionManager.mTestClient.initialized);
            testConnMonitor.setConnected(true);
            notification.notify(Event.NETWORK_CONNECTED);
            assertTrue(connectionManager.mTestClient.initialized);
            assertEquals(2, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(1) == Event.CASCADE_CONNECTING);
            notification.mNotifications.clear();


        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        } finally {
            TestAnonClient.numOfInitializationsToFail = 0;
        }

    }

    @Test
    public void testAutomaticReconnectAfterSuccessfulInitialize() {

        TestNotificationDispatcher notification = new TestNotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();
        TestConnectivityMonitor testConnMonitor = new TestConnectivityMonitor();
        testConnMonitor.setConnected(true);

        try {
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    testConnMonitor,
                    config,
                    testHTTPFactory);

            assertEquals(1, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            notification.mNotifications.clear();

            connectionManager.connectionEstablished(testCascade);
            assertEquals(1, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTED);
            notification.mNotifications.clear();

            testConnMonitor.setConnected(false);

            /**
             * The connection gets interrupted while the device is still connected to the
             * Wifi or mobile network.
             */
            connectionManager.connectionError(new AnonServiceException(testCascade, "Test service exception", 0));
            connectionManager.waitForReconnectTask();
            assertEquals(2, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(0) == Event.CASCADE_CONNECTING);
            assertTrue(notification.mNotifications.get(1) == Event.CASCADE_DISCONNECTED);
            notification.mNotifications.clear();

            connectionManager.mTestClient.initialized = false;
            testConnMonitor.setConnected(true);
            notification.notify(Event.NETWORK_CONNECTED);
            assertTrue(connectionManager.mTestClient.initialized);
            assertEquals(2, notification.mNotifications.size());
            assertTrue(notification.mNotifications.get(1) == Event.CASCADE_CONNECTING);
            notification.mNotifications.clear();

        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        } finally {
            TestAnonClient.numOfInitializationsToFail = 0;
        }

    }

    private MixCascade getDummyCascade() {
        try {
            String xml = "        <MixCascade context=\"jondonym\" id=\"F30905FCD73B6B30CB5FEFD3250FD66EF4B32591\" serial=\"1500154699723\">\n" +
                    "            <Name>Dresden-Dresden</Name>\n" +
                    "            <Network>\n" +
                    "                <ListenerInterfaces>\n" +
                    "                    <ListenerInterface>\n" +
                    "                        <Type>RAW/TCP</Type>\n" +
                    "                        <Port>6544</Port>\n" +
                    "                        <Host>141.76.46.165</Host>\n" +
                    "                        <IP>141.76.46.165</IP>\n" +
                    "                    </ListenerInterface>\n" +
                    "                </ListenerInterfaces>\n" +
                    "            </Network>\n" +
                    "            <Mixes count=\"2\">\n" +
                    "                <Mix id=\"F30905FCD73B6B30CB5FEFD3250FD66EF4B32591\" serial=\"1495732588740\">\n" +
                    "                    <MixType>FirstMix</MixType>\n" +
                    "                    <Name>First Mix of Dresden-Dresden</Name>\n" +
                    "                    <Location>\n" +
                    "                        <City>Dresden</City>\n" +
                    "                        <State>Saxony</State>\n" +
                    "                        <Country>de</Country>\n" +
                    "                        <Position>\n" +
                    "                            <Geo>\n" +
                    "                                <Longitude>13.762</Longitude>\n" +
                    "                                <Latitude>51.053</Latitude>\n" +
                    "                            </Geo>\n" +
                    "                        </Position>\n" +
                    "                    </Location>\n" +
                    "                    <Operator>\n" +
                    "                        <Organisation>JAP-Team</Organisation>\n" +
                    "                        <URL>http://anon.inf.tu-dresden.de/</URL>\n" +
                    "                        <Country>de</Country>\n" +
                    "                        <OrganisationalUnit>Department of Computer Science, Institute for System Architecture</OrganisationalUnit>\n" +
                    "                    </Operator>\n" +
                    "                    <ListenerInterfaces>\n" +
                    "                        <ListenerInterface>\n" +
                    "                            <NetworkProtocol>RAW/TCP</NetworkProtocol>\n" +
                    "                            <Host>141.76.46.165</Host>\n" +
                    "                            <IP>141.76.46.165</IP>\n" +
                    "                            <Port>6544</Port>\n" +
                    "                        </ListenerInterface>\n" +
                    "                    </ListenerInterfaces>\n" +
                    "                    <Software>\n" +
                    "                        <Version>00.11.17</Version>\n" +
                    "                    </Software>\n" +
                    "                    <PaymentReminderProbability>-1</PaymentReminderProbability>\n" +
                    "                    <LastUpdate>1500155540078</LastUpdate>\n" +
                    "                </Mix>\n" +
                    "                <Mix id=\"75ACF4F101510607BA3E9E348821D8697BE8FC58\" serial=\"1495732671831\">\n" +
                    "                    <MixType payment=\"False\">LastMix</MixType>\n" +
                    "                    <Name>Last Mix of Dresden-Dresden</Name>\n" +
                    "                    <Location>\n" +
                    "                        <City>Dresden</City>\n" +
                    "                        <State>Saxony</State>\n" +
                    "                        <Country>de</Country>\n" +
                    "                        <Position>\n" +
                    "                            <Geo>\n" +
                    "                                <Longitude>13.762</Longitude>\n" +
                    "                                <Latitude>51.053</Latitude>\n" +
                    "                            </Geo>\n" +
                    "                        </Position>\n" +
                    "                    </Location>\n" +
                    "                    <Operator>\n" +
                    "                        <Organisation>JAP-Team</Organisation>\n" +
                    "                        <URL>http://anon.inf.tu-dresden.de/</URL>\n" +
                    "                        <Country>de</Country>\n" +
                    "                        <OrganisationalUnit>Department of Computer Science, Institute for System Architecture</OrganisationalUnit>\n" +
                    "                    </Operator>\n" +
                    "                    <ListenerInterfaces>\n" +
                    "                        <ListenerInterface>\n" +
                    "                            <NetworkProtocol>RAW/TCP</NetworkProtocol>\n" +
                    "                            <Port>46544</Port>\n" +
                    "                            <Host>141.76.45.33</Host>\n" +
                    "                        </ListenerInterface>\n" +
                    "                    </ListenerInterfaces>\n" +
                    "                    <Proxies>\n" +
                    "                        <Proxy>\n" +
                    "                            <VisibleAddresses>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.33</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.34</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.35</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.33</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.34</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.35</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                            </VisibleAddresses>\n" +
                    "                        </Proxy>\n" +
                    "                    </Proxies>\n" +
                    "                    <Software>\n" +
                    "                        <Version>00.11.19</Version>\n" +
                    "                    </Software>\n" +
                    "                    <LastUpdate>1500155542631</LastUpdate>\n" +
                    "                    <MixProtocolVersion>0.6</MixProtocolVersion>\n" +
                    "                    <FlowControl withUpstreamFlowControl=\"true\">\n" +
                    "                        <UpstreamSendMe>80</UpstreamSendMe>\n" +
                    "                        <DownstreamSendMe>80</DownstreamSendMe>\n" +
                    "                    </FlowControl>\n" +
                    "                    <RSAKeyValue>\n" +
                    "                        <Modulus>rB4/VxKrBWSHLUIYY/Pq2qRuq8k+NrnM4sex4R42BzU2rlVnB8EPNnAMgfCxpUFw\n" +
                    "62Rvvxr22uUTb1xcDWaBEcjC4W9if3KVMwru40kfG+p77+G0/mSURwQYGGT9JzBe\n" +
                    "3pb/Yo9eLnerTtCXe7pNZgC6zXPThnz3ODo494/FaFU=</Modulus>\n" +
                    "                        <Exponent>AQAB</Exponent>\n" +
                    "                    </RSAKeyValue>\n" +
                    "                    <Compatibility version=\"00.11.19\">\n" +
                    "                        <Flags>\n" +
                    "                            <NewFlowControl></NewFlowControl>\n" +
                    "                            <NewChannelEncryption></NewChannelEncryption>\n" +
                    "                            <WithIntegrityCheck></WithIntegrityCheck>\n" +
                    "                        </Flags>\n" +
                    "                    </Compatibility>\n" +
                    "                    <Nonce>1d+zYXdF5RlaU2l30wBKPQ==</Nonce>\n" +
                    "                    <KeepAlive>\n" +
                    "                        <SendInterval>65000</SendInterval>\n" +
                    "                        <ReceiveInterval>75000</ReceiveInterval>\n" +
                    "                    </KeepAlive>\n" +
                    "                </Mix>\n" +
                    "            </Mixes>\n" +
                    "            <Payment required=\"false\"></Payment>\n" +
                    "            <LastUpdate>1500993297300</LastUpdate>\n" +
                    "        </MixCascade>";
            SignatureVerifier.getInstance().setCheckSignatures(false);
            Document doc = anon.util.XMLUtil.readXMLDocument(new StringReader(xml));
            return new MixCascade(doc.getDocumentElement());
        } catch (Exception e) {
            System.out.println("FAILED TO GENERATE DUMMY CASCADE: " + e.getMessage());
            return null;
        }
    }
}
