package anonvpn.anon_next.core.networking;

import anonvpn.anon_next.core.CascadeConnectionManager;
import anonvpn.anon_next.core.TestAnonClient;
import anonvpn.anon_next.core.TestPacket;
import anonvpn.anon_next.core.TestTCPChannel;
import anonvpn.anon_next.core.entities.IPv4Packet;
import anonvpn.anon_next.core.networking.mocks.TestHTTPClientSocketFactory;
import anonvpn.anon_next.core.networking.mocks.TestOutputStream;
import anonvpn.anon_next.core.networking.mocks.TestTunnelingInterface;
import anonvpn.anon_next.core.persistence.IConfig;
import anonvpn.anon_next.core.persistence.mocks.TestConfig;
import org.junit.Test;

import java.io.IOException;
import java.io.OutputStream;
import java.net.Inet4Address;
import java.nio.ByteBuffer;

import static org.junit.Assert.*;

/**
 * Example local unit test, which will execute on the development machine (host).
 *
 * @see <a href="http://d.android.com/tools/testing">Testing documentation</a>
 */
public class SenderThreadTest {

    @Test
    public void TCPHandling() throws Exception {

        TestConfig config = new TestConfig();
        TestTunnelingInterface tunDevice = new TestTunnelingInterface();
        TestOutputStream mixOut = new TestOutputStream();
        TestAnonClient mAnonClient = new TestAnonClient();
        TestHTTPClientSocketFactory mMixSocketFactory = new TestHTTPClientSocketFactory();

        SenderThread testThread = new SenderThread(tunDevice, mixOut, mAnonClient, mMixSocketFactory);

        // Prepare tunneling interface
        Inet4Address testSrc = (Inet4Address) Inet4Address.getByName("113.60.50.10");
        Inet4Address testDst = (Inet4Address) Inet4Address.getByName("113.60.50.11");
        ByteBuffer testSyn = TestPacket.generateTCPSynPacket(testSrc, testDst,(short) 5000, (short) 6000);
        IPv4Packet testPkt = new IPv4Packet(testSyn.array());
        tunDevice.mAvailablePackets.add(testPkt);

        // Start sender Thread
        assertTrue(mAnonClient.mTestTCPChannels.size() == 0);
        assertTrue(tunDevice.mWrittenPackets.size() == 0);

        testThread.start();
        Thread.sleep(500);
        testThread.terminate();
        testThread.join();

        assertTrue(mAnonClient.mTestTCPChannels.size() == 1);
        assertTrue(tunDevice.mWrittenPackets.size() == 1);

        TestTCPChannel testChannel = mAnonClient.mTestTCPChannels.get(0);

        assertEquals("113.60.50.11", testChannel.mHostname);
        assertEquals(6000, testChannel.mPort);
    }

    @Test
    public void testDNSHandlingNoFakeDNSNoFakeSocks() throws Exception {

        // 1. Prepare sender thread
        IConfig config = new TestConfig();
        TestTunnelingInterface tunDevice = new TestTunnelingInterface();
        TestOutputStream mixOut = new TestOutputStream();
        TestAnonClient anonClient = new TestAnonClient();
        TestHTTPClientSocketFactory mixFactory = new TestHTTPClientSocketFactory();

        CascadeConnectionManager.USE_FAKE_DNS = false;
        CascadeConnectionManager.USE_UDP_SOCKS = false;

        SenderThread testThread = new SenderThread(tunDevice, mixOut, anonClient, mixFactory);

        // 2. Prepare tunneling interface
        Inet4Address testSrc = (Inet4Address) Inet4Address.getByName("113.60.50.10");
        Inet4Address testDst = (Inet4Address) Inet4Address.getByName("113.60.50.11");
        ByteBuffer testSyn = TestPacket.generateDNSPacket(testSrc, testDst,(short) 5000, (short) 6000, "test.dud.inf.tu-dresden.de");
        IPv4Packet testPkt = new IPv4Packet(testSyn.array());
        tunDevice.mAvailablePackets.add(testPkt);

        testThread.start();
        Thread.sleep(500);
        testThread.terminate();
        testThread.join();

        assertEquals(0, anonClient.mTestUDPChannels.size());
        assertTrue(mixOut.mWrittenBytes.size() > 0);
        assertEquals(0, tunDevice.mWrittenPackets.size());
    }
    @Test
    public void testFakeDNSNoDNSSocks() throws Exception {

        // 1. Prepare sender thread
        IConfig config = new TestConfig();
        TestTunnelingInterface tunDevice = new TestTunnelingInterface();
        TestOutputStream mixOut = new TestOutputStream();
        TestAnonClient anonClient = new TestAnonClient();
        TestHTTPClientSocketFactory mixFactory = new TestHTTPClientSocketFactory();

        CascadeConnectionManager.USE_FAKE_DNS = true;
        CascadeConnectionManager.USE_UDP_SOCKS = false;

        assertEquals(0,anonClient.mTestUDPChannels.size());
        SenderThread testThread = new SenderThread(tunDevice, mixOut, anonClient, mixFactory);

        // 2. Prepare tunneling interface
        Inet4Address testSrc = (Inet4Address) Inet4Address.getByName("113.60.50.10");
        Inet4Address testDst = (Inet4Address) Inet4Address.getByName("10.0.0.1");

        // TODO GenerateDNSPacket does not yet generate valid DNS packets
        ByteBuffer testSyn = TestPacket.generateDNSPacket(testSrc, testDst,(short) 5000, (short) 6000, "test.dud.inf.tu-dresden.de");
        IPv4Packet testPkt = new IPv4Packet(testSyn.array());
        tunDevice.mAvailablePackets.add(testPkt);
        assertEquals(0, tunDevice.mWrittenPackets.size());

        testThread.start();
        Thread.sleep(500);
        testThread.terminate();
        testThread.join();

        assertEquals(0, anonClient.mTestUDPChannels.size());
        assertEquals(0, mixOut.mWrittenBytes.size());
        assertEquals(1, tunDevice.mWrittenPackets.size());
    }

    @Test
    public void testDNSSocksNoFakeDNS() throws Exception {

        // 1. Prepare sender thread
        IConfig config = new TestConfig();
        TestTunnelingInterface tunDevice = new TestTunnelingInterface();
        TestOutputStream mixOut = new TestOutputStream();
        TestAnonClient anonClient = new TestAnonClient();
        TestHTTPClientSocketFactory mixFactory = new TestHTTPClientSocketFactory();

        CascadeConnectionManager.USE_FAKE_DNS = false;
        CascadeConnectionManager.USE_UDP_SOCKS = true;

        SenderThread testThread = new SenderThread(tunDevice, mixOut, anonClient, mixFactory);

        // 2. Prepare tunneling interface
        Inet4Address testSrc = (Inet4Address) Inet4Address.getByName("113.60.50.10");
        Inet4Address testDst = (Inet4Address) Inet4Address.getByName("10.0.0.1");

        // TODO GenerateDNSPacket does not yet generate valid DNS packets
        ByteBuffer testRequest = TestPacket.generateDNSPacket(testSrc, testDst,(short) 5000, (short) 53, "test.dud.inf.tu-dresden.de");
        IPv4Packet testPkt = new IPv4Packet(testRequest.array());
        tunDevice.mAvailablePackets.add(testPkt);

        testThread.start();
        Thread.sleep(1000);
        testThread.terminate();
        testThread.join();

        assertEquals(1, anonClient.mTestUDPChannels.size());
        assertEquals(testPkt.getLength() - 20 - 8, ((TestOutputStream) anonClient.mTestUDPChannels.get(0).getOutputStream()).mWrittenBytes.size());
        assertEquals(0, mixOut.mWrittenBytes.size());
        assertEquals(0, tunDevice.mWrittenPackets.size());
    }

}