package anonvpn.anon_next.core.networking;

import org.junit.Test;
import org.w3c.dom.Document;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.net.ConnectException;
import java.util.LinkedList;

import HTTPClient.HTTPClientSocketFactory;
import anon.AnonChannel;
import anon.AnonServerDescription;
import anon.IServiceContainer;
import anon.client.AnonClient;
import anon.crypto.SignatureVerifier;
import anon.error.AnonServiceException;
import anon.infoservice.MixCascade;
import anon.infoservice.SimpleMixCascadeContainer;
import anon.terms.TermsAndConditionConfirmation;
import anonvpn.anon_next.core.CascadeConnectionManager;
import anonvpn.anon_next.core.networking.mocks.TestConfig;
import anonvpn.anon_next.core.networking.mocks.TestHTTPClientSocketFactory;
import anonvpn.anon_next.core.networking.mocks.TestInternetSocket;
import anonvpn.anon_next.core.networking.mocks.TestNetworkFactory;
import anonvpn.anon_next.core.networking.mocks.TestTunnelingInterface;
import anonvpn.anon_next.core.notification.Event;
import anonvpn.anon_next.core.notification.NotificationDispatcher;
import anonvpn.anon_next.core.persistence.IConfig;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertNotNull;
import static junit.framework.Assert.assertTrue;
import static junit.framework.TestCase.assertFalse;

/**
 * Created by martin on 8/14/17.
 */

public class CascadeConnectionManagerTest {

    class TestChannel implements AnonChannel {

        public Boolean mClosed = false;

        @Override
        public InputStream getInputStream() {
            return null;
        }

        @Override
        public OutputStream getOutputStream() {
            return null;
        }

        @Override
        public int getOutputBlockSize() {
            return 0;
        }

        @Override
        public void close() {
            mClosed = true;
        }

        @Override
        public boolean isClosed() {
            return mClosed;
        }
    }

    class TestAnonClient extends AnonClient {

        public MixCascade mUsedCascade = null;
        public TestChannel mTestChannel = null;
        public boolean initialized = false;

        @Override
        public void initialize(final AnonServerDescription a_mixCascade,
                               final IServiceContainer a_serviceContainer,
                               final TermsAndConditionConfirmation termsConfirmation,
                               final boolean a_bIsReconnect) throws AnonServiceException {
            mUsedCascade = ((MixCascade) a_mixCascade);
            initialized = true;
        }

        @Override
        public AnonChannel createChannel(int a_type) throws ConnectException {
            mTestChannel = new TestChannel();
            return mTestChannel;

        }

        @Override
        public boolean isConnected() {
            return true;
        }

        @Override
        public void shutdown(boolean resetTransferredBytes) {
            initialized = false;
        }
    }

    class TestCascadeConnectionManager extends CascadeConnectionManager {

        public TestAnonClient mTestClient;
        public boolean mSenderStarted = false;
        public boolean mReceiverStarted = false;


        public TestCascadeConnectionManager(NotificationDispatcher notification,
                                            INetworkFactory networkFactory,
                                            IConfig config,
                                            HTTPClientSocketFactory mixSocketFactory) {
            super(notification, networkFactory, config, mixSocketFactory);
        }

        @Override
        protected AnonClient getNewAnonClient() {
            if (mTestClient == null) {
                mTestClient = new TestAnonClient();
            }
            return mTestClient;
        }

        @Override
        protected void startReceiverThread() {
            mReceiverStarted = true;
        }

        @Override
        protected void startSenderThread() {
            mSenderStarted = true;
        }

        @Override
        protected void stopReceiverThread() {
            mReceiverStarted = false;
        }

        @Override
        protected void stopSenderThread() {
            mSenderStarted = false;
        }
    }

    @Test
    public void testCascadeSetup() {

        NotificationDispatcher notification = new NotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();


        /**
         *
         * The actual test
         *
         */

        try {
            MixCascade testCascade = getDummyCascade();
            assertNotNull(testCascade);
            config.setChosenMixCascade(testCascade);
            assertNotNull(config.getChosenMixCascade());

            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    config,
                    testHTTPFactory);

            // Check if our test client has been initialized with the correct cascade
            assertNotNull(connectionManager.mTestClient);
            assertNotNull(connectionManager.mTestClient.mUsedCascade);
            assertTrue(connectionManager.mTestClient.initialized);
            assertTrue(connectionManager.mTestClient.mUsedCascade.compareMixIDs(testCascade));

        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        }
    }

    @Test
    public void testChannelInteraction() {

        NotificationDispatcher notification = new NotificationDispatcher();
        TestTunnelingInterface tunDev = new TestTunnelingInterface();
        InputStream iSocketInput = new InputStream() {
            @Override
            public int read() throws IOException {
                return 0;
            }
        };
        OutputStream iSocketOutput = new OutputStream() {
            @Override
            public void write(int b) throws IOException {
                return;
            }
        };
        TestInternetSocket iSocket = new TestInternetSocket(iSocketInput, iSocketOutput);
        INetworkFactory networkFactory = new TestNetworkFactory(tunDev, iSocket);
        TestConfig config = new TestConfig();
        TestHTTPClientSocketFactory testHTTPFactory = new TestHTTPClientSocketFactory();

        /**
         *
         * The actual test
         *
         */

        try {
            MixCascade testCascade = getDummyCascade();
            config.setChosenMixCascade(testCascade);

            TestCascadeConnectionManager connectionManager = new TestCascadeConnectionManager(notification,
                    networkFactory,
                    config,
                    testHTTPFactory);

            assertFalse(connectionManager.mReceiverStarted);
            assertFalse(connectionManager.mSenderStarted);
            assertTrue(connectionManager.mTestClient.mTestChannel == null);

            connectionManager.connectionEstablished(testCascade);

            assertTrue(connectionManager.mReceiverStarted);
            assertTrue(connectionManager.mSenderStarted);
            assertNotNull(connectionManager.mTestClient.mTestChannel);

            notification.notify(Event.VPN_DISABLED);
            assertFalse(connectionManager.mReceiverStarted);
            assertFalse(connectionManager.mSenderStarted);
            assertFalse(connectionManager.mTestClient.initialized);


        } catch (Exception e) {
            System.out.println("TEST THROWN UNEXPECTED EXCEPTION: " + e.getMessage());
            e.printStackTrace();
            assertFalse(true);
        }
    }


    private MixCascade getDummyCascade() {
        try {
            String xml = "        <MixCascade context=\"jondonym\" id=\"F30905FCD73B6B30CB5FEFD3250FD66EF4B32591\" serial=\"1500154699723\">\n" +
                    "            <Name>Dresden-Dresden</Name>\n" +
                    "            <Network>\n" +
                    "                <ListenerInterfaces>\n" +
                    "                    <ListenerInterface>\n" +
                    "                        <Type>RAW/TCP</Type>\n" +
                    "                        <Port>6544</Port>\n" +
                    "                        <Host>141.76.46.165</Host>\n" +
                    "                        <IP>141.76.46.165</IP>\n" +
                    "                    </ListenerInterface>\n" +
                    "                </ListenerInterfaces>\n" +
                    "            </Network>\n" +
                    "            <Mixes count=\"2\">\n" +
                    "                <Mix id=\"F30905FCD73B6B30CB5FEFD3250FD66EF4B32591\" serial=\"1495732588740\">\n" +
                    "                    <MixType>FirstMix</MixType>\n" +
                    "                    <Name>First Mix of Dresden-Dresden</Name>\n" +
                    "                    <Location>\n" +
                    "                        <City>Dresden</City>\n" +
                    "                        <State>Saxony</State>\n" +
                    "                        <Country>de</Country>\n" +
                    "                        <Position>\n" +
                    "                            <Geo>\n" +
                    "                                <Longitude>13.762</Longitude>\n" +
                    "                                <Latitude>51.053</Latitude>\n" +
                    "                            </Geo>\n" +
                    "                        </Position>\n" +
                    "                    </Location>\n" +
                    "                    <Operator>\n" +
                    "                        <Organisation>JAP-Team</Organisation>\n" +
                    "                        <URL>http://anon.inf.tu-dresden.de/</URL>\n" +
                    "                        <Country>de</Country>\n" +
                    "                        <OrganisationalUnit>Department of Computer Science, Institute for System Architecture</OrganisationalUnit>\n" +
                    "                    </Operator>\n" +
                    "                    <ListenerInterfaces>\n" +
                    "                        <ListenerInterface>\n" +
                    "                            <NetworkProtocol>RAW/TCP</NetworkProtocol>\n" +
                    "                            <Host>141.76.46.165</Host>\n" +
                    "                            <IP>141.76.46.165</IP>\n" +
                    "                            <Port>6544</Port>\n" +
                    "                        </ListenerInterface>\n" +
                    "                    </ListenerInterfaces>\n" +
                    "                    <Software>\n" +
                    "                        <Version>00.11.17</Version>\n" +
                    "                    </Software>\n" +
                    "                    <PaymentReminderProbability>-1</PaymentReminderProbability>\n" +
                    "                    <LastUpdate>1500155540078</LastUpdate>\n" +
                    "                </Mix>\n" +
                    "                <Mix id=\"75ACF4F101510607BA3E9E348821D8697BE8FC58\" serial=\"1495732671831\">\n" +
                    "                    <MixType payment=\"False\">LastMix</MixType>\n" +
                    "                    <Name>Last Mix of Dresden-Dresden</Name>\n" +
                    "                    <Location>\n" +
                    "                        <City>Dresden</City>\n" +
                    "                        <State>Saxony</State>\n" +
                    "                        <Country>de</Country>\n" +
                    "                        <Position>\n" +
                    "                            <Geo>\n" +
                    "                                <Longitude>13.762</Longitude>\n" +
                    "                                <Latitude>51.053</Latitude>\n" +
                    "                            </Geo>\n" +
                    "                        </Position>\n" +
                    "                    </Location>\n" +
                    "                    <Operator>\n" +
                    "                        <Organisation>JAP-Team</Organisation>\n" +
                    "                        <URL>http://anon.inf.tu-dresden.de/</URL>\n" +
                    "                        <Country>de</Country>\n" +
                    "                        <OrganisationalUnit>Department of Computer Science, Institute for System Architecture</OrganisationalUnit>\n" +
                    "                    </Operator>\n" +
                    "                    <ListenerInterfaces>\n" +
                    "                        <ListenerInterface>\n" +
                    "                            <NetworkProtocol>RAW/TCP</NetworkProtocol>\n" +
                    "                            <Port>46544</Port>\n" +
                    "                            <Host>141.76.45.33</Host>\n" +
                    "                        </ListenerInterface>\n" +
                    "                    </ListenerInterfaces>\n" +
                    "                    <Proxies>\n" +
                    "                        <Proxy>\n" +
                    "                            <VisibleAddresses>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.33</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.34</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.35</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.33</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.34</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                                <VisibleAddress>\n" +
                    "                                    <Host>141.76.45.35</Host>\n" +
                    "                                </VisibleAddress>\n" +
                    "                            </VisibleAddresses>\n" +
                    "                        </Proxy>\n" +
                    "                    </Proxies>\n" +
                    "                    <Software>\n" +
                    "                        <Version>00.11.19</Version>\n" +
                    "                    </Software>\n" +
                    "                    <LastUpdate>1500155542631</LastUpdate>\n" +
                    "                    <MixProtocolVersion>0.6</MixProtocolVersion>\n" +
                    "                    <FlowControl withUpstreamFlowControl=\"true\">\n" +
                    "                        <UpstreamSendMe>80</UpstreamSendMe>\n" +
                    "                        <DownstreamSendMe>80</DownstreamSendMe>\n" +
                    "                    </FlowControl>\n" +
                    "                    <RSAKeyValue>\n" +
                    "                        <Modulus>rB4/VxKrBWSHLUIYY/Pq2qRuq8k+NrnM4sex4R42BzU2rlVnB8EPNnAMgfCxpUFw\n" +
                    "62Rvvxr22uUTb1xcDWaBEcjC4W9if3KVMwru40kfG+p77+G0/mSURwQYGGT9JzBe\n" +
                    "3pb/Yo9eLnerTtCXe7pNZgC6zXPThnz3ODo494/FaFU=</Modulus>\n" +
                    "                        <Exponent>AQAB</Exponent>\n" +
                    "                    </RSAKeyValue>\n" +
                    "                    <Compatibility version=\"00.11.19\">\n" +
                    "                        <Flags>\n" +
                    "                            <NewFlowControl></NewFlowControl>\n" +
                    "                            <NewChannelEncryption></NewChannelEncryption>\n" +
                    "                            <WithIntegrityCheck></WithIntegrityCheck>\n" +
                    "                        </Flags>\n" +
                    "                    </Compatibility>\n" +
                    "                    <Nonce>1d+zYXdF5RlaU2l30wBKPQ==</Nonce>\n" +
                    "                    <KeepAlive>\n" +
                    "                        <SendInterval>65000</SendInterval>\n" +
                    "                        <ReceiveInterval>75000</ReceiveInterval>\n" +
                    "                    </KeepAlive>\n" +
                    "                </Mix>\n" +
                    "            </Mixes>\n" +
                    "            <Payment required=\"false\"></Payment>\n" +
                    "            <LastUpdate>1500993297300</LastUpdate>\n" +
                    "        </MixCascade>";
            SignatureVerifier.getInstance().setCheckSignatures(false);
            Document doc = anon.util.XMLUtil.readXMLDocument(new StringReader(xml));
            return new MixCascade(doc.getDocumentElement());
        } catch (Exception e) {
            System.out.println("FAILED TO GENERATE DUMMY CASCADE: " + e.getMessage());
            return null;
        }
    }
}
