/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
   may be used to endorse or promote products derived from this software without specific
  prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */

package anon;

import java.io.InputStream;
import java.io.OutputStream;

/**
 * An AnonChannel can be used to send and receive data. There are different types of
 * channels and the transmitted data must match the (proxy-)protocol implied by the type
 * of the channel. A new channel is created via {@link AnonService#createChannel(int)  createChannel}
 * <p>
 * It provides an {@link OutputStream} and {@link InputStream} implementation
 * to send and receive data.
 *
 * @version 1.2
 */
public interface AnonChannel
{
	/** Constant for the channel type: HTTP-Proxy. Such a channel could be used to transmit
	 * data which conforms to the HTTP-Proxy-Protocol.
	 */
	public final static int HTTP = 0;

	/** Constant for the channel type: SOCKS-Proxy. Such a channel could be used to transmit
	* data which conforms to the SOCKS-Proxy-Protocol.
	*/
	public final static int SOCKS = 1;

	/** Constant for the channel type: VPN-like Service. Such a channel could be used to transmit
	* data which conforms to the AN.ON project proprietary VPN-like-Protocol.
	*/
	public final static int VPN = 2;

	/** Constant for the channel type: SMTP-Proxy. Such a channel could be used to transmit
	* data which conforms to the SMTP-Protocol.
	*/
	public final static int SMTP = 3;

	/** Constant for the channel type: TCP. Such a channel could be used to transmit
	* data to a target Host:Port over TCP.
	*/
	public final static int TCP = 4;

	/** Constant for the channel type: UDP. Such a channel could be used to transmit
	*  data to a target Host:Port using UDP.
	*  Note: the client automatically select the best suited anonymisation/transport protocol 
	*  for transporting UDP packets based on the features the currently chosen cascade offers.
	*/
	public static final int UDP = 5;

	/** Constant for the channel type: UDP. Such a channel could be used to transmit
	*  data to a target Host:Port using UDP.
	*  Note: this channel will use the TCP-based anonymisation protocol.
	*/
	public static final int UDP_OVER_TCP = 6;

	/** Constant for the channel type: UDP. Such a channel could be used to transmit
	*  data to a target Host:Port using UDP.
	*  Note: this channel will use the UDP-based anonymisation protocol.
	*/
	public static final int UDP_OVER_UDP = 7;
	
	/** The returned InputStream could be used to receive data.
	 *  @return InputStream, useful for receiving data
	 */
	public InputStream getInputStream();

	/** The returned OutputStream could be used to send data.
	 *  @return OutputStream, useful for sending data
	 */
	public OutputStream getOutputStream();

	/** Returns a value that indicates the current optimum size of data to write. Because often
	 *  anon services transport the data fragmented into fixed sized packets for optimum performance
	 *  it may be good to send data according to the packet size to avoid unneccessary overheads.
	 *
	 * @return the current optimum size for output data. If 1 is returned the size does not matter.
	 */
	public int getOutputBlockSize();

	/** Closes the channel and releases all resources used. */
	public void close();

	/** Returns if the channel is already closed, i.e. no data can be sent / recieved any more on this channel.
	 *
	 * @return <code>true</code>, if the channel is already closed
	 */
	public boolean isClosed();
}
