/*
 Copyright (c) 2000- 2004, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
  may be used to endorse or promote products derived from this software without specific
  prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon;

import java.net.ConnectException;

import HTTPClient.HTTPClientSocketFactory;
import anon.error.AnonServiceException;
import anon.infoservice.IMutableProxyInterface;
import anon.terms.TermsAndConditionConfirmation;

/** This interface allows access to anonymization regardless of the underlying
 *  mechanics. See {@link anon.AnonServiceFactory AnonServiceFactory} for
 *  details regarding the available service types.
 *  <p>
 *  An AnonService needs to be initialized with the proper parameters before use
 *  and allows the creation of {@link AnonChannel} objects, which provide In-
 *  and OutputStreams to receive and send data, through it
 *
 *  @see AnonService#createChannel(int)
 */
public interface AnonService
{
	///The version of the AnonLib
	final String ANONLIB_VERSION = "00.21.001";

	/** Initializes the underlying AnonService, using a subclass of
	 * {@link anon.AnonServerDescription AnonServerDescription} to give
	 * specific details.
	 * This method must be called before any Channels can be created.
	 *
	 *  @param serviceDesc Parameters describing the service configuration.
	 *
	 *  @return E_SUCCESS, if the connection could be established
	 *  @return E_ALREADY_CONNECTED, if this AnonService is already connected to a server
	 *  @return E_INVALID_SERVICE, if anonService is not a valid server
	 *  @return E_PROTOCOL_NOT_SUPPORTED, if the chosen AnonService uses a Protocol, which this version of
	 *                                        the AnonLib does not understand
	 *  @return E_CONNECT, if a general connection error occurred
	 *
	 */
	void initialize(AnonServerDescription serviceDesc, IServiceContainer a_serviceContainer,
			TermsAndConditionConfirmation termsConfirmation, boolean a_bIsReconnect)
		throws AnonServiceException;

	/** Sets the settings for the proxy, which should be used to establish network connections
	 * @return E_SUCCESS, if ok
	 * @return E_UNKNOWN, if an error occurred
	 */
	int setProxy(IMutableProxyInterface a_Proxy);

	/** Sets the socket factory to be used in creating sockets to connect to the AnonService.
	 * @param socketFactory socket factory used to create sockets, if NULL the default socket factory of the JRE is used.
	 */
	void setSocketFactory(HTTPClientSocketFactory socketFactory);

	/** Disconnects form the server.*/
	void shutdown(boolean a_bResetTransferredBytes);

	/** Returns true if this Anon Service is connected, e.g. initialized and usable*/
	boolean isConnected();

	/** Creates a new AnonChannel, which could be used for transmitting data. There is a
	 *  limit of 50 channels per AnonService-connection, in order to prevent Denial of Service-attacks
	 *
	 *  @param type the type of the created channel; See {@link anon.AnonChannel AnonChannel}
	 *
	 *  @return {@link anon.AnonChannel AnonChannel}, useful for data transmission
	 *
	 *  @throws ConnectException, if the Channel could not be created
	 *  @throws TooManyOpenChannels, if there a too many open channels for this AnonService
	 */
	AnonChannel createChannel(int type) throws ConnectException;

	/** Adds an AnonServiceEventListener. This listener will receive events like:
	 *  ... For more information see {@link AnonServiceEventListener AnonServiceEventListener}.
	 *  @param l Listener to add
	 */
	void addEventListener(AnonServiceEventListener l);

	/** Removes an AnonServiceEventListener. This Listener will not receive any Events anymore.
	 *  @param l Listener, which will be removed
	 */
	void removeEventListener(AnonServiceEventListener l);

	void removeEventListeners();
}
