package anon.anonudp;

import java.io.FileInputStream;
import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.UnknownHostException;
import java.util.Vector;

import anon.anonudp.exception.DecryptionFailed;
import anon.anonudp.exception.EncryptionFailed;
import anon.anonudp.mixchannel.Channel;
import anon.anonudp.mixchannel.IPv4AndPort;
import anon.anonudp.mixmessage.crypto.EccGroup713;
import anon.anonudp.mixmessage.crypto.LinkEncryption;
import anon.anonudp.mixmessage.crypto.PublicKey;
import anon.anonudp.mixpacket.DataPacket;
import anon.anonudp.mixpacket.IPacket;
import anon.client.KeyExchangeManager;
import anon.infoservice.ListenerInterface;
import anon.infoservice.MixCascade;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

public class UDPMultiplexer
	{
		private DatagramSocket m_socketToMix;
		private DatagramPacket m_datagrampacketSentMixPacket;
		private final byte[] m_datagrampacketSentMixPacketData;

		private final LinkEncryption m_linkEncryption;

		private MixCascade mixCascade;

		private IPv4AndPort firstMix;

		private PublicKey[] publicKeys;
		private UDPMultiplexerReceiver receiver;

		public UDPMultiplexer(MixCascade mixCascade, KeyExchangeManager keyExchangeManager) throws Exception
			{
				this.mixCascade = mixCascade;

				this.setFirstMixAddress(this.mixCascade);
				this.setPublicKeys(keyExchangeManager);

				m_socketToMix = new DatagramSocket();
				m_socketToMix.connect(firstMix.getHost(), firstMix.getPort());
				m_datagrampacketSentMixPacketData=new byte[DataPacket.SIZE+LinkEncryption.OVERHEAD];
				m_datagrampacketSentMixPacket=new DatagramPacket(m_datagrampacketSentMixPacketData,m_datagrampacketSentMixPacketData.length);

				receiver = new UDPMultiplexerReceiver(this.m_socketToMix);

				receiver.start();

				m_linkEncryption = new LinkEncryption(new byte[EccGroup713.SYMMETRIC_KEY_LENGTH]);
			}

		// mocked to work for now
		private void setFirstMixAddress(MixCascade cascade) throws UnknownHostException
			{
				String udpHost = System.getProperty("UDP_MIX_HOST", null);
				String udpPort = System.getProperty("UDP_MIX_PORT", null);
				if (udpHost != null && udpPort != null)
					{
						this.firstMix = new IPv4AndPort(udpHost, Integer.parseInt(udpPort));
						return;
					}

				Vector firstMixListenerInterfaces = this.mixCascade.getListenerInterfaces();
				ListenerInterface udpInterface = null;

				for (int i = 0; i < firstMixListenerInterfaces.size(); ++i)
					{
						if (((ListenerInterface) (firstMixListenerInterfaces.get(i)))
								.getProtocol() == ListenerInterface.PROTOCOL_TYPE_RAW_UDP)
							{
								udpInterface = (ListenerInterface) (firstMixListenerInterfaces.get(i));
								this.firstMix = new IPv4AndPort(udpInterface.getHost(), udpInterface.getPort());
								return;
							}
					}

				throw new RuntimeException("No suitable UDP interface found for first Mix.");

			}

		// mocked to work for now
		private void setPublicKeys(KeyExchangeManager keyExchangeManager) throws IOException
			{
				/*int mixCount = this.mixCascade.getNumberOfMixes();

				this.publicKeys = new PublicKey[mixCount];

				for (int i = 0; i < mixCount; ++i) {
					//this.publicKeys[i] = keyExchangeManager.getMixParameters()[i].getMixCipher().getPublicKey();
				}
				*/

				String pathPrefix = System.getProperty("UDP_MIX_KEY_PATH", "c:/temp/");
				LogHolder.log(LogLevel.DEBUG, LogType.NET, "Try to read UDP Mix keys from: " + pathPrefix);

				PublicKey mix1 = readPublicKeyFromFile(pathPrefix + "Mix1.pub");
				PublicKey mix2 = readPublicKeyFromFile(pathPrefix + "Mix2.pub");
				PublicKey mix3 = readPublicKeyFromFile(pathPrefix + "Mix3.pub");

				this.publicKeys = new PublicKey[] { mix1, mix2, mix3 };
			}

		public void sendToMix(IPacket packet) throws EncryptionFailed, IOException
			{
				LogHolder.log(LogLevel.DEBUG,LogType.NET,"UDPMultiplexer - try to sent a UDP mix packet to first Mix");
				int encLen= m_linkEncryption.encrypt(packet,m_datagrampacketSentMixPacketData);
				m_datagrampacketSentMixPacket.setLength(encLen);
				m_socketToMix.send(m_datagrampacketSentMixPacket);
				LogHolder.log(LogLevel.DEBUG,LogType.NET,"UDPMultiplexer - sent a UDP mix packet to first Mix");
			}

		private void packetToChannel(IPacket mixPacket) throws DecryptionFailed
			{
				int channelID = mixPacket.getChannelID();
				try
					{
						Channel channel = (Channel) Channel.table.get(new Integer(channelID));
						channel.response(mixPacket);
					}
				catch (NullPointerException npe)
					{
						System.out
								.println("The response mix packet is for a channel not known to the multiplexer. Invalid channel id.");
					}
			}

		final private class UDPMultiplexerReceiver implements Runnable
			{
				final DatagramSocket m_udpsocketFromMix;
				volatile boolean m_bRun = true;
				private Thread m_Thread;

				UDPMultiplexerReceiver(DatagramSocket socketToMix)
					{
						m_udpsocketFromMix = socketToMix;
					}

				public void start()
					{
						m_Thread = new Thread(this, "UDPMuiltiplexer Recevier Thread");
						m_Thread.setDaemon(true);
						m_bRun = true;
						m_Thread.start();
					}

				public void run()
					{
						final byte[] buffer = new byte[LinkEncryption.OVERHEAD + DataPacket.SIZE];
						DatagramPacket response = new DatagramPacket(buffer, buffer.length);
						while (this.m_bRun)
							{
								try
									{
										response.setLength(buffer.length);
										m_udpsocketFromMix.receive(response);
									}
								catch (IOException e)
									{
										e.printStackTrace();
										continue;
									}
								try
									{
										IPacket responseMixPacket = m_linkEncryption.decrypt(response.getData());
										packetToChannel(responseMixPacket);
									}
								catch (DecryptionFailed decryptionFailed)
									{
										decryptionFailed.printStackTrace();
									}
							}
					}

				public void stop()
					{
						m_bRun = false;
						try
							{
								m_Thread.join();
							}
						catch (Exception e)
							{
							}
					}
			}

		public PublicKey[] getPublicKeys()
			{
				return this.publicKeys;
			}

		private static PublicKey readPublicKeyFromFile(String path) throws IOException
			{
				FileInputStream binaryKey = new FileInputStream(path);

				int publicKeyLength = PublicKey.SIZE;

				byte[] keyBytes = new byte[publicKeyLength];

				binaryKey.read(keyBytes);

				binaryKey.close();

				return PublicKey.fromBytes(keyBytes);
			}

		public void stop()
			{
				receiver.stop();
				m_socketToMix.close();
			}
	}
