/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.anonudp.client;

import java.io.IOException;

import anon.anonudp.UDPMultiplexer;
import anon.anonudp.exception.EncryptionFailed;
import anon.anonudp.exception.PacketCreationFailed;
import anon.anonudp.mixchannel.Channel;
import anon.anonudp.mixchannel.IPv4AndPort;
import anon.anonudp.mixpacket.DataPacket;
import anon.anonudp.mixpacket.IPacket;
import anon.client.AbstractAnonUDPChannel;
import anon.client.UDPPacketStreamDesynchronisationException;
import anon.client.UDPPacketToLargeException;

public class AnonPureUDPChannel extends AbstractAnonUDPChannel
	{
		private Channel underlyingChannel;
		private UDPMultiplexer multiplexer;

		private IPv4AndPort destination;

		private static final int READ_INTERVAL = 10; // milliseconds

		public AnonPureUDPChannel(UDPMultiplexer multiplexer)
			{
				super();
				this.multiplexer = multiplexer;
			}

		public void setDestination(String host, int port)
			{
				if (this.destination == null)
					{
						try
							{
								this.destination = new IPv4AndPort(host, port);
								this.underlyingChannel = new Channel(this.destination, this.multiplexer.getPublicKeys());
							}
						catch (IOException e)
							{
								throw new RuntimeException("Wasn't able to find IP-Address for given host: " + host);
							}
					}
				else
					{
						throw new RuntimeException("Destination was already set.");
					}
			}

		public void writePacket(byte[] payload, int offset, int length) throws IOException
			{
				if (length == payload.length && offset == 0)
					{
						this.writePacket(payload);
						return;
					}

				byte[] payloadPart = new byte[length];

				System.arraycopy(payload, offset, payloadPart, 0, length);

				this.writePacket(payloadPart);
			}

		public void writePacket(byte[] payload) throws IOException
			{
				IPacket[] packets;

				try
					{
						packets = this.underlyingChannel.request(payload);
					}
				catch (Exception e)
					{
						throw new IOException("Packet creation failed.");
					}

				for (int i = 0; i < packets.length; i++)
					{
						try
							{
								this.multiplexer.sendToMix(packets[i]);
							}
						catch (Exception e)
							{
								throw new IOException("Link encryption failed.");
							}
					}
			}

		public int readPacket(byte[] payload, int off, int len)
				throws IOException, UDPPacketStreamDesynchronisationException, UDPPacketToLargeException
			{

				while (!this.underlyingChannel.hasNext())
					{
						try
							{
								Thread.sleep(READ_INTERVAL);
							}
						catch (InterruptedException e)
							{
								return -1;
							}
					}

				byte[] nextPacketPayload = this.underlyingChannel.next();
				if (nextPacketPayload.length > len)
					{
						throw new UDPPacketToLargeException();
					}
				System.arraycopy(nextPacketPayload, 0, payload, off, nextPacketPayload.length);

				return nextPacketPayload.length;
			}

		public int getOutputBlockSize()
			{
				return DataPacket.SIZE;
			}

		public void close()
			{
				// TODO Auto-generated method stub
			}

		public boolean isClosed()
			{
				// TODO Auto-generated method stub
				return false;
			}

	}
