package anon.anonudp.mixmessage;

import java.io.ByteArrayOutputStream;
import java.io.IOException;

import anon.anonudp.Constants;
import anon.anonudp.mixmessage.crypto.PublicKey;
import anon.anonudp.mixpacket.InitPacket;

public class Fragment
	{
		public static final int DATA_OVERHEAD = 0;
		private static final int INIT_OVERHEAD = PublicKey.SIZE + InitPacket.CHANNEL_KEY_ONION_SIZE
				+ InitPacket.PAYLOAD_SIZE;

		private static final int DUMMY_PAYLOAD_SIZE = 0;
		public static final int DATA_PAYLOAD_SIZE = 274;
		public static final int INIT_PAYLOAD_SIZE = DATA_PAYLOAD_SIZE - (INIT_OVERHEAD - DATA_OVERHEAD);

		static final int ID_SIZE = 2;
		private static final int INDEX_SIZE = 1;
		private static final int HEADER_SIZE = ID_SIZE + INDEX_SIZE;

		public static final int SIZE_DATA = HEADER_SIZE + DATA_PAYLOAD_SIZE;
		public static final int SIZE_INIT = HEADER_SIZE + INIT_PAYLOAD_SIZE;

		static final int SINGLE_FRAGMENT_MESSAGE_ID = 0;
		static final int SINGLE_FRAGMENT_FRAGMENT_NUMBER = 0;

		private static final int BIT_HAS_PADDING = 0x01;
		private static final int BIT_LAST_FRAGMENT = 0x02;

		private int m_iMessageID;
		private int m_iFragmentNumber;
		private final boolean m_bisLast;
		private byte[] payload;

		private boolean bytesCached;
		private byte[] byteArrayCache;

		private Padding padding;

		public Fragment()
			{
				this(SINGLE_FRAGMENT_MESSAGE_ID, SINGLE_FRAGMENT_FRAGMENT_NUMBER, new byte[0], DUMMY_PAYLOAD_SIZE);
			}

		public Fragment(int message_id, int fragment_number, byte[] payload, int payload_limit)
			{
				if (payload_limit != INIT_PAYLOAD_SIZE && payload_limit != DATA_PAYLOAD_SIZE
						&& payload_limit != DUMMY_PAYLOAD_SIZE)
					throw new IllegalArgumentException(
							"Payload limit is not an accepted value. Use Fragment.*_PAYLOAD_SIZE instead.");

				this.bytesCached = false;
				this.byteArrayCache = new byte[HEADER_SIZE + payload_limit];

				// TODO: Isn't this only necessary for 275 Byte fragments?
				if (fragment_number == 0 && payload.length <= payload_limit + 1)
					{
						this.m_iMessageID = SINGLE_FRAGMENT_MESSAGE_ID;

						payload_limit += 1;
					}
				else
					{
						this.m_iMessageID = message_id;
					}

				this.m_bisLast = payload.length <= payload_limit;

				this.m_iFragmentNumber = fragment_number;

				this.padding = new Padding(payload_limit - payload.length);

				if (payload.length > payload_limit)
					this.payload = Util.copyOf(payload, payload_limit);
				else
					this.payload = payload;
			}

		public Fragment(byte[] fragment) throws IllegalArgumentException
			{
				this(fragment, SIZE_DATA);
			}

		public Fragment(byte[] fragment, int length) throws IllegalArgumentException
			{
				this.bytesCached = true;
				this.byteArrayCache = new byte[fragment.length];

				System.arraycopy(fragment, 0, this.byteArrayCache, 0, fragment.length);

				int current_offset = 0;

				this.m_iMessageID = fragment[current_offset];
				this.m_iMessageID <<= Constants.BYTE_SIZE;

				++current_offset;

				this.m_iMessageID |= (fragment[current_offset] & 0x00FF);
				this.m_iMessageID >>= 2;

				this.m_bisLast = (fragment[current_offset] & BIT_LAST_FRAGMENT) > 0;

				if (m_iMessageID == 0 && !m_bisLast)
					throw new IllegalArgumentException(
							"Message should only contain one fragment, but the given fragment was not the last.");

				boolean has_padding = (fragment[current_offset] & BIT_HAS_PADDING) > 0;

				++current_offset;

				if (m_iMessageID == 0)
					{
						this.m_iFragmentNumber = 0;
					}
				else
					{
						this.m_iFragmentNumber = fragment[current_offset] & 0xff;

						++current_offset;
					}

				if (has_padding)
					{
						this.padding = new Padding(fragment, current_offset);

						current_offset += this.getPadding_bytes().length;
					}
				else
					{
						this.padding = new Padding(0);
					}

				// we only receive data fragments, so SIZE_DATA is the length of meaningful bytes, the rest
				int payload_length = length - current_offset - this.padding.getLength();

				this.payload = Util.copyOfRange(fragment, current_offset, current_offset + payload_length);
				if(this.payload==null)
					{
						throw new IllegalArgumentException(
								"Something is wrong with this fragment (payload) [ID="+m_iMessageID+"]: Current offset: "+current_offset+ "payload len: "+payload_length);

					}
			}

		int getMessage_id()
			{
				return m_iMessageID;
			}

		int getFragment_number()
			{
				return m_iFragmentNumber;
			}

		boolean isLast()
			{
				return m_bisLast;
			}

		public byte[] getPayload()
			{
				return payload;
			}

		int getPadding_length()
			{
				return this.padding.getLength();
			}

		byte[] getPadding_bytes()
			{
				return this.padding.getLengthAsBytes();
			}

		public byte[] toBytes() throws IOException
			{
				if (!this.bytesCached)
					{
						ByteArrayOutputStream bos = new ByteArrayOutputStream();

						int message_id_and_flags = this.m_iMessageID;

						message_id_and_flags <<= 2;

						if (this.m_bisLast)
							message_id_and_flags |= BIT_LAST_FRAGMENT;

						if (this.getPadding_length() > 0)
							message_id_and_flags |= BIT_HAS_PADDING;

						bos.write(message_id_and_flags >> Constants.BYTE_SIZE & 0xFF);
						bos.write(message_id_and_flags & 0xFF);

						if (this.m_iMessageID != 0)
							bos.write(this.m_iFragmentNumber);

						bos.write(this.getPadding_bytes());

						bos.write(this.payload);

						bos.write(this.padding.getPaddingBytes());

						if (bos.size() != this.byteArrayCache.length)
							{
								throw new IOException("Generated byte array was not the correct size of a fragment.");
							}

						this.byteArrayCache = bos.toByteArray();

						this.bytesCached = true;
					}

				return this.byteArrayCache;
			}
	}
