package anon.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/** Abstract base class for UDP channels. This abstract class implements the IO-Streams  based on
 * readPacket() and writePacket() which need to be implemented by derived classes.
 * @author sk13
 *
 */

public abstract class AbstractAnonUDPChannel implements IAnonUDPChannel
	{
		private AnonUDPChannelOutputStream m_anonUDPOutputStream = null;
		private AnonUDPChannelInputStream m_anonUDPInputStream = null;
		protected String m_strDstHost;
		protected int m_iDstPort;
		protected int m_iSrcPort;

		
		/** Output stream which can be sued to send data to the recipient. Note that the write methods will no
		 * preserve any packet boundaries. The implementation makes use of the writePacket() method.
		 * @author sk13
		 *
		 */
		private class AnonUDPChannelOutputStream extends OutputStream
			{

				AnonUDPChannelOutputStream()
					{
					}

				/**
				 * Sends the bytes in array b to the recipient. No packet boundaries are preserved meaning that
				 * the bytes might be sent in one ore more UDP packets.
				 */
				public void write(byte[] b) throws IOException
					{
						write(b, 0, b.length);
					}

				/**
				 * Send a single byte using to the recipient.
				 */
				public void write(int b) throws IOException
					{
						byte bar[] = new byte[1];
						bar[0] = (byte) (b);
						write(bar, 0, 1);
					}

				/**
				 * Sends bytes in array b to the recipient.
				 */

				public void write(byte[] b, int off, int len) throws IOException
					{
						writePacket(b, off, len);
					}

				public void flush() throws IOException
					{
					}

				public void close() throws IOException
					{
					}

			}

		private class AnonUDPChannelInputStream extends InputStream
			{
				private int m_iRemainingLenUDPPayload;
				private int m_iOffsetUDPPayload;
				private byte[] m_arOneByte = new byte[1];
				private byte[] m_buffUDPPacket = new byte[0xFFFF];

				AnonUDPChannelInputStream()
					{
						m_iRemainingLenUDPPayload = -1;
					}

				public void close() throws IOException
					{
					}

				public synchronized int read() throws IOException
					{
						int ret = read(m_arOneByte, 0, 1);
						if (ret != 1)
							{
								return -1;
							}
						return m_arOneByte[0];
					}

				public synchronized int read(byte[] b) throws IOException
					{
						return read(b, 0, b.length);
					}

				public synchronized int read(byte[] b, int off, int len) throws IOException
					{
						int ret = -1;
						if (m_iRemainingLenUDPPayload <= 0)
							{//read new UDP packet payload
								m_iRemainingLenUDPPayload = readPacket(m_buffUDPPacket, 0, m_buffUDPPacket.length);
								if (m_iRemainingLenUDPPayload < 0)
									return -1; // stream closed
								m_iOffsetUDPPayload = 0;
							}
						ret = Math.min(m_iRemainingLenUDPPayload, len);
						System.arraycopy(m_buffUDPPacket, m_iOffsetUDPPayload, b, off, ret);
						if (ret >= 0)
							{
								m_iRemainingLenUDPPayload -= ret;
								m_iOffsetUDPPayload += ret;
							}
						return ret;
					}
			}

		protected AbstractAnonUDPChannel()
			{
				m_anonUDPInputStream = new AnonUDPChannelInputStream();
				m_anonUDPOutputStream = new AnonUDPChannelOutputStream();
			}

		/***
		 * @param host
		 *          the domain name of the target host
		 * @param port
		 *          the target port number
		 */
		public void setDestination(String host, int port)
			{
				m_strDstHost = host;
				m_iDstPort = port;
			}

		/***
		 * @param port
		 *          the target port number
		 */
		public void setSourcePort(int port)
			{
				m_iSrcPort = port;
			}

		/**
		 * Sets the time to wait before actually sending some data over this channel
		 * in case the packet belonging to the underlying anonymisation protocol is
		 * not full yet.
		 * <p>
		 * This method can be used for performance optimisations / efficiency. One
		 * use case would be if you know that this channel will sent data
		 * immediately after connection establishment. Therefore you would call
		 * <code>setSentAutoFlushDelay</code> with some <code>delayvalue > 0</code>
		 * just before you call the <code>connect</code> method. If you call
		 * <code>Outputstream.write</code> within <code>delayvalue</code>
		 * milliseconds after the <code>connect</code> call then the data you intend
		 * to sent will be piggybacked on the connect request. This will save some
		 * roundtrip-time delay.
		 * <p>
		 * Additionally, using this method correctly could decrease the bandwidth
		 * overhead introduced by the underlying anonymisation protocol. Many
		 * anonymisation protocols fragment the data stream into fixed sized packets
		 * which are filled with dummy data if not enough real data is available.
		 * Therefore waiting a certain amount of time for additional real data can
		 * increase the channel utilisation.
		 *
		 * <p>
		 * Note: This method will not do anything at the moment as the
		 * implementation is not ready yet.
		 *
		 * @param msAutoFlushDelay
		 *          the amount of time in milliseconds to wait for new data before
		 *          the current sent buffer is flushed. Setting this value to
		 *          <code>0</code> will always auto flush immediately. Setting this
		 *          value to <code>-1</code> will only flush the sent buffer, if the
		 *          sent buffer is full, i.e. getOutputBufgferSize many bytes are
		 *          contained in the buffer.
		 */
		public void setSentAutoFlushDelay(int msAutoFlushDelay)
			{

			}

		public void writePacket(byte[] payload) throws IOException
			{
				writePacket(payload, 0, payload.length);
			}

		public InputStream getInputStream()
			{
				return m_anonUDPInputStream;
			}

		public OutputStream getOutputStream()
			{
				return m_anonUDPOutputStream;
			}

	}
