package anon.client;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

import anon.infoservice.AbstractMixCascadeContainer;
import anon.infoservice.Database;
import anon.infoservice.MixCascade;
import anon.pay.PayAccountsFile;
import anon.util.JAPMessages;
import anon.util.Util;

/**
 * This class returns a new random cascade from all currently available cascades every time
 * getNextCascade() is called. If all available cascades have been returned once, this class starts
 * again by choosing the random cascades from all available ones.
 *
 * @author Rolf Wendolsky
 */
public abstract class AbstractAutoSwitchedMixCascadeContainer extends AbstractMixCascadeContainer
{
	public static final MixCascade INITIAL_DUMMY_SERVICE;
	
	private static final long MINIMUM_NEXT_WAITING_TIME = 250; // wait some time until the next cascade is searched to prevent high processor power drain
	
	private long m_lLastNextRequest  = 0l;

	static
	{
		MixCascade initial;
		try
		{
			initial = new MixCascade("-", "-", "0.0.0.0", 6544, System.currentTimeMillis())
			{
				public String getName()
				{
					return JAPMessages.getString("noCascadesAvail");
				}
				
				public Vector getDecomposedCascadeName()
				{
					return Util.toVector(getName());
				}
			};
			initial.showAsTrusted(true);
			initial.setUserDefined(false, null);
		}
		catch (Exception a_e)
		{
			initial = null;
		}
		INITIAL_DUMMY_SERVICE = initial;
	}
	
	private Hashtable m_alreadyTriedCascades;
	private Random m_random;
	private MixCascade m_initialCascade;
	private MixCascade m_currentCascade;
	private boolean m_bKeepCurrentCascade;
	private boolean m_bSkipInitialCascade;
	private String m_strStartupServiceId;
	private boolean m_bInitialRun;
	
	public AbstractAutoSwitchedMixCascadeContainer(boolean a_bSkipInitialCascade, MixCascade a_initialCascade, String a_strStartupServiceId)
	{
		if (a_initialCascade == null)
		{
			if (INITIAL_DUMMY_SERVICE == null)
			{
				throw new NullPointerException("Initial cascade is null!");
			}
			a_initialCascade = INITIAL_DUMMY_SERVICE; 
			a_initialCascade.showAsTrusted(true);
		}
		
		if (a_initialCascade == INITIAL_DUMMY_SERVICE)
		{
			m_bInitialRun = true;
		}
		else
		{
			m_bInitialRun = false;
		}
		m_strStartupServiceId = a_strStartupServiceId;
		m_bSkipInitialCascade = a_bSkipInitialCascade;
		m_alreadyTriedCascades = new Hashtable();
		m_random = new Random(System.currentTimeMillis());
		m_random.nextInt();
		m_initialCascade = a_initialCascade;
		m_currentCascade = m_initialCascade;
		m_bKeepCurrentCascade = false;
	}

	public final MixCascade getInitialCascade()
	{
		return m_initialCascade;
	}

	/**
	 * Clears the list of recently tried cascades, allowing them to be selectable
	 * by {@link #getNextCascade()} again.
	 */
	public void reset()
	{
		m_alreadyTriedCascades.clear();
	}

	/**
	 * Explicitly chooses the next cascade at random, by calling {@link #reset()}
	 * beforehand. If auto-reconnect is active, this method is equal to
	 * {@link #getNextCascade()}.
	 *
	 * @return A random {@link MixCascade} from all available.
	 */
	public final MixCascade getNextRandomCascade()
	{
		synchronized (m_alreadyTriedCascades)
		{
			reset();
			return getNextCascade(true, false);
		}
	}

	/**
	 * Chooses a random Cascade from all available, not already tried ones, until
	 * all have been tried once, or {@link #reset()} is called.
	 */
	public final MixCascade getNextCascade()
	{
		synchronized (m_alreadyTriedCascades)
		{
			try 
			{
				Thread.sleep(Math.max(0, m_lLastNextRequest + MINIMUM_NEXT_WAITING_TIME - System.currentTimeMillis()));
			} 
			catch (InterruptedException e) 
			{
				// ignore
			}
			
			
			m_lLastNextRequest = System.currentTimeMillis();
			
			return getNextCascade(false, false);
		}
	}
	
	private final MixCascade getNextCascade(boolean a_bForceNextRandom, boolean a_bUseSkippedServices)
	{
		synchronized (m_alreadyTriedCascades)
		{
			// if there are no trusted cascades, keep the current one
			if (!TrustModel.getCurrentTrustModel().hasTrustedCascades())
			{
				m_alreadyTriedCascades.put(m_currentCascade.getId(), m_currentCascade);
			}
			else if (!isServiceAutoSwitched() && !a_bForceNextRandom && (m_strStartupServiceId == null ||
					!m_strStartupServiceId.equals(INITIAL_DUMMY_SERVICE.getId())) &&
				(m_currentCascade != INITIAL_DUMMY_SERVICE || 
					isSuitableCascade((MixCascade)Database.getInstance(MixCascade.class).getEntryById(m_strStartupServiceId))))
			{
				/*
				 * if the service is not auto switched, we need to do it manually
				 * if we don't force random, we need to keep the already tested ones in
				 */
				reset();
				m_bKeepCurrentCascade = false;
				
				if (m_initialCascade == INITIAL_DUMMY_SERVICE)
				{
					if (m_strStartupServiceId == null && 
						m_currentCascade != INITIAL_DUMMY_SERVICE)
					{
						m_initialCascade = m_currentCascade; 
					}
					else
					{
						MixCascade startupService = 
							(MixCascade)Database.getInstance(MixCascade.class).getEntryById(m_strStartupServiceId); 
						if (startupService != null)
						{
							m_initialCascade = startupService;
						}
					}
				}
				
				if (m_currentCascade == INITIAL_DUMMY_SERVICE)
				{
					update(m_initialCascade, false);
				}
			}
			else if (m_bKeepCurrentCascade && 
					!a_bForceNextRandom && TrustModel.getCurrentTrustModel().isTrusted(m_currentCascade) &&
					m_currentCascade != INITIAL_DUMMY_SERVICE)
			{
				// do not check if this cascade has been used before
				m_bKeepCurrentCascade = false;
				m_alreadyTriedCascades.put(m_currentCascade.getId(), m_currentCascade);
			}
			else
			{
				MixCascade currentCascade = null;
				Vector availableCascades;
				boolean forward = true;

				availableCascades = Database.getInstance(MixCascade.class).getEntryList();
				if (availableCascades.size() > 0)
				{
					int chosenCascadeIndex = m_random.nextInt();
					if (chosenCascadeIndex < 0)
					{
						// only positive numbers are allowed
						chosenCascadeIndex *= -1;
						// move backward
						forward = false;
					}

					// chose an index from the vector
					chosenCascadeIndex %= availableCascades.size();
					/* Go through all indices until a suitable MixCascade is found or the original index
					 * is reached.
					 */
					int i;
					Vector vecSuitableFound = new Vector();
					for (i = 0; i < availableCascades.size(); i++)
					{
						currentCascade = (MixCascade) availableCascades.elementAt(chosenCascadeIndex);
						// this is the logic that decides whether to use a cascade or not
						if (m_alreadyTriedCascades.containsKey(currentCascade.getId()))
						{
							currentCascade = null;
						}
						else
						{
							if (!isSuitableCascade(currentCascade))
							{
								m_alreadyTriedCascades.put(currentCascade.getId(), currentCascade);
								currentCascade = null;
							}
							else
							{
								// choose one-mix-services as the very last service of all...
								if (!a_bUseSkippedServices &&
									currentCascade.getNumberOfOperators() <= 1)
								{	
									Hashtable hashAvailable = 
										Database.getInstance(MixCascade.class).getEntryHash();
									Enumeration enumTried = m_alreadyTriedCascades.keys();
									MixCascade tmpCascade;
									while (enumTried.hasMoreElements())
									{
										hashAvailable.remove(enumTried.nextElement());
									}
									
									enumTried = hashAvailable.elements();
									while (enumTried.hasMoreElements())
									{
										tmpCascade = (MixCascade)enumTried.nextElement();
										if (tmpCascade.getNumberOfOperators() > 1 && 
											isSuitableCascade(tmpCascade))
										{
											// skip this service...
											currentCascade = null;
											break;
										}
									}
								}
								
								if (currentCascade != null)
								{
									double iProbability;

									/*StatusInfo status = StatusInfo.getStatusInfo(currentCascade);
									if (status != null)
									{
										iProbability = 
											Math.max(0.05, Math.pow(
													(double)(currentCascade.getDistribution() + Math.min(status.getAnonLevel(), 2)), 2.0) / 
													Math.pow((double)(MixCascade.DISTRIBUTION_MAX + 2), 2.0));
									}
									else
									*/
									{
										iProbability = 
											Math.max(0.05, Math.pow((double)currentCascade.getDistribution(), 2.0) / 
													Math.pow((double)MixCascade.DISTRIBUTION_MAX, 2.0));
									}
									
									// found a suitable cascade
									
	//								System.out.println(currentCascade.getName() + ": " + iProbability);
									
									if (m_random.nextDouble() <= iProbability)
									{
//										System.out.println("Won: " + currentCascade.getName());
										break;
									}
									else
									{
										if (!vecSuitableFound.contains(currentCascade))
										{
											vecSuitableFound.addElement(currentCascade);
										}
										currentCascade = null;
									}
								}
							}
						}
						if (forward)
						{
							chosenCascadeIndex = (chosenCascadeIndex + 1) % availableCascades.size();
						}
						else
						{
							chosenCascadeIndex -= 1;
							if (chosenCascadeIndex < 0)
							{
								chosenCascadeIndex = availableCascades.size() - 1;
							}
						}
					}
					if (currentCascade == null)
					{
						if (vecSuitableFound.size() > 0)
						{
							// take the best of the services found
							for (i = 0; i < vecSuitableFound.size(); i++)
							{
								if ((currentCascade == null || 
									currentCascade.getDistribution() < 
									((MixCascade)vecSuitableFound.elementAt(i)).getDistribution()) &&
									isSuitableCascade((MixCascade)vecSuitableFound.elementAt(i)))
								{
									currentCascade = ((MixCascade)vecSuitableFound.elementAt(i));
								}
							}
//							System.out.println("Best: " + currentCascade);
						}
					}
				}

				// Couldn't find any without using already tried cascades
				if (currentCascade == null && !a_bUseSkippedServices)
				{
					return getNextCascade(a_bForceNextRandom, true);
				}

				if (currentCascade == null)
				{
					// no suitable cascade was found
					/*
					if (m_alreadyTriedCascades.size() == 0)
					{
						//
						try 
						{
							Thread.sleep(100);
						} 
						catch (InterruptedException e) 
						{
							LogHolder.log(LogLevel.ERR, LogType.NET, e);
						}
					}
					*/	
					
					reset();
					
					availableCascades = Database.getInstance(MixCascade.class).getEntryList(true);
					for (int i = 0; i < availableCascades.size(); i++)
					{
						// just take the first suitable cascade...
						if (isSuitableCascade((MixCascade)availableCascades.elementAt(i)))
						{
							currentCascade = (MixCascade)availableCascades.elementAt(i);
							LogHolder.log(LogLevel.WARNING, LogType.MISC, 
									"Tried all services, but did not find any suitable... Now using random service: " + 
									currentCascade);
							break;
						}
					}
					
					if (currentCascade == null)
					{
						currentCascade = m_currentCascade;
						
						if (currentCascade != INITIAL_DUMMY_SERVICE)
						{
							LogHolder.log(LogLevel.WARNING, LogType.MISC, 
									"Tried all services, but did not find any suitable... Now re-using current service: " + 
									currentCascade);
						}
					}
				}
				
				m_alreadyTriedCascades.put(currentCascade.getId(), currentCascade);
				update(currentCascade, false);
			}

			if (m_bSkipInitialCascade || m_initialCascade == INITIAL_DUMMY_SERVICE)
			{
				m_initialCascade = m_currentCascade;
			}
			// this only happens for the first call
			m_bSkipInitialCascade = false;
		}

		return m_currentCascade;
	}

	public abstract boolean isServiceAutoSwitched();
	public abstract boolean isReconnectedAutomatically();
	public abstract boolean hasUserAllowedPaidServices(String a_PIID);
	
	public final String getStartupServiceId()
	{
		return m_strStartupServiceId;
	}

	private final boolean isSuitableCascade(MixCascade a_cascade)
	{
		if (a_cascade == null)
		{
			return false;
		}
		
		TrustModel model = TrustModel.getCurrentTrustModel();
		
		if (!model.isPaymentForced()) // Force is stronger!
		{	
			if (a_cascade.isPayment())
			{
					// if no valid account exists, do not allow automatic connections to
					// paid services if the user does not want to connect to them or if an account in "waiting
					// for payment" state exists
				if (PayAccountsFile.getInstance().getChargedAccount(a_cascade.getPIID()) == null &&
						(!hasUserAllowedPaidServices(a_cascade.getPIID()) || 
							PayAccountsFile.getInstance().getAccountWaitingForTransaction(a_cascade.getPIID()) != null))
				{
					// do not connect to payment for new users
					return false;
				}
			}
			else if (!PayAccountsFile.getInstance().isNewUser() && // this user already had or has a premium account
					model.getAttribute(TrustModel.ForcePremiumIfChargedAccountAttribute.class).getTrustCondition() == TrustModel.TRUST_IF_TRUE &&
					hasUserAllowedPaidServices(a_cascade.getPIID()) && 
					PayAccountsFile.getInstance().getAccountWaitingForTransaction(a_cascade.getPIID()) == null &&
					model.hasPremiumCascades())
			{
				// we will switch to a premium service if auto switching is allowed
				return false;
			}
		}

		if (m_initialCascade != null && m_bSkipInitialCascade && a_cascade.equals(m_initialCascade))
		{
			return false;
		}

		return model.isTrusted(a_cascade);
	}
	
	private void update(MixCascade a_cascade, boolean a_bEnforceInitialRun)
	{
		m_bInitialRun = false;
		
		synchronized (m_alreadyTriedCascades)
		{
			m_bKeepCurrentCascade = true;
			if (m_currentCascade == null && a_cascade != null ||
				m_currentCascade != null && a_cascade == null ||
				!m_currentCascade.equals(a_cascade) ||
				!getTrustModel().isTrusted(m_currentCascade)) // this may happen if all cascades of a filter are blacklisted
			{
				setChanged();
			}
			m_currentCascade = a_cascade;
			if (m_initialCascade == null)
			{
				m_initialCascade = m_currentCascade;
			}
			
			if (a_bEnforceInitialRun)
			{
				m_bInitialRun = true;
			}
			
			notifyObservers(m_currentCascade);
			/*
			if (hasChanged())
			{
				Thread tNotify = new Thread(new Runnable()
				{
					public void run()
					{
						synchronized(this)
						{
							setChanged();
							notifyObservers(m_currentCascade);
						}
					}
				});
				tNotify.setDaemon(true);
				tNotify.start();
			}*/
		}
	}
	
	public final MixCascade getCurrentCascade()
	{
		return m_currentCascade;
	}
	
	public final boolean setCurrentCascade(MixCascade a_cascade)
	{
		if (!getTrustModel().isTrusted(a_cascade))
		{
			return false;
		}
		
		update(a_cascade, true);
		
		return true;
	}

	public final void keepCurrentService(boolean a_bKeepCurrentCascade)
	{
		synchronized (m_alreadyTriedCascades)
		{
			m_bKeepCurrentCascade = a_bKeepCurrentCascade;
		}
	}

	public final ITrustModel getTrustModel()
	{
		return TrustModel.getCurrentTrustModel();
	}
}
