/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.IOException;

import anon.ErrorCodes;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;
import anon.IServiceContainer;

/**
 * The base class for control channels. These are used for messages between the
 * infrastructure components (mixes, JAP).
 * <p>
 * Separate channels are necessary, so that these messages are distinct from
 * user traffic and don't have to compete for bandwidth.
 * </p>
 *
 * @author Stefan Lieske
 */
public abstract class AbstractControlChannel extends AbstractChannel
{
	private IServiceContainer m_serviceContainer;
	private boolean m_bSendingPacket = false;


	/**
	 * Sets up the control channel base and registers it with the given
	 * multiplexer.
	 *
	 * @param a_channelId
	 * @param a_parentMultiplexer
	 * @param a_serviceContainer
	 */
	public AbstractControlChannel(int a_channelId, Multiplexer a_parentMultiplexer,
								  IServiceContainer a_serviceContainer)
	{
		super(a_channelId, a_parentMultiplexer);
		m_serviceContainer = a_serviceContainer;
		if (m_serviceContainer  == null)
		{
			m_serviceContainer = new IServiceContainer()
			{
				public void keepCurrentService(boolean a_bKeepCurrentService)
				{
				}

				public boolean isServiceAutoSwitched()
				{
					return false;
				}

				public boolean isReconnectedAutomatically()
				{
					return false;
				}

				public ITrustModel getTrustModel()
				{
					return new BasicTrustModel();
				}

				public void reset()
				{

				}
			};
		}
		/* register the channel */
		a_parentMultiplexer.getChannelTable().registerControlChannel(a_channelId, this);
	}

	public boolean isSending()
	{
		return m_bSendingPacket;
	}

	/**
	 * Sends the supplied message in as many MixPackets as necessary.
	 * @param a_message
	 * @return
	 */

	public int sendRawMessage(byte[] a_message)
	{
		try
		{
			int bytesToSend = a_message.length;
			m_bSendingPacket = true;

			do
			{
				/* we also send empty packages (maybe this is necessary for interaction) */
				MixPacket currentMixPacket = createEmptyMixPacket();

				int currentPacketLength = Math.min(bytesToSend, MixPacket.getPayloadSize());

				currentMixPacket.setChannelFlags( (short) currentPacketLength);

				System.arraycopy(a_message, a_message.length - bytesToSend, currentMixPacket.getPayloadData(),
								 0, currentPacketLength);

				sendPacket(currentMixPacket);

				bytesToSend = bytesToSend - currentPacketLength;
			}
			while (bytesToSend > 0 && !Thread.currentThread().isInterrupted());
			return ErrorCodes.E_SUCCESS;
		}
		catch (IOException e)
		{
			return ErrorCodes.E_UNKNOWN;
		}
		finally
		{
			m_bSendingPacket = false;
		}
	}

	public void processReceivedPacket(MixPacket a_mixPacket)
	{
		int packetDataLength = a_mixPacket.getChannelFlags();
		if ( (packetDataLength > MixPacket.getPayloadSize()) || (packetDataLength < 0))
		{
			/* something is wrong here -> ignore the packet */
			LogHolder.log(LogLevel.ERR, LogType.NET,
						  "AbstractControlChannel: processReceivedPacket(): Invalid packet length.");
		}
		else
		{
			byte[] packetData = new byte[packetDataLength];
			System.arraycopy(a_mixPacket.getPayloadData(), 0, packetData, 0, packetDataLength);
			processPacketData(packetData);
		}
	}

	protected final IServiceContainer getServiceContainer()
	{
		return m_serviceContainer;
	}

	protected abstract void processPacketData(byte[] a_packetData);

}
