/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;

import anon.AnonChannel;
import anon.TooMuchDataForPacketException;

/**
 * @author Stefan Lieske
 */
public abstract class AbstractDataChain implements AnonChannel, Observer, Runnable
	{

		private DataChainInputStreamImplementation m_inputStream;

		private DataChainOutputStreamImplementation m_outputStream;

		private Vector m_messageQueuesNotifications;

		private IDataChannelCreator m_channelCreator;

		private DataChainErrorListener m_errorListener;

		private boolean m_chainClosed;

		private Thread m_downstreamThread;

		private IntegrityErrorListener m_integrityErrorListener;

		private class DataChainOutputStreamImplementation extends OutputStream
			{

				private boolean m_closed;

				private Object m_internalStreamSynchronization;

				public DataChainOutputStreamImplementation()
					{
						m_closed = false;
						m_internalStreamSynchronization = new Object();
					}

				public void write(int a_byte) throws IOException
					{
						byte[] byteAsArray = { (byte) a_byte };
						write(byteAsArray);
					}

				public void write(byte[] a_buffer, int a_offset, int a_length) throws IOException
					{
						synchronized (m_internalStreamSynchronization)
							{
								if (m_closed)
									{
										throw (new IOException("Stream is closed."));
									}
								byte[] dataToSend = new byte[a_length];

								// copy bytes into dataToSend
								System.arraycopy(a_buffer, a_offset, dataToSend, 0, a_length);

								DataChainSendOrderStructure order = new DataChainSendOrderStructure(dataToSend);
								synchronized (order.getSynchronizationObject())
									{
										orderPacket(order);
										/*
										 * check whether we have a single-threaded implementation
										 * (processing of the order is already done with the call of
										 * orderPacket()) or multi-threaded implementation (we have
										 * to wait until the processing-thread has processed the
										 * order)
										 */
										if (!order.isProcessingDone())
											{
												/* multi-threaded implementation */
												try
													{
														order.getSynchronizationObject().wait();
													}
												catch (InterruptedException e)
													{
														/*
														 * Attention: We will not set the number of
														 * transferred bytes in the thrown Exception because
														 * we can't know it.
														 */
														throw (new InterruptedIOException("InterruptedException: " + e.toString()));
													}
											}
										/* processing is done -> check for exceptions */
										if (order.getThrownException() != null)
											{
												/* throw that exception */
												throw (order.getThrownException());
											}
										/* check whether all bytes could have been sent */
										if (order.getProcessedBytes() < a_length)
											{
												/* throw an exception with the number of processed bytes */
												throw (new TooMuchDataForPacketException(order.getProcessedBytes()));
											}
									}
							}
					}

				public void close() throws IOException
					{
						if (!m_closed)
							{
								synchronized (m_internalStreamSynchronization)
									{
										m_closed = true;
										outputStreamClosed();
									}
							}
					}
			}

		public AbstractDataChain(IDataChannelCreator a_channelCreator, DataChainErrorListener a_errorListener,
				IntegrityErrorListener a_integrityErrorListener)
			{
				m_channelCreator = a_channelCreator;
				m_errorListener = a_errorListener;
				m_integrityErrorListener = a_integrityErrorListener;
				m_inputStream = new DataChainInputStreamImplementation();
				m_outputStream = new DataChainOutputStreamImplementation();
				m_messageQueuesNotifications = new Vector();
				m_chainClosed = false;
				m_downstreamThread = new Thread(this, "AbstractDataChain: Downstream-Organizer Thread");
				m_downstreamThread.setDaemon(true);
				m_downstreamThread.start();
			}

		public InputStream getInputStream()
			{
				return m_inputStream;
			}

		public OutputStream getOutputStream()
			{
				return m_outputStream;
			}

		public boolean isClosed()
			{
				return m_chainClosed;
			}

		public void close()
			{
				if (!m_chainClosed)
					{
						/*
						 * First: Prevent loops which may occur when closing the
						 * output-stream (calls of outputStreamClosed() could re-call this
						 * method).
						 */
						m_chainClosed = true;
						try
							{
								getOutputStream().close();
							}
						catch (IOException e)
							{
							}
						try
							{
								getInputStream().close();
							}
						catch (IOException e)
							{
							}
						closeDataChain();
						/*
						 * wait for stop of the downstream thread after all channels are
						 * closed
						 */
						try
							{
								m_downstreamThread.join();
							}
						catch (InterruptedException e)
							{
							}
					}
			}

		public void update(Observable a_observedObject, Object a_message)
			{
				if (a_observedObject instanceof InternalChannelMessageQueue)
					{
						synchronized (m_messageQueuesNotifications)
							{
								m_messageQueuesNotifications.addElement(a_observedObject);
								m_messageQueuesNotifications.notify();
							}
					}
			}

		protected Vector getMessageQueuesNotificationsList()
			{
				return m_messageQueuesNotifications;
			}

		protected void addInputStreamQueueEntry(DataChainInputStreamQueueEntry a_entry)
			{
				m_inputStream.addToQueue(a_entry);
			}

		protected AbstractDataChannel createDataChannel()
			{
				return m_channelCreator.createDataChannel(this);
			}

		protected void interruptDownstreamThread()
			{
				m_downstreamThread.interrupt();
			}

		protected void propagateConnectionError()
			{
				m_errorListener.dataChainErrorSignaled(null);
			}

		protected void propagateIntegrityError(int a_iErrorCode)
			{
				m_integrityErrorListener.integrityErrorSignaled(a_iErrorCode);
			}

		public abstract int getOutputBlockSize();

		/**
		 * Creates the packet payload from the data stored in the given order.
		 * Additional information may be added depending on the type of data
		 * chain that is implementing this method.
		 * <p>
		 * The processed data will be stored in the channel cell of the order.
		 *</p>
		 *
		 * @param a_order
		 */

		public abstract void createPacketPayload(DataChainSendOrderStructure a_order);

		public abstract void run();

		protected abstract void orderPacket(DataChainSendOrderStructure a_order);

		protected abstract void outputStreamClosed() throws IOException;

		protected abstract void closeDataChain();

	}
