/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.IOException;
import java.security.SecureRandom;
import anon.client.crypto.MixCipherChain;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

/**
 * This class is the abstract base class for channels sending user data to other
 * mixes.
 *
 * @author Stefan Lieske
 */
public abstract class AbstractDataChannel extends AbstractChannel
	{

		protected final MixCipherChain m_mixCipherChain;

		private AbstractDataChain m_parentDataChain;

		protected final InternalChannelMessageQueue m_channelMessageQueue;

		private boolean m_bWithIntegrityCheck;
		private boolean m_bDebug = false;

		private final static SecureRandom ms_secureRandom = new SecureRandom();

		public AbstractDataChannel(int a_channelId, Multiplexer a_parentMultiplexer, AbstractDataChain a_parentDataChain,
				MixCipherChain a_mixCipherChain)
			{
				super(a_channelId, a_parentMultiplexer);
				m_parentDataChain = a_parentDataChain;
				m_mixCipherChain = a_mixCipherChain;
				m_channelMessageQueue = new InternalChannelMessageQueue();
				m_bWithIntegrityCheck = a_parentMultiplexer.isProtocolWithIntegrityCheck();
				m_bDebug = a_parentMultiplexer.isDebug();
			}

		public InternalChannelMessageQueue getChannelMessageQueue()
			{
				return m_channelMessageQueue;
			}

		public AbstractDataChain getDataChain()
			{
				return m_parentDataChain;
			}

		public MixCipherChain getMixCipherChain()
			{
				return m_mixCipherChain;
			}

		public void processReceivedPacket(MixPacket a_mixPacket)
			{

				/* catch all dummies (shouldn't be used any more in the future) */
				if ((a_mixPacket.getChannelFlags() & MixPacket.FLAG_CHANNEL_DUMMY) == MixPacket.FLAG_CHANNEL_DUMMY)
					{
						/*
						 * we haven't ordered any dummies -> normally we shouldn't
						 * receive any
						 */
						LogHolder.log(LogLevel.INFO, LogType.NET,
								"AbstractDataChannel: processReceivedPacket(): Catched an unexpected dummy-paket on channel '"
										+ Integer.toString(a_mixPacket.getChannelId()) + "'.");
					}
				else
					{
						/* further handling is done by the channel-implementation */
						handleReceivedPacket(a_mixPacket);
					}
			}

		public int getNextPacketRecommandedOutputBlocksize()
			{
				int recommandedSize = 0;
				synchronized (m_mixCipherChain)
					{
						recommandedSize = MixPacket.getPayloadSize() - m_mixCipherChain.getNextPacketEncryptionOverhead();
					}
				return recommandedSize;
			}

		/**
		 * Creates an empty MixPacket, sets the channel flags and encrypts the
		 * data inside the data in the SendOrder for the mix chain in this
		 * channel.
		 *
		 * @param a_order
		 * @param a_channelFlags
		 */

		protected void createAndSendMixPacket(DataChainSendOrderStructure a_order, short a_channelFlags)
			{
				MixPacket packet = createEmptyMixPacket();
				if (m_bDebug)
					{
						a_channelFlags |= MixPacket.FLAG_DEBUG;
					}
				packet.setChannelFlags(a_channelFlags);
				synchronized (m_mixCipherChain)
					{
						// create a buffer with the maximum length minus the
						// encryption overhead
						byte[] buff = new byte[MixPacket.getPayloadSize() - m_mixCipherChain.getNextPacketEncryptionOverhead()];

						// fill the channel cell with random bytes
						ms_secureRandom.nextBytes(buff);
						a_order.setChannelCell(buff);

						if (a_order.getOrderData() != null)
							{
								m_parentDataChain.createPacketPayload(a_order);
							}
						else //HACK: set the payloadsize to '0', because we need it for placing the MAC in case of the integrity protection
							{
							/*
							 * messages without order-data are channel-internal and needs no
							 * interaction with the parent chain
							 */
								if (m_bWithIntegrityCheck)
									{
										a_order.getChannelCell()[0] = 0;
										a_order.getChannelCell()[1] = 0;
									}
							}
						/*if (m_bWithIntegrityCheck)
							{
								byte[] extendedChannelCell = new byte[a_order.getChannelCell().length + DataChannelCipher.getMACSize()];
								System.arraycopy(a_order.getChannelCell(), 0, extendedChannelCell, 0, a_order.getChannelCell().length);
								a_order.setChannelCell(extendedChannelCell);
							}*/

//						LogHolder.log(LogLevel.INFO, LogType.NET, "Packet to send: (" + a_order.getChannelCell().length + ") " + Arrays.toString(a_order.getChannelCell()));

						// encrypt the payload for sending through the mix chain
						byte[] encryptedPacketPayload = m_mixCipherChain.encryptPacket(a_order.getChannelCell(),
								MixPacket.getPayloadSize(), packet.getSendCallbackHandlers());

						System.arraycopy(encryptedPacketPayload, 0, packet.getPayloadData(), 0,
								encryptedPacketPayload.length/*packet.getPayloadData().length*/);
						try
							{
								sendPacket(packet);
							}
						catch (IOException e)
							{
								a_order.setThrownException(e);
							}
					}
				a_order.processingDone();
			}

		public abstract boolean processSendOrder(DataChainSendOrderStructure a_order);

		public abstract void organizeChannelClose() throws IOException;

		protected abstract void handleReceivedPacket(MixPacket a_mixPacket);

	}
