/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.IOException;
import java.net.ConnectException;
import java.net.Socket;
import java.security.SecureRandom;
import java.util.Enumeration;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;
import HTTPClient.HTTPClientSocketFactory;
import HTTPClient.IHTTPClientDNSResolver;
import HTTPClient.HTTPConnection;
import HTTPClient.ThreadInterruptedIOException;
import anon.AnonChannel;
import anon.AnonLib;
import anon.AnonServerDescription;
import anon.AnonService;
import anon.AnonServiceEventListener;
import anon.ErrorCodes;
import anon.IServiceContainer;
import anon.NotConnectedToMixException;
import anon.anonudp.UDPMultiplexer;
import anon.anonudp.client.AnonPureUDPChannel;
import anon.client.replay.ReplayControlChannel;
import anon.client.replay.TimestampUpdater;
import anon.error.AlreadyConnectedException;
import anon.error.AnonServiceException;
import anon.error.ConnectionEstablishmentTimeoutException;
import anon.error.IntegrityCheckException;
import anon.error.InvalidServiceException;
import anon.error.ParseServiceException;
import anon.error.ServiceInterruptedException;
import anon.infoservice.Database;
import anon.infoservice.HTTPConnectionFactory;
import anon.infoservice.IMutableProxyInterface;
import anon.infoservice.ImmutableProxyInterface;
import anon.infoservice.ListenerInterface;
import anon.infoservice.MixCascade;
import anon.infoservice.RandomListenerInterfaceSwitcher;
import anon.pay.AIControlChannel;
import anon.pay.IAIEventListener;
import anon.pay.PayAccount;
import anon.proxy.DirectProxy;
import anon.terms.TermsAndConditionConfirmation;
import anon.terms.TermsAndConditionsReadException;
import anon.transport.connection.ConnectionException;
import anon.transport.connection.IStreamConnection;
import anon.transport.connection.SocketConnection;
import anon.util.JobQueue;
import anon.util.XMLParseException;

/**
 * This class implements an {@link AnonService} using mix cascades to ensure
 * anonymous communication.
 *
 * @author Stefan Lieske
 */
public class AnonClient implements AnonService, Observer, DataChainErrorListener, IntegrityErrorListener
	{
		private static boolean ENABLE_CONTROL_CHANNEL_TEST = false;

		private static boolean ms_bIsTestInstance = false;
		private static boolean ms_bIsTestInstanceSet = false;

		public static final int DEFAULT_LOGIN_TIMEOUT = 30000;
		private static final int FAST_LOGIN_TIMEOUT = 4000; // try the fast timeout
																												// first
		private static final int CONNECT_TIMEOUT = 8000;

		public static final long DEFAULT_BLOCK_TIMEOUT = 180000l;
		private long m_msBlockTimeout = DEFAULT_BLOCK_TIMEOUT;

		private static int m_loginTimeout = DEFAULT_LOGIN_TIMEOUT;
		private static int m_loginTimeoutFastAvailable;

		private static boolean ms_bBlockOnHttpError;

		private static int ms_preferredConnectionPort = 0;
		private static Vector ms_vecTimedOutPorts = new Vector();

		private Multiplexer m_multiplexer; //handels the normal protocol (TCP based)
		private UDPMultiplexer m_udpMultiplexer; //handels the UDP over UDP protocol

		private static JobQueue ms_queuePacketCount = new JobQueue("JobQueue: AnonClient Packet count updater");

		private IMutableProxyInterface m_proxyInterface;

		private Object m_internalSynchronization;

		private IServiceContainer m_serviceContainer;

		private Thread m_threadInitialise;
		private Object SYNC_SHUTDOWN = new Object();

		private Object m_internalSynchronizationForSocket;

		private Object m_internalSynchronizationForDummyTraffic;

		private SocketHandler m_socketHandler;

		private Vector m_eventListeners;

		private PacketCounter m_packetCounter;

		private DummyTrafficControlChannel m_dummyTrafficControlChannel;

		private int m_dummyTrafficInterval = DummyTrafficControlChannel.DT_MAX_INTERVAL_MS;

		private KeyExchangeManager m_keyExchangeManager;

		private IStreamConnection m_streamConnection;

		private boolean m_connected;

		private MixCascade m_currentService;

		private DirectProxy m_directProxy;

		private HTTPClientSocketFactory m_socketFactory=null;

		private IHTTPClientDNSResolver m_dnsResolver=null;

		private boolean m_bDebug=false;
		static
			{
				resetInternalLoginTimeout();
			}

		public AnonClient()
			{
				this(null);
			}

		public AnonClient(DirectProxy a_directProxy)
			{
				m_directProxy = a_directProxy;
				m_socketHandler = null;
				m_multiplexer = null;
				m_packetCounter = null;
				m_dummyTrafficControlChannel = null;
				m_dummyTrafficInterval = -1;
				m_keyExchangeManager = null;
				m_streamConnection = null;
				m_internalSynchronization = new Object();
				m_internalSynchronizationForSocket = new Object();
				m_internalSynchronizationForDummyTraffic = new Object();
				m_eventListeners = new Vector();
				m_connected = false;
				if(ms_queuePacketCount==null)
					ms_queuePacketCount = new JobQueue("JobQueue: AnonClient Packet count updater");
				m_proxyInterface = new IMutableProxyInterface.DummyMutableProxyInterface();
				m_bDebug=false;
			}

		public AnonClient(DirectProxy a_directProxy, IStreamConnection a_theConnection, MixCascade a_currentService)
			{
				this(a_directProxy);
				m_streamConnection = a_theConnection;
				m_currentService = a_currentService;
				TrustModel.cleanAttributeWhitelist(m_currentService);
			}

		/** if bDebug==true this client operates in debug modes. Which means for instance, that all keys are set to zero!*/
		public void setDebug(boolean bDebug)
		{
			m_bDebug=bDebug;
		}

		public static void setTestInstance(boolean a_bTestInstance)
			{
				if (!ms_bIsTestInstanceSet)
					{
						ms_bIsTestInstance = a_bTestInstance;
						ms_bIsTestInstanceSet = true;
					}
			}

		public static boolean isTestInstance()
			{
				return ms_bIsTestInstance;
			}

		private static interface StatusThread extends Runnable
			{
				public AnonServiceException getStatus();
			}

		public MixCascade getCurrentService()
			{
				return m_currentService;
			}

		// the synchronized should prevent two methods initializing at the same time
		public synchronized void initialize(final AnonServerDescription a_mixCascade,
				final IServiceContainer a_serviceContainer, final TermsAndConditionConfirmation termsConfirmation,
				final boolean a_bIsReconnect) throws AnonServiceException
			{
				if (!(a_mixCascade instanceof MixCascade))
					{
						throw new InvalidServiceException(a_mixCascade);
					}

				if (isConnected())
					{
						throw new AlreadyConnectedException(a_mixCascade);
					}

				synchronized (m_internalSynchronization)
					{
						m_currentService = (MixCascade) a_mixCascade;
						TrustModel.cleanAttributeWhitelist(m_currentService);
						m_serviceContainer = a_serviceContainer;
					}

				StatusThread run = new StatusThread()
					{
						AnonServiceException status = null;

						public AnonServiceException getStatus()
							{
								return status;
							}

						public void run()
							{
								synchronized (m_internalSynchronization)
									{
										if (isConnected())
											{
												LogHolder.log(LogLevel.ERR, LogType.NET, "AnonClient was already connected when connecting!");
												/*
												 * shutdown(); if (isConnected()) {
												 */
												status = new AlreadyConnectedException(a_mixCascade);
												synchronized (m_threadInitialise)
													{
														m_threadInitialise.notifyAll();
													}
												return;
											}

										IStreamConnection connectionToMixCascade = null;
										if (m_streamConnection != null)
											{
												connectionToMixCascade = m_streamConnection;
												m_streamConnection = null;
											}
										else
											{
												try
													{
														a_serviceContainer.getTrustModel().checkTrust((MixCascade) a_mixCascade);

														connectionToMixCascade = connectMixCascade((MixCascade) a_mixCascade, m_proxyInterface
																.getProxyInterface(false).getProxyInterface(), a_serviceContainer, a_bIsReconnect);
													}
												catch (InterruptedException a_e)
													{
														status = new ServiceInterruptedException(a_mixCascade);
														synchronized (m_threadInitialise)
															{
																m_threadInitialise.notifyAll();
															}
														return;
													}
												catch (AnonServiceException a_e)
													{
														status = a_e;
														synchronized (m_threadInitialise)
															{
																m_threadInitialise.notifyAll();
															}
														return;
													}
											}
										if (connectionToMixCascade == null)
											{
												status = new AnonServiceException(a_mixCascade, "Could not connect to service " + a_mixCascade
														+ "  for an unknown reason.", ErrorCodes.E_CONNECT);
												synchronized (m_threadInitialise)
													{
														m_threadInitialise.notifyAll();
													}
												return;
											}
										try
											{
												initializeProtocol(connectionToMixCascade, (MixCascade) a_mixCascade, a_serviceContainer,	termsConfirmation);
											}
										catch (AnonServiceException a_e)
											{
												status = a_e;
											}
										synchronized (m_threadInitialise)
											{
												m_threadInitialise.notifyAll();
											}
									}
							}
					};
				synchronized (SYNC_SHUTDOWN)
					{
						m_threadInitialise = new Thread(run);
					}
				m_threadInitialise.start();
				try
					{
						m_threadInitialise.join();
					}
				catch (InterruptedException ex)
					{
						synchronized (m_threadInitialise)
							{
								while (m_threadInitialise.isAlive())
									{
										m_threadInitialise.interrupt();
										try
											{
												m_threadInitialise.wait(500);
											}
										catch (InterruptedException ex1)
											{
											}
									}
							}
						throw new ServiceInterruptedException(a_mixCascade);
					}

				if (run.getStatus() != null)
					{
						throw run.getStatus();
					}
			}

		public static void setLoginTimeout(int a_loginTimeoutMS)
			{
				if (a_loginTimeoutMS >= 1000)
					{
						m_loginTimeout = a_loginTimeoutMS;
					}
			}

		public static void setBlockOnHttpConnectionError(boolean a_bBlock)
			{
				ms_bBlockOnHttpError = a_bBlock;
			}

		public static boolean isBlockedOnHttpConnectionError()
			{
				return ms_bBlockOnHttpError;
			}

		private static void resetInternalLoginTimeout()
			{
				int maxLoginTimeoutFastAvailable = DEFAULT_LOGIN_TIMEOUT / 1000;

				m_loginTimeoutFastAvailable = Math.max(m_loginTimeout / 1000, m_loginTimeout / FAST_LOGIN_TIMEOUT);

				if (m_loginTimeoutFastAvailable > maxLoginTimeoutFastAvailable)
					{
						m_loginTimeoutFastAvailable = maxLoginTimeoutFastAvailable;
					}
			}

		private static int getInternalLoginTimeout(IServiceContainer a_serviceContainer)
			{
				if (a_serviceContainer != null && m_loginTimeoutFastAvailable > 0
						&& a_serviceContainer.isReconnectedAutomatically() && a_serviceContainer.isServiceAutoSwitched())
					{
						m_loginTimeoutFastAvailable--;
						return FAST_LOGIN_TIMEOUT;
					}

				return m_loginTimeout;
			}

		public static int getLoginTimeout()
			{
				return m_loginTimeout;
			}

		public int setProxy(IMutableProxyInterface a_proxyInterface)
			{
				synchronized (m_internalSynchronization)
					{
						if (a_proxyInterface == null)
							{
								m_proxyInterface = new IMutableProxyInterface.DummyMutableProxyInterface();
							}
						else
							{
								m_proxyInterface = a_proxyInterface;
							}
					}
				return ErrorCodes.E_SUCCESS;
			}

		public void shutdown(boolean a_bResetTransferredBytes)
			{
				synchronized (m_internalSynchronizationForSocket)
					{
						if (m_socketHandler != null)
							{
								m_socketHandler.deleteObservers();
							}
					}

				synchronized (m_internalSynchronization)
					{
						if (m_multiplexer != null)
							{
								m_multiplexer.close();
							}
					}

				synchronized (m_internalSynchronization)
				{
					if (m_udpMultiplexer != null)
						{
							m_udpMultiplexer.stop();
							m_udpMultiplexer=null;
						}
				}

				synchronized (m_internalSynchronizationForSocket)
					{
						if (m_socketHandler != null)
							{
								m_socketHandler.closeSocket();
								m_socketHandler = null;
							}
					}
				synchronized (SYNC_SHUTDOWN)
					{
						if (m_threadInitialise != null)
							{
								synchronized (m_threadInitialise)
									{
										while (m_threadInitialise.isAlive())
											{
												m_threadInitialise.interrupt();
												try
													{
														m_threadInitialise.wait(100);
													}
												catch (InterruptedException e)
													{
														break;
													}
											}
									}
							}
					}
				synchronized (m_internalSynchronization)
					{
						if (m_multiplexer != null)
							{
								m_multiplexer.deleteObservers();
							}
						m_multiplexer = null;
						m_connected = false;

						synchronized (m_internalSynchronizationForDummyTraffic)
							{
								if (m_dummyTrafficControlChannel != null)
									{
										m_dummyTrafficControlChannel.stop();
										m_dummyTrafficControlChannel = null;
									}
							}

						if (m_packetCounter != null)
							{
								m_packetCounter.deleteObserver(this);
								if (a_bResetTransferredBytes)
									{
										m_packetCounter = null;
									}
							}
						if (m_keyExchangeManager != null)
							{
								m_keyExchangeManager.removeCertificateLock();
								m_keyExchangeManager = null;
							}
						/*
						 * synchronized (m_eventListeners) { final Enumeration
						 * eventListenersList = m_eventListeners.elements(); Thread
						 * notificationThread = new Thread(new Runnable() { public void
						 * run() { while (eventListenersList.hasMoreElements()) { (
						 * (AnonServiceEventListener)
						 * (eventListenersList.nextElement())).packetMixed(0); } } },
						 * "AnonClient: Zero PacketMixed notification (after shutdown)");
						 * notificationThread.setDaemon(true); notificationThread.start(); }
						 */
						if (m_directProxy != null)
							{
								m_directProxy.start();
							}
					}
			}

		public boolean isConnected()
			{
				return m_connected;
			}

		public boolean isSendingControlMessage()
			{
				Multiplexer multiplexer = m_multiplexer;

				if (multiplexer == null)
					{
						return false;
					}

				return multiplexer.isSendingControlMessage();
			}

		public AnonChannel createChannel(int a_type) throws ConnectException
			{
				Multiplexer multiplexer = null;
				KeyExchangeManager keyExchangeManager = null;
				synchronized (m_internalSynchronization)
					{
						if (m_multiplexer == null)
							{
								/* not connected */
								throw (new NotConnectedToMixException(
										"AnonClient: createChannel(): The AN.ON client is currently not connected to a mixcascade."));
							}
						/*
						 * we don't want to block everything, if chain-creation is waiting
						 * for free resources -> only get the pointers to the necessary
						 * objects within the synchronized part
						 */
						multiplexer = m_multiplexer;
						keyExchangeManager = m_keyExchangeManager;
					}
				if(a_type==AnonChannel.TCP)
					{
						AnonChannel channel=createChannel(AnonChannel.SOCKS);
						return new AnonTCPChannel(channel);
					}
				if(a_type==AnonChannel.UDP_OVER_TCP)
					{
						
						AnonChannel channel=createChannel(AnonChannel.VPN);
						return new AnonUDPChannel(channel);
					}
				if(a_type==AnonChannel.UDP_OVER_UDP)
					{
						if(m_udpMultiplexer!=null)			
							return new AnonPureUDPChannel(m_udpMultiplexer);
						return null;
					}
				if(a_type==AnonChannel.UDP)
					{
						if(m_udpMultiplexer!=null)			
							{
								return new AnonPureUDPChannel(m_udpMultiplexer);					
							}
						AnonChannel channel=createChannel(AnonChannel.VPN);
						return new AnonUDPChannel(channel);
					}

				FixedRatioChannelsDescription channelsDescription = keyExchangeManager.getFixedRatioChannelsDescription();
				if (channelsDescription == null)
					{
						/* old protocol with unlimited channels */
						return (new SingleChannelDataChain(multiplexer.getChannelTable(), this, this, a_type,
								keyExchangeManager.isChainProtocolWithFlowControl(),
								keyExchangeManager.isChainProtocolWithUpstreamFlowControl(), keyExchangeManager.getUpstreamSendMe(),
								keyExchangeManager.getDownstreamSendMe(), keyExchangeManager.isProtocolWithEnhancedChannelEncryption(),
								keyExchangeManager.isProtocolWithIntegrityCheck()));

					}
				return (new TypeFilterDataChain(new SequentialChannelDataChain(multiplexer.getChannelTable(), this, this,
						channelsDescription.getChainTimeout()), a_type));
			}

		public void addEventListener(AnonServiceEventListener a_eventListener)
			{
				synchronized (m_eventListeners)
					{
						m_eventListeners.addElement(a_eventListener);
					}

			}

		public void removeEventListeners()
			{
				m_eventListeners.removeAllElements();
			}

		public void removeEventListener(AnonServiceEventListener a_eventListener)
			{
				synchronized (m_eventListeners)
					{
						m_eventListeners.removeElement(a_eventListener);
					}
			}

		private void reconnect(final MixCascade a_cascade, final IOException a_argument)
			{
				/* shutdown everything */
				Thread threadShutdown= new Thread(new Runnable()
					{
						/**
						 * @todo We use a thread to avoid deadlocks - is this really the best
						 *       solution?
						 */
						public void run()
							{
								shutdown(!m_serviceContainer.isReconnectedAutomatically());
								synchronized (m_eventListeners)
									{
										final Enumeration eventListenersList = m_eventListeners.elements();
										Thread notificationThread = new Thread(new Runnable()
											{
												public void run()
													{
														if (a_argument != null)
															{
																LogHolder.log(LogLevel.WARNING, LogType.NET, a_argument);
															}
														while (eventListenersList.hasMoreElements())
															{
																((AnonServiceEventListener) (eventListenersList.nextElement()))
																		.connectionError(new AnonServiceException(a_cascade, "Reconnect..."));
															}
													}
											}, "ConnectionError notification");
										notificationThread.setDaemon(true);
										notificationThread.start();
									}
							}
					},"AN.ONClient - MixCascadeConnection Shutdown thread");
				threadShutdown.setDaemon(true);
				threadShutdown.start();
			}

		public void update(Observable a_object, final Object a_argument)
			{
				if ((a_object == m_socketHandler) && (a_argument instanceof IOException))
					{
						reconnect(getCurrentService(), (IOException) a_argument);
					}
				else if (a_object == m_packetCounter)
					{
						JobQueue.Job notificationThread = new JobQueue.Job(true)
							{
								public void runJob()
									{
										Vector eventListeners = m_eventListeners;
										PacketCounter packetCounter = m_packetCounter;
										if (eventListeners != null && packetCounter != null)
											{
												synchronized (eventListeners)
													{
														Enumeration eventListenersList = eventListeners.elements();
														while (eventListenersList.hasMoreElements())
															{
																((AnonServiceEventListener) (eventListenersList.nextElement()))
																		.packetMixed(packetCounter.getProcessedPackets()
																				* (long) (MixPacket.getPacketSize()));
															}
													}
											}
									}
							};
						ms_queuePacketCount.addJob(notificationThread);
					}
			}

		/**
		 * @todo signals proxy connection error without recovery
		 */
		public void dataChainErrorSignaled(final AnonServiceException a_e)
			{
				AnonServiceException e = a_e;
				if (e == null)
					{
						e = new AnonServiceException(getCurrentService(), "Proxy at the last mix was nuked!");
					}
				final AnonServiceException exception = e;

				synchronized (m_eventListeners)
					{
						final Enumeration eventListenersList = m_eventListeners.elements();
						Thread notificationThread = new Thread(new Runnable()
							{
								public void run()
									{

										while (eventListenersList.hasMoreElements())
											{
												((AnonServiceEventListener) (eventListenersList.nextElement())).dataChainErrorSignaled(exception);
											}
									}
							}, "AnonClient: DataChainErrorSignaled notification");
						notificationThread.setDaemon(true);
						notificationThread.start();
					}
			}

		public void integrityErrorSignaled(final int a_iErrorCode)
			{
				synchronized (m_eventListeners)
					{
						final Enumeration eventListenersList = m_eventListeners.elements();
						Thread notificationThread = new Thread(new Runnable()
							{
								public void run()
									{
										while (eventListenersList.hasMoreElements())
											{
												((AnonServiceEventListener) (eventListenersList.nextElement())).integrityErrorSignaled(new IntegrityCheckException(
														getCurrentService(), a_iErrorCode));
											}
									}
							}, "AnonClient: integrityErrorSignaled notification");
						notificationThread.setDaemon(true);
						notificationThread.start();
					}
			}

		/** Set the interval after which a dummy is sent
		 *
		 * @param a_interval The new dummy traffic interval in milliseconds or @see DummyTrafficControlChannel.DT_DISABLE,
     *                   if dummy traffic shall be disabled.
		 */

		public void setDummyTraffic(int a_interval)
			{
				synchronized (m_internalSynchronizationForDummyTraffic)
					{
						m_dummyTrafficInterval = a_interval;
						if (m_dummyTrafficControlChannel != null)
							{
								m_dummyTrafficControlChannel.setDummyTrafficInterval(a_interval);
							}
					}
			}

		/**
		 * Sets the time [in ms] an interface of a Cascade is block, if a connection
		 * error occurs)
		 *
		 * @param ms
		 *          block timeout in ms
		 */
		public void setInterfaceBlockTimout(long msTimeOut)
			{
				m_msBlockTimeout = msTimeOut;
			}

		/** Return the TCP connection used to connect to the cascade. Needed in certain applications (e.g. VPN scenarios)
		 *  to make sure that this connection is not routed over the MixCascade...
		 * @return Socket used for connection to MixCascade
		 */
		public SocketHandler getConnectionToMixCascade()
		{
			return m_socketHandler;
		}


		private IStreamConnection connectMixCascade(final MixCascade a_mixCascade,
				ImmutableProxyInterface a_proxyInterface, final IServiceContainer a_serviceContainer,
				final boolean a_bIsReconnect) throws InterruptedException, AnonServiceException
			{
				LogHolder.log(LogLevel.DEBUG, LogType.NET, "Trying to connect to MixCascade '" + a_mixCascade.toString()
						+ "'...");

				Thread notificationThread = new Thread(new Runnable()
					{
						public void run()
							{
								synchronized (m_eventListeners)
									{
										Enumeration eventListenersList = m_eventListeners.elements();
										while (eventListenersList.hasMoreElements())
											{
												((AnonServiceEventListener) (eventListenersList.nextElement())).connecting(a_mixCascade,
														a_bIsReconnect);
											}
									}
							}
					}, "AnonClient: Connecting notification");
				notificationThread.setDaemon(true);
				notificationThread.start();

				Socket connectedSocket = null;
				HTTPConnection connection=null;

				RandomListenerInterfaceSwitcher interfaceSwitcher = new RandomListenerInterfaceSwitcher(
						a_mixCascade.getListenerInterfaces(), ms_preferredConnectionPort, 2, ms_vecTimedOutPorts);
				ListenerInterface nextInterface;
				boolean bTried = false;

				while ((nextInterface = interfaceSwitcher.getNextInterface()) != null && nextInterface.isValid()
						&& (connectedSocket == null) && (!Thread.currentThread().isInterrupted()))
					{
						// recheck the trust for faster switch of services
						a_serviceContainer.getTrustModel().checkTrust(a_mixCascade);

						bTried = true;

						LogHolder.log(LogLevel.INFO, LogType.NET, "Trying cascade connection (Preferred port: "
								+ ms_preferredConnectionPort + ") to interface: " + nextInterface);

						/*
						 * try out all interfaces of the mixcascade until we have a
						 * connection
						 */
						try
							{
								connection = HTTPConnectionFactory.getInstance().createHTTPConnection(nextInterface, a_proxyInterface);
								connection.setTimeout(CONNECT_TIMEOUT);
								connection.setSocketFactory(m_socketFactory);
								connection.setDNSResolver(m_dnsResolver);
								connectedSocket = connection.Connect();
								/*
								 * connectedSocket = new Socket(a_mixCascade.
								 * getListenerInterface(i).getHost(), a_mixCascade.
								 * getListenerInterface(i).getPort());
								 */
							}
						catch (Exception a_e)
							{
								if (a_e instanceof ThreadInterruptedIOException)
									{
										/*
										 * Thread.interrupt() was called while
										 * connection-establishment -> stop all activities
										 */
										LogHolder.log(LogLevel.NOTICE, LogType.NET, "Interrupted while connecting to MixCascade '"
												+ a_mixCascade.toString() + "'.");
										throw new InterruptedException();
									}

								int loglevel = LogLevel.ERR;
								try
									{
										if (a_mixCascade.getListenerInterface(0).getHost().equals("0.0.0.0"))
											{
												loglevel = LogLevel.INFO;
											}
									}
								catch (Exception a_e2)
									{
										// ignore
									}

								if (ms_bBlockOnHttpError || ListenerInterface.isBlockingRecommended(a_e))
									{
										nextInterface.blockInterface(m_msBlockTimeout);
									}

								LogHolder.log(loglevel, LogType.NET, "Error while connecting to MixCascade " + a_mixCascade.toString()
										+ " via " + nextInterface.toString() + "!", a_e);
							}
					}
				if (connectedSocket != null)
					{
						if (ms_preferredConnectionPort != connectedSocket.getPort())
							{
								ms_preferredConnectionPort = connectedSocket.getPort();
								LogHolder.log(LogLevel.NOTICE, LogType.NET, "New preferred cascade connection port is: "
										+ ms_preferredConnectionPort);
							}
						LogHolder.log(LogLevel.DEBUG, LogType.NET, "Connection to MixCascade '" + a_mixCascade.toString()
								+ "' successfully established - starting key-exchange...");
						return new SocketConnection(connectedSocket);
					}
				else
					{
						int loglevel = LogLevel.ERR;
						try
							{
								if (a_mixCascade.getListenerInterface(0).getHost().equals("0.0.0.0"))
									{
										loglevel = LogLevel.INFO;
									}
							}
						catch (Exception a_e)
							{
								// ignore
							}
						String strNoInterfaces = "";
						if (!bTried)
							{
								strNoInterfaces = " No listener interfaces available!";
							}

						LogHolder.log(loglevel, LogType.NET, "Failed to connect to MixCascade '" + a_mixCascade.toString() + "'."
								+ strNoInterfaces);
						return null;
					}
			}

		private void initializeProtocol(IStreamConnection a_connectionToMixCascade, final MixCascade a_mixCascade,
				final IServiceContainer a_serviceContainer, final TermsAndConditionConfirmation termsConfirmation)
				throws AnonServiceException
			{
				synchronized (m_internalSynchronization)
					{
						try
							{
								try
									{
										/* limit timeouts while login procedure */
										a_connectionToMixCascade.setTimeout(getInternalLoginTimeout(a_serviceContainer));
									}
								catch (ConnectionException e)
									{
										/* ignore it */
									}

								synchronized (m_internalSynchronizationForSocket)
									{
										if (m_socketHandler != null)
											{
												m_socketHandler.deleteObservers();
											}
										m_socketHandler = new SocketHandler(a_connectionToMixCascade);
									}

								final Vector exceptionCache = new Vector();
								Thread loginThread = new Thread(new Runnable()
									{
										public void run()
											{
												boolean tcRetry = true;
												int tctry = 0;
												try
													{
														while (tcRetry)
															{
																try
																	{
																		/*
																		 * if (ms_preferredConnectionPort == 80 ||
																		 * ms_preferredConnectionPort == 443) {
																		 * throw new
																		 * java.net.SocketTimeoutException("Test");
																		 * }
																		 */
																		m_keyExchangeManager = new KeyExchangeManager(m_socketHandler.getInputStream(),
																				m_socketHandler.getOutputStream(), (MixCascade) a_mixCascade,
																				a_serviceContainer.getTrustModel(),m_bDebug);
																		tcRetry = false;
																	}
																catch (TermsAndConditionsReadException tcie)
																	{
																		if (!termsConfirmation.confirmTermsAndConditions(tcie.getOperators(),
																				tcie.getTermsTermsAndConditonsToRead()))
																			{
																				a_serviceContainer.keepCurrentService(false);
																				throw new InterruptedException("Client rejected T&C after reading.");
																			}

																		tctry++;
																		if (tctry > 1)
																			{
																				LogHolder.log(LogLevel.ERR, LogType.NET,
																						"Requesting t&cs after the first try is not allowed!");
																				throw new InterruptedException("A second tc request must never be sent.");
																			}

																		m_socketHandler = new SocketHandler(connectMixCascade((MixCascade) a_mixCascade,
																				m_proxyInterface.getProxyInterface(false).getProxyInterface(),
																				a_serviceContainer, true));

																	}
															}
													}
												catch (Exception a_e)
													{
														exceptionCache.addElement(a_e);
													}
											}
									}, "Login Thread");
								loginThread.start();
								try
									{
										// this trick is needed to interrupt the key exchange read
										// operation
										loginThread.join();
									}
								catch (InterruptedException a_e)
									{
										throw a_e;
									}

								if (exceptionCache.size() > 0)
									{
										throw (Exception) exceptionCache.firstElement();
									}
							}
						catch (AnonServiceException e)
							{
								LogHolder.log(LogLevel.ERR, LogType.NET, e);
								closeSocketHandler();
								throw e;
							}
						catch (InterruptedException a_e)
							{
								LogHolder.log(LogLevel.INFO, LogType.NET, a_e);
								closeSocketHandler();
								throw new ServiceInterruptedException(a_mixCascade);
							}
						catch (XMLParseException a_e)
							{
								LogHolder.log(LogLevel.ERR, LogType.NET, a_e);
								closeSocketHandler();
								throw new ParseServiceException(a_mixCascade, a_e.getMessage());
							}
						catch (Exception a_e)
							{
								LogHolder.log(LogLevel.ERR, LogType.NET, a_e);
								closeSocketHandler();

								// java.net.SocketTimeoutException
								Class classSocketTimeoutException = null;
								try
									{
										classSocketTimeoutException = Class.forName("java.net.SocketTimeoutException");
									}
								catch (ClassNotFoundException e1)
									{
										LogHolder.log(LogLevel.INFO, LogType.NET,
												"Old JRE does not have SocketTimeoutException. Please update your Java ASAP!");
									}
								if (classSocketTimeoutException != null && classSocketTimeoutException.isInstance(a_e))
									{
										if (ms_preferredConnectionPort > 0)
											{
												ms_vecTimedOutPorts.addElement(new Integer(ms_preferredConnectionPort));
											}
										ms_preferredConnectionPort = 0;
										throw new ConnectionEstablishmentTimeoutException(a_mixCascade);
									}
								else
									{
										throw new AnonServiceException(a_mixCascade, a_e.getMessage());
									}
							}

						try
							{
								/* try to set infinite timeout */
								a_connectionToMixCascade.setTimeout(0);
							}
						catch (ConnectionException e)
							{
								/* ignore it */
								LogHolder.log(LogLevel.DEBUG, LogType.NET,"AN.ONClient initializeProtocol() - could not set socket to non-blocking after key exchange!");
							}

						m_multiplexer = new Multiplexer(m_socketHandler.getInputStream(), m_socketHandler.getOutputStream(), m_keyExchangeManager, new SecureRandom());
						m_socketHandler.addObserver(this);
						if (m_packetCounter != null)
							{
								// there has been a previous connection
								m_packetCounter = new PacketCounter(m_packetCounter.getProcessedPackets());
							}
						else
							{
								m_packetCounter = new PacketCounter();
							}
						if(!AnonLib.DISABLE_PACKET_COUNTING)
							{
									m_multiplexer.addObserver(m_packetCounter);
							}
						m_packetCounter.addObserver(this);
						synchronized (m_internalSynchronizationForDummyTraffic)
							{
								m_dummyTrafficControlChannel = new DummyTrafficControlChannel(m_multiplexer, a_serviceContainer);
								m_dummyTrafficControlChannel.setDummyTrafficInterval(m_dummyTrafficInterval);
							}
						// the Test Control Channel if used...
						if (ENABLE_CONTROL_CHANNEL_TEST)
							{
								TestControlChannel t = new TestControlChannel(m_multiplexer, a_serviceContainer);
								t.setMessageInterval(30000);
							}

						/* maybe we have to start some more services */
						try
							{
								finishInitialization(m_multiplexer, m_keyExchangeManager, m_packetCounter, a_serviceContainer,	m_keyExchangeManager.getConnectedCascade());
							}
						catch (AnonServiceException a_e)
							{
								shutdown(!a_serviceContainer.isReconnectedAutomatically());
								throw a_e;
							}

							// try
							{
								/* try to set infinite timeout */
								// a_connectedSocket.setSoTimeout(0);
							}
							// catch (SocketException e)
							{
								/* ignore it */
							}

						if (Database.getInstance(MixCascade.class).getEntryById(m_keyExchangeManager.getConnectedCascade().getId()) == null
								|| !a_mixCascade.getId().equals(m_keyExchangeManager.getConnectedCascade().getId()))
							{
								m_currentService = m_keyExchangeManager.getConnectedCascade();
							}
						else
							{
								m_currentService = a_mixCascade;
							}
						TrustModel.cleanAttributeWhitelist(m_currentService);

						if (m_directProxy != null)
							{
								m_directProxy.stop();
							}

						connectionEstablished(m_currentService);
					}
			}

		private void connectionEstablished(final AnonServerDescription a_serverDescription)
			{
				resetInternalLoginTimeout();

				m_connected = true;
				TrustModel.cleanAttributeWhitelist(null);

				Thread notificationThread = new Thread(new Runnable()
					{
						public void run()
							{
								synchronized (m_eventListeners)
									{
										Enumeration eventListenersList = m_eventListeners.elements();
										while (eventListenersList.hasMoreElements())
											{
												((AnonServiceEventListener) (eventListenersList.nextElement())).connectionEstablished(a_serverDescription);
											}
									}
							}
					}, "AnonClient: ConnectionEstablished notification");
				notificationThread.setDaemon(true);
				notificationThread.start();

				LogHolder.log(LogLevel.INFO, LogType.NET, "Connected to MixCascade '" + a_serverDescription.toString() + "'!");
				/* AnonClient successfully started */
			}

		private void finishInitialization(Multiplexer a_multiplexer, KeyExchangeManager a_keyExchangeManager,
				PacketCounter a_packetCounter, IServiceContainer a_serviceContainer, MixCascade a_cascade)
				throws AnonServiceException
			{
				if (a_keyExchangeManager.isProtocolWithTimestamp())
					{
						/* initialize replay-prevention */
						MixParameters[] mixesWithReplayDetection = a_keyExchangeManager.getMixParameters();
						if (a_keyExchangeManager.getFirstMixSymmetricCipher() != null)
							{
								/* the first mix doesn't need timestamps */
								mixesWithReplayDetection = new MixParameters[a_keyExchangeManager.getMixParameters().length - 1];
								for (int i = 0; i < a_keyExchangeManager.getMixParameters().length - 1; i++)
									{
										mixesWithReplayDetection[i] = a_keyExchangeManager.getMixParameters()[i + 1];
									}
							}
						try
							{
								new TimestampUpdater(mixesWithReplayDetection, new ReplayControlChannel(a_multiplexer,
										a_serviceContainer));
							}
						catch (Exception e)
							{
								LogHolder.log(LogLevel.ERR, LogType.NET, "Fetching of timestamps failed - closing connection.", e);
								throw new AnonServiceException(a_cascade, "Fetching of timestamps failed - closing connection.");
							}
					}

				// aiControlChannel.addAIListener(a_aiListener)
				if (a_keyExchangeManager.isPaymentRequired())
					{
						AIControlChannel aiControlChannel = new AIControlChannel(a_multiplexer, a_packetCounter,
								a_serviceContainer, a_cascade);

						aiControlChannel.addAIListener(new IAIEventListener()
							{
								public void accountEmpty(PayAccount a_account, MixCascade a_cascade)
									{
										LogHolder.log(LogLevel.WARNING, LogType.PAY, "Account empty: " + a_account + " Reconnect!");
										// close the connection due to an empty account; if set,
										// reconnection will be done automatically
										reconnect(a_cascade, null);
									}

								public void accountChanged(PayAccount a_account, MixCascade a_cascade)
									{
										LogHolder.log(LogLevel.WARNING, LogType.PAY, "Account changed: " + a_account + " Reconnect!");
										reconnect(a_cascade, null);
									}
							});
						aiControlChannel.setAILoginTimeout(m_loginTimeout);
						aiControlChannel.sendAccountCert();
					}
				if(a_keyExchangeManager.supportsUDPOverUDP())//Cascade support the UDP over UDP protocol
					{
						try
							{
								m_udpMultiplexer=new UDPMultiplexer(a_cascade,a_keyExchangeManager);
							}
						catch (Exception e)
							{
								throw new AnonServiceException(a_cascade, "'Initialisation of UDPMultiplexer failed - closing connection.");
							}
					}
			}

		private void closeSocketHandler()
			{
				synchronized (m_internalSynchronizationForSocket)
					{
						if (m_socketHandler != null)
							{
								/*
								 * we have to check for null because the socket could be closed
								 * by another thread in the meantime
								 */
								m_socketHandler.closeSocket();
								m_socketHandler = null;
							}
					}
			}

		public void setSocketFactory(HTTPClientSocketFactory socketFactory)
			{
				m_socketFactory=socketFactory;
			}

		public void setDNSResolver(IHTTPClientDNSResolver a_dnsResolver)
			{
				m_dnsResolver=a_dnsResolver;
			}


		/** Only for internal cleanup!*/
		public static void tearDown()
			{
				ms_queuePacketCount.stop();
				ms_queuePacketCount=null;
			}

	}
