/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import anon.AnonChannel;

/***
 * This class can be used to create an anonymous TCP channel to a given
 * host:port. You first create an unconnected channel of type TCP using
 * <code>AnonService.createChannel(AnonChannel.TCP)</code>. Afterwards you
 * connect to the target host using {@link #connect(String, int) connect}. Data
 * transmission can be done by using the Input-/Outputstream returned by
 * <code>getInputStream()</code> / <code>getOutputStream()</code>.
 * <p>
 * Hint: for performance optimisation / efficiency have a look at
 * {@link #setSentAutoFlushDelay(int)}.
 * <p>
 * Example: <code><pre>
 *  AnonService anonservice=AnonServiceFactory.getAnonServiceInstance(AnonServiceFactory.SERVICE_ANON);
 *  AnonTCPChannel anonchannel=(AnonTCPChannel)anonservice.createChannel(AnonChannel.TCP);
 *  anonchannel.connect("desti.nation",80);
 *  InputStream in=anonchannel.getInputStream();
 *  OutputStream out=anonchannel.getOutputStream();
 *  in.read();
 *  out.write(...);
 *  anonchannel.close();
 *</pre></code>
 * <p>
 * Note: Since the AnonTCPChannel implementation uses an anonymous SOCKS
 * channel, an AnonTCPChannel can only be established using a cascade which
 * supports SOCKS.
 *
 * @author sk13
 *
 */
public class AnonTCPChannel implements AnonChannel
	{

		private AnonChannel m_channelUnderlyingChannel = null;

		class AnonTCPChannelOutputStream extends OutputStream
			{
				OutputStream m_Out;

				AnonTCPChannelOutputStream(OutputStream out)
					{
						m_Out = out;
					}

				public void write(byte[] b) throws IOException
					{
						write(b, 0, b.length);
					}

				public void write(int b) throws IOException
					{
						m_Out.write(b);
					}

				public void write(byte[] b, int off, int len) throws IOException
					{
						while (len > 0)
							{
								int i = Math.min(m_channelUnderlyingChannel.getOutputBlockSize(), len);
								m_Out.write(b, off, i);
								off += i;
								len -= i;
							}
					}

				public void flush() throws IOException
					{
						m_Out.flush();
					}

				public void close() throws IOException
					{
						m_Out.close();
					}

			}

		protected AnonTCPChannel(AnonChannel a_channelUnderlyingChannel)
			{
				m_channelUnderlyingChannel = a_channelUnderlyingChannel;
			}

		/***
		 * Establishes an anonymous TCP connection to the given host:port.
		 *
		 * @param hostname
		 *          the domain name of the target host
		 * @param port
		 *          the target port number
		 * @throws IOException
		 *           if the connection could not be established
		 */
		public void connect(String hostname, int port) throws IOException
			{
				//Establish a TCP connection to the specified Host:Port using SOCKS (over AN.ON)
				//Send SOCKS Auth Request and Target in one packet...
				OutputStream out = getOutputStream();
				int lenHostname = hostname.length();
				byte[] req = new byte[lenHostname + 10];
				//Auth sub request
				req[0] = 5; //SOCKS version 5;
				req[1] = 1; //we know about one auth method...
				req[2] = 0; //...and this is the empty method
				//CONNECT request
				req[3] = 5; //SOCKS version 5
				req[4] = 1; //TCP-CONNECT command
				//req[5]=0;  //reserved byte = 0
				req[6] = 3; //target is given as hostname
				req[7] = (byte) lenHostname; // length of hostname
				byte[] bar_hostname = hostname.getBytes();
				System.arraycopy(bar_hostname, 0, req, 8, lenHostname); // the bytes of the hostname itself
				req[8 + lenHostname] = (byte) ((port >> 8) & 0x00FF); // high byte of port
				req[9 + lenHostname] = (byte) (port & 0x00FF); //low byte of port
				out.write(req);
				out.flush();

				//Read response from SOCKS-Proxy
				InputStream in = getInputStream();
				int ret = in.read(req, 0, 2); // Auth response --> we ignore it for now
				//CONNECT reply
				ret = in.read(); // SOCKS-Version
				int result = in.read(); // reply --> 0 means success
				if (result != 0)
					{
						close();
						throw (new IOException("Connect to: "+hostname+" not successful!"));
					}
				//read over the remaining fields of the reply
				in.read(); // reserved field
				int atype = in.read();
				if (atype == 1) //IPv4
					{
						in.read(req, 0, 6); // IPv4 addr + port;
					}
				else if (atype == 4) //IPv6
					{
						in.read(req, 0, 18); // IPv6 addr + port;
					}
				else if (atype == 3) // hostname
					{
						int len = in.read(); //hostname length
						in.read(req, 0, len); // the bytes of teh hostname
					}
				else
					{
						close();
						throw (new IOException("Connect not successful!"));
					}
			}

		/**
		 * Sets the time to wait before actually sending some data over this channel
		 * in case the packet belonging to the underlying anonymisation protocol is
		 * not full yet.
		 * <p>
		 * This method can be used for performance optimisations / efficiency. One
		 * use case would be if you know that this channel will sent data
		 * immediately after connection establishment. Therefore you would call
		 * <code>setSentAutoFlushDelay</code> with some <code>delayvalue > 0</code>
		 * just before you call the <code>connect</code> method. If you call
		 * <code>Outputstream.write</code> within <code>delayvalue</code>
		 * milliseconds after the <code>connect</code> call then the data you intend
		 * to sent will be piggybacked on the connect request. This will save some
		 * roundtrip-time delay.
		 * <p>
		 * Additionally, using this method correctly could decrease the bandwidth
		 * overhead introduced by the underlying anonymisation protocol. Many
		 * anonymisation protocols fragment the data stream into fixed sized packets
		 * which are filled with dummy data if not enough real data is available.
		 * Therefore waiting a certain amount of time for additional real data can
		 * increase the channel utilisation.
		 *
		 * <p>
		 * Note: This method will not do anything at the moment as the
		 * implementation is not ready yet.
		 *
		 * @param msAutoFlushDelay
		 *          the amount of time in milliseconds to wait for new data before
		 *          the current sent buffer is flushed. Setting this value to
		 *          <code>0</code> will always auto flush immediately. Setting this
		 *          value to <code>-1</code> will only flush the sent buffer, if the
		 *          sent buffer is full, i.e. getOutputBufgferSize many bytes are
		 *          contained in the buffer.
		 */
		public void setSentAutoFlushDelay(int msAutoFlushDelay)
			{

			}

		public InputStream getInputStream()
			{
				return m_channelUnderlyingChannel.getInputStream();
			}

		public OutputStream getOutputStream()
			{
				return new AnonTCPChannelOutputStream(m_channelUnderlyingChannel.getOutputStream());
			}

		public int getOutputBlockSize()
			{
				return m_channelUnderlyingChannel.getOutputBlockSize();

			}

		public void close()
			{
				m_channelUnderlyingChannel.close();
			}

		public boolean isClosed()
			{
				return m_channelUnderlyingChannel.isClosed();
			}

	}
