/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import anon.AnonChannel;

/***
 * This class can be used to create an anonymous UDP channel to a given
 * host:port. You first create a channel of type UDP using
 * <code>AnonService.createChannel(AnonChannel.UDP)</code>. Afterwards you set
 * the destination host and port as well as the source port. Data transmission
 * can be done by using the Input-/Outputstream returned by
 * <code>getInputStream()</code> / <code>getOutputStream()</code>.
 * <p>
 * Hint: for performance optimization / efficiency have a look at
 * {@link #setSentAutoFlushDelay(int)}.
 * <p>
 * Example: <code><pre>
 *  AnonService anonservice=AnonServiceFactory.getAnonServiceInstance(AnonServiceFactory.SERVICE_ANON);
 *  AnonUDPChannel anonchannel=(AnonUDPChannel)anonservice.createChannel(AnonChannel.UDP);
 *  anonchannel.setDst("desti.nation",80);
 *  anonchannel.setSrcPort(5678);
 *  InputStream in=anonchannel.getInputStream();
 *  OutputStream out=anonchannel.getOutputStream();
 *  out.write(...);
 *  in.read(...)
 *  anonchannel.close();
 *</pre></code>
 * <p>
 * Note: Since the AnonUDPChannel implementation uses an anonymous VPN UDP
 * protocol channel, an AnonUDPChannel can only be established using a cascade
 * which supports VPN.
 *
 * @author sk13
 *
 */
public class AnonUDPChannel extends AbstractAnonUDPChannel
	{
		private final static int UDP_PROTOCOL_HEADER_SIZE = 2;
		private AnonChannel m_channelUnderlyingChannel = null;

		private boolean m_bConnectSent = false;
		private OutputStream m_Out;
		private InputStream m_In;
		private int m_iRemainingLenUDPPayload;

		protected AnonUDPChannel(AnonChannel a_channelUnderlyingChannel)
			{
				super();
				m_iRemainingLenUDPPayload = -1;
				m_channelUnderlyingChannel = a_channelUnderlyingChannel;
				m_Out = m_channelUnderlyingChannel.getOutputStream();
				m_In = m_channelUnderlyingChannel.getInputStream();
			}

		/***
		 * Creates a connect message to the host given in setDestination() and saves
		 * it into the given byte array.
		 */
		private void getConnectMsg(byte req[])
			{
				//Establish a UDP protocol connection to the specified Host:Port using VPN UDP protocol (over AN.ON)
				//Send connection request in one packet...
				int lenHostname = m_strDstHost.length();
				//Auth sub request
				req[0] = 1; //UDP protocol version 1;
				//CONNECT request
				req[1] = 4; //UDP-CONNECT command
				req[2] = 3; //target is given as hostname
				req[3] = (byte) lenHostname; // length of hostname

				byte[] bar_hostname = m_strDstHost.getBytes();
				System.arraycopy(bar_hostname, 0, req, 4, lenHostname); // the bytes of the hostname itself

				req[4 + lenHostname] = (byte) ((m_iDstPort >> 8) & 0x00FF); // high byte of dst port
				req[5 + lenHostname] = (byte) (m_iDstPort & 0x00FF); //low byte of dst port
				req[6 + lenHostname] = (byte) ((m_iSrcPort >> 8) & 0x00FF); // high byte of src port
				req[7 + lenHostname] = (byte) (m_iSrcPort & 0x00FF); //low byte of src port
			}

		/***
		 * Returns the size of the UDP protocol connect msg.
		 */
		private int getConnectMsgSize()
			{
				return m_strDstHost.length() + 8;
			}

		/**
		 * Sets the time to wait before actually sending some data over this channel
		 * in case the packet belonging to the underlying anonymisation protocol is
		 * not full yet.
		 * <p>
		 * This method can be used for performance optimisations / efficiency. One
		 * use case would be if you know that this channel will sent data
		 * immediately after connection establishment. Therefore you would call
		 * <code>setSentAutoFlushDelay</code> with some <code>delayvalue > 0</code>
		 * just before you call the <code>connect</code> method. If you call
		 * <code>Outputstream.write</code> within <code>delayvalue</code>
		 * milliseconds after the <code>connect</code> call then the data you intend
		 * to sent will be piggybacked on the connect request. This will save some
		 * roundtrip-time delay.
		 * <p>
		 * Additionally, using this method correctly could decrease the bandwidth
		 * overhead introduced by the underlying anonymisation protocol. Many
		 * anonymisation protocols fragment the data stream into fixed sized packets
		 * which are filled with dummy data if not enough real data is available.
		 * Therefore waiting a certain amount of time for additional real data can
		 * increase the channel utilisation.
		 *
		 * <p>
		 * Note: This method will not do anything at the moment as the
		 * implementation is not ready yet.
		 *
		 * @param msAutoFlushDelay
		 *          the amount of time in milliseconds to wait for new data before
		 *          the current sent buffer is flushed. Setting this value to
		 *          <code>0</code> will always auto flush immediately. Setting this
		 *          value to <code>-1</code> will only flush the sent buffer, if the
		 *          sent buffer is full, i.e. getOutputBufgferSize many bytes are
		 *          contained in the buffer.
		 */
		public void setSentAutoFlushDelay(int msAutoFlushDelay)
			{

			}

		public void writePacket(byte[] payload, int offset, int len) throws IOException
			{
				/*
				 * Prepare first message to be sent, by declaring its
				 * size the first two bytes.
				 *
				 * If no initial connect message was sent yet, send it
				 * first.
				 */
				int bytes_written;

				if (!m_bConnectSent)
					{
						int connectMsgSize = getConnectMsgSize();
						// buffer for connect msg, udp protocol header and actual packet
						// from parameter
						byte msg[] = new byte[connectMsgSize + UDP_PROTOCOL_HEADER_SIZE + len]; // assumes OutputBlockSize > connectMsgSize +2!!

						// copy connect message into buffer
						getConnectMsg(msg);

						// add header for the actual message after the connect message
						// body (basically only the length)
						msg[connectMsgSize] = (byte) ((len >> 8) & 0x00FF);
						msg[connectMsgSize + 1] = (byte) (len & 0x00FF);

						// append bytes of the actual message
						System.arraycopy(payload, offset, msg, connectMsgSize + UDP_PROTOCOL_HEADER_SIZE, len);
						payload = msg;
						offset = 0;

						// add the connect message and header bytes to
						// the length to be sent further down
						len += connectMsgSize + UDP_PROTOCOL_HEADER_SIZE;

						m_bConnectSent = true;
					}
				else
					{
						//XXX What happens, when the method returns 1?
						byte[] firstPacket = new byte[m_channelUnderlyingChannel.getOutputBlockSize()];

						firstPacket[0] = (byte) ((len >> 8) & 0x00FF); // add len header
						firstPacket[1] = (byte) (len & 0x00FF);

						// add as many bytes of b as possible to the 2 header bytes
						bytes_written = Math.min(firstPacket.length - UDP_PROTOCOL_HEADER_SIZE, len);

						System.arraycopy(payload, offset, firstPacket, UDP_PROTOCOL_HEADER_SIZE, bytes_written);

						// AbstractDataChain - DataChainOutputStreamImplementation
						m_Out.write(firstPacket, 0, bytes_written + UDP_PROTOCOL_HEADER_SIZE);

						len -= bytes_written;
						offset += bytes_written;
					}

				// send out left over bytes
				while (len > 0)
					{
						bytes_written = Math.min(m_channelUnderlyingChannel.getOutputBlockSize(), len);
						m_Out.write(payload, offset, bytes_written);
						offset += bytes_written;
						len -= bytes_written;
					}
			}

		/**
		 * Reads the payload of a whole UDP packet sent from the peer. If the
		 * provided buffer is too small a UDPPacketToLargeException is thrown.
		 *
		 * Note: Do not mix calls to getInputStream.read() and readPacket() --
		 * otherwise the UDP packet stream might get out of sync leading to a
		 * UDPPacketStreamDesynchronisationException.
		 *
		 * @param payload
		 * @param off
		 * @param len
		 * @return
		 * @throws IOException
		 * @throws UDPPacketStreamDesynchronisationException
		 * @throws UDPPacketToLargeException
		 */
		public int readPacket(byte[] payload, int off, int len)
				throws IOException, UDPPacketStreamDesynchronisationException, UDPPacketToLargeException
			{
				int ret = -1;

				if (m_iRemainingLenUDPPayload <= 0)
					{//read new UDP packet payload
						m_iRemainingLenUDPPayload = m_In.read();

						if (m_iRemainingLenUDPPayload < 0)
							return -1; // stream closed

						// read first header byte; highest byte of packet length
						m_iRemainingLenUDPPayload <<= 8;

						//read second header byte; lowest byte of packet length
						ret = m_In.read();

						if (ret < 0)
							return -1; // stream closed

						// assemble the packet length
						m_iRemainingLenUDPPayload |= (ret & 0x00FF);
						m_iRemainingLenUDPPayload &= 0x00FFFF;
						//					System.out.println("len: " + len + " p_len: " + m_iRemainingLenUDPPayload);
					}
				else
					{
						// there are still bytes left from a previous packet
						throw new UDPPacketStreamDesynchronisationException();
					}

				// the packet is larger than the alloted space in the byte array
				if (m_iRemainingLenUDPPayload > len)
					throw new UDPPacketToLargeException();

				len = m_iRemainingLenUDPPayload;

				while (m_iRemainingLenUDPPayload > 0)
					{
						ret = m_In.read(payload, off, m_iRemainingLenUDPPayload);

						if (ret >= 0)
							{
								m_iRemainingLenUDPPayload -= ret;
								off += ret;
							}
						else
							return ret;
					}
				return len;
			}

		public int getOutputBlockSize()
			{
				return m_channelUnderlyingChannel.getOutputBlockSize() - UDP_PROTOCOL_HEADER_SIZE;

			}

		public void close()
			{
				m_channelUnderlyingChannel.close();
			}

		public boolean isClosed()
			{
				return m_channelUnderlyingChannel.isClosed();
			}

	}
