package anon.client;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;

final class DataChainInputStreamImplementation extends InputStream
	{

		private boolean m_closed;

		DataChainInputStreamQueueEntry m_queueEntriesHead;
		DataChainInputStreamQueueEntry m_queueEntriesTail;

		Object m_oSync = null;

		DataChainInputStreamImplementation()
			{
				m_queueEntriesHead = null;
				m_queueEntriesTail = null;
				m_closed = false;
				m_oSync = new Object();
			}

		public int read() throws IOException
			{
				byte[] buffer = new byte[1];
				int bytesRead = 0;
				do
					{
						bytesRead = read(buffer);
					}
				while (bytesRead == 0);
				int returnedByte = -1;
				if (bytesRead == 1)
					{
						returnedByte = (buffer[0]&0x00FF);
					}
				return returnedByte;
			}

		public int read(byte[] a_buffer, int a_offset, int a_length) throws IOException
			{
				int bytesRead = 0;
				/*
				 * to prevent later exceptions, we check (and - if necessary -
				 * alter) offset and length
				 */
				if (a_buffer.length < a_offset)
					{
						a_offset = a_buffer.length;
					}
				if (a_buffer.length < a_offset + a_length)
					{
						a_length = a_buffer.length - a_offset;
					}
				if (a_length > 0)
					{
						synchronized (m_oSync)
							{
								if (m_closed)
									{
										throw (new IOException("Stream is closed."));
									}
								while (m_queueEntriesHead == null)
									{
										if (m_closed)
											{
												throw (new IOException("Stream is closed."));
											}
										/*
										 * nothing is currently available -> wait for available
										 * queue-entries
										 */
										try
											{
												m_oSync.wait();
											}
										catch (InterruptedException e)
											{
												throw (new InterruptedIOException("InterruptedException: " + e.toString()));
											}
									}
								/* we have an entry -> read data from there */

								switch (m_queueEntriesHead.getType())
									{
										case DataChainInputStreamQueueEntry.TYPE_STREAM_END:
											{
												/* end of stream reached -> do nothing */
												bytesRead = -1;
												break;
											}
										case DataChainInputStreamQueueEntry.TYPE_DATA_AVAILABLE:
											{
												/* data-available entry */
												while ((m_queueEntriesHead != null)
														&& (m_queueEntriesHead.getType() == DataChainInputStreamQueueEntry.TYPE_DATA_AVAILABLE)
														&& (bytesRead < a_length))
													{
														/* read some more bytes */
														int bytesToRead = Math.min(a_length - bytesRead,m_queueEntriesHead.getAvailableBytes());
														
														m_queueEntriesHead.getData(a_buffer,a_offset + bytesRead,bytesToRead);
														
														//System.arraycopy(currentEntry.getData(), currentEntry.getAlreadyReadBytes(), a_buffer,
														//		a_offset + bytesRead, bytesToRead);
														bytesRead += bytesToRead;
														if (m_queueEntriesHead.getAvailableBytes()==0)
															{
																/*
																 * the data of the current entry was
																 * read completely -> remove the entry
																 * from the queue
																 */
																m_queueEntriesHead = m_queueEntriesHead.m_Next;
																if (m_queueEntriesHead == null)//removed last element
																	{
																		m_queueEntriesTail = null;
																	}
															}
													}
												break;
											}
										case DataChainInputStreamQueueEntry.TYPE_IO_EXCEPTION:
											{
												/* we shall throw an IOException */
												IOException exceptionToThrow = m_queueEntriesHead.getIOException();
												m_queueEntriesHead = m_queueEntriesHead.m_Next;
												if (m_queueEntriesHead == null)//removed last element
													{
														m_queueEntriesTail = null;
													}
												if (exceptionToThrow != null)
													{
														throw exceptionToThrow;
													}
												break;
											}
									}
							}
					}

				return bytesRead;
			}

		public int available() throws IOException
			{
				int availableBytes = 0;
				synchronized (m_oSync)
					{
						if (m_closed)
							{
								throw (new IOException("Stream is closed."));
							}
						DataChainInputStreamQueueEntry currentEntry = m_queueEntriesHead;
						while (currentEntry != null)
							{
								if (currentEntry.getType() == DataChainInputStreamQueueEntry.TYPE_DATA_AVAILABLE)
									{
										/* we have found a data entry */
										availableBytes+=currentEntry.getAvailableBytes();
										/* try to get the next entry */
										currentEntry = currentEntry.m_Next;
									}
								else
									{
										/*
										 * stop processing of the queue because we have
										 * found an non-data entry
										 */
										currentEntry = null;
									}
							}
					}
				return availableBytes;
			}

		public void close()
			{
				if (!m_closed)
					{
						synchronized (m_oSync)
							{
								m_closed = true;
								/* clear all entries in the queue */
								m_queueEntriesHead = null;
								m_queueEntriesTail = null;
								/* wake up waiting threads (normally there should be no one) */
								m_oSync.notifyAll();
							}
					}
			}

		public void addToQueue(DataChainInputStreamQueueEntry a_entry)
			{
				if (a_entry.getType() == DataChainInputStreamQueueEntry.TYPE_DATA_AVAILABLE && a_entry.getAvailableBytes() == 0)
					{
						return;
					}
				synchronized (m_oSync)
					{
						if (m_closed)
							{
								/* only add entries, if the stream isn't closed */
								return;
							}
						else if (m_queueEntriesTail != null	&& m_queueEntriesTail.getType() == DataChainInputStreamQueueEntry.TYPE_STREAM_END)
							{
								/* don't add entries after a stream-end */
								return;
							}

						a_entry.m_Next = null;
						if (m_queueEntriesTail != null)
							{
								m_queueEntriesTail.m_Next = a_entry;
								m_queueEntriesTail=a_entry;
							}
						else
							{
								m_queueEntriesTail = a_entry;
								m_queueEntriesHead = a_entry;
							}
						/* wake up a waiting thread */
						m_oSync.notifyAll();
					}
			}
	}