/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.IOException;

/**
 * @author Stefan Lieske
 */
public class DataChainSendOrderStructure
	{

		private byte[] m_orderData;

		/** A multi purpose object that can be set to be anything */
		private Object m_additionalProtocolData;

		private int m_processedBytes;

		private IOException m_thrownException;

		/**
		 * A byte array formatted for the channel that is used to send it.
		 * Contains the user data plus any additional information the channel
		 * needs.
		 * <p>
		 * This will be encrypted for the mix chain and send over the socket.
		 * </p>
		 *
		 * @see {@link anon.client.SequentialChannelDataChain#createPacketPayload(DataChainSendOrderStructure)}
		 * @see {@link anon.client.SingleChannelDataChain#createPacketPayload(DataChainSendOrderStructure)}
		 */
		private byte[] m_channelCell;

		/** Object used to synchronize processing and sending of the order. */
		private Object m_internalSynchronization;

		/** Whether processing on this order has finished or not. */
		private boolean m_processingDone;

		public DataChainSendOrderStructure(byte[] a_orderData)
			{
				m_orderData = a_orderData;
				m_processedBytes = 0;
				m_thrownException = null;
				m_channelCell = null;
				m_internalSynchronization = new Object();
				m_processingDone = false;
				m_additionalProtocolData = null;
			}

		public byte[] getOrderData()
			{
				return m_orderData;
			}

		/**
		 * Returns the additional protocol data as an Object type.
		 * <p>
		 * This needs to be cast to whatever type the protocol data is expected
		 * to be.
		 * </p>
		 * @return protocol data as an Object, null if none was set
		 */
		public Object getAdditionalProtocolData()
			{
				return m_additionalProtocolData;
			}

		/**
		 * Sets any object as additional protocol data.
		 *
		 * @param a_protocolData protocol data as Object
		 */

		public void setAdditionalProtocolData(Object a_protocolData)
			{
				m_additionalProtocolData = a_protocolData;
			}

		/**
		 * Declares the processing of this order as having finished and notifies
		 * threads waiting on it.
		 *
		 * @see {@link anon.client.DataChainSendOrderStructure#m_internalSynchronization m_internalSynchronization}
		 * @see {@link anon.client.DataChainSendOrderStructure#isProcessingDone() isProcessingDone()}
		 * @see {@link anon.client.DataChainSendOrderStructure#getSynchronizationObject() getSynchronizationObject()}
		 */
		public void processingDone()
			{
				synchronized (m_internalSynchronization)
					{
						m_processingDone = true;
						/* wake up waiting threads */
						m_internalSynchronization.notify();
					}
			}

		/**
		 * Returns whether or not the processing was finished on this order
		 * already.
		 *
		 * @return true, if processing is finished, false otherwise
		 */
		public boolean isProcessingDone()
			{
				return m_processingDone;
			}

		/**
		 * Returns the internal synchronization object. Threads can wait on this
		 * object, to be notified when processing of this order was completed.
		 *
		 * @return the internal synchronization object
		 */
		public Object getSynchronizationObject()
			{
				return m_internalSynchronization;
			}

		/**
		 * Sets the exception that occurred, when the processing of an order
		 * ended in an error.
		 *
		 * @param a_thrownException
		 */
		public void setThrownException(IOException a_thrownException)
			{
				m_thrownException = a_thrownException;
			}

		/**
		 * Returns the exception that occurred while processing this order.
		 *
		 * @return the exception that occurred while processing this order.
		 */
		public IOException getThrownException()
			{
				return m_thrownException;
			}

		/**
		 * Set how many bytes of the order data have been processed already.
		 *
		 * @param a_processedBytes
		 */
		public void setProcessedBytes(int a_processedBytes)
			{
				m_processedBytes = a_processedBytes;
			}

		/**
		 * Returns the number of bytes of the order data, that have already
		 * been processed.
		 *
		 * @return
		 */
		public int getProcessedBytes()
			{
				return m_processedBytes;
			}

		/**
		 * Set the channel cell byte buffer, that will be transmitted over the
		 * channel.
		 * <p>
		 * This is usually called by the data chain, when preparing the payload.
		 * </p>
		 *
		 * @see {@link anon.client.DataChainSendOrderStructure#m_channelCell m_channelCell}
		 * @see {@link anon.client.AbstractDataChain#createPacketPayload(DataChainSendOrderStructure)}
		 *
		 * @param a_channelCell
		 */

		public void setChannelCell(byte[] a_channelCell)
			{
				m_channelCell = a_channelCell;
			}

		/**
		 * Returns the channel cell buffer, that will be transmitted over the
		 * channel.
		 *
		 * @see {@link anon.client.DataChainSendOrderStructure#m_channelCell m_channelCell}
		 *
		 * @return
		 */
		public byte[] getChannelCell()
			{
				return m_channelCell;
			}

	}
