/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.util.Observable;

/**
 * This class holds messages, that are channel specific and distinct from the
 * user traffic, i. e. a message signaling the closing of the channel.
 * <p>
 * These messages are added by the corresponding channel upon reception from
 * the multiplexer and processed accordingly.
 * </p>
 *
 * @author Stefan Lieske
 */
final public class InternalChannelMessageQueue extends Observable
	{

		private InternalChannelMessage m_messageQueueHead;
		private InternalChannelMessage m_messageQueueTail;
		private Object m_oSync=new Object();

		/**
		 * Initializes an empty queue.
		 */
		public InternalChannelMessageQueue()
			{
				m_messageQueueHead = null;
				m_messageQueueTail = null;
			}

		/**
		 * Adds (only) the given message to the tail of the queue and notify all
		 * observers.
		 *
		 * @param a_message
		 */
		public void addChannelMessage(InternalChannelMessage a_message)
			{
				synchronized (m_oSync)
					{
						a_message.m_Next=null;

						// head -> ... -> ... -> tail -> null

						if(m_messageQueueHead==null) // queue is empty
							{
								m_messageQueueHead=a_message;
								m_messageQueueTail=a_message;
							}
						else
							{
								// head -> ... -> ... -> tail -> a_message
								m_messageQueueTail.m_Next=a_message;
								// head -> ... -> ... -> old_tail -> tail (a_message)
								m_messageQueueTail=a_message;
							}
						m_oSync.notify();
					}
				setChanged();
				notifyObservers(a_message);
			}

		/**
		 * Returns the next message without removing it.
		 *
		 * @return the next message, may be null
		 */

		public InternalChannelMessage getFirstMessage()
			{
				InternalChannelMessage nextMessage = null;
				synchronized (m_oSync)
					{
						if (m_messageQueueHead!=null)
							{
								nextMessage = m_messageQueueHead;
							}
					}
				return nextMessage;
			}

		/*public void removeFirstMessage()
			{
				synchronized (m_oSync)
					{
						if (m_messageQueueHead!=null)
							{
								m_messageQueueHead=m_messageQueueHead.m_Next;
								if(m_messageQueueHead==null)//we removed the last element
									{
										m_messageQueueTail=null;
									}
							}
					}
			}*/

		/**
		 * This method will return the current message queue head. If there is
		 * no message to return this method will block until a message is
		 * available or an exception is thrown. The message is removed from the
		 * queue.
		 *
		 * @return internal channel message
		 * @throws InterruptedException
		 */

		public InternalChannelMessage waitForNextMessage() throws InterruptedException
			{
				InternalChannelMessage returnedMessage=null;
				synchronized (m_oSync)
					{
						while (m_messageQueueHead == null)
							{
								m_oSync.wait();
							}
						returnedMessage = m_messageQueueHead;

						//advance queue
						m_messageQueueHead=m_messageQueueHead.m_Next;

						if(m_messageQueueHead==null)//we removed the last element
								{
									m_messageQueueTail=null;
								}
					}
				return returnedMessage;
			}

	}
