/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.EOFException;
import java.io.InputStream;
import java.io.IOException;
import java.security.SecureRandom;
import java.util.Vector;

import anon.client.crypto.ISymCipher;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

/**
 * A MixPacket representation.
 *
 * @author Stefan Lieske
 */
public class MixPacket
	{

		private static final int PACKET_SIZE = 998;
		private static final int HEADER_SIZE = 6; // length of channel id + flags
		private static final int PAYLOAD_SIZE = PACKET_SIZE - HEADER_SIZE; // PACKET_SIZE - header

		// channel flags
		public static final short FLAG_CHANNEL_DATA  = 0x0000;

		public static final short FLAG_CHANNEL_CLOSE = 0x0001;

		public static final short FLAG_CHANNEL_OPEN  = 0x0008;
		public static final short FLAG_CHANNEL_DUMMY = 0x0010;
		public static final short FLAG_DEBUG  = 0x0020; //used only for debugging purposes to flag certain MixPackets


		private final static SecureRandom ms_secureRandom = new SecureRandom();

		private int m_channelId;

		private short m_channelFlags;

		private byte[] m_payloadData;

		private Vector m_sendCallbackHandlers;

		public static int getPacketSize()
			{
				return PACKET_SIZE;
			}

		/**
		 * Returns the size of the internal payload buffer.
		 *
		 * @return length of internal payload byte buffer
		 */

		public static int getPayloadSize()
			{
				return PAYLOAD_SIZE;
			}

		/**
		 * Tries to read a mix packet from the given InputStream and decrypts
		 * its payload using the given ISymCipher.
		 *
		 * This Constructor blocks until input data is available, end of file is
		 * detected, or an exception is thrown.
		 *
		 * @param a_inputStream
		 * @param a_inputStreamCipher
		 * @throws IOException
		 */

		public MixPacket(InputStream a_inputStream, ISymCipher a_inputStreamCipher) throws IOException
			{
				m_sendCallbackHandlers = new Vector();
				/* read the packet from the origin stream */
				byte[] rawPacket = new byte[PACKET_SIZE];


				int pos=0;
				int length=PACKET_SIZE;
				try
					{
						while(length>0)
							{
								int ret=a_inputStream.read(rawPacket, pos, length);
								if(ret<0)
									{
										throw new EOFException();
									}
								else
									{
										pos+=ret;
										length-=ret;
									}
							}
						// /Do packet Timing logging --> note: this is just for doing some
						// kind of Web-Fingerprinting Attac Experiments and will be removed
						// afterwards...
						//LogHolder.log(LogLevel.DEBUG, LogType.TRANSPORT, "PacketReceived at time: " + System.currentTimeMillis());

					}
				catch (EOFException eofe)
					{
						LogHolder.log(LogLevel.WARNING, LogType.NET, Thread.currentThread().getName()
								+ ": received a truncated packet from a mix: ", eofe);
						throw eofe;
					}
				catch (IOException a_e)
					{
						try
							{
								Class classSocketTimeoutException = Class.forName("java.net.SocketTimeoutException");
								if (classSocketTimeoutException.isAssignableFrom(a_e.getClass()))
									{
										// make another try to evade timeout
										while(length>0)
											{
												int ret=a_inputStream.read(rawPacket, pos, length);
												if(ret<0)
													{
														throw new EOFException();
													}
												else
													{
														pos+=ret;
														length-=ret;
													}
											}
									}
								else
									{
										throw a_e;
									}
							}
						catch (ClassNotFoundException e)
							{
								// TODO Automatisch erstellter Catch-Block
								throw a_e;
							}
						// java.net.SocketTimeoutException
					}
				/* do stream-decryption */
				if (a_inputStreamCipher != null)
					{
						a_inputStreamCipher.encryptAES1(rawPacket, 0, rawPacket, 0, 16);
					}
				/* read channel-id and channel-flags from the packet */
				m_channelId =  ( rawPacket[0] &0x00FF) <<24 | (rawPacket[1] &0x00FF) <<16 | (rawPacket[2] &0x00FF) <<8 | (rawPacket[3] &0x00FF) ;
				m_channelFlags =(short) ((rawPacket[4] &0x00FF) <<8 | (rawPacket[5] &0x00FF)) ;
				m_payloadData = new byte[PAYLOAD_SIZE];
				System.arraycopy(rawPacket, HEADER_SIZE, m_payloadData, 0, PAYLOAD_SIZE);
			}

		/**
		 * Creates a Mixpacket with the given channel ID and random bytes as the
		 * payload.
		 *
		 * @param a_channelId
		 */

		public MixPacket(int a_channelId)
			{
				m_sendCallbackHandlers = new Vector();
				m_channelId = a_channelId;
				m_channelFlags = 0;
				m_payloadData = new byte[PAYLOAD_SIZE];
				/* initialize payload with random bytes */
				ms_secureRandom.nextBytes(m_payloadData);
			}

		/**
		 * Returns the identifier of the channel the mix packet is supposed to
		 * be sent in.
		 *
		 * @return
		 */

		public int getChannelId()
			{
				return m_channelId;
			}

		public short getChannelFlags()
			{
				return m_channelFlags;
			}

		public void setChannelFlags(short a_channelFlags)
			{
				m_channelFlags = a_channelFlags;
			}

		/**
		 * Returns the internal payload data buffer without headers.
		 *
		 * Size of the internal buffer is equal to MixPacket.PAYLOAD_SIZE.
		 *
		 * @return The internal payload data buffer without headers.
		 */

		public byte[] getPayloadData()
			{
				return m_payloadData;
			}

		/**
		 * Returns the byte array representation of the mix packet with
		 * channel ID, channel flags and the payload. This gets encrypted and
		 * sent over the network.
		 *
		 * @see {@link anon.client.Multiplexer#sendPacket(MixPacket)}
		 * @return The byte representation of the mix packet.
		 */

		public byte[] getRawPacket()
			{
				byte[] rawPacket = new byte[PACKET_SIZE];
				rawPacket[0]=(byte)((m_channelId>>24) & 0x00FF);
				rawPacket[1]=(byte)((m_channelId>>16) & 0x00FF);
				rawPacket[2]=(byte)((m_channelId>>8) & 0x00FF);
				rawPacket[3]=(byte)((m_channelId) & 0x00FF);
				rawPacket[4]=(byte)((m_channelFlags>>8) & 0x00FF);
				rawPacket[5]=(byte)((m_channelFlags) & 0x00FF);
				System.arraycopy(m_payloadData, 0, rawPacket, HEADER_SIZE, PAYLOAD_SIZE);
				return rawPacket;
			}

		public Vector getSendCallbackHandlers()
			{
				return m_sendCallbackHandlers;
			}

	}
