/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.IOException;
import java.io.OutputStream;
import java.security.SecureRandom;
import java.util.Enumeration;
import java.util.Observable;
import java.util.Vector;

import anon.client.crypto.ControlChannelCipher;
import anon.client.crypto.ISymCipher;
import anon.util.Base64;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

/**
 * This class, using a table of active data and control channels, sends and receives
 * mix packets over a TCP socket connection to the first mix.
 * <br />
 * Channels use this class to send mix packets to the first mix and get notified
 * by it, when new packages arrive for them.
 *
 * @see {@link anon.client.Multiplexer#run()}
 * @see {@link anon.client.Multiplexer#sendPacket(MixPacket)}
 * @see {@link anon.client.AbstractDataChannel}
 * @see {@link anon.client.AbstractControlChannel}
 *
 * @author Stefan Lieske
 */
public class Multiplexer extends Observable implements Runnable
	{
		private Vector m_sendJobQueue;
		private Vector m_controlMessageQueue;

		private Object m_waitQueueObject;

		private ChannelTable m_channelTable;

		private InputStream m_inputStream;

		private OutputStream m_outputStream;

		private ISymCipher m_inputStreamCipher;

		private ISymCipher m_outputStreamCipher;

		private ControlChannelCipher m_controlchannelCiper;

		private Object m_internalEventSynchronization;

		private boolean m_bClosed = false;

		private boolean m_bWithIntegrityCheck;
		private boolean m_bDebug = false;

		public Multiplexer(InputStream a_inputStream, OutputStream a_outputStream, KeyExchangeManager a_keyExchangeManager,	SecureRandom a_channelIdGenerator)
			{
				m_internalEventSynchronization = new Object();
				m_sendJobQueue = new Vector();
				m_controlMessageQueue = new Vector();
				m_waitQueueObject = new Object();

				m_channelTable = new ChannelTable(new DefaultDataChannelFactory(a_keyExchangeManager, this),a_channelIdGenerator,a_keyExchangeManager.getMaxOpenChannels());
				m_inputStream = a_inputStream;
				m_inputStreamCipher = a_keyExchangeManager.getMultiplexerInputStreamCipher();
				m_outputStream = a_outputStream;
				m_outputStreamCipher = a_keyExchangeManager.getMultiplexerOutputStreamCipher();
				m_controlchannelCiper = a_keyExchangeManager.getControlChannelCipher();
				m_bWithIntegrityCheck = a_keyExchangeManager.isProtocolWithIntegrityCheck();
				m_bDebug = a_keyExchangeManager.isDebug();
				Thread downstreamThread = new Thread(this, "Multiplexer: Receive-Thread");
				downstreamThread.setDaemon(true);
				downstreamThread.start();
			}

		public boolean isSendingControlMessage()
			{
				return m_channelTable.isSendingControlMessage();
			}

		/***
		 * checks if this MixPacket could be sent right now, or if it needs to be
		 * delayed because of the upstream flow control. in the later case the call
		 * to this function will block until we packet could be sent.
		 * 
		 * @param a_mixPacket
		 *          the MixPacket in question
		 */
		/*private final void handleUpstreamFlowControl(MixPacket a_mixPacket)
			{
				AbstractAnonChannel channel = m_channelTable.getChannel(a_mixPacket.getChannelId());
				if(!(channel instanceof AbstractDataChannel)) //Upstream Flow control is currently only supported for the "old" protocol....
					{
						return;
					}
				AbstractDataChain chain=((AbstractDataChannel)channel).getDataChain();
				if(!(chain instanceof SingleChannelDataChain))//Upstream Flow control is currently only supported for the "old" protocol....
					{
						return;
					}
				SingleChannelDataChain dataChain=(SingleChannelDataChain)chain;
				///Todo: check if packet contains Download SEND_ME Flag --> if yes: do not count the paket!, i.e. send even if upstream flow control forbids to sent it...

			}
		*/

		/**
		 * checks if this MixPacket could be sent right now, or if it needs to be
		 * delayed because of the upstream flow control. in the later case the call
		 * to this function will block until the packet could be sent.
		 *
		 * @param a_mixPacket
		 *          the MixPacket in question
		 */
		public void sendPacket(MixPacket a_mixPacket) throws IOException
			{
				Object ownSynchronizationObject = new Object();
				boolean bIsControlChPacket = m_channelTable.isControlChannelId(a_mixPacket.getChannelId());
				Vector waitQueue = bIsControlChPacket ? m_controlMessageQueue : m_sendJobQueue;

//				LogHolder.log(LogLevel.INFO, LogType.NET, "Packet to send: (" + MixPacket.getPayloadSize() + ") " + Arrays.toString(a_mixPacket.getPayloadData()));

				/*	if(!bIsControlChPacket)
						{
								handleUpstreamFlowControl(a_mixPacket);
						}
					*/
				synchronized (ownSynchronizationObject)
					{
						boolean waitForAccess = false;

						synchronized (m_waitQueueObject)
							{
								if (!bIsControlChPacket)
									{
										if ((m_controlMessageQueue.size() > 0) || (m_sendJobQueue.size() > 0))
											{
												/* data channel packets have to wait until
												 * control channel packets are processed
												 */
												waitForAccess = true;
											}
									}
								else
									{
										if (m_controlMessageQueue.size() > 0)
											{
												/* control channel packets have higher priority and are processed
												 * before any data traffic can be transmitted so cost confirmations
												 * can be delivered on time in case of data congestion.
												 */
												waitForAccess = true;
												LogHolder.log(LogLevel.WARNING, LogType.NET, "Control channel congestion");
											}
									}
								waitQueue.addElement(ownSynchronizationObject);
							}
						if (waitForAccess)
							{
								try
									{
										ownSynchronizationObject.wait();
									}
								catch (InterruptedException e)
									{
										/* stop waiting, if we get interrupted and remove ourself from the send-queue */
										Object nextLockObject = null;
										synchronized (m_waitQueueObject)
											{
												waitQueue.removeElement(ownSynchronizationObject);

												/* just in this moment we should get notified -> notify the next waiting
												 * thread, if there is one
												 */
												/*control channel messages have higher priority */
												if (m_controlMessageQueue.size() > 0)
													{
														/* first wake up control channel packets */
														nextLockObject = m_controlMessageQueue.firstElement();
													}
												else if (m_sendJobQueue.size() > 0)
													{
														/* there are more threads waiting to send packets */
														nextLockObject = m_sendJobQueue.firstElement();
													}
											}
										if (nextLockObject != null)
											{
												synchronized (nextLockObject)
													{
														/* wake up the next of the waiting threads */
														nextLockObject.notify();
													}
											}
										/* now we have cleaned up everything -> throw the origin InterruptedException
										 * as an InterruptedIOException
										 */
										throw (new InterruptedIOException(e.toString()));
									}
							}
					}

				/* first call all SendCallbackHandlers to finalize the packet */
				Enumeration sendCallbackHandlers = a_mixPacket.getSendCallbackHandlers().elements();
				while (sendCallbackHandlers.hasMoreElements())
					{
						((ISendCallbackHandler) (sendCallbackHandlers.nextElement())).finalizePacket(a_mixPacket);
					}
				byte[] packetData = a_mixPacket.getRawPacket();
				/* do stream-encoding */
				if (m_outputStreamCipher != null)
					{
						m_outputStreamCipher.encryptAES1(packetData, 0, packetData, 0, 16);
					}
				/* now we can send the packet */
				try
					{
						m_outputStream.write(packetData);
						m_outputStream.flush();
						///Do packet Timing logging --> note: this is just for doing some kind of Web-Fingerprinting Attac Experiments and will be removed afterwards...
						LogHolder.log(LogLevel.DEBUG, LogType.TRANSPORT, "PacketSent at time: " + System.currentTimeMillis());
						synchronized (m_internalEventSynchronization)
							{
								setChanged();
								if (bIsControlChPacket)
									{
										/* we've sent a packet on a control channel */
										notifyObservers(PacketProcessedEvent.PPE_CONTROL_PACKET_SENT);
									}
								else
									{
										/* we've sent a packet on a data channel */
										notifyObservers(PacketProcessedEvent.PPE_DATA_PACKET_SENT);
									}
							}
					}
				finally
					{
						/* always wake up the next waiting thread */
						Object nextLockObject = null;

						synchronized (m_waitQueueObject)
							{
								/* remove our lock-object from the job-queue */
								waitQueue.removeElement(ownSynchronizationObject);

								/* first handle control channel packets that have higher priority */
								if (m_controlMessageQueue.size() > 0)
									{
										/* there are more threads waiting to send packets */
										nextLockObject = m_controlMessageQueue.firstElement();
									}
								else if (m_sendJobQueue.size() > 0)
									{
										/* if there no control channel packets to be processed wake
										 * up other Threads for data transmission
										 */
										nextLockObject = m_sendJobQueue.firstElement();
									}
								else
									{
										m_waitQueueObject.notify();
									}
							}
						if (nextLockObject != null)
							{
								synchronized (nextLockObject)
									{
										/* wake up the next of the waiting threads */
										nextLockObject.notify();
									}
							}
					}
			}

		protected void close()
			{
				m_bClosed = true;
			}

		/**
		 * This method continuously tries to read MixPackets from the
		 * InputStream, deciphers them, checks their channel ID and relays the
		 * packets to be processed by that channel.
		 *
		 */
		public void run()
			{
				int iPacketCounter = 0;
				try
					{
						while (true)
							{

								/* if something is available, we read packet from stream */
								MixPacket receivedPacket = new MixPacket(m_inputStream, m_inputStreamCipher);

								iPacketCounter++;
								int channelId = receivedPacket.getChannelId();

								if (m_bDebug)
									{
										LogHolder.log(LogLevel.DEBUG, LogType.NET,
												"AN.ON debug packet received (PacketCounter: " + iPacketCounter + ", Channel:" + channelId
														+ "): "
														+ Base64.encode(receivedPacket.getPayloadData(), 0, MixPacket.getPayloadSize(), false));
									}
								else
									{
										LogHolder.log(LogLevel.DEBUG, LogType.NET,"AN.ON packet received - Channel:" + channelId);
									}
								AbstractChannel channel = m_channelTable.getChannel(channelId);

								if (channel != null)
									{
										synchronized (m_internalEventSynchronization)
											{
												setChanged();
												if (m_channelTable.isControlChannelId(channelId))
													{
														// we've received a packet on a control channel
														notifyObservers(PacketProcessedEvent.PPE_CONTROL_PACKET_RECEIVED);
													}
												else
													{
														// we've received a packet on a data channel
														notifyObservers(PacketProcessedEvent.PPE_DATA_PACKET_RECEIVED);
													}
											}
										channel.processReceivedPacket(receivedPacket);
									}
								else
									{
										/* we don't know a channel with the specified ID - maybe it's already closed */
										if (LogHolder.isLogged(LogLevel.INFO, LogType.NET))
											{
												LogHolder.log(LogLevel.INFO, LogType.NET,
														"Received a packet for unknown channel '"
																+ Long.toString(((long) channelId) & 0x00000000FFFFFFl)
																+ "' Maybe we have already closed it and do not want to get more data for it.");
											}
										synchronized (m_internalEventSynchronization)
											{
												setChanged();
												if (m_channelTable.isControlChannelId(channelId))
													{
														/* we've discarded a control channel packet */
														notifyObservers(PacketProcessedEvent.PPE_CONTROL_PACKET_DISCARDED);
													}
												else
													{
														/* we've discarded a data channel packet */
														notifyObservers(PacketProcessedEvent.PPE_DATA_PACKET_DISCARDED);
													}
											}
									}
								//Thread.yield();
							} // end of the while true
					}
				catch (IOException e)
					{
						/* end of input stream handling */
						if (m_bClosed)
							{
								if (LogHolder.isLogged(LogLevel.NOTICE, LogType.NET))
									{
										LogHolder.log(LogLevel.NOTICE, LogType.NET, Thread.currentThread().getName() + ": terminated!", e);
									}
							}
						else
							{
								LogHolder.log(LogLevel.EXCEPTION, LogType.NET, Thread.currentThread().getName() + ": terminated!", e);
							}
					}

				/* close the channel-table (notifies also all open channels) */
				m_channelTable.closeChannelTable();
			}

		public ChannelTable getChannelTable()
			{
				return m_channelTable;
			}

		public ControlChannelCipher getControlChannelCipher()
			{
				return m_controlchannelCiper;
			}

		public boolean isProtocolWithIntegrityCheck()
			{
				return m_bWithIntegrityCheck;
			}

		public boolean isDebug()
			{
				return m_bDebug;
			}

	}
