/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import anon.ErrorCodes;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

/**
 * @author Stefan Lieske
 */
final public class SingleChannelDataChain extends AbstractDataChain
	{

		private static final int FLAG_FLOW_CONTROL = 0x08000;

		private static final int CLOSE_CELL_CONNECTION_ERROR = 1;

		private static final short FLAG_INTEGRITY_ERROR = (short) 0x4000;

		private static final int CHAIN_HEADER_LEN = 3;

		private int m_chainType;

		private boolean m_supportFlowControl;
		private boolean m_supportUpstreamFlowControl;

		private AbstractDataChannel m_associatedChannel;

		private boolean m_firstUpstreamPacket;

		private int m_downstreamSendMeCount;
		private int m_upstreamSendMeCount;
		private int m_downstreamSendMeLimit;
		private int m_upstreamSendMeLimit;

		private Object m_oSyncUpstreamFlowControl; // /Synchronization Object used
																								// for upstream flow control;

		//private boolean m_bEnhancedChannelEncryption;
		private boolean m_bWithIntegrityCheck;

		private class InvalidChainCellException extends Exception
			{

				/**
			 *
			 */
			private static final long serialVersionUID = 1L;

				public InvalidChainCellException(String a_message)
					{
						super(a_message);
					}

			}

		final private class ChainCell
			{

				private static final int DATALENGTH_MASK = 0x03FF;

				// private static final short FLAG_CONNECTION_ERROR = (short) 0x4000;

				private byte[] m_rawData; // payload header + payload

				private int m_payloadLen;
				private int m_payloadType;

				private boolean m_flowControlFlagSet;
				private boolean m_integrityErrorFlagSet;

				// private boolean m_connectionErrorFlagSet;

				public ChainCell(byte[] a_rawData) throws InvalidChainCellException
					{
						if (a_rawData.length < CHAIN_HEADER_LEN)
							{
								throw (new InvalidChainCellException(
										"SingleChannelDataChain: ChainCell: Constructor: Length of ChainCell must be at least 3 bytes."));
							}
						int lengthAndFlagsField = (a_rawData[0] << 8 | (a_rawData[1] & 0x0FF)) & 0x0000FFFF;
						m_payloadType = a_rawData[2] & 0x00FF;
						m_flowControlFlagSet = false;
						m_integrityErrorFlagSet = false;
						// m_connectionErrorFlagSet = false;
						int flags = lengthAndFlagsField & (~DATALENGTH_MASK);
						if (m_supportFlowControl)
							{
								if ((flags & FLAG_FLOW_CONTROL) == FLAG_FLOW_CONTROL)
									{
										m_flowControlFlagSet = true;
									}
							}
						if ((flags & FLAG_INTEGRITY_ERROR) == FLAG_INTEGRITY_ERROR)
							{
								m_integrityErrorFlagSet = true;
							}
						// if ( (flags & FLAG_CONNECTION_ERROR) == FLAG_CONNECTION_ERROR)
						// {
						// m_connectionErrorFlagSet = true;
						// }
						m_payloadLen = lengthAndFlagsField & DATALENGTH_MASK;
						/*
						 * data is starting at byte 3 (0 and 1 are length and flags, 2 is
						 * the type and can be ignored)
						 */
						int dataOffset = CHAIN_HEADER_LEN;
						if (dataOffset + m_payloadLen > a_rawData.length)
							{
								throw (new InvalidChainCellException(	"SingleChannelDataChain: ChainCell: Constructor: ChainCell has invalid length-field."));
							}
						m_rawData=a_rawData;
					}

				public byte[] getRawData()
					{
						return m_rawData;
					}

				public int getPayloadType()
					{
						return m_payloadType;
					}

				public int getPayloadLength()
					{
						return m_payloadLen;
					}

				public boolean isFlowControlFlagSet()
					{
						return m_flowControlFlagSet;
					}

				public boolean isIntegrityErrorFlagSet()
					{
						return m_integrityErrorFlagSet;
					}

				// public boolean isConnectionErrorFlagSet()
				// {
				// return m_connectionErrorFlagSet;
				// }
			}


		private InternalChannelMessageQueue m_internalChannelMessageQueue;

		public SingleChannelDataChain(IDataChannelCreator a_channelCreator, DataChainErrorListener a_errorListener,
				IntegrityErrorListener a_integrityErrorListener, int a_chainType, boolean a_supportFlowControl,
				boolean a_supportUpstreamFlowControl, int upstreamSendMe, int downstreamSendMe,
				boolean a_enhancedChannelEncryption, boolean a_withIntegrityCheck)
			{
				super(a_channelCreator, a_errorListener, a_integrityErrorListener);
				m_chainType = a_chainType;
				m_supportFlowControl = a_supportFlowControl;
				m_supportUpstreamFlowControl = a_supportUpstreamFlowControl;
				//m_bEnhancedChannelEncryption = a_enhancedChannelEncryption;
				m_bWithIntegrityCheck = a_withIntegrityCheck;
				/* create the channel */
				m_associatedChannel = createDataChannel();
				/* observe the channel-message queue */
				//m_associatedChannel.getChannelMessageQueue().addObserver(this);
				m_internalChannelMessageQueue=m_associatedChannel.getChannelMessageQueue();
				m_firstUpstreamPacket = true;
				m_downstreamSendMeCount = 0;
				m_downstreamSendMeLimit = downstreamSendMe;
				m_upstreamSendMeCount = 0;
				m_upstreamSendMeLimit = upstreamSendMe;
				m_oSyncUpstreamFlowControl = new Object();
			}

		public int getOutputBlockSize()
			{
				int channelPacketSize = 0;
				synchronized (m_associatedChannel)
					{
						channelPacketSize = m_associatedChannel.getNextPacketRecommandedOutputBlocksize();
					}
				return Math.max(0, channelPacketSize - CHAIN_HEADER_LEN);
			}
		/**
		 * Takes the order data of the given order, adds the length, the flow
		 * control flag and chain type to the front and copies that data to the
		 * channel cell buffer of the order.
		 */

		public void createPacketPayload(DataChainSendOrderStructure a_order)
			{
				if (a_order.getOrderData() != null)
					{
						// figure out the length of the payload
						// -3, 2 for the length and 1 for the chain type
					  // channel cell is filled with max. amount of random bytes right now
						int dataLength = Math.min(a_order.getOrderData().length, a_order.getChannelCell().length - CHAIN_HEADER_LEN);
						int dataLengthField = dataLength;

						if (m_supportFlowControl)
							{
								/* check whether we shall send a flow-control flag */
								if (a_order.getAdditionalProtocolData() instanceof Boolean)
									{
										if (((Boolean) (a_order.getAdditionalProtocolData())).booleanValue())
											{
												/* we shall send a packet with flow-control flag */
												dataLengthField = dataLengthField | FLAG_FLOW_CONTROL;
											}
									}
							}

						// construct the packet, length, chain type, data
						ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
						DataOutputStream dataStream = new DataOutputStream(byteStream);
						try
							{
								dataStream.writeShort(dataLengthField);
								dataStream.flush();
								if (m_firstUpstreamPacket)
									{
										byteStream.write(m_chainType);
										m_firstUpstreamPacket = false;
									}
								else
									{
										byteStream.write(0);
									}
								byteStream.write(a_order.getOrderData(), 0, dataLength);
								byteStream.flush();
							}
						catch (IOException e)
							{
								/* cannot happen */
							}

						// copy to order and set number of processed bytes
						System.arraycopy(byteStream.toByteArray(), 0, a_order.getChannelCell(), 0, byteStream.toByteArray().length);
						a_order.setProcessedBytes(dataLength);
					}
			}

		public void run()
			{
				//Vector messageQueuesNotifications = getMessageQueuesNotificationsList();
				try
					{
						while(m_internalChannelMessageQueue==null)//startup race dealy...
							Thread.sleep(10);
						while (!Thread.interrupted())
							{
								InternalChannelMessage currentMessage = null;
								/*InternalChannelMessageQueue messageQueue = null;
								synchronized (messageQueuesNotifications)
									{
										while (messageQueuesNotifications.size() == 0)
											{
												messageQueuesNotifications.wait();
											}
										messageQueue = (InternalChannelMessageQueue) (messageQueuesNotifications.firstElement());
										currentMessage = messageQueue.getFirstMessage();
										//
										// we will process the message immediately -> remove it from
										// the queue
										//
										messageQueue.removeFirstMessage();
										// remove also the notification //
										messageQueuesNotifications.removeElementAt(0);
									}
									*/
								currentMessage=m_internalChannelMessageQueue.waitForNextMessage();
								byte[] packetData=currentMessage.getMessageData();
								if (packetData!=null&&!currentMessage.getMixCipherChain().decryptPacket(packetData))
									{
										closeDataChain();
										propagateIntegrityError(ErrorCodes.E_INTEGRITY_ERROR_DOWNSTREAM);
									}
								else
									{

										switch (currentMessage.getMessageCode())
											{
												case InternalChannelMessage.CODE_PACKET_RECEIVED:
													{
														try
															{
																m_downstreamSendMeCount++;
																ChainCell dataCell = new ChainCell(packetData);
																if (m_supportFlowControl && /*
																														 * dataCell.
																														 * isFlowControlFlagSet
																														 * ()
																														 */m_downstreamSendMeCount >= m_downstreamSendMeLimit)
																	{
																		LogHolder.log(LogLevel.DEBUG, LogType.NET,
																				"FlowControl: Will sent sendme - and download packet counter is: "
																						+ m_downstreamSendMeCount);
																		/*
																		 * send a response-packet with the flow-control
																		 * flag
																		 */
																		DataChainSendOrderStructure order = new DataChainSendOrderStructure(new byte[0]);
																		order.setAdditionalProtocolData(new Boolean(true));
																		orderPacket(order);
																		m_downstreamSendMeCount = 0;
																	}
																// / Temporary check to see if we receive the
																// Upstream SENDME ack packets
																if (m_supportUpstreamFlowControl && dataCell.isFlowControlFlagSet())
																	{
																		synchronized (this.m_oSyncUpstreamFlowControl)
																			{
																				LogHolder.log(LogLevel.DEBUG, LogType.NET,
																						"got sendme - and upstream packet counter is: " + m_upstreamSendMeCount);
																				m_upstreamSendMeCount = Math.max(0,
																						m_upstreamSendMeCount - this.m_upstreamSendMeLimit);
																				m_oSyncUpstreamFlowControl.notifyAll();
																			}
																	}

																/* add data to the datastream, skip chain header */
																addInputStreamQueueEntry(new DataChainInputStreamQueueEntry(
																		DataChainInputStreamQueueEntry.TYPE_DATA_AVAILABLE, dataCell.getRawData(),CHAIN_HEADER_LEN,dataCell.getPayloadLength()));
															}
														catch (InvalidChainCellException e)
															{
																addInputStreamQueueEntry(
																		new DataChainInputStreamQueueEntry(new IOException(e.toString())));
															}
														break;
													}
												case InternalChannelMessage.CODE_CHANNEL_CLOSED:
													{
														addInputStreamQueueEntry(new DataChainInputStreamQueueEntry(DataChainInputStreamQueueEntry.TYPE_STREAM_END));

														try
															{
																if (packetData != null)
																	{
																		ChainCell dataCell = new ChainCell(packetData);
																		if (dataCell.getPayloadLength() == 0
																				&& dataCell.getPayloadType() == CLOSE_CELL_CONNECTION_ERROR)
																			{
																				addInputStreamQueueEntry(new DataChainInputStreamQueueEntry(new IOException(
																						"SingleChannelDataChain: run(): Last mix signaled connection error.")));
																				propagateConnectionError();
																			}
																		if (dataCell.isIntegrityErrorFlagSet() && m_bWithIntegrityCheck)
																			{
																				propagateIntegrityError(ErrorCodes.E_INTEGRITY_ERROR_UPSTREAM);
																			}
																	}
															}
														catch (InvalidChainCellException e)
															{
																addInputStreamQueueEntry(
																		new DataChainInputStreamQueueEntry(new IOException(e.toString())));
															}
														/* stop observing the message-queue of the channel */
														//messageQueue.deleteObserver(this);
														/*
														 * interrupt this thread because the only channel of the
														 * DataChain was closed
														 */
														Thread.currentThread().interrupt();
														break;
													}
												case InternalChannelMessage.CODE_CHANNEL_EXCEPTION:
													{
														addInputStreamQueueEntry(new DataChainInputStreamQueueEntry(new IOException(
																"SingleChannelDataChain: run(): Channel signaled an exception - closing chain.")));
														break;
													}
											}
									}
							}
					}
				catch (InterruptedException e)
					{
					}

			}

		protected void orderPacket(DataChainSendOrderStructure a_order)
			{
				if (m_supportUpstreamFlowControl)
					{
						// Check if this packet has the download SENDME FLOW CONTROL
						// FLAG set --> if so do not count and sent it immediately
						if (!(a_order.getAdditionalProtocolData() instanceof Boolean)
								|| !((Boolean) (a_order.getAdditionalProtocolData())).booleanValue())
							{

								synchronized (m_oSyncUpstreamFlowControl)
									{
										while (m_upstreamSendMeCount > 2 * this.m_upstreamSendMeLimit && !(this.isClosed()))
											{
												try
													{
														this.m_oSyncUpstreamFlowControl.wait();
													}
												catch (Exception e)
													{
													}
											}
										m_upstreamSendMeCount++; // Do not know yet if this is
																							// really the right place...
									}

							}
					}
				synchronized (m_associatedChannel)
					{

						m_associatedChannel.processSendOrder(a_order);
					}
			}

		protected void outputStreamClosed() throws IOException
			{
				/* close the whole data-chain */
				close();
			}

		protected void closeDataChain()
			{
				synchronized (m_associatedChannel)
					{
						try
							{
								m_associatedChannel.organizeChannelClose();
							}
						catch (IOException e)
							{
								/*
								 * channel is locally closed in every case -> IOException
								 * doesn't matter
								 */
							}
					}
				synchronized (this.m_oSyncUpstreamFlowControl)
					{
						this.m_oSyncUpstreamFlowControl.notifyAll();
					}
			}

	}
