/*
 Copyright (c) 2000 - 2006, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

  - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.client;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Observable;
import java.util.Observer;
import java.lang.reflect.Method;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import anon.error.ServiceSignatureException;
import anon.error.TrustException;
import anon.infoservice.BlacklistedCascadeIDEntry;
import anon.infoservice.Database;
import anon.infoservice.IServiceContextContainer;
import anon.infoservice.ListenerInterface;
import anon.infoservice.MixCascade;
import anon.infoservice.MixInfo;
import anon.infoservice.ServiceOperator;
import anon.infoservice.StatusInfo;
import anon.infoservice.PerformanceEntry;
import anon.infoservice.PerformanceInfo;
import anon.pay.PayAccountsFile;
import anon.pay.PaymentInstanceDBEntry;
import anon.util.IXMLEncodable;
import anon.util.JAPMessages;
import anon.util.XMLParseException;
import anon.util.XMLUtil;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;


/**
 * This is the general trust model for JAP.
 *
 * @author Rolf Wendolsky
 */
public class TrustModel extends BasicTrustModel implements IXMLEncodable
{
	public static final Integer NOTIFY_TRUST_MODEL_CHANGED = new Integer(0);
	public static final Integer NOTIFY_TRUST_MODEL_ADDED = new Integer(1);
	public static final Integer NOTIFY_TRUST_MODEL_REMOVED = new Integer(2);

	/**
	 * Unreserved IDs may be used by controller classes to define more models.
	 * IDs lower than FIRST_UNRESERVED_MODEL_ID may not be used.
	 */
	public static final int FIRST_UNRESERVED_MODEL_ID = 5;
	
	public static final String XML_ELEMENT_NAME = "TrustModel";
	public static final String XML_ELEMENT_CONTAINER_NAME = "TrustModels";
	private static final String XML_ATTR_CURRENT_TRUST_MODEL = "currentTrustModel";
	private static final String XML_ATTR_NAME = "name";

	public static final int TRUST_MODEL_DEFAULT_INDEX = 0;
	

	public static final TrustModel TRUST_MODEL_ALL_SERVICES;

	private static TrustModel TRUST_MODEL_CUSTOM_FILTER;
	
	private static MixCascade ms_cascadeWhitelisted = null;
	
	/**
	 * Always trust the cascade, regardless of the attribute
	 */
	public static final int TRUST_ALWAYS = 0;

	/**
	 * Only trust the cascade if the attribute is NOT true
	 */
	public static final int TRUST_IF_NOT_TRUE = 1;

	/**
	 * Only trust the cascade if the attribute is true
	 */
	public static final int TRUST_IF_TRUE = 2;

	/**
	 * Only trust the cascade if the attribute is greater than a specified value
	 */
	public static final int TRUST_IF_AT_LEAST = 3;

	/**
	 * Only trust the cascade if the attribute is lesser than a specified value
	 */
	public static final int TRUST_IF_AT_MOST = 5;

	/**
	 * Only trust the cascade if the attribute value is not within the specified list
	 */
	public static final int TRUST_IF_NOT_IN_LIST = 6;

	/**
	 * Reserved for future use
	 */
	public static final int TRUST_RESERVED = 7;

	private static final String MSG_SERVICES_WITH_COSTS = TrustModel.class.getName() + "_servicesWithCosts";
	public static final String MSG_SERVICES_WITHOUT_COSTS = TrustModel.class.getName() + "_servicesWithoutCosts";
	private static final String MSG_SERVICES_USER_DEFINED = TrustModel.class.getName() + "_servicesUserDefined";
	private static final String MSG_CASCADES_FILTER = TrustModel.class.getName() + "_servicesFilter";
	private static final String MSG_ALL_SERVICES = TrustModel.class.getName() + "_allServices";
	private static final String MSG_DEFAULT_FILTER = TrustModel.class.getName() + ".defaultFilter";
	private static final String MSG_SERVICES_BUSINESS = TrustModel.class.getName() + "_servicesBusiness";
	private static final String MSG_SERVICES_TEST = TrustModel.class.getName() + "_servicesTest";
	private static final String MSG_SERVICES_PREMIUM_PRIVATE = TrustModel.class.getName() + "_servicesPremiumPrivate";
	public static final String MSG_PI_UNAVAILABLE = TrustModel.class.getName() + "_piUnavailable";
	public static final String MSG_BLACKLISTED = TrustModel.class.getName() + "_blacklisted";
	
	private static final String MSG_EXCEPTION_NO_SOCKS = TrustModel.class.getName() + "_exceptionNoSocks";
	private static final String MSG_EXCEPTION_DATA_RETENTION = TrustModel.class.getName() + "_exceptionDataRetention";
	private static final String MSG_EXCEPTION_PAY_CASCADE = TrustModel.class.getName() + "_exceptionPayCascade";
	public static final String MSG_EXCEPTION_FREE_CASCADE = TrustModel.class.getName() + "_exceptionFreeCascade";
	private static final String MSG_EXCEPTION_WRONG_SERVICE_CONTEXT = TrustModel.class.getName() + "_wrongServiceContext";
	public static final String MSG_EXCEPTION_NOT_ENOUGH_MIXES = TrustModel.class.getName() + "_exceptionNotEnoughMixes";
	public static final String MSG_EXCEPTION_TOO_MANY_MIXES = TrustModel.class.getName() + "_exceptionTooManyMixes";
	private static final String MSG_EXCEPTION_EXPIRED_CERT = TrustModel.class.getName() + "_exceptionExpiredCert";
	private static final String MSG_EXCEPTION_NOT_USER_DEFINED = TrustModel.class.getName() + "_exceptionNotUserDefined";
	private static final String MSG_EXCEPTION_TOO_FEW_COUNTRIES = TrustModel.class.getName() + "_exceptionTooFewCountries";
	private static final String MSG_EXCEPTION_TOO_MANY_COUNTRIES = TrustModel.class.getName() + "_exceptionTooManyCountries";
	private static final String MSG_EXCEPTION_NOT_INTERNATIONAL = TrustModel.class.getName() + "_exceptionNotInternational";
	private static final String MSG_EXCEPTION_INTERNATIONAL = TrustModel.class.getName() + "_exceptionInternational";
	private static final String MSG_EXCEPTION_NOT_ENOUGH_ANON = TrustModel.class.getName() + "_exceptionNotEnoughAnon";
	private static final String MSG_EXCEPTION_BLACKLISTED = TrustModel.class.getName() + "_exceptionBlacklisted";
	private static final String MSG_EXCEPTION_NOT_ENOUGH_SPEED = TrustModel.class.getName() + "_exceptionNotEnoughSpeed";
	private static final String MSG_EXCEPTION_RESPONSE_TIME_TOO_HIGH = TrustModel.class.getName() + "_exceptionResponseTimeTooHigh";	
	
	
	private static Vector ms_trustModels = new Vector();
	private static TrustModel ms_currentTrustModel;
	
	private static String ms_strContext = IServiceContextContainer.CONTEXT_JONDONYM;
	
	private static boolean ms_bFreeAllowed = true;

	public static class InnerObservable extends Observable
	{
		public void setChanged()
		{
			super.setChanged();
		}
	}
	private static InnerObservable ms_trustModelObservable = new InnerObservable();

	
	private Hashtable m_trustAttributes = new Hashtable();

	private String m_strName;
	private long m_id;

	private boolean m_bEditable;

	public static abstract class TrustAttribute implements IXMLEncodable
	{
		public static final int CATEGORY_DEFAULT = 0;

		public static final String XML_ELEMENT_NAME = "TrustAttribute";
		public static final String XML_VALUE_ELEMENT_NAME = "ConditionValue";
		public static final String XML_VALUE_CONTAINER_ELEMENT_NAME = "ConditionValueList";
		public static final String XML_ATTR_NAME = "name";
		public static final String XML_ATTR_TRUST_CONDITION = "trustCondition";
		public static final String XML_ATTR_CONDITION_VALUE = "conditonValue";
		public static final String XML_ATTR_IGNORE_NO_DATA = "ignoreNoData";
		
		private int m_category;
		private boolean m_bIgnoreNoDataAvailable;
		private int m_trustCondition;
		private Object m_conditionValue;

		protected TrustAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			m_trustCondition = a_trustCondition;
			m_conditionValue = a_conditionValue;
			m_category = CATEGORY_DEFAULT;
			m_bIgnoreNoDataAvailable = a_bIgnoreNoDataAvailable;
		}
		
		public static int getDefaultCondition()
		{
			return TRUST_ALWAYS;
		}
		
		public static Object getDefaultValue()
		{
			return new Integer(0);
		}
		
		public boolean isNoDataIgnored()
		{
			return m_bIgnoreNoDataAvailable;
		}
		
		public final int getCategory()
		{
			return m_category;
		}

		public int getTrustCondition()
		{
			return m_trustCondition;
		}

		public Object getConditionValue()
		{
			return m_conditionValue;
		}

		public boolean isTrusted(MixCascade a_cascade)
		{
			try
			{
				checkTrust(a_cascade);
				return true;
			}
			catch(TrustException a_e)
			{
				// ignore
			}
			catch (ServiceSignatureException a_e)
			{
				// ignore
			}
			return false;						
		}
		
		public abstract void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException;
		
		public Element toXmlElement(Document a_doc)
		{
			if(a_doc == null) return null;

			Element el = a_doc.createElement(XML_ELEMENT_NAME);
			XMLUtil.setAttribute(el, XML_ATTR_NAME, this.getClass().getName());
			XMLUtil.setAttribute(el, XML_ATTR_TRUST_CONDITION, m_trustCondition);
			XMLUtil.setAttribute(el, XML_ATTR_IGNORE_NO_DATA, m_bIgnoreNoDataAvailable);

			if (m_conditionValue instanceof Integer)
			{
				XMLUtil.setAttribute(el, XML_ATTR_CONDITION_VALUE, ((Integer) m_conditionValue).intValue());
			}
			else if (m_conditionValue instanceof Vector)
			{
				Vector vec = (Vector) m_conditionValue;
				Element list = a_doc.createElement(XML_VALUE_CONTAINER_ELEMENT_NAME);
				for(int i = 0; i < vec.size(); i++)
				{
					Element e = a_doc.createElement(XML_VALUE_ELEMENT_NAME);
					XMLUtil.setValue(e, ((ServiceOperator) vec.elementAt(i)).getId());
					list.appendChild(e);
				}
				el.appendChild(list);
			}

			return el;
		}

		public static TrustAttribute fromXmlElement(Element a_e) throws XMLParseException
		{
			if(a_e == null) return null;

			XMLUtil.assertNodeName(a_e, XML_ELEMENT_NAME);
			XMLUtil.assertNotNull(a_e, XML_ATTR_NAME);

			String name = XMLUtil.parseAttribute(a_e, XML_ATTR_NAME, null);
			int trustCondition = XMLUtil.parseAttribute(a_e, XML_ATTR_TRUST_CONDITION, TRUST_ALWAYS);
			int conditionValue = XMLUtil.parseAttribute(a_e, XML_ATTR_CONDITION_VALUE, 0);
			boolean bIgnoreNoData = XMLUtil.parseAttribute(a_e, XML_ATTR_IGNORE_NO_DATA, false);

			TrustAttribute attr;

			try
			{
				Object value = null;

				if (trustCondition == TRUST_IF_NOT_IN_LIST)
				{
					Node n = XMLUtil.getFirstChildByName(a_e, XML_VALUE_CONTAINER_ELEMENT_NAME);
					XMLUtil.assertNotNull(n);

					NodeList list = n.getChildNodes();
					value = new Vector();

					for (int i = 0; i < list.getLength(); i++)
					{
						// look for a matching ServiceOperator database entry
						ServiceOperator op = (ServiceOperator) Database.getInstance(ServiceOperator.class).getEntryById(XMLUtil.parseValue(list.item(i),null));
						if(op != null) ((Vector) value).addElement(op);
					}
				}
				else
				{
					value = new Integer(conditionValue);
				}
				
				if (name != null && name.startsWith("jap.TrustModel"))
				{
					// backwards compatibility
					name = TrustModel.class.getName() + name.substring(("jap.TrustModel").length(), name.length());
				}
				attr = getInstance(Class.forName(name), trustCondition, value, bIgnoreNoData);
				if (attr == null)
				{
					throw new XMLParseException(XML_ELEMENT_NAME, "Could not create TrustAttribute + " + name + "!");
				}
			}
			catch(Exception ex)
			{
				LogHolder.log(LogLevel.NOTICE, LogType.DB, ex);
				throw new XMLParseException(XML_ELEMENT_NAME, ex.getMessage());
			}
			
			return attr;
		}
		
		public static TrustAttribute getInstance(Class a_attr, int a_trustCondition, Object a_conditionValue, 
				boolean a_bIgnoreNoDataAvailable)
		{
			try
			{
				return (TrustAttribute) a_attr.getConstructor(new Class[] { int.class, Object.class, boolean.class })
			 	.newInstance(new Object[] {new Integer(a_trustCondition), a_conditionValue, new Boolean(a_bIgnoreNoDataAvailable)});
			}
			catch(Exception ex) 
			{
				LogHolder.log(LogLevel.EXCEPTION, LogType.MISC, "Could not create " + a_attr);
				return null;
			}			
		}
	}

	public static class ContextAttribute extends TrustAttribute
	{
		public ContextAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			String strContextCascade = a_cascade.getContext();
			String strContextModel = ms_strContext;
			
			if (getTrustCondition() == TRUST_IF_TRUE && !strContextCascade.equals(strContextModel))
			{
				throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_WRONG_SERVICE_CONTEXT));
			}
			else if (getTrustCondition() == TRUST_IF_NOT_TRUE && strContextCascade.equals(strContextModel))
			{
				throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_WRONG_SERVICE_CONTEXT));
			}
			else if (!(strContextCascade.equals(strContextModel) ||
					(strContextModel.startsWith(IServiceContextContainer.CONTEXT_JONDONYM) &&
					strContextCascade.equals(IServiceContextContainer.CONTEXT_JONDONYM_PREMIUM))))
			{
				throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_WRONG_SERVICE_CONTEXT));
			}
		}
	};
	
	public static class ForcePremiumIfExistingUserAttribute extends TrustAttribute
	{
		public ForcePremiumIfExistingUserAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}
		
		public static int getDefaultCondition()
		{
			return TRUST_ALWAYS;
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{				
			if (getTrustCondition() == TRUST_IF_TRUE)
			{
				if (a_cascade.isPayment())
				{
					if (a_cascade.getNumberOfOperators() < 2)
					{
						throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NOT_ENOUGH_MIXES) + ": " + a_cascade.getNumberOfOperators());
					}
					
					if (!a_cascade.isSocks5Supported())
					{
						throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NO_SOCKS));
					}		
				}
				else if (!a_cascade.isPayment() && !PayAccountsFile.getInstance().isNewUser())
				{
					throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_FREE_CASCADE));
				}
			}
		}
	};
	
	public static class ForcePremiumIfChargedAccountAttribute extends TrustAttribute
	{
		public ForcePremiumIfChargedAccountAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}
		
		public static int getDefaultCondition()
		{
			return TRUST_IF_TRUE;
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{				
			if (getTrustCondition() == TRUST_IF_TRUE)
			{
				if (a_cascade.isPayment())
				{
					/*
					if (a_cascade.getNumberOfOperators() < 2)
					{
						throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NOT_ENOUGH_MIXES));
					}
					*/
					/*
					if (!a_cascade.isSocks5Supported())
					{
						throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NO_SOCKS));
					}*/				
				}
				else if (!a_cascade.isPayment())
				{
					Vector vecEntries = Database.getInstance(PaymentInstanceDBEntry.class).getEntryList();
					PaymentInstanceDBEntry entry;
					
					for (int i = 0; i < vecEntries.size(); i++)
					{
						entry = (PaymentInstanceDBEntry)vecEntries.elementAt(i);
						if (!entry.isTest() && 
							PayAccountsFile.getInstance().getChargedAccount(entry.getId()) != null)
						{
							throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_FREE_CASCADE));
						}
					}
				}
			}
		}
	};
	
	
	public static class PremiumChargedAccountAttribute extends TrustAttribute
	{
		public PremiumChargedAccountAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{	
			if (getTrustCondition() != TRUST_ALWAYS && a_cascade.isPayment())
			{
				if (PayAccountsFile.getInstance().getChargedAccount(a_cascade.getPIID()) == null)
				{
					throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_PAY_CASCADE));
				}
			}
		}
	};
	
	
	public static class UsePremiumAttribute extends TrustAttribute
	{
		public UsePremiumAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			if (a_cascade.isPayment())
			{
				if (getTrustCondition() == TRUST_IF_NOT_TRUE)
				{
					throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_PAY_CASCADE));
				}
			}
			else if (getTrustCondition() == TRUST_IF_TRUE)
			{
				throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_FREE_CASCADE));
			}
		}
	};
	
	public static class DataRetentionAttribute extends TrustAttribute
	{
		public DataRetentionAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			if (a_cascade.getDataRetentionInformation() != null && getTrustCondition() == TRUST_IF_NOT_TRUE)
			{
				throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_DATA_RETENTION));
			}
		}
	};
	
	
	public static class SocksAttribute extends TrustAttribute
	{
		public SocksAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			if (!a_cascade.isSocks5Supported() && getTrustCondition() == TRUST_IF_TRUE)
			{
				throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NO_SOCKS));
			}
		}
	};

	public static class NumberOfMixesAttribute extends TrustAttribute
	{
		public NumberOfMixesAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}
		
		public static int getDefaultCondition()
		{
			return TRUST_ALWAYS;
		}
		
		public static Object getDefaultValue()
		{
			return new Integer(1);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			Object conditionValue = getConditionValue();
			
			if (conditionValue == null)
			{
				return;
			}
			
			int minMixes = getIntegerConditionValue(conditionValue);
			
			if (getTrustCondition() == TRUST_IF_AT_LEAST && (a_cascade == null ||  
				a_cascade.getNumberOfOperators() < minMixes))
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NOT_ENOUGH_MIXES) + ": " + a_cascade.getNumberOfOperators()));
			}
			else if (getTrustCondition() == TRUST_IF_AT_MOST && a_cascade != null &&  
				a_cascade.getNumberOfOperators() > minMixes)
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_TOO_MANY_MIXES) + ": " + a_cascade.getNumberOfOperators()));
			}
		}
	};

	public static class UserDefinedAttribute extends TrustAttribute
	{
		public UserDefinedAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}
		
		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			if (getTrustCondition() == TRUST_IF_TRUE)
			{
				if (a_cascade.isUserDefined())
				{
					//if (a_cascade.getCertificate() == null && (a_cascade.getNumberOfMixes() == 0 ||
					//	a_cascade.getNumberOfMixes() == 1 &&
					//	(a_cascade.getMixInfo(0) == null || a_cascade.getMixInfo(0).getCertificate() == null)))
					{
						// not yet connected; do not make further tests
						return;
					}
				}
				else
				{
					throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NOT_USER_DEFINED));
				}
			}
		}
	};

	public static class InternationalAttribute extends TrustAttribute
	{
		public InternationalAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			super(a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			Object conditionValue = getConditionValue();
			
			if (conditionValue == null)
			{
				return;
			}
			
			if (getTrustCondition() == TRUST_IF_AT_LEAST && a_cascade.getNumberOfCountries() < getIntegerConditionValue(conditionValue))
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_TOO_FEW_COUNTRIES) + ": " + a_cascade.getNumberOfCountries()));
			}
			else if (getTrustCondition() == TRUST_IF_AT_MOST && a_cascade.getNumberOfCountries() > getIntegerConditionValue(conditionValue))
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_TOO_MANY_COUNTRIES) + ": " + a_cascade.getNumberOfCountries()));
			}
		}
	};

	/*public static class NewAttribute extends TrustAttribute
	{
		public NewAttribute(int a_trustCondition, Object a_conditionValue)
		{
			super(a_trustCondition, a_conditionValue);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, SignatureException
		{
			if (m_trustCondition == TRUST_IF_TRUE)
			{
				if (Database.getInstance(NewCascadeIDEntry.class).getEntryById(
					 a_cascade.getMixIDsAsString()) == null)
				{
					throw (new TrustException("only new cascades are accepted!"));
				}
			}
			else if(m_trustCondition == TRUST_IF_NOT_TRUE)
			{
				if (Database.getInstance(NewCascadeIDEntry.class).getEntryById(a_cascade.getMixIDsAsString()) != null)
				{
					throw (new TrustException("No new cascades are allowed"));
				}
			}
		}
	};*/

	public static class AnonLevelAttribute extends TrustAttribute
	{
		public AnonLevelAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			// MUST always be TRUST_IF_AT_LEAST
			super(TRUST_IF_AT_LEAST, a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			Object conditionValue = getConditionValue();
			
			StatusInfo info = (StatusInfo) Database.getInstance(StatusInfo.class).getEntryById(a_cascade.getId());
			if(getTrustCondition() == TRUST_IF_AT_LEAST && (info == null || 
				(conditionValue != null && info.getAnonLevel() < getIntegerConditionValue(conditionValue))))
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NOT_ENOUGH_ANON)));
			}
		}
	}

	public static class OperatorBlacklistAttribute extends TrustAttribute
	{
		public OperatorBlacklistAttribute(int a_trustCondtion, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			// MUST always be TRUST_IF_NOT_IN_LIST and value must be a vector
			super (TRUST_IF_NOT_IN_LIST, (a_conditionValue == null || 
					!(a_conditionValue instanceof Vector)) ? new Vector() : a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException, ServiceSignatureException
		{
			ServiceOperator mixOperator;
			ServiceOperator blacklistedOperator;
			
			Object conditionValue = getConditionValue();
			
			if (conditionValue == null || !(conditionValue instanceof Vector))
			{
				return;
			}
			
			
			if (getTrustCondition() == TRUST_IF_NOT_IN_LIST)
			{
				for (int i = 0; i < a_cascade.getNumberOfMixes(); i++)
				{
					Vector list = (Vector) conditionValue;
					MixInfo mixInfo=a_cascade.getMixInfo(i);
					
					mixOperator = null;
					if (mixInfo != null)
					{
						mixOperator = mixInfo.getServiceOperator();
					}
					if (mixOperator != null && list.contains(mixOperator))
					{
						throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_BLACKLISTED));
					}
					if (mixOperator !=null && mixOperator.getOrganization() != null)
					{
						// additionally, test if this operator is known with another ID but the same name
						for (int j = 0; j < list.size(); j++)
						{
							blacklistedOperator = (ServiceOperator)list.elementAt(j);
							if (blacklistedOperator.getOrganization() == null)
							{
								continue;
							}
							if (blacklistedOperator.getOrganization().equals(mixOperator.getOrganization()))
							{
								throw new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_BLACKLISTED));
							}
						}
					}
				}
			}
		}
	}
	
	private static int getIntegerConditionValue(Object a_conditionValue)
	{
		if (a_conditionValue == null || !(a_conditionValue instanceof Integer))
		{
			return 0;
		}
		return ((Integer) a_conditionValue).intValue();
	}

	public static class SpeedAttribute extends TrustAttribute
	{
		public SpeedAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			// MUST always be TRUST_IF_AT_LEAST
			super(TRUST_IF_AT_LEAST, a_conditionValue, a_bIgnoreNoDataAvailable);
		}

		public void checkTrust(MixCascade a_cascade) throws TrustException
		{
			if (Database.getInstance(PerformanceInfo.class).getNumberOfEntries() == 0)
			{
				// no data available for any service, ignore
				return;
			}
			
			PerformanceEntry entry = PerformanceInfo.getLowestCommonBoundEntry(a_cascade.getId());
			Object condidtionValue = getConditionValue();
			
			if (condidtionValue == null || getIntegerConditionValue(condidtionValue) <= 0)
			{
				// do not test speed, as all speed values are accepted
				return;
			}
			else if (entry == null || entry.getBound(PerformanceEntry.SPEED).getBound() == Integer.MAX_VALUE)  
			{
				if (isNoDataIgnored())
				{
					// no performance data available; do not test
					return;
				}
				else
				{
					throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NOT_ENOUGH_SPEED)));
				}
			}	

			if (getTrustCondition() == TRUST_IF_AT_LEAST && (entry == null || 
				entry.getBound(PerformanceEntry.SPEED).getBound() < getIntegerConditionValue(condidtionValue)))
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_NOT_ENOUGH_SPEED)));
			}
		}
	}
	
	public static class DelayAttribute extends TrustAttribute
	{
		public DelayAttribute(int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
		{
			// MUST always be TRUST_IF_AT_MOST
			super(TRUST_IF_AT_MOST, a_conditionValue, a_bIgnoreNoDataAvailable);
		}
		
		public void checkTrust(MixCascade a_cascade) throws TrustException
		{
			if (Database.getInstance(PerformanceInfo.class).getNumberOfEntries() == 0)
			{
				// no data available for any service, ignore
				return;
			}
			
			PerformanceEntry entry = PerformanceInfo.getLowestCommonBoundEntry(a_cascade.getId());
			Object conditionValue = getConditionValue();
				
			
			if (conditionValue == null || getIntegerConditionValue(conditionValue) == Integer.MAX_VALUE)
			{
				// do not test delay, as all delay values are accepted
				return;
			}
			
			if (entry == null || entry.getBound(PerformanceEntry.DELAY).getBound() == 0)
			{
				if (isNoDataIgnored())
				{
					// no performance data available; do not test
					return;
				}
				else
				{
					throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_RESPONSE_TIME_TOO_HIGH)));
				}
			}
			
			if (getTrustCondition() == TRUST_IF_AT_MOST && (entry == null || 
					entry.getBound(PerformanceEntry.DELAY).getBound() < 0 || 
					entry.getBound(PerformanceEntry.DELAY).getBound() > getIntegerConditionValue(conditionValue)))
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_RESPONSE_TIME_TOO_HIGH)));
			}
		}
		
		public static Object getDefaultValue()
		{
			return new Integer(Integer.MAX_VALUE);
		}
	}	
	
	static
	{
		// Initialize basic trust models
		TrustModel model;
		
		model = new TrustModel(MSG_ALL_SERVICES, -1, true);
		model.setAttribute(ContextAttribute.class, TRUST_RESERVED);
		TRUST_MODEL_ALL_SERVICES = model;
		//ms_trustModels.addElement(model); // this model is not "officially" available

		addCustomFilter();
		
		setCurrentTrustModel((TrustModel)ms_trustModels.elementAt(0));
	}

	public static String getContext()
	{
		return ms_strContext;
	}
	
	// TODO remove
	public static boolean updateContext(String a_context)
	{
		synchronized (ms_trustModels)
		{
			boolean bChanged = false;
			if (ms_strContext != a_context && a_context != null && (ms_strContext == null || !ms_strContext.equals(a_context)))
			{
				// changed!
				if ((ms_strContext == null && !a_context.equals(IServiceContextContainer.CONTEXT_JONDONYM)) || ms_strContext != null)
				{
					//changed from default to other context or the other way round
					bChanged = true;
				}
				ms_strContext = a_context;
			}
			
			if (!bChanged)
			{
				return false;
			}
			
			TrustModel temp = ms_currentTrustModel;
			
			
			ms_trustModelObservable.setChanged();
			setCurrentTrustModel(temp.getId());
			
			return true;
		}
	}
	
	/**
	 * Creates a new TrustModel object with the specified name
	 *
	 * @param a_strName	The name of the TrustModel
	 */
	public TrustModel(String a_strName, long a_id)
	{
		this(a_strName, a_id, false);
		setEditable(true);
	}
	
	private TrustModel(String a_strName, long a_id, boolean b_allowReservedID)
	{
		//m_id = ms_trustModels.size();
		if (!b_allowReservedID && a_id < FIRST_UNRESERVED_MODEL_ID)
		{
			throw new IllegalArgumentException("Trust model ID " + a_id + " is reserved!");
		}
		m_id = a_id;
		m_strName = a_strName == null ? "Default trust model" : a_strName;
	}

	/**
	 * Creates a TrustModel object from another TrustModel object
	 *
	 * @param a_trustModel	The TrustModel object to copy
	 */
	public TrustModel(TrustModel a_trustModel)
	{
		setEditable(true);
		clone(a_trustModel);
		if (!a_trustModel.isEditable())
		{
			setEditable(false);
		}
	}

	/**
	 * Creates a TrustModel object from an XML element
	 *
	 * @param a_trustModelElement	The XML element which holds the TrustModel data
	 *
	 * @throws XMLParseException
	 */
	public TrustModel(Element a_trustModelElement) throws XMLParseException
	{
		XMLUtil.assertNodeName(a_trustModelElement, XML_ELEMENT_NAME);

		XMLUtil.assertNotNull(a_trustModelElement, XML_ATTR_ID);
		//XMLUtil.assertNotNull(a_trustModelElement, XML_ATTR_NAME);

		m_id = XMLUtil.parseAttribute(a_trustModelElement, XML_ATTR_ID, -1l);
		if (m_id != FIRST_UNRESERVED_MODEL_ID)
		{
			m_id = FIRST_UNRESERVED_MODEL_ID;
			//throw new XMLParseException("Reserved model ID: " + m_id);
		}
		//m_strName = XMLUtil.parseAttribute(a_trustModelElement, XML_ATTR_NAME, null);
		m_strName = JAPMessages.getString(MSG_DEFAULT_FILTER);
		m_bEditable = true;

		for (int i = 0; i < a_trustModelElement.getChildNodes().getLength(); i++)
		{
			Element el = (Element) a_trustModelElement.getChildNodes().item(i);
			try
			{
				setAttribute(TrustAttribute.fromXmlElement(el));
			}
			catch (XMLParseException a_e)
			{
				LogHolder.log(LogLevel.ERR, LogType.MISC, a_e);
			}
		}
		// set defaults
		getAttribute(NumberOfMixesAttribute.class);
		getAttribute(ForcePremiumIfChargedAccountAttribute.class);
	}
	
	public void unblockInterfacesFromDatabase()
	{
		ListenerInterface.unblockInterfacesFromDatabase(MixCascade.class);
		synchronized (ms_trustModels)
		{
			ms_trustModelObservable.setChanged();
			ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
		}
	}
	
	public void copyAttributes(TrustModel a_trustModel)
	{
		if (a_trustModel == null)
		{
			throw new NullPointerException("No argument given!");
		}
		
		if (!isEditable())
		{
			throw new IllegalStateException("Trust model not editable!");
		}

		synchronized(m_trustAttributes)
		{
			m_trustAttributes = (Hashtable) a_trustModel.m_trustAttributes.clone();
		}
		synchronized (ms_trustModels)
		{
			if (ms_trustModels.contains(this))
			{
				ms_trustModelObservable.setChanged();
				ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
			}
		}
	}
	
	
	public void clone(TrustModel a_trustModel)
	{
		if (a_trustModel == null)
		{
			throw new NullPointerException("No argument given!");
		}
		
		if (!isEditable())
		{
			throw new IllegalStateException("Trust model not editable!");
		}

		m_id = a_trustModel.m_id;
		m_strName = a_trustModel.m_strName;

		copyAttributes(a_trustModel);
		m_bEditable = a_trustModel.m_bEditable;
	}

	public static Observable getObservable()
	{
		return ms_trustModelObservable;
	}

	public static void addModelObserver(Observer a_observer)
	{
		ms_trustModelObservable.addObserver(a_observer);
	}

	public static void deleteModelObserver(Observer a_observer)
	{
		ms_trustModelObservable.deleteObserver(a_observer);
	}

	/**
	 * Compares two TrustModel objects
	 *
	 * @return true if the id's of the two objects match, false otherwise
	 */
	public boolean equals(Object a_trustModel)
	{
		if (a_trustModel == null || !(a_trustModel instanceof TrustModel))
		{
			return false;
		}
		return getId() == ((TrustModel)a_trustModel).getId();
	}

	public int hashCode()
	{
		return (int)getId();
	}
	
	public static void setFreeAllowed(boolean a_bFreeAllowed)
	{
		ms_bFreeAllowed = a_bFreeAllowed;
	}
	
	/*
	 * That means: Do not allow to get any free services loaded!
	 */
	public static boolean isFreeAllowed()
	{
		return ms_bFreeAllowed;
	}

	/**
	 * Allow the given service even if the current filter has attributes that would otherwise disallow its usage.
	 * @param a_cascade
	 */
	public static void allowAttributeWhitelist(MixCascade a_cascade)
	{
		synchronized (ms_trustModels)
		{
			if (a_cascade != null && (ms_cascadeWhitelisted == null ||
					!ms_cascadeWhitelisted.getId().equals(a_cascade)))
			{
				ms_cascadeWhitelisted = a_cascade;
				ms_trustModelObservable.setChanged();
				ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
			}
		}
	}
	
	public static void cleanAttributeWhitelist(MixCascade a_currentCascade)
	{
		synchronized (ms_trustModels)
		{
			if (a_currentCascade == null || (ms_cascadeWhitelisted != null &&
				!ms_cascadeWhitelisted.getId().equals(a_currentCascade.getId())))
			{
				ms_cascadeWhitelisted = null;
				ms_trustModelObservable.setChanged();
				ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
			}
		}
	}
	
	private static boolean addTrustModel(TrustModel a_trustModel)
	{
		synchronized (ms_trustModels)
		{
			if (a_trustModel != null && !ms_trustModels.contains(a_trustModel))
			{
				if (a_trustModel.getId() == FIRST_UNRESERVED_MODEL_ID)
				{
					TRUST_MODEL_CUSTOM_FILTER = a_trustModel;
				}
				if (ms_currentTrustModel != null && 
					ms_currentTrustModel.getId() == a_trustModel.getId())
				{
					ms_currentTrustModel = a_trustModel;
				}
				
				ms_trustModels.insertElementAt(a_trustModel, 0);
				ms_trustModelObservable.setChanged();
				ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_ADDED);
				return true;
			}
		}
		return false;
	}

	// TODO make private; we do not need to define trust models from outside any longer!
	public static TrustModel removeTrustModel(TrustModel a_trustModel)
	{
		if (a_trustModel == null)
		{
			return null;
		}
		
		if (!a_trustModel.isEditable())
		{
			// non-editable models must not be removed
			return null;
		}
		synchronized (ms_trustModels)
		{
			if (a_trustModel != null && ms_trustModels.removeElement(a_trustModel))
			{
				if (getCurrentTrustModel() == a_trustModel)
				{
					restoreDefault();
				}
				ms_trustModelObservable.setChanged();
				ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_REMOVED);
				return a_trustModel;
			}
		}
		return null;
	}

	public TrustAttribute setAttribute(Class a_attr, int a_trustCondition)
	{
		return setAttribute(a_attr, a_trustCondition, false);
	}
	
	private TrustAttribute setAttribute(Class a_attr, int a_trustCondition, boolean a_bIgnoreNoDataAvailable)
	{
		return setAttribute(a_attr, a_trustCondition, (Object)null, a_bIgnoreNoDataAvailable);
	}

	public TrustAttribute setAttribute(Class a_attr, int a_trustCondition, int a_conditionValue)
	{
		return setAttribute(a_attr, a_trustCondition, a_conditionValue, false);
	}
	
	public TrustAttribute setAttribute(Class a_attr, int a_trustCondition, Object a_conditionValue)
	{
		return setAttribute(a_attr, a_trustCondition, a_conditionValue, false);
	}
	
	private TrustAttribute setAttribute(Class a_attr, int a_trustCondition, int a_conditionValue, boolean a_bIgnoreNoDataAvailable)
	{
		return setAttribute(a_attr, a_trustCondition, new Integer(a_conditionValue), a_bIgnoreNoDataAvailable);
	}
	
	public TrustAttribute setAttribute(Class a_attr, int a_trustCondition, Vector a_conditionValue)
	{
		return setAttribute(a_attr, a_trustCondition, a_conditionValue, false);
	}

	private void setEditable(boolean a_bEditable)
	{
		m_bEditable = a_bEditable;
	}

	public boolean isEditable()
	{
		return m_bEditable;
	}


	private TrustAttribute setAttribute(Class a_attr, int a_trustCondition, Object a_conditionValue, boolean a_bIgnoreNoDataAvailable)
	{
		 return setAttribute(TrustAttribute.getInstance(a_attr, a_trustCondition, a_conditionValue, a_bIgnoreNoDataAvailable));
	}
	
	public void removeAllAttributes()
	{
		if (!isEditable())
		{
			throw new IllegalStateException("Trust model not editable!");
		}
		synchronized(m_trustAttributes)
		{
			if (m_trustAttributes.size() > 0)
			{
				ms_trustModelObservable.setChanged();
				m_trustAttributes.clear();
			}
		}
		
		ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
	}

	private TrustAttribute setAttribute(TrustAttribute a_attr)
	{
		if (a_attr != null)
		{
			synchronized (ms_trustModels)
			{
				TrustAttribute value;
	
				synchronized(m_trustAttributes)
				{
					if ((value = (TrustAttribute) m_trustAttributes.get(a_attr)) != null)
					{
						if (value.getTrustCondition() != a_attr.getTrustCondition() || value.getConditionValue() != a_attr.getConditionValue())
						{
							m_trustAttributes.put(a_attr, value);
							if (ms_trustModels.contains(this))
							{
								ms_trustModelObservable.setChanged();
							}
						}
					}
		
					m_trustAttributes.put(a_attr.getClass(), a_attr);
				}
				
				if (ms_trustModels.contains(this))
				{
					ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
				}
			}
		}
		
		return a_attr;
	}

	public TrustAttribute getAttribute(Class a_attr)
	{
		if (a_attr == null)
		{
			return null;
		}
		
		synchronized(m_trustAttributes)
		{
			TrustAttribute attr = (TrustAttribute) m_trustAttributes.get(a_attr);
			if(attr == null)
			{
				Object defaultValue = new Integer(0);
				int iDefaultCondition = TRUST_ALWAYS;
				try
				{
					Method getDefaultValue = a_attr.getMethod("getDefaultValue", (Class[]) null);
					defaultValue = getDefaultValue.invoke(null, (Object[]) null);
				}
				catch(Exception ex)
				{
					LogHolder.log(LogLevel.EXCEPTION, LogType.MISC, "Exception occured while trying to get the default value of a TrustAttribute: ", ex);
				}
				
				try
				{
					Method getDefaultCondition = a_attr.getMethod("getDefaultCondition", (Class[]) null);
					iDefaultCondition = ((Integer)getDefaultCondition.invoke(null, (Object[]) null)).intValue();
				}
				catch(Exception ex)
				{
					LogHolder.log(LogLevel.EXCEPTION, LogType.MISC, "Exception occured while trying to get the default value of a TrustAttribute: ", ex);
				}
				
				return setAttribute(a_attr, iDefaultCondition, defaultValue);
			}

			return attr;
		}
	}

	public static Vector getTrustModels()
	{
		return (Vector)ms_trustModels.clone();
	}
	
	public static void setCurrentTrustModel(long a_id)
	{
		if (a_id < 0)
		{
			return;
		}

		synchronized (ms_trustModels)
		{
			for (int i = 0; i < ms_trustModels.size(); i++)
			{
				if (((TrustModel)ms_trustModels.elementAt(i)).getId() == a_id)
				{
					ms_currentTrustModel = (TrustModel)ms_trustModels.elementAt(i);
					ms_trustModelObservable.setChanged();
					ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
					break;
				}
			}
			
		}
	}

	public static void setCurrentTrustModel(TrustModel a_trustModel)
	{
		if (a_trustModel == null)
		{
			return;
		}
		synchronized (ms_trustModels)
		{
			if (!ms_trustModels.contains(a_trustModel))
			{
				ms_trustModels.addElement(a_trustModel);
			}
			if (ms_currentTrustModel != a_trustModel)
			{
				ms_currentTrustModel = a_trustModel;
				ms_trustModelObservable.setChanged();
			}
			ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
		}
	}
	
	public static TrustModel getTrustModelDefault()
	{
		return (TrustModel)ms_trustModels.elementAt(0);
	}
	
	public static TrustModel getCurrentTrustModel()
	{
		//synchronized (ms_trustModels)
		{
			return ms_currentTrustModel;
		}
	}

	public static TrustModel getCustomFilter()
	{
		return TRUST_MODEL_CUSTOM_FILTER;
	}
	
	public static void fromXmlElement(Element a_container)
	{
		int trustModelsAdded = 0;

		if (a_container != null && a_container.getNodeName().equals(XML_ELEMENT_CONTAINER_NAME))
		{
			NodeList elements = a_container.getElementsByTagName(XML_ELEMENT_NAME);
			for (int i = 0; i < elements.getLength(); i++)
			{
				synchronized (ms_trustModels)
				{
					try
					{
						TrustModel model = new TrustModel((Element) elements.item(i));
						boolean bReplaced = false;
						
						for (int j = 0; j < ms_trustModels.size(); j++)
						{
							if (((TrustModel)ms_trustModels.elementAt(j)).getId() == 
								model.getId())
							{
								if (model.getId() == FIRST_UNRESERVED_MODEL_ID)
								{
									TRUST_MODEL_CUSTOM_FILTER = model;
								}
								ms_trustModels.removeElementAt(j);
								ms_trustModels.insertElementAt(model, j);
								bReplaced = true;
								break;
							}
						}
						
						if (model.getId() == ms_currentTrustModel.getId())
						{
							ms_currentTrustModel = model;
						}
						
						if (bReplaced)
						{
							ms_trustModelObservable.setChanged();
							ms_trustModelObservable.notifyObservers(NOTIFY_TRUST_MODEL_CHANGED);
						}
						else
						{
							addTrustModel(model);
						}
						trustModelsAdded++;
					}
					catch (Exception a_e)
					{
						LogHolder.log(LogLevel.EXCEPTION, LogType.MISC, "Could not load trust model from XML!", a_e);
					}
				}
			}
			
			if (trustModelsAdded == 0)
			{
				addCustomFilter();
			}
			
			setCurrentTrustModel(XMLUtil.parseAttribute(a_container, XML_ATTR_CURRENT_TRUST_MODEL, 0l));
		}
	}
	
	private static void addCustomFilter()
	{
		synchronized (ms_trustModels)
		{
			removeTrustModel(TRUST_MODEL_CUSTOM_FILTER);
			
			TrustModel model = new TrustModel(MSG_DEFAULT_FILTER, FIRST_UNRESERVED_MODEL_ID, false);
			//model.setAttribute(NumberOfMixesAttribute.class, TRUST_IF_AT_LEAST, 1);
			model.setAttribute(ForcePremiumIfChargedAccountAttribute.class, TRUST_IF_TRUE);
			model.setEditable(true);
			addTrustModel(model);
		}
	}
	
	public static void restoreDefault()
	{
		addCustomFilter();
		setCurrentTrustModel(TRUST_MODEL_DEFAULT_INDEX);
	}

	public static Element toXmlElement(Document a_doc, String a_xmlContainerName)
	{
		if (a_doc == null || a_xmlContainerName == null)
		{
			return null;
		}

		Element container = a_doc.createElement(a_xmlContainerName);
		XMLUtil.setAttribute(container, XML_ATTR_CURRENT_TRUST_MODEL, getCurrentTrustModel().getId());
		synchronized (ms_trustModels)
		{
			for (int i = 0; i < ms_trustModels.size(); i++)
			{
				// Only write custom trust models to the configuration file
				if(((TrustModel)ms_trustModels.elementAt(i)).isEditable())
				{
					container.appendChild(((TrustModel)ms_trustModels.elementAt(i)).toXmlElement(a_doc));
				}
			}
		}
		return container;
	}

	public void setName(String a_strName) throws IllegalArgumentException
	{
		if (a_strName == null || a_strName.trim().length() == 0)
		{
			throw new IllegalArgumentException("Invalid name for trust model!");
		}
		if (a_strName.trim().equals(JAPMessages.getString(MSG_CASCADES_FILTER)))
		{
			m_strName = "";
		}
		else
		{
			m_strName = a_strName;
		}
	}

	/**
	 * Returns the localized name of the TrustModel
	 *
	 * @return The localized name of the TrustModel
	 */
	public String getName()
	{
		if (m_strName == null || m_strName.trim().length() == 0)
		{
			return JAPMessages.getString(MSG_CASCADES_FILTER);
		}
		return JAPMessages.getString(m_strName);
	}

	public String toString()
	{
		return getName();
	}

	public long getId()
	{
		return m_id;
	}


	public Element toXmlElement(Document a_doc)
	{
		if (a_doc == null)
		{
			return null;
		}

		Element elemTrustModel = a_doc.createElement(XML_ELEMENT_NAME);

		//XMLUtil.setAttribute(elemTrustModel, XML_ATTR_SHOW_WARNING, m_bSh owWarning);
		XMLUtil.setAttribute(elemTrustModel, XML_ATTR_ID, m_id);
		//XMLUtil.setAttribute(elemTrustModel, XML_ATTR_NAME, m_strName);

		synchronized(m_trustAttributes)
		{
			Enumeration e = m_trustAttributes.elements();
			while (e.hasMoreElements()) 
			{
				TrustAttribute attr = (TrustAttribute) e.nextElement();
				elemTrustModel.appendChild(attr.toXmlElement(a_doc));
			}
		}

		return elemTrustModel;
	}

	public boolean isFreeServicesForced()
	{
		TrustAttribute attr = getAttribute(UsePremiumAttribute.class);

		return (attr == null) ? false : (attr.getTrustCondition() == TRUST_IF_NOT_TRUE);
	}
	
	
	public boolean isPaymentForced()
	{
		if (!TrustModel.isFreeAllowed())
		{
			return true;
		}
		
		TrustAttribute attr = getAttribute(UsePremiumAttribute.class);

		return (attr == null) ? false : (attr.getTrustCondition() == TRUST_IF_TRUE);
	}

	public boolean isAdded()
	{
		return true;
	}

	
	
	
	public int countTrustedCascades()
	{
		Vector list = Database.getInstance(MixCascade.class).getEntryList();
		int iTrusted = 0;
		for (int i = 0; i < list.size(); i++)
		{
			if (isTrusted((MixCascade)list.elementAt(i)))
			{
				iTrusted++;
			}
		}

		return iTrusted;
	}
	
	/**
	 * Checks if the current trust model has any trusted cascades at all
	 * @return true if the trust model has trusted cascades, false otherwise
	 */
	public boolean hasTrustedCascades()
	{
		Vector list = Database.getInstance(MixCascade.class).getEntryList();

		for (int i = 0; i < list.size(); i++)
		{
			if (isTrusted((MixCascade)list.elementAt(i))) return true;
		}

		return false;
	}
	
	public boolean hasPremiumCascades()
	{
		Vector list = Database.getInstance(MixCascade.class).getEntryList();

		for (int i = 0; i < list.size(); i++)
		{
			if (((MixCascade)list.elementAt(i)).isPayment() &&
				isTrusted((MixCascade)list.elementAt(i))) return true;
		}

		return false;
	}
	
	public boolean hasFreeCascades()
	{
		Vector list = Database.getInstance(MixCascade.class).getEntryList();

		for (int i = 0; i < list.size(); i++)
		{
			if (!((MixCascade)list.elementAt(i)).isPayment() &&
				isTrusted((MixCascade)list.elementAt(i))) return true;
		}

		return false;
	}

	public static boolean isBlacklisted(MixCascade a_cascade)
	{
		if (a_cascade == null)
		{
			return false;
		}
		return Database.getInstance(BlacklistedCascadeIDEntry.class).getEntryById(
				  a_cascade.getMixIDsAsString()) != null;
	}
	
	public static boolean areListenerInterfacesBlocked(MixCascade a_cascade)
	{
		if (a_cascade == null)
		{
			return false;
		}
		return a_cascade.areListenerInterfacesBlocked();
	}
	
	public static boolean isNoPaymentInstanceFound(MixCascade a_cascade)
	{
		if (a_cascade == null)
		{
			return false;
		}
		return a_cascade.isPayment() && PayAccountsFile.getInstance().getBI(a_cascade.getPIID()) == null &&
			PayAccountsFile.getInstance().getChargedAccount(a_cascade.getPIID()) != null;
	}
	
	public Vector readUntrustedAttributeText(MixCascade a_cascade)
	{
		Vector vecReasons = new Vector();
		if (!isTrusted(a_cascade))
		{
			synchronized (m_trustAttributes)
			{
				Enumeration e = m_trustAttributes.elements();
				while (e.hasMoreElements()) {
					TrustAttribute attr = (TrustAttribute) e.nextElement();
					
					try 
					{
						attr.checkTrust(a_cascade);
					} 
					catch (TrustException e1) 
					{
						vecReasons.addElement(e1.getMessage());
					} 
					catch (ServiceSignatureException e1) 
					{
						vecReasons.addElement(e1.getMessage());
					}
				}
			}
		}
		
		return vecReasons;
	}
	
	public void checkTrust(MixCascade a_cascade, boolean a_bCheckUserDefined) 
		throws TrustException, ServiceSignatureException
	{
		if (a_cascade == null)
		{
			throw (new TrustException(a_cascade, "Null cascade!"));
		}

		if (isBlacklisted(a_cascade))
		{
			throw (new TrustException(a_cascade, JAPMessages.getString(MSG_BLACKLISTED)));
		}
		
		/*
		 *  Allow unverified cascades in any filter, but NOT when connecting.
		 */
		if (a_bCheckUserDefined || !a_cascade.isUserDefined())
		{
			if (!ms_bFreeAllowed && !a_cascade.isPayment())
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_EXCEPTION_FREE_CASCADE)));
			}
			
			if (isNoPaymentInstanceFound(a_cascade))
			{
				throw (new TrustException(a_cascade, JAPMessages.getString(MSG_PI_UNAVAILABLE)));
			}
			
			super.checkTrust(a_cascade, a_bCheckUserDefined);
		}
		else if (a_cascade.getNumberOfOperators() <= 0) 
		{
			// do not check attributes for unknown cascades, as they don't have any...
			return;
		}
		
		MixCascade cascadeWhitelist = ms_cascadeWhitelisted;
		
		if (cascadeWhitelist == null || !cascadeWhitelist.getId().equals(a_cascade.getId()))
		{
			synchronized (m_trustAttributes)
			{
				Enumeration e = m_trustAttributes.elements();
				while (e.hasMoreElements()) {
					TrustAttribute attr = (TrustAttribute) e.nextElement();
					attr.checkTrust(a_cascade);
				}
			}
		}
	}
}
