/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client.crypto;

import mybouncycastle.org.bouncycastle.crypto.engines.AESFastEngine;
import mybouncycastle.org.bouncycastle.crypto.modes.GCMBlockCipher;
import mybouncycastle.org.bouncycastle.crypto.params.AEADParameters;
import mybouncycastle.org.bouncycastle.crypto.params.KeyParameter;

/**
 * @author Lukas Braune
 */
final public class DataChannelCipher implements IDataChannelChipher
	{

		private GCMBlockCipher m_sendEngine;
		private GCMBlockCipher m_recvEngine;

		private long m_EncMsgCounter;
		private long m_DecMsgCounter;

		private MyAEADParameters m_sendKey;
		private MyAEADParameters m_recvKey;
		private byte[] m_sendKeyBytes;
		private byte[] m_recvKeyBytes;
		private byte[] m_sendIV;
		private byte[] m_recvIV;

		private boolean m_bSingleKey = false;

		private byte[] m_decryptHelperBuff=new byte[32];
		
		public DataChannelCipher()
			{
				m_sendEngine = new GCMBlockCipher(new AESFastEngine());
				m_recvEngine = new GCMBlockCipher(new AESFastEngine());
				m_sendIV = new byte[12];
				m_recvIV = new byte[12];
				m_EncMsgCounter = 0;
				m_DecMsgCounter = 0;
			}

		synchronized private int setSendKey(byte[] key, int offset, int len)
			{
				try
					{
						m_sendKeyBytes=new byte[key.length];
						System.arraycopy(key, 0, m_sendKeyBytes, 0, key.length);
						m_sendKey = new MyAEADParameters(new KeyParameter(key, offset, len), 128, m_sendIV, null);
						m_sendEngine.init(true, m_sendKey);
						//We set the AES key to null to avoid that the AES key schedule is calculate for
						//every GCM invocation
						m_sendKey = new MyAEADParameters(null, 128, m_sendIV, null);
						m_EncMsgCounter = 0;
						return 0;
					}
				catch (Exception e)
					{
						return -1;
					}
			}

		synchronized private int setRecvKey(byte[] key, int offset, int len)
			{
				try
					{
						m_recvKeyBytes=new byte[key.length];
						System.arraycopy(key, 0, m_recvKeyBytes, 0, key.length);
						m_recvKey = new MyAEADParameters(new KeyParameter(key, offset, len), 128, m_recvIV, null);
						m_recvEngine.init(false, m_recvKey);
						//We set the AES key to null to avoid that the AES key schedule is calculate for
						//every GCM invocation
						m_recvKey = new MyAEADParameters(null, 128, m_recvIV, null);
						m_DecMsgCounter = 0;
						return 0;
					}
				catch (Exception e)
					{
						return -1;
					}
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#setEncryptionKeysAES(byte[])
		 */
		synchronized public int setEncryptionKeysAES(byte[] keys)
			{
				return setEncryptionKeysAES(keys,0,keys.length);
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#setEncryptionKeysAES(byte[], int, int)
		 */
		synchronized public int setEncryptionKeysAES(byte[] keys,int offset,int keylen)
			{
				int ret1, ret2;
				try
					{
						if (keylen == 16)
							{
								m_bSingleKey = true;
								ret1 = setSendKey(keys, offset, 16);
								ret2 = setRecvKey(keys, offset, 16);
							}
						else
							{
								m_bSingleKey = false;
								ret1 = setSendKey(keys, offset, 16);
								ret2 = setRecvKey(keys, offset+16, 16);
							}
						if ((ret1 == 0) && (ret2 == 0))
							{
								return 0;
							}
						else
							{
								return -1;
							}
					}
				catch (Exception e)
					{
						return -1;
					}
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#getKeys()
		 */
		public byte[] getKeys()
			{
				try
					{
						if ((m_sendKeyBytes == null) || (m_recvKeyBytes == null))
							{
								return null;
							}
						else if (m_bSingleKey)
							{
								return m_sendKeyBytes;
							}
						else
							{
								byte[] keys = new byte[32];
								System.arraycopy(m_sendKeyBytes, 0, keys, 0, 16);
								System.arraycopy(m_recvKeyBytes, 0, keys, 16, 16);
								return keys;
							}
					}
				catch (Exception e)
					{
						return null;
					}
			}

		private static void createIV(byte[] iv, long counter)
			{
				iv[8] = (byte) (counter >> 24 & 0x00ff);
				iv[9] = (byte) (counter >> 16 & 0x00ff);
				iv[10] = (byte) (counter >> 8 & 0x00ff);
				iv[11] = (byte) (counter & 0x00ff);
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#encryptGCM1(byte[], int, byte[], int, int)
		 */
		public void encryptGCM1(byte[] from, int ifrom, byte[] to, int ito, int len) throws Exception
			{
				DataChannelCipher.createIV(m_sendIV, m_EncMsgCounter);
				m_EncMsgCounter++;
				m_sendEngine.init(true, m_sendKey);
				int outlen = m_sendEngine.processBytes(from, ifrom, len, to, ito);
				m_sendEngine.doFinal(to, ito + outlen);
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#decryptGCM2(byte[], int, byte[], int, int)
		 */
		public void decryptGCM2(byte[] from, int ifrom, byte[] to, int ito, int len) throws Exception
			{
				DataChannelCipher.createIV(m_recvIV, m_DecMsgCounter);
				m_DecMsgCounter++;
				m_recvEngine.init(false, m_recvKey);
				/* decrypt only the first 2 bytes to get the payload length */
				//Note: We have to process a full block + MAC_SIZE here!
				m_recvEngine.processBytes(from, ifrom, 32, m_decryptHelperBuff,0);
				int payloadLen = ((m_decryptHelperBuff[0] << 8) & 0x003F00) | (m_decryptHelperBuff[1] & 0x00FF);
				/* decrypt the remaining packet now */
				m_recvEngine.reset();
				int declen = m_recvEngine.processBytes(from, ifrom, payloadLen + 19, to, ito); //Size= payloadlen+ payloadHeaderSize+MacSize
				m_recvEngine.doFinal(to, ito + declen);
			}

		public static int getMACSize()
			{
				return 16;
			}
	}
