/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client.crypto;

import java.lang.reflect.Constructor;
import java.security.spec.AlgorithmParameterSpec;

import javax.crypto.Cipher;
import javax.crypto.spec.SecretKeySpec;



/**
 * @author Lukas Braune
 */
final public class DataChannelCipherSDK implements IDataChannelChipher
	{

		private Cipher m_sendEngine;
		private Cipher m_recvEngine;

		private long m_EncMsgCounter;
		private long m_DecMsgCounter;

		private SecretKeySpec  m_sendKey;
		private SecretKeySpec  m_recvKey;
		private byte[] m_sendIV;
		private byte[] m_recvIV;

		private boolean m_bSingleKey = false;

		private byte[] m_decryptHelperBuff=new byte[64];
		
		private static Constructor ms_constructorGCMSpec;
		private static Object[] ms_constructorargsGCMSpec;
		
		static
			{
				try
					{
						Class c=Class.forName("javax.crypto.spec.GCMParameterSpec");
						Class args[]= {int.class,byte[].class};
						ms_constructorGCMSpec=c.getConstructor(args);
						ms_constructorargsGCMSpec=new Object[2];
						ms_constructorargsGCMSpec[0]=new Integer(128);
					}
				catch (Exception e)
					{
						ms_constructorGCMSpec=null;
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
			}
		
		public DataChannelCipherSDK()
		{
			this(null);
		}

		public DataChannelCipherSDK(String provider)
			{
				try
					{
						if(provider!=null)
							{
								m_sendEngine = Cipher.getInstance("AES/GCM/NoPadding",provider);
								m_recvEngine = Cipher.getInstance("AES/GCM/NoPadding",provider);
							}
						else
							{
								m_sendEngine = Cipher.getInstance("AES/GCM/NoPadding");
								m_recvEngine = Cipher.getInstance("AES/GCM/NoPadding");								
							}
					}
				catch (Exception e)
					{
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				m_sendIV = new byte[12];
				m_recvIV = new byte[12];
				m_EncMsgCounter = 0;
				m_DecMsgCounter = 0;
			}

		synchronized private int setSendKey(byte[] key, int offset, int len)
			{
				try
					{
						m_sendKey =new SecretKeySpec(key, offset, len,"AES");
						m_EncMsgCounter = 0;
						return 0;
					}
				catch (Exception e)
					{
						return -1;
					}
			}

		synchronized private int setRecvKey(byte[] key, int offset, int len)
			{
				try
					{
						m_recvKey = new SecretKeySpec(key, offset, len,"AES");
						m_DecMsgCounter = 0;
						return 0;
					}
				catch (Exception e)
					{
						return -1;
					}
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#setEncryptionKeysAES(byte[])
		 */
		synchronized public int setEncryptionKeysAES(byte[] keys)
			{
				return setEncryptionKeysAES(keys,0,keys.length);
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#setEncryptionKeysAES(byte[], int, int)
		 */
		synchronized public int setEncryptionKeysAES(byte[] keys,int offset,int keylen)
			{
				int ret1, ret2;
				try
					{
						if (keylen == 16)
							{
								m_bSingleKey = true;
								ret1 = setSendKey(keys, offset, 16);
								ret2 = setRecvKey(keys, offset, 16);
							}
						else
							{
								m_bSingleKey = false;
								ret1 = setSendKey(keys, offset, 16);
								ret2 = setRecvKey(keys, offset+16, 16);
							}
						if ((ret1 == 0) && (ret2 == 0))
							{
								return 0;
							}
						else
							{
								return -1;
							}
					}
				catch (Exception e)
					{
						return -1;
					}
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#getKeys()
		 */
		public byte[] getKeys()
			{
				try
					{
						if ((m_sendKey == null) || (m_recvKey == null))
							{
								return null;
							}
						else if (m_bSingleKey)
							{
								return m_sendKey.getEncoded();
							}
						else
							{
								byte[] keys = new byte[32];
								System.arraycopy(m_sendKey.getEncoded(), 0, keys, 0, 16);
								System.arraycopy(m_recvKey.getEncoded(), 0, keys, 16, 16);
								return keys;
							}
					}
				catch (Exception e)
					{
						return null;
					}
			}

		private static void createIV(byte[] iv, long counter)
			{
				iv[8] = (byte) (counter >> 24 & 0x00ff);
				iv[9] = (byte) (counter >> 16 & 0x00ff);
				iv[10] = (byte) (counter >> 8 & 0x00ff);
				iv[11] = (byte) (counter & 0x00ff);
			}
		
		private static AlgorithmParameterSpec createGCMParameterSpec(byte[] iv)
		{
			try
				{
					ms_constructorargsGCMSpec[1]=iv;
					return (AlgorithmParameterSpec)ms_constructorGCMSpec.newInstance(ms_constructorargsGCMSpec);
				}
			catch (Exception e)
				{
					// TODO Auto-generated catch block
					e.printStackTrace();
					return null;
				}
		}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#encryptGCM1(byte[], int, byte[], int, int)
		 */
		public void encryptGCM1(byte[] from, int ifrom, byte[] to, int ito, int len) throws Exception
			{
				DataChannelCipherSDK.createIV(m_sendIV, m_EncMsgCounter);
				m_EncMsgCounter++;
				AlgorithmParameterSpec spec = createGCMParameterSpec(m_sendIV);
				m_sendEngine.init(Cipher.ENCRYPT_MODE, m_sendKey, spec);
				int outlen = m_sendEngine.update(from, ifrom, len, to, ito);
				m_sendEngine.doFinal(to, ito + outlen);
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.IDataChannelChipher#decryptGCM2(byte[], int, byte[], int, int)
		 */
		public void decryptGCM2(byte[] from, int ifrom, byte[] to, int ito, int len) throws Exception
			{
				DataChannelCipherSDK.createIV(m_recvIV, m_DecMsgCounter);
				m_DecMsgCounter++;
				AlgorithmParameterSpec spec = createGCMParameterSpec(m_recvIV);
				m_recvEngine.init(Cipher.DECRYPT_MODE, m_recvKey, spec);

				// decrypt only the first 2 bytes to get the payload length //
				//Note: We have to process a full block + MAC_SIZE here!
		//		int i=m_recvEngine.update(from, ifrom, len, to,ito);
				try {
				m_recvEngine.doFinal(from,ifrom,32,this.m_decryptHelperBuff,0);
				}
				catch(Exception e)
					{
						e.printStackTrace();
					}
				int payloadLen = ((m_decryptHelperBuff[0] << 8) & 0x003F00) | (m_decryptHelperBuff[1] & 0x00FF);
				// decrypt the remaining packet now //
				System.arraycopy(m_decryptHelperBuff, 0, to, ito, 32);
				int declen = m_recvEngine.update(from, ifrom+32, payloadLen + 19, to, ito+32); //Size= payloadlen+ payloadHeaderSize+MacSize
				m_recvEngine.doFinal(to, ito + declen+16);
			}

		public static int getMACSize()
			{
				return 16;
			}
	}
