/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
  may be used to endorse or promote products derived from this software without specific
  prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.client.crypto;

import java.security.SecureRandom;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;
import anon.ErrorCodes;

final public class KeyPool implements Runnable
	{
		private SecureRandom m_SecureRandom;

		//private KeyList keys;
		private KeyList m_keylistPool;
		private KeyList m_keylistAktKey;
		private int m_iKeySize;
		private int m_iPoolSize;
		private Object l1;
		private Object l2;
		private volatile boolean m_bRun;
		private static KeyPool m_KeyPool = null;
		private Thread m_KeyPoolThread = null;
		private boolean m_bDebug = false; //if m_bDebug ==true then all keys =0!

		private final class KeyList
			{
				public byte[] key;
				public KeyList next;

				public KeyList(int iKeySize)
					{
						key = new byte[iKeySize];
						next = null;
					}
			}

		private KeyPool(int poolsize, int keylength, boolean bDebug)
			{
				m_bDebug = bDebug;
				m_iKeySize = keylength;
				m_iPoolSize = poolsize;
				m_keylistPool = null;
				m_keylistAktKey = null;
				l1 = new Object();
				l2 = new Object();
				m_bRun = true;
				m_KeyPoolThread = new Thread(this, "JAP - KeyPool");
				m_KeyPoolThread.setDaemon(true);
				m_KeyPoolThread.setPriority(Thread.MIN_PRIORITY);
				m_KeyPoolThread.start();
			}

		/**
		 * Warning: if bDebug==true then all keys are set to 0! This is for debug
		 * purposes only!
		 */
		public static synchronized KeyPool start(boolean bDebug)
			{
				if (m_KeyPool == null)
					{
						m_KeyPool = new KeyPool(20, 16, bDebug);
					}
				else if (m_KeyPool.m_bDebug != bDebug)
					{
						m_KeyPool.stop_internal();
						m_KeyPool = new KeyPool(20, 16, bDebug);
					}
				return m_KeyPool;
			}
		
		public static synchronized void stop()
		{
			if(m_KeyPool!=null)
				{
					m_KeyPool.stop_internal();
					m_KeyPool=null;
				}
			
		}

		private void stop_internal()
			{
				m_bRun = false;
				synchronized (l1)
					{
						l1.notify();
					}
			}

		public void run()
			{
				byte[] seed = null;
				try
					{
						seed = SecureRandom.getSeed(20);
						m_SecureRandom = new SecureRandom(seed);
					}
				catch (Throwable e)
					{
						m_SecureRandom = new SecureRandom();
					}
				seed = null;
				KeyList tmpKey;
				m_keylistPool = new KeyList(m_iKeySize);
				for (int i = 1; i < m_iPoolSize; i++)
					{
						tmpKey = new KeyList(m_iKeySize);
						tmpKey.next = m_keylistPool;
						m_keylistPool = tmpKey;
					}
				//keys=null;
				m_keylistAktKey = null;
				m_bRun = true;
				while (m_bRun)
					{
						//LogHolder.log(LogLevel.DEBUG, LogType.MISC, "JAPKeyPool:run() loop");
						if (m_keylistPool != null)
							{
								synchronized (this)
									{
										if (!m_bDebug)
											{
												m_SecureRandom.nextBytes(m_keylistPool.key);
											}
										//tmpKey ausketten aus pool
										tmpKey = m_keylistPool;
										m_keylistPool = m_keylistPool.next;

										//einketten in keys..
										//	tmpKey.next=keys;
										//	keys=tmpKey;

										// aktueller=kopf....
										tmpKey.next = m_keylistAktKey;
										m_keylistAktKey = tmpKey;
										synchronized (l2)
											{
												l2.notify();
											}

									}
							}
						else
							{
								try
									{
										synchronized (l1)
											{
												l1.wait();
											}
									}
								catch (InterruptedException e)
									{
										LogHolder.log(LogLevel.DEBUG, LogType.MISC, "JAPKeyPool:run() waiting interrupted!");
									}
							}
					}
			}

		public static int getKey(byte[] key)
			{
				return getKey(key, 0);
			}

		public static int getKey(byte[] key, int offset)
			{
				if (m_KeyPool == null || key == null || (key.length - offset) < m_KeyPool.m_iKeySize)
					{
						return ErrorCodes.E_UNKNOWN;
					}
				//LogHolder.log(LogLevel.DEBUG, LogType.MISC, "JAPKeyPool:getKey()");
				if (m_KeyPool.m_keylistAktKey == null)
					{
						try
							{
								synchronized (m_KeyPool.l2)
									{
										m_KeyPool.l2.wait();
									}
							}
						catch (InterruptedException e)
							{
								LogHolder.log(LogLevel.DEBUG, LogType.MISC, "JAPKeyPool:getKey() waiting interrupted!");
							}
					}
				synchronized (m_KeyPool)
					{
						KeyList tmpKey;
						System.arraycopy(m_KeyPool.m_keylistAktKey.key, 0, key, offset, m_KeyPool.m_iKeySize);
						tmpKey = m_KeyPool.m_keylistAktKey;
						//if(aktKey.next!=null)
						m_KeyPool.m_keylistAktKey = m_KeyPool.m_keylistAktKey.next;
						//else
						//	aktKey=keys;
						tmpKey.next = m_KeyPool.m_keylistPool;

						m_KeyPool.m_keylistPool = tmpKey;
					}
				synchronized (m_KeyPool.l1)
					{
						m_KeyPool.l1.notify();
					}
				return ErrorCodes.E_SUCCESS;
			}
	}
