/*
 * Copyright (c) 2006, The JAP-Team
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *   - Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of the University of Technology Dresden, Germany nor
 *     the names of its contributors may be used to endorse or promote
 *     products derived from this software without specific prior written
 *     permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
package anon.client.crypto;

import java.util.Vector;

/**
 * Holds the information on the different ciphers the mixes in the chain use and
 * allows the encryption of packets to be sent and decryption of packets
 * received from the mix chain.
 *
 * @author Stefan Lieske
 */
public class MixCipherChain
	{

		private IMixCipher[] m_cipherChain;

		public MixCipherChain(IMixCipher[] a_cipherChain)
			{
				m_cipherChain = a_cipherChain;
			}

		/**
		 * Encrypts the buffer for all mixes in the chain, last to first.
		 *
		 * @param a_plainPacket The plain text packet to be encrypted.
		 * @param a_virtualPacketLength The minimal length of the resulting packet.
		 * @param a_sendCallbackHandlers
		 * @return
		 */

		public byte[] encryptPacket(byte[] a_plainPacket, int a_virtualPacketLength, Vector a_sendCallbackHandlers)
			{
				byte[] encryptedPacket = a_plainPacket;

				// encrypt the packet for all mixes, back to front
				for (int i = (m_cipherChain.length - 1); i >= 0; i--)
					{
						encryptedPacket = m_cipherChain[i].encrypt(encryptedPacket, a_virtualPacketLength, a_sendCallbackHandlers);
					}
				return encryptedPacket;
			}

		/**
		 * Returns the number of bytes the encryption of the next packet will
		 * take up.
		 * <p>
		 * Note: This is not the same as the size of the encrypted packet itself.
		 * </p>
		 *
		 * @return amount of overhead bytes
		 */
		public int getNextPacketEncryptionOverhead()
			{
				int overhead = 0;
				for (int i = (m_cipherChain.length - 1); i >= 0; i--)
					{
						overhead = overhead + m_cipherChain[i].getNextPacketEncryptionOverhead();
					}
				return overhead;
			}

		/**
		 * Decrypts the given packet and stores the plain text back into the
		 * given byte buffer.
		 *
		 * @param m_packet byte array representing an encrypted packet
		 *
		 * @return true if the decryption was successful, false otherwise
		 */
		public boolean decryptPacket(byte[] m_packet)
			{
				boolean bSuccess = true;
				for (int i = 0; i < m_cipherChain.length; i++)
					{
						bSuccess = bSuccess && m_cipherChain[i].decrypt(m_packet);
					}
				return bSuccess;
			}

	}
