/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.client.crypto;

import mybouncycastle.org.bouncycastle.crypto.engines.AESFastEngine;
import mybouncycastle.org.bouncycastle.crypto.params.KeyParameter;

final public class SymCipher implements ISymCipher
	{
		public static final int NO_ENCRYPTION = 0;
		public static final int OFB=1;
		public static final int CTR=2;
		public static final int UNKNOWN_ALGORITHM = -1;


		AESFastEngine m_aesEngine1;
		AESFastEngine m_aesEngine2;

		byte[] m_iv1 = null;

		byte[] m_iv2 = null;

		byte[] m_aesKeys;

		/**
		 * Creates a new SymCipher object with two AES engines and two zero'd
		 * initialization vectors.
		 */

		public SymCipher()
			{
				m_aesEngine1 = new AESFastEngine();
				m_aesEngine2 = new AESFastEngine();
				m_aesKeys = null;
				m_iv1 = new byte[16];
				for (int i = 0; i < 16; i++)
					{
						m_iv1[i] = 0;
					}
				m_iv2 = new byte[16];
				for (int i = 0; i < 16; i++)
					{
						m_iv2[i] = 0;
					}
			}

		/**
		 * Sets the AES encryption key to the first 16 byte of the the key buffer.
		 */
		synchronized public int setEncryptionKeyAES(byte[] key)
			{
				return setEncryptionKeyAES(key, 0, 16);
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.ISymCipher#setEncryptionKeyAES(byte[], int, int)
		 */

		synchronized public int setEncryptionKeyAES(byte[] key, int offset, int len)
			{
				try
					{
						m_aesKeys = new byte[16];
						System.arraycopy(key, offset, m_aesKeys, 0, 16);
						m_aesEngine1.init(true, new KeyParameter(m_aesKeys));
						m_aesEngine2.init(true, new KeyParameter(m_aesKeys));
						if (len == 16)
							{
								for (int i = 0; i < 16; i++)
									{
										m_iv1[i] = 0;
										m_iv2[i] = 0;
									}
							}
						else
							{
								/* if the length is >16 it is assumed, that the
								 * following 16 bytes are meant to initialize
								 * the initialization vectors
								 */
								for (int i = 0; i < 16; i++)
									{
										m_iv1[i] = key[16 + i + offset];
										m_iv2[i] = key[16 + i + offset];
									}
							}

						return 0;
					}
				catch (Exception e)
					{
						m_aesKeys = null;

						return -1;
					}
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.ISymCipher#setEncryptionKeysAES(byte[])
		 */
		synchronized public int setEncryptionKeysAES(byte[] keys)
			{
				try
					{
						if (keys.length == 16)
							{
								return setEncryptionKeyAES(keys);
							}
						else
							{
								m_aesKeys = new byte[32];
								System.arraycopy(keys, 0, m_aesKeys, 0, 32);
								m_aesEngine1.init(true, new KeyParameter(m_aesKeys, 0, 16));
								m_aesEngine2.init(true, new KeyParameter(m_aesKeys, 16, 16));
								for (int i = 0; i < 16; i++)
									{
										m_iv1[i] = 0;
										m_iv2[i] = 0;
									}

								return 0;
							}
					}
				catch (Exception e)
					{
						m_aesKeys = null;

						return -1;
					}
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.ISymCipher#getKeys()
		 */
		public byte[] getKeys()
			{
				return m_aesKeys;
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.ISymCipher#setIV2(byte[])
		 */
		synchronized public void setIV2(byte[] buff)
			{
				for (int i = 0; i < 16; i++)
					{
						m_iv2[i] = buff[i];
					}
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.ISymCipher#encryptAES1(byte[], int, byte[], int, int)
		 */
		/**
		 * The AES encryption uses a Output Feedback Block Cipher. This method
		 * uses Key 1 and IV 1.
		 */
		public int encryptAES1(byte[] src, int s_offset, byte[] dest, int d_offset, int len)
			{
				len = s_offset + len;
				while (s_offset < len - 15)
					{
						/*
						 * Encrypt the current IV and replace it with the cipher
						 * text. Then XOR it with the src and copy it into dest
						 * in 16 Byte increments.
						 */

						synchronized (m_aesEngine1)
							{
								m_aesEngine1.processBlock(m_iv1, 0, m_iv1, 0);
							}
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[0]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[1]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[2]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[3]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[4]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[5]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[6]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[7]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[8]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[9]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[10]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[11]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[12]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[13]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[14]);
						dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[15]);
					}
				if (s_offset < len)
					{
						synchronized (m_aesEngine1)
							{
								m_aesEngine1.processBlock(m_iv1, 0, m_iv1, 0);
							}
						len -= s_offset;
						for (int k = 0; k < len; k++)
							{
								dest[d_offset++] = (byte) (src[s_offset++] ^ m_iv1[k]);
							}
					}
				return 0;
			}

		/* (non-Javadoc)
		 * @see anon.client.crypto.ISymCipher#encryptAES2(byte[])
		 */
		/**
		 * The AES encryption uses a Output Feedback Block Cipher. This method
		 * uses Key 2 and IV 2. The IVs get initialized to the same value, but
		 * allow for 2 different OFB Streams.
		 * The buffer gets overwritten with the encrypted packet.
		 */
		public int encryptAES2(byte[] buff)
			{
				int i = 0;
				int len = buff.length;
				while (i < len - 15)
					{
						synchronized (m_aesEngine2)
							{
								m_aesEngine2.processBlock(m_iv2, 0, m_iv2, 0);
							}
						buff[i++] ^= m_iv2[0];
						buff[i++] ^= m_iv2[1];
						buff[i++] ^= m_iv2[2];
						buff[i++] ^= m_iv2[3];
						buff[i++] ^= m_iv2[4];
						buff[i++] ^= m_iv2[5];
						buff[i++] ^= m_iv2[6];
						buff[i++] ^= m_iv2[7];
						buff[i++] ^= m_iv2[8];
						buff[i++] ^= m_iv2[9];
						buff[i++] ^= m_iv2[10];
						buff[i++] ^= m_iv2[11];
						buff[i++] ^= m_iv2[12];
						buff[i++] ^= m_iv2[13];
						buff[i++] ^= m_iv2[14];
						buff[i++] ^= m_iv2[15];
					}
				if (i < len)
					{
						synchronized (m_aesEngine2)
							{
								m_aesEngine2.processBlock(m_iv2, 0, m_iv2, 0);
							}
						len -= i;
						for (int k = 0; k < len; k++)
							{
								buff[i++] ^= m_iv2[k];
							}
					}
				return 0;
			}

		public static int getAlgorithmForName(String strAlg)
			{
				if(strAlg==null)
					return UNKNOWN_ALGORITHM;
				else if(strAlg.equals("AES/OFB/ANON"))
					return OFB;
				else if(strAlg.equals("AES/CTR"))
					return CTR;
				else if(strAlg.equals("NULL"))
					return NO_ENCRYPTION;
				return UNKNOWN_ALGORITHM;
			}

	}
