/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.client.crypto;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;

import javax.crypto.Cipher;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

final public class SymCipherMultithreadedCTR implements ISymCipher
	{
		private Cipher m_aesEngine1;
		private Cipher m_aesEngine2;

		private byte[] m_iv1 = null;

		private byte[] m_iv2 = null;

		private byte[] m_aesKeys;

		private final static int OPAD_LEN=256*16*16;
		private byte[] m_arCounters1;
		private byte[] m_arOpads1;
		private int m_nrThreads=2;
		private int m_iOpadPerThread;
		private EncryptionThread[] m_arEncryptionThreads;
		private Object m_oSyncEncryptionThreads = new Object();
		private int m_OpadIndex = 0;
		private volatile boolean[] m_arRunningThread;

		private volatile int m_iRunningEncryptionThreads;

		private synchronized void incRunningEncryptionthreads()
			{
				m_iRunningEncryptionThreads++;
			}

		private synchronized void decRunningEncryptionthreads()
			{
				m_iRunningEncryptionThreads--;
			}

		private class EncryptionThread implements Runnable
			{
				private byte[] m_arCounters;
				private byte[] m_arOpads;
				private int m_lowerIndex;
				private int m_Len;
				private volatile boolean m_bRun;
				private Cipher m_aesEngine;
				private Thread m_Thread;
				private Object m_oSync;
				private volatile boolean m_bFinished;
	
				protected EncryptionThread(Cipher aesEngine, byte[] counters, byte[] opads, int lowerIndex, int len)
					{
						m_arCounters = counters;
						m_arOpads = opads;
						m_lowerIndex = lowerIndex;
						m_Len = len;
						m_aesEngine = aesEngine;
						m_oSync=new Object();
						m_bFinished=false;
					}

				public void run()
					{
						int lowerArrayIndex = m_lowerIndex;
						while (m_bRun)
							{
								try
									{
										//Encrypt counter, creating opads
										m_aesEngine.update(m_arCounters, lowerArrayIndex, m_Len, m_arOpads, lowerArrayIndex);

										//increment counters
										/*int b = m_arCounters[lowerArrayIndex + 2] + 1;
										for (int i = lowerArrayIndex + 2; i <= upperArrayIndex; i += 4)
											{
												m_arCounters[i] = (byte) b;
											}
										if (b == 0)//overrun 
											{
												b = m_arCounters[lowerArrayIndex + 1] + 1;
												for (int i = lowerArrayIndex + 1; i <= upperArrayIndex; i += 4)
													{
														m_arCounters[i] = (byte) b;
													}
											}
											*/
										//wait for opads to be consumed
										synchronized(m_oSync)
										{
											m_bFinished=true;
											m_oSync.notify();
										}
										synchronized(m_oSync)
											{
											m_oSync.wait();
										}
									}
								catch (Exception e)
									{
										// TODO Auto-generated catch block
										e.printStackTrace();
									}
							}

					}

				protected synchronized void start()
					{
						m_bRun = true;
						m_Thread = new Thread(this, "Multihreads Sym CTR - Encryption Thread");
						m_Thread.setDaemon(true);
						m_Thread.start();
					}

				protected synchronized void stop()
					{
						m_bRun = false;
						synchronized(m_oSyncEncryptionThreads)
						{
							m_oSyncEncryptionThreads.notifyAll();
						}
						try
							{
								m_Thread.join();
							}
						catch (Exception e)
							{
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
					}
				
				protected void waitForEncryptionFinished()
				{
					synchronized(m_oSync)
					{
						while(!m_bFinished)
							{
								try
									{
										m_oSync.wait();
									}
								catch (InterruptedException e)
									{
										// TODO Auto-generated catch block
										e.printStackTrace();
									}
							}
					}
				}
				
				protected void continueEncryption()
				{
					synchronized(m_oSync)
					{
						m_bFinished=false;
						m_oSync.notify();
					}
				}
			}

		public SymCipherMultithreadedCTR()
			{
				try
					{
						m_aesEngine1 = Cipher.getInstance("AES/ECB/NoPadding");
						m_aesEngine2 = Cipher.getInstance("AES/ECB/NoPadding");
					}
				catch (Exception e)
					{
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				m_aesKeys = null;
				m_iv1 = new byte[16];
				for (int i = 0; i < 16; i++)
					{
						m_iv1[i] = 0;
					}
				m_iv2 = new byte[16];
				for (int i = 0; i < 16; i++)
					{
						m_iv2[i] = 0;
					}

			}

		synchronized public int setEncryptionKeyAES(byte[] key)
			{
				return setEncryptionKeyAES(key, 0, 16);
			}

		/**
		 * 
		 * @param key
		 *          byte[]
		 * @param offset
		 *          int
		 * @param len
		 *          int if len==16 --> only the key is set; if len==32 --> key and
		 *          IV is set
		 * @return int
		 */
		synchronized public int setEncryptionKeyAES(byte[] key, int offset, int len)
			{
				try
					{
						m_aesKeys = new byte[16];
						System.arraycopy(key, offset, m_aesKeys, 0, 16);

						if (len == 16)
							{
								m_arCounters1 = new byte[OPAD_LEN];
								m_arOpads1 = new byte[OPAD_LEN];
								int b = 0;
								for (int i = 3; i < 256 * 16; i += 4)
									{
										m_arCounters1[i] = (byte) b;
										b++;
									}
								for (int i = 0; i < 16; i++)
									{
										m_iv1[i] = 0;
										m_iv2[i] = 0;
									}
							}
						else
							{
								for (int i = 0; i < 16; i++)
									{
										m_iv1[i] = key[i + 16 + offset];
										m_iv2[i] = key[i + 16 + offset];
									}
							}

						SecretKeySpec aesKey1 = new SecretKeySpec(m_aesKeys, "AES");
						SecretKeySpec aesKey2 = new SecretKeySpec(m_aesKeys, "AES");
						m_aesEngine1.init(Cipher.ENCRYPT_MODE, aesKey1);
						m_aesEngine2.init(Cipher.ENCRYPT_MODE, aesKey2);

						m_arEncryptionThreads = new EncryptionThread[m_nrThreads];
						m_arRunningThread=new boolean[m_arEncryptionThreads.length];
						int lowerIndex = 0;
						m_iOpadPerThread = OPAD_LEN / m_arEncryptionThreads.length;
						for (int i = 0; i < m_arEncryptionThreads.length; i++)
							{
								SecretKeySpec aesKey = new SecretKeySpec(m_aesKeys, "AES");
								//m_aesEngine1 = Cipher.getInstance("AES/ECB/NoPadding");
								//m_aesEngine1.init(Cipher.ENCRYPT_MODE, aesKey);
								m_arRunningThread[i]=true;
								m_arEncryptionThreads[i] = new EncryptionThread(m_aesEngine1, m_arCounters1, m_arOpads1, lowerIndex,m_iOpadPerThread
										);
								lowerIndex += m_iOpadPerThread;
								m_arEncryptionThreads[i].start();
							}
						return 0;
					}
				catch (Exception e)
					{
						m_aesKeys = null;

						return -1;
					}
			}

		/**
		 * Set both keys to different value, if keys.length==2* KEY_SIZe, else set
		 * both keys to the same value
		 * 
		 * @param keys
		 *          byte[]
		 * @return int
		 */
		synchronized public int setEncryptionKeysAES(byte[] keys)
			{
				try
					{
						if (keys.length == 16)
							{
								return setEncryptionKeyAES(keys);
							}
						else
							{
								m_aesKeys = new byte[32];
								System.arraycopy(keys, 0, m_aesKeys, 0, 32);
								for (int i = 0; i < 16; i++)
									{
										m_iv1[i] = 0;
										m_iv2[i] = 0;
									}

								SecretKeySpec aesKey1 = new SecretKeySpec(m_aesKeys, 0, 16, "AES");
								SecretKeySpec aesKey2 = new SecretKeySpec(m_aesKeys, 16, 16, "AES");
								m_aesEngine1.init(Cipher.ENCRYPT_MODE, aesKey1);
								m_aesEngine2.init(Cipher.ENCRYPT_MODE, aesKey2);
								return 0;
							}
					}
				catch (Exception e)
					{
						m_aesKeys = null;

						return -1;
					}
			}

		/**
		 * Returns the currently used key(s) for encryption. If both keys are set,
		 * than an array of 32 byte is returned, otherwise an array of 16 bytes
		 * 
		 * @return The current key(s) used for encryption or null, if no key is set.
		 */
		public byte[] getKeys()
			{
				return m_aesKeys;
			}

		synchronized public void setIV2(byte[] buff)
			{
				for (int i = 0; i < 16; i++)
					{
						m_iv2[i] = buff[i];
					}

				IvParameterSpec ivSpec2 = new IvParameterSpec(m_iv2);
				SecretKeySpec aesKey2;
				if (m_aesKeys.length == 16)
					{
						aesKey2 = new SecretKeySpec(m_aesKeys, "AES");

					}
				else
					{
						aesKey2 = new SecretKeySpec(m_aesKeys, 16, 16, "AES");
					}
				try
					{
						m_aesEngine2.init(Cipher.DECRYPT_MODE, aesKey2, ivSpec2);
					}
				catch (InvalidKeyException e)
					{
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				catch (InvalidAlgorithmParameterException e)
					{
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
			}

		public int encryptAES1(byte[] from, int ifrom, byte[] to, int ito, int len)
			{
				try
					{
						while (len > 0)
							{
								int remainingOpad = m_iOpadPerThread-(m_OpadIndex % m_iOpadPerThread);
								int nrThread=m_OpadIndex / m_iOpadPerThread;
								int l = Math.min(remainingOpad, len);
								
								m_arEncryptionThreads[nrThread].waitForEncryptionFinished();
								for (int i = l; i > 0; i--)
									{
										to[ito++] = (byte) (from[ifrom++] ^ this.m_arOpads1[m_OpadIndex++]);
									}
								len -= l;
								remainingOpad-=l;
								if(remainingOpad==0)
									m_arEncryptionThreads[nrThread].continueEncryption();
								if(m_OpadIndex==OPAD_LEN)
									{
										m_OpadIndex=0;
									}
							}
					}
				catch (Exception e)
					{
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				return 0;
			}

		public int encryptAES2(byte[] buff)
			{
				try
					{
						int len = buff.length;
						if (true/*len%16==0*/)
							{
								m_aesEngine2.update(buff, 0, len, buff, 0);
							}
						else
							{
								int len1 = len & 0x0FFFFFF0;
								m_aesEngine2.update(buff, 0, len1, buff, 0);
								int len2 = len - len1;
								byte[] tmpBuff = new byte[16];
								System.arraycopy(buff, len1, tmpBuff, 0, len2);
								m_aesEngine2.update(tmpBuff, 0, 16, tmpBuff, 0);
								System.arraycopy(tmpBuff, 0, buff, len1, len2);
							}
					}
				catch (Exception e)
					{
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				return 0;
			}

			
		public static void main(String argv[])
		{
			byte[] k=new byte[16];
			byte[] inBuff=new byte[1600];
			byte[] outBuff=new byte[1600];
			byte[] outBuff2=new byte[1600];
	
			SymCipherMultithreadedCTR c5=new SymCipherMultithreadedCTR();
			c5.setEncryptionKeyAES(k);
			long l1=System.currentTimeMillis();
			c5.setEncryptionKeyAES(k);
					for (int i = 0; i < 10000000; i++)
				c5.encryptAES1(inBuff, 0, outBuff, 0, 1600);
			long l2=System.currentTimeMillis();
			System.out.println("System Multithreaded CTR mode encryption takes: " + (l2 - l1) + " ms");
		Object o=new Object();
/*		synchronized(o)
		{
			try
				{
					o.wait();
				}
			catch (InterruptedException e)
				{
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
		}*/
	}
	
	}
