/*
 Copyright (c) 2000 - 2004, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

  - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
  may be used to endorse or promote products derived from this software without specific
  prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */

package anon.forward.client;

import java.util.Observable;

import anon.util.IProgressCapsule;

/**
 * This class provide a observable for the initializing status bar 
 * The main problem is the number a packet that we should become, we can't know this information
 * We use a arbitral maximum, based on experiment
 */
public class ProgressCounter extends Observable {
	
	/**
	 * Internal implementation of a ProgressCapsule
	 * This Class is trivial, It's only a counter and two constants (the maximum and the minimum)
	 * 
	 */
	private class ProgressCount implements IProgressCapsule {
		/**
		 *  the maximal value for the number of packets
		 *  The maximum is set when we initialize the Capsule 
		 */
		private int m_maxPacket;
		
		/**
		 *  The current value of the counter
		 */
		private int m_counterValue; 
				
		/**
		 * The internal status of the ProgressCapsule
		 * @see IProgressCapsule for the constants
		 */
		private int m_status;
		
		/**
		 * A new capsule has a status "ONGOING"
		 * The maximum can be set by an other function, the minimum is always 0
		 * @param a_max The maximum at the beginning of capsule life
		 */
		ProgressCount (int a_max) {
			m_maxPacket = a_max;
			reset();
		}
		
		public void reset()
		{
			m_counterValue = MINIMUM_PACKET;
			m_status = IProgressCapsule.PROGRESS_ONGOING;
		}
		
		public int getMinimum() {
			return MINIMUM_PACKET;
		}
		
		public int getValue() {
			return m_counterValue; 
		}
		
		public int getStatus() {
			return m_status; 
		}
		public void setStatus(int a_status) {
			m_status = a_status;
		}
		
		public String getMessage()
		{
			return null;
		}
		
		public int getMaximum() {
			return m_maxPacket;
		}
		
		public void setMaximum(int a_maximum) {
			m_maxPacket = a_maximum;
		}
		public void incrValue(int a_incrValue) {
			m_counterValue += a_incrValue;
		}
	}
	
	private ProgressCount m_capsule; 
	
	/**
	 * This is the maximum net size of a protocol message in bytes. Bigger messages are not accepted
	 * and causes an exception.
	 */
	private static final int MAXIMUM_PROTOCOLMESSAGE_SIZE = 1000000;
	
	/**
	 *  The minimum value (0 actually...)
	 */
	private final static int MINIMUM_PACKET = 0;
	
	/** 
	 * The XML connection offer can't be bigger than the Maximum protocol size
	 */
	public ProgressCounter() {
		m_capsule = new ProgressCount(MAXIMUM_PROTOCOLMESSAGE_SIZE);
	}
	
	/**
	 * Set the length, when we know the real size of the XML file
	 * @param a_maximum
	 */
	public void setMax(int a_maximum) {
		m_capsule.setMaximum(a_maximum);
	}
	/**
	 * Increment the counter and update the status of the observer
	 */
	public void incrValue() {
		incrValue(1);
	}
	/**
	 * Increment the counter and update the status of the observer
	 * @param a_incrValue How many bits that we increment
	 */
	public void incrValue(int a_incrValue) {
		synchronized(m_capsule) {
			if (m_capsule.getStatus() == IProgressCapsule.PROGRESS_ONGOING) {		
				m_capsule.incrValue(a_incrValue);
				// Update the status on the GUI
				setChanged();
				notifyObservers(m_capsule);
			}
		}
	}
	
	/**
	 * Close the capsule when the connection is etablished
	 */
	public void close() {
		synchronized(m_capsule) {
			m_capsule.setStatus(IProgressCapsule.PROGRESS_FINISHED);
		}
	}
	
	
}