/*
Copyright (c) 2008 The JAP-Team, JonDos GmbH

All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
       this list of conditions and the following disclaimer in the documentation and/or
       other materials provided with the distribution.
 * Neither the name of the University of Technology Dresden, Germany, nor the name of
       the JonDos GmbH, nor the names of their contributors may be used to endorse or
       promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package anon.forward.server;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

import anon.forward.server.ForwardScheduler;
import anon.forward.server.IServerManager;
import anon.transport.connection.ChunkConnectionAdapter;
import anon.transport.connection.IChunkConnection;
import anon.transport.connection.SkypeConnection;

import com.skype.Application;
import com.skype.ApplicationListener;
import com.skype.Skype;
import com.skype.SkypeException;
import com.skype.Stream;

/**
 * {@link IServerManager} um Verbindungsanforderungen ueber Skype
 * entgegenzunehmen.
 */
public class SkypeServerManager implements IServerManager
	{

		/**
		 * Eigentlicher Listener um ueber Verbindungsanfragen informiert zu werden.
		 * <p>
		 * Wird an eine bestimmte Application gebunden um fuer diese Anfragen
		 * entgegenzunehmen.
		 */
		private class RequestListener implements ApplicationListener
			{

				/**
				 * Behandelt, seitens Skype, neu eingerichtet Verbindungen.
				 */
				public void connected(Stream a_connectionStream) throws SkypeException
					{
						// use the same synchronization, as start and shutdown of
						// SkypeServerManager
						synchronized (SkypeServerManager.this)
							{
								if (!m_isListning || (m_scheduler == null))
									{
										a_connectionStream.disconnect();
										return;
									}
								IChunkConnection baseConnection = new SkypeConnection(a_connectionStream);
								m_scheduler.handleNewConnection(new ChunkConnectionAdapter(baseConnection));

							}
					}

				public void disconnected(Stream a_connectionStream) throws SkypeException
					{
						// disconnects are handled by the Connection itself
					}

			}

		/**
		 * Der Applications Name, und damit die Application an welcher gelauscht
		 * wird.
		 */
		private final String m_appName;

		/** Der Scheduler, welcher die eigentliche Verbindungen verwalted. */
		private ForwardScheduler m_scheduler;

		/** Das Application Object, sofern es eingerichtet werden konnte. */
		private Application m_application;

		/** Der asynchrone Listner fuer neue Verbindungen. */
		private RequestListener m_listner;

		/** Warten wir auf Verbindungsanfragen? */
		private boolean m_isListning;

		/**
		 * Erstellt einen neunen {@link SkypeServerManager} mit dem angegeben
		 * Application Name.
		 * 
		 * @param a_applicationName
		 *          Der Name der Application an welcher auf Verbindungsanfragen
		 *          gelauscht werden soll.
		 */
		public SkypeServerManager(String a_applicationName)
			{
				m_appName = a_applicationName;
				m_isListning = false;
			}

		/**
		 * Die Eindeutige ID des Servermanager.
		 * <p>
		 * verwendet {@link #toString()}, da diese genuegend Eindeutigkeit umsetzt.
		 */
		public Object getId()
			{
				return toString();
			}

		/**
		 * Schliesst den {@link ServerManager} indem neue verbindungsanfragen
		 * ignoriert werden.
		 * <p>
		 * Da die Skype API das entfernen einer registrierten Applikation nicht
		 * vorsieht, kann nur der listener entfernt werden. Um die Moeglichkeit des
		 * Kontaktierens vollstaendig auszuschliessen, muesste Skype geschlossen
		 * werden.
		 */
		public synchronized void shutdown()
			{
				if (!m_isListning)
					return;
				m_application.removeApplicationListener(m_listner);
				try
					{
						m_application.finish();
					}
				catch (Exception e)
					{
						LogHolder.log(LogLevel.EXCEPTION, LogType.TRANSPORT, e);
					}
				m_scheduler = null;
				m_listner = null;
				m_isListning = false;
			}

		/**
		 * Startet des {@link IServerManager} indem versucht wird eine Appliction
		 * mit dem uebergebene Namen zu registrieren und an dieser auf
		 * Verbindungswuensche zu lauschen.
		 */
		public synchronized void startServerManager(ForwardScheduler a_scheduler) throws Exception
			{
				if (m_isListning)
					return;
				m_scheduler = a_scheduler;
				m_listner = new RequestListener();
				try
					{
						m_application = Skype.addApplication(m_appName);
						m_application.addApplicationListener(m_listner);
						m_isListning = true;
					}
				catch (SkypeException e)
					{
						LogHolder.log(LogLevel.ERR, LogType.TRANSPORT, "Could not Start Skype forwarding Server.");
						shutdown();
						throw e;
					}

			}

		/**
		 * String Repraesentation des jeweiligen {@link SkypeServerManager} in der
		 * Form "skype:app(ApplicationName)"
		 */
		// @Override
		public String toString()
			{
				return "skype:app(" + m_appName + ")";
			}

	}
