package anon.infoservice;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import anon.util.Base64;
import anon.util.XMLUtil;

public class ANONAddress extends AbstractDistributableDatabaseEntry
	{
		public static final String XML_ELEMENT_NAME = "ANONAddress";
		private static final long MAX_STORAGE_TIME_IN_MILLI_SECONDS = 24 * 7 * 3600 * 1000; //max time for storing an ANONAddress (currently: 1 week);
		private final static int MAX_ADDRESS_SIZE=32; //how many bytes can be the <Address> element
		private final static int MAX_BLINDED_KEY_SIZE=32; //how many bytes can be the <BlindedKey> element
		private final static int MAX_XML_STRUCT_SIZE=MAX_ADDRESS_SIZE+MAX_BLINDED_KEY_SIZE+200; //maximal size of teh whole XML structure to make DoS attacks a little bit harder
		
		private String m_strID;

		/**
		 * Stores the XML structure of this ANONAddress.
		 */
		private Element m_xmlStructure;

		/** The same as above as byte[] */
		private byte[] m_bytesPostData;

		private long m_lastUpdate;
		
		private String m_strAddress;
		private byte [] m_arBlindedKey;

		public ANONAddress(String id, Element a_elemAnonAddressRootNode) throws Exception
			{
				super(0);
				m_strID = id;

				if (a_elemAnonAddressRootNode == null || !XML_ELEMENT_NAME.equals(a_elemAnonAddressRootNode.getNodeName()))
					throw new Exception("ANONAddress - wrong root XML node name");

				Node elemAddress = XMLUtil.getFirstChildByName(a_elemAnonAddressRootNode, "Address");
				m_strAddress = XMLUtil.parseValue(elemAddress, (String) null);
				Node elemBlindedKey = XMLUtil.getFirstChildByName(a_elemAnonAddressRootNode, "BlindedKey");
				String strBlindedKey = XMLUtil.parseValue(elemBlindedKey, (String) null);
				if(strBlindedKey!=null)
					m_arBlindedKey=Base64.decode(strBlindedKey);
				if (m_strAddress == null&& m_arBlindedKey==null)
					throw new Exception("ANONAddress - not <Address> or <BlindedKey> information given");
				if (m_strAddress!=null&&m_strAddress.length()>MAX_ADDRESS_SIZE)
					throw new Exception("ANONAddress - <Address> to large");
				if (m_arBlindedKey!=null&&m_arBlindedKey.length>MAX_BLINDED_KEY_SIZE)
					throw new Exception("ANONAddress - <BlindedKey> to large");

				Node elemLastUpdate = XMLUtil.getFirstChildByName(a_elemAnonAddressRootNode, "LastUpdate");
				m_lastUpdate = XMLUtil.parseValue(elemLastUpdate, -1L);
				if (m_lastUpdate < 0)
					throw new Exception("ANONAddress - not <LastUpdate> information given");
				m_lastUpdate *= 1000; //make it milli seconds

				m_xmlStructure = a_elemAnonAddressRootNode;
				m_bytesPostData = super.getPostData();
				if (m_bytesPostData.length>MAX_XML_STRUCT_SIZE)
					throw new Exception("ANONAddress - XML struct to large");
				
				Node elemExpire = XMLUtil.getFirstChildByName(a_elemAnonAddressRootNode, "Expire");
				long lExpire = XMLUtil.parseValue(elemExpire, -1L);
				lExpire *= 1000;// make it milli seconds
				if (lExpire < 0 || lExpire > System.currentTimeMillis() + MAX_STORAGE_TIME_IN_MILLI_SECONDS)
					lExpire = System.currentTimeMillis() + MAX_STORAGE_TIME_IN_MILLI_SECONDS;
				setExpireTime(lExpire);
			}
		
		public ANONAddress(String id,String addr,byte[] arBlindedKey)
		{
			super(System.currentTimeMillis()+MAX_STORAGE_TIME_IN_MILLI_SECONDS);
			m_strID = id;
			m_lastUpdate=System.currentTimeMillis();
			m_strAddress=addr;
			m_arBlindedKey=arBlindedKey;
			m_xmlStructure = toXMLElement();
			m_bytesPostData = super.getPostData();
		}

		private Element toXMLElement()
		{
			Document doc=XMLUtil.createDocument();
			Element elemRoot=XMLUtil.createChildElement(doc, XML_ELEMENT_NAME);
			if(m_strAddress!=null)
				{
					Element elemAddr=XMLUtil.createChildElement(elemRoot, "Address");
					XMLUtil.setValue(elemAddr,m_strAddress);
				}
			if(m_arBlindedKey!=null)
				{
					Element elemAddr=XMLUtil.createChildElement(elemRoot, "BlindedKey");
					String strBlindedKey=Base64.encode(m_arBlindedKey, false);
					XMLUtil.setValue(elemAddr,strBlindedKey);
				}
			Element elemLastUpdate=XMLUtil.createChildElement(elemRoot, "LastUpdate");
			XMLUtil.setValue(elemLastUpdate,m_lastUpdate/1000);
			Element elemExpire=XMLUtil.createChildElement(elemRoot, "Expire");
			XMLUtil.setValue(elemExpire,getExpireTime()/1000);
			return elemRoot;
		}
		
		/**
		 * This returns the filename (InfoService command), where this ANONAddress
		 * information is posted to other InfoServices, It's '/anonaddress/id'.
		 *
		 * @return The filename where this ANONaddress entry is posted to other
		 *         InfoServices when this entry is forwarded.
		 */
		public String getPostFile()
			{
				return "/anonaddress/" + m_strID;
			}

		/**
		 * Returns the XML structure of this ANONaddress.
		 *
		 * @return The XML node of this ANONaddress.
		 */
		public Element getXmlStructure()
			{
				return m_xmlStructure;
			}

		public byte[] getPostData()
			{
				return m_bytesPostData;
			}

		public String getId()
			{
				return m_strID;
			}

		public long getLastUpdate()
			{
				return m_lastUpdate;
			}

		public long getVersionNumber()
			{
				return m_lastUpdate;
			}

		public static String getXmlElementName()
			{
				return XML_ELEMENT_NAME;
			}

		public String getAddress()
			{
					return m_strAddress;
			}

		public byte[] getBlindeyKey()
			{
					return m_arBlindedKey;
			}

	}
