package anon.infoservice;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import anon.util.IXMLEncodable;
import anon.util.XMLParseException;
import anon.util.XMLUtil;

/**
 * Stores all exit addresses of one <code>MixCasacde</code>.
 * 
 * @author Christian Banse
 */
public class MixCascadeExitAddresses extends AbstractDatabaseEntry implements IXMLEncodable
{
	/**
	 * The time-to-live of the exit address.
	 */
	public static final long EXIT_ADDRESS_TTL = 1000*60*60*24*7l;
	
	public static final long EXIT_ADDRESS_TTL_INFOSERVICE = 1000*60*60*24l;
	
	/**
	 * The entry's container XML element name.
	 */
	public static final String XML_ELEMENT_CONTAINER_NAME = "ExitAddressesList";

	/**
	 * The XML element name.
	 */
	public static final String XML_ELEMENT_NAME = "ExitAddresses";
	
	/**
	 * The exit address XML element name.
	 */
	public static final String XML_ELEMENT_ADDRESS_NAME = "ExitAddress";
	
	/**
	 * The last update XML attribute name.
	 */
	private static final String XML_ATTR_LAST_UPDATE = "lastUpdate";
	
	public static final String XML_ATTR_DISTRIBUTION = "distribution";
	
	private static final String XML_ATTR_INFO_SERVICE = "infoservice";
	
	/**
	 * The payment XML attribute name.
	 */
	public static final String XML_ATTR_PAYMENT = "payment";
	
	/**
	 * The time of the last update.
	 */
	private long m_lastUpdate;
	
	/**
	 * The cascade id.
	 */
	private String m_strCascadeId = null;
	
	/**
	 * The list of addresses.
	 */
	private Hashtable m_tblAddresses = new Hashtable();
	
	private int m_distribution = MixCascade.DISTRIBUTION_MIN;
	
	private static class InfoServiceID
	{
		private static final String OWN_ID = "OWN_ID";
		private String m_ID;
		private long m_lLastUpdate;
		private InetAddress m_addInet;
		
		private InfoServiceID(InetAddress a_addInet, String a_ID, long a_lLastUpdate)
		{
			m_ID = a_ID;
			m_addInet = a_addInet;
			m_lLastUpdate = a_lLastUpdate;
		}
		
		public boolean equals(Object a_object)
		{
			if (a_object == null)
			{
				return false;
			}
			if (!(a_object instanceof InfoServiceID))
			{
				return false;
			}
			if (((InfoServiceID)a_object).getID().equals(getID()))
			{
				return true;
			}
			return false;
		}
		
		public int hashCode()
		{
			return m_ID.hashCode();
		}
		
		public InetAddress getInetAddress()
		{
			return m_addInet;
		}
		
		public String getID()
		{
			return m_ID;
		}
		
		public long getLastUpdate()
		{
			return m_lLastUpdate;
		}
		
		public String toString()
		{
			return m_ID + " / " + m_addInet;
		}
	}
	
	/**
	 * Constructs a new <code>MixCascadeExitAddress</code> object from a given 
	 * mix cascade id.
	 * 
	 * @param a_strCascadeId The mix cascade id.
	 */
	private MixCascadeExitAddresses(String a_cascadeID, long a_lastUpdate)
	{
		super(a_lastUpdate + EXIT_ADDRESS_TTL);
		
		m_strCascadeId = a_cascadeID;
		
		m_lastUpdate = a_lastUpdate;
	}
	
	/**
	 * Only used in client context.
	 * @param a_xmlElement
	 * @throws XMLParseException
	 */
	public MixCascadeExitAddresses(Element a_xmlElement) throws XMLParseException
	{
		super(System.currentTimeMillis() + EXIT_ADDRESS_TTL);
		
		InetAddress address;
		long lTime;
		
		
		XMLUtil.assertNodeName(a_xmlElement, XML_ELEMENT_NAME);
		XMLUtil.assertNotNull(a_xmlElement, XML_ATTR_ID);
		m_strCascadeId = XMLUtil.parseAttribute(a_xmlElement, XML_ATTR_ID, null);
		m_distribution = XMLUtil.parseAttribute(a_xmlElement, XML_ATTR_DISTRIBUTION, MixCascade.DISTRIBUTION_MAX);
		
		MixCascade cascade = (MixCascade)Database.getInstance(MixCascade.class).getEntryById(m_strCascadeId);
		if (cascade != null)
		{
			m_distribution = cascade.getDistribution();
		}
		
		m_lastUpdate = System.currentTimeMillis();
		
		NodeList lstAddresses = a_xmlElement.getElementsByTagName(XML_ELEMENT_ADDRESS_NAME);
		for (int i = 0; i < lstAddresses.getLength(); i++)
		{
			lTime = XMLUtil.parseAttribute(lstAddresses.item(i), 
					XML_ATTR_LAST_UPDATE, System.currentTimeMillis());
			XMLUtil.assertNotNull(lstAddresses.item(i));
			
			try 
			{
				address = InetAddress.getByName(XMLUtil.parseValue(lstAddresses.item(i), null));
			} 
			catch (UnknownHostException e) 
			{
				LogHolder.log(LogLevel.WARNING, LogType.NET, e);
				continue;
			}
			
			//strID = XMLUtil.parseAttribute(lstAddresses.item(i), XML_ATTR_INFO_SERVICE, null);
			addInetAddress(address, m_distribution, lTime, null);
		}
	}
	
	public String getId() 
	{
		return m_strCascadeId;
	}
	
	public int getDistribution() 
	{
		return m_distribution;
	}
	
	public long getLastUpdate() 
	{
		return m_lastUpdate;
	}
	
	public long getVersionNumber() 
	{
		return m_lastUpdate;
	}
	
	public String createExitAddressAsString()
	{
		synchronized (m_tblAddresses)
		{
			Enumeration enumer = m_tblAddresses.keys();
			InetAddress addr;
			InfoServiceID id;
			Hashtable hashID;
			
			if (m_tblAddresses.size() == 1)
			{
				//System.out.println("One: " + getId());
				addr = ((InetAddress)enumer.nextElement());
				
				hashID = (Hashtable)m_tblAddresses.get(addr);
				enumer = hashID.elements();
				while (enumer.hasMoreElements())
				{
					if (((InfoServiceID)enumer.nextElement()).getLastUpdate() <
							System.currentTimeMillis() - EXIT_ADDRESS_TTL)
					{
						continue;
					}
					return addr.getHostAddress();
				}
				
				return null;
			}
			else if (m_tblAddresses.size() > 1)
			{
				//System.out.println("More: " + getId());
				
				// it may not be clear which address we have...
				boolean bIPv6 = false;
				String strCurrentIP;
				
				String strNextIP = null;
				Enumeration enumerIDs;
				
				int index;
				
				Vector vecSuitableAddresses = new Vector();
				while (enumer.hasMoreElements())
				{
					addr = (InetAddress)enumer.nextElement();
					
					
					hashID = (Hashtable)m_tblAddresses.get(addr);
					enumerIDs = hashID.elements();
					
					/*
					if (getId().equals("F30905FCD73B6B30CB5FEFD3250FD66EF4B32591"))
					{
						System.out.println("Current (" + hashID.size() + "): " + addr);
					}*/
					
					int iCounter = 0;
					while (enumerIDs.hasMoreElements())
					{
						id = (InfoServiceID)enumerIDs.nextElement();
						if (id.getLastUpdate() >= 
							System.currentTimeMillis() - EXIT_ADDRESS_TTL)
						{
							iCounter++;
							if (id.getID() == InfoServiceID.OWN_ID)
							{
								iCounter = Integer.MAX_VALUE;
								break;
							}
						}
					}
					
					if (iCounter > 1)
					{
						vecSuitableAddresses.addElement(addr);
					}
				}
				
				if (vecSuitableAddresses.size() == 0)
				{
					// we cannot decide which of the addresses is valid
					return null;
				}
				
				strCurrentIP = ((InetAddress)vecSuitableAddresses.elementAt(0)).getHostAddress();
				if (vecSuitableAddresses.size() == 1)
				{
					return strCurrentIP;
				}
				
				/*
				if (getId().equals("F30905FCD73B6B30CB5FEFD3250FD66EF4B32591"))
				for (int i = 0; i < vecSuitableAddresses.size(); i++)
				{
					System.out.println("Suitable:" + ((InetAddress)vecSuitableAddresses.elementAt(i)));
				}*/
				
				
				
				
				for (int i = 1; i < vecSuitableAddresses.size(); i++)
				{
					strNextIP = ((InetAddress)vecSuitableAddresses.elementAt(i)).getHostAddress();
					index = strNextIP.lastIndexOf(".");
					if (index < 0)
					{
						index = strNextIP.lastIndexOf(":");
						bIPv6 = true;
					}
					
					strNextIP = strNextIP.substring(0, index);
					if (!strCurrentIP.startsWith(strNextIP))
					{
						return null;
					}
				}
				return strNextIP + (bIPv6 ? ":" : ".") + "*";
			}
		}
		return null;
	}
	
	public static MixCascadeExitAddresses addInetAddress(String a_cascadeID, 
			InetAddress a_IPAddress, int a_cascadeDistribution, String a_strInfoServiceID)
	{
		synchronized (Database.getInstance(MixCascadeExitAddresses.class))
		{
			MixCascadeExitAddresses exit = null;
			MixCascadeExitAddresses exitNew = null;
			long lastUpdate = System.currentTimeMillis();
			if (a_cascadeID != null && a_cascadeID.trim().length() > 0 && a_IPAddress != null)
			{
				exit = (MixCascadeExitAddresses) Database.getInstance(MixCascadeExitAddresses.class).
					getEntryById(a_cascadeID);
				
				if (exit != null)
				{
					lastUpdate = exit.getLastUpdate() + 1; // force a database update
				}
				exitNew = new MixCascadeExitAddresses(a_cascadeID, lastUpdate);
				
				if (exit != null)
				{
					synchronized (exit.m_tblAddresses)
					{
						Enumeration enumer = exit.m_tblAddresses.keys();
						InetAddress addr;
						InfoServiceID id;
						Hashtable hashID;
						Enumeration enumID;
						while (enumer.hasMoreElements())
						{
							addr = (InetAddress)enumer.nextElement();
							hashID = (Hashtable)exit.m_tblAddresses.get(addr);
							
							/*
							if (exit.getId().equals("F30905FCD73B6B30CB5FEFD3250FD66EF4B32591"))
							{
								System.out.println("Kept: " + addr);
							}*/
							
							enumID = hashID.elements();
							while (enumID.hasMoreElements())
							{
								id = (InfoServiceID)enumID.nextElement();
								exitNew.addInetAddress(addr, a_cascadeDistribution, id.getLastUpdate(), 
										id.getID());
							}
							//exitNew.m_tblAddresses.put(addr, exit.m_tblAddresses.get(addr));
						}
					}
				}
				
			
				if (exitNew.addInetAddress(a_IPAddress, a_cascadeDistribution, a_strInfoServiceID))
				{	
					if (Database.getInstance(MixCascadeExitAddresses.class).update(exitNew))
					{
						/*
						if (exitNew.getId().equals("F30905FCD73B6B30CB5FEFD3250FD66EF4B32591"))
						{
							System.out.println("Updated: " + a_IPAddress);
						}*/
						
						return exitNew;
					}
				}
			}
			return exit;
		}
	}
	
	public static boolean isValidAddress(InetAddress a_address)
	{
		return isValidAddress(a_address, "isAnyLocalAddress") ||
			isValidAddress(a_address, "isLoopbackAddress") ||
			isValidAddress(a_address, "isLinkLocalAddress") ||
			isValidAddress(a_address, "isMulticastAddress") ||
			isValidAddress(a_address, "isSiteLocalAddress");
	}
	
	private static boolean isValidAddress(InetAddress a_address, String a_methodName)
	{
		try
		{
			return !((Boolean)InetAddress.class.getMethod(a_methodName, 
					(Class[])null).invoke(a_address, (Object[])null)).booleanValue();
		}
		catch (Exception a_e)
		{
//			 maybe java too old
			return true;
		}
	}
	
	private boolean addInetAddress(InetAddress a_addr, int a_cascadeDistribution,
			String a_strInfoServiceID)
	{
		return addInetAddress(a_addr, a_cascadeDistribution, System.currentTimeMillis(),
				a_strInfoServiceID);
	}
	
	/**
	 * Adds an address to the list.
	 * 
	 * @param a_addr The address to add.
	 * @return True, if the list has changed, false otherwise.
	 */
	private boolean addInetAddress(InetAddress a_addr, int a_cascadeDistribution, 
			long a_timestamp, String a_strInfoServiceID)
	{
		boolean bChanged = false;
		
		if (a_strInfoServiceID == null)
		{
			a_strInfoServiceID = InfoServiceID.OWN_ID;
		}
		
		synchronized (m_tblAddresses)
		{
			// enumerate through all entries and remove obsolete ones
			Enumeration e = m_tblAddresses.keys();
			Enumeration enumIDs;
			Vector vecIDsToRemove;
			InfoServiceID id;
			Hashtable hashID;
			
			vecIDsToRemove = new Vector();
			while(e.hasMoreElements())
			{
				InetAddress addr = (InetAddress) e.nextElement();
				hashID = (Hashtable)m_tblAddresses.get(addr);
				enumIDs = hashID.elements();
				while (enumIDs.hasMoreElements())
				{
					id = (InfoServiceID)enumIDs.nextElement();
					if (id.getLastUpdate() < System.currentTimeMillis() - EXIT_ADDRESS_TTL)
					{
						LogHolder.log(LogLevel.WARNING, LogType.DB, 
								"Exit address expired: " + id);
						// exit address expired, remove it
						vecIDsToRemove.addElement(id);
						bChanged = true;
					}
				}
			}
			
			for (int i = 0; i < vecIDsToRemove.size(); i++)
			{
				id = (InfoServiceID)vecIDsToRemove.elementAt(i);
				hashID = (Hashtable)m_tblAddresses.get(id.getInetAddress());
				hashID.remove(id.getID());
				if (hashID.size() == 0)
				{
					m_tblAddresses.remove(id.getInetAddress());
				}
			}
			
			
			if (m_distribution != a_cascadeDistribution)
			{
				m_distribution = a_cascadeDistribution;
				bChanged = true;
			}
			
			
			if (a_timestamp < System.currentTimeMillis() - EXIT_ADDRESS_TTL)
			{
				return false;
			}
			
			
			//a_timestamp = Math.max(a_timestamp, System.currentTimeMillis());
			boolean bAdded = false;
			// add the address to the list
			if (!m_tblAddresses.containsKey(a_addr))
			{
				hashID = new Hashtable();
				hashID.put(a_strInfoServiceID, new InfoServiceID(a_addr, 
						a_strInfoServiceID, a_timestamp));
				bChanged = true;
				bAdded = true;
				
				/*
				if (getId().equals("F30905FCD73B6B30CB5FEFD3250FD66EF4B32591"))
				{
					System.out.println("Added1:" + a_addr + " (" + a_strInfoServiceID + ")");
				}*/
			}
			else
			{
				hashID = (Hashtable)m_tblAddresses.get(a_addr);
				if (!hashID.contains(a_strInfoServiceID) ||
						((InfoServiceID)hashID.get(a_strInfoServiceID)).getLastUpdate() < a_timestamp)
				{
					hashID.put(a_strInfoServiceID, new InfoServiceID(a_addr, 
							a_strInfoServiceID, a_timestamp));
					bChanged = true;
					bAdded = true;
					/*
					if (getId().equals("F30905FCD73B6B30CB5FEFD3250FD66EF4B32591"))
					{
						System.out.println("Added2:" + a_addr);
					}*/
				}
			}
			m_tblAddresses.put(a_addr, hashID);
			
			/*
			if (bAdded)
			{
				LogHolder.log(LogLevel.ALERT, LogType.MISC, "Added: " + a_addr + "  (" + a_strInfoServiceID + ")");
			}*/
			
			/*
			if (bAdded && getId().equals("F30905FCD73B6B30CB5FEFD3250FD66EF4B32591"))
			{
				hashID = ((Hashtable)m_tblAddresses.get(a_addr));
				System.out.println("Size:" + m_tblAddresses.size());
			}*/
		}
			
			
		return bChanged;
	}
	
	public Element toXmlElement(Document a_doc)
	{
		Element elem = a_doc.createElement(XML_ELEMENT_NAME);
		XMLUtil.setAttribute(elem, XML_ATTR_ID, getId());
		XMLUtil.setAttribute(elem, XML_ATTR_DISTRIBUTION, getDistribution());
		//XMLUtil.setAttribute(elem, XML_ATTR_PAYMENT, m_bPayment);
		
		synchronized (m_tblAddresses)
		{
			Enumeration e = m_tblAddresses.keys();
			InfoServiceID id;
			Hashtable hashID;
			Element el;
			Enumeration enumIDs;
			long timestamp;
			String strID;
			Vector vecAddrToDelete = new Vector();
			
			while (e.hasMoreElements())
			{
				InetAddress addr = (InetAddress) e.nextElement();
				
				//LogHolder.log(LogLevel.ALERT, LogType.MISC, "Try: " + addr );
				
				hashID = (Hashtable) m_tblAddresses.get(addr);
				if (hashID.containsKey(InfoServiceID.OWN_ID) || hashID.size() > 1)
				{
					strID = null;
					timestamp = 0;
					if (hashID.size() > 1)
					{
						hashID.remove(InfoServiceID.OWN_ID);
						if (hashID.size() > 1)
						{
							timestamp = 0;
							enumIDs = hashID.elements();
							while (enumIDs.hasMoreElements())
							{
								id = (InfoServiceID)enumIDs.nextElement();
								if (id.getLastUpdate() > timestamp)
								{
									timestamp = id.getLastUpdate();
									strID = id.getID();
								}
							}
						}
					}
					else if (hashID.containsKey(InfoServiceID.OWN_ID))
					{
						id = (InfoServiceID)hashID.get(InfoServiceID.OWN_ID);
						timestamp = id.getLastUpdate();
						strID = InfoServiceID.OWN_ID;
					}
					//LogHolder.log(LogLevel.ALERT, LogType.MISC, "Show: " + addr);
					
					if (timestamp >= System.currentTimeMillis() - EXIT_ADDRESS_TTL)
					{
						el = a_doc.createElement(XML_ELEMENT_ADDRESS_NAME);
						XMLUtil.setAttribute(el, XML_ATTR_LAST_UPDATE, timestamp);
						if (strID != null && !strID.equals(InfoServiceID.OWN_ID))
						{
							XMLUtil.setAttribute(el, XML_ATTR_INFO_SERVICE, strID);
						}
						
						XMLUtil.setValue(el, addr.getHostAddress());
						elem.appendChild(el);
					}
					else
					{
						vecAddrToDelete.addElement(addr);
					}
				}
				/*
				else if (hashID.size() == 1)
				{	
					LogHolder.log(LogLevel.ALERT, LogType.MISC, "Failed: " + addr 
							+ "  (" + ((InfoServiceID)hashID.elements().nextElement()).getID() + ")");
				}*/
			}
			
			for (int i = 0; i < vecAddrToDelete.size(); i++)
			{
				m_tblAddresses.remove(vecAddrToDelete.elementAt(i));
			}
		}
		
		return elem;
	}
}
