/*
Copyright (c) 2007, The JAP-Team, JonDos GmbH
All rights reserved.
Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the name of
the JonDos GmbH, nor the names of their contributors may be used to endorse or
promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.infoservice.update;

import anon.util.Updater;
import anon.pay.xml.XMLAccountInfo;
import anon.pay.xml.XMLBalance;
import logging.LogLevel;
import logging.LogHolder;
import java.util.Enumeration;
import anon.pay.PayAccountsFile;
import anon.pay.PayAccount;
import logging.LogType;


public class AccountUpdater extends Updater
{
	private boolean m_successfulUpdate = false;
	private long m_lLastUpdate = Long.MAX_VALUE;
	private boolean m_bInternalCalculation;

	public AccountUpdater(final boolean a_bInternalCalculation)
	{
		super(new IUpdateInterval()
		{
			public long getUpdateInterval()
			{
				// we differentiate between active account and other accounts later
				return PayAccount.ACCOUNT_MIN_UPDATE_INTERVAL_MS;
			}
		}
		,new ObservableInfo(PayAccountsFile.getInstance())
		{
			public Integer getUpdateChanged()
			{
				return PayAccountsFile.CHANGED_AUTO_UPDATE;
			}

			public boolean isUpdateDisabled()
			{
				if (a_bInternalCalculation)
				{
					return PayAccountsFile.getInstance().isBalanceAutoUpdateEnabled();
				}
				
				return !PayAccountsFile.getInstance().isBalanceAutoUpdateEnabled();
			}
			
			public boolean updateImmediately()
			{
				return true;
			}
		}
		);
		
		m_bInternalCalculation = a_bInternalCalculation;
	}

	/**
	 * getUpdatedClass
	 *
	 * @return Class
	 */
	public Class getUpdatedClass()
	{
		return XMLAccountInfo.class;
	}

	/**
	 * Does the update and should tell if it was successful or not.
	 *
	 */
	protected void updateInternal()
	{
		m_successfulUpdate = false;
		//System.out.println("account update");

		if (Thread.currentThread().isInterrupted())
		{
			// this thread is being stopped; ignore this error
			m_successfulUpdate = true;
			return;
		}

		Enumeration accounts = PayAccountsFile.getInstance().getAccounts();
		boolean bTried = false;
		while (accounts.hasMoreElements() && !Thread.currentThread().isInterrupted())
		{
			PayAccount account = (PayAccount) accounts.nextElement();
			XMLAccountInfo accountInfo;
			try
			{
				if (!m_bInternalCalculation && account.shouldUpdateAccountInfo())
				{
					bTried = true;
					LogHolder.log(LogLevel.DEBUG, LogType.PAY,
								  "Fetching statement for account: " + account.getAccountNumber());
					accountInfo = account.fetchAccountInfo(false);
					if (accountInfo != null)
					{
						m_successfulUpdate = true;
						m_lLastUpdate = System.currentTimeMillis();
					}
				}
				else if (m_bInternalCalculation)
				{
					// this is necessary to update capped flat rate volumes if balance update is disabled
					accountInfo = account.getAccountInfo();
					XMLBalance balance;
					
					if (accountInfo != null)
					{
						balance = accountInfo.getBalance();
						if (!account.hasExpired() && balance != null &&
								(account.getCurrentSpent() < 0 || (	// negative value for spent - we might have lost CCs?
								balance.getVolumeBytesMonthly() > 0 &&
								(balance.getOverusageDate() != null || 
										accountInfo.getLastBalanceUpdateLocalTime().getTime() > System.currentTimeMillis())))) // we switched to the past
						{
							// with overusage, we definitely need automatic balance updates!
							accountInfo = account.fetchAccountInfo(true);
							if (accountInfo != null)
							{
								m_successfulUpdate = true;
								m_lLastUpdate = System.currentTimeMillis();
							}
						}
						else
						{
							// do only local calculation
							accountInfo.checkMonthlyBytesUpdatedOn();
							m_successfulUpdate = true;
							m_lLastUpdate = System.currentTimeMillis();
						}
					}
				}
			}
			catch (Exception e)
			{
				LogHolder.log(LogLevel.ERR, LogType.PAY,
							  "Could not fetch statement for account: " + account.getAccountNumber(), e);
				
				if (account.getAccountInfo() != null)
				{
					account.getAccountInfo().checkMonthlyBytesUpdatedOn();
				}
			}
		}

		if (Thread.currentThread().isInterrupted())
		{
			// this thread is being stopped; ignore this error
			m_successfulUpdate = true;
			return;
		}
		else if (PayAccountsFile.getInstance().getNumAccounts() == 0)
		{
			// no reason to show an error...
			m_successfulUpdate = true;
		}
		else if (!m_bInternalCalculation && !bTried)
		{
			// no reason to show an error...
			m_successfulUpdate = true;
		}
	}
	
	public final long getLastUpdate()
	{
		return m_lLastUpdate;
	}
	
	/**
	 * wasUpdateSuccessful
	 *
	 * @return boolean
	 */
	protected boolean wasUpdateSuccessful()
	{
		return m_successfulUpdate;
	}
}
