/*
 Copyright (c) 2000 - 2006, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.pay;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import anon.crypto.ByteSignature;

import anon.client.AnonClient;
import anon.client.ChannelTable;
import anon.client.Multiplexer;
import anon.client.PacketCounter;
import anon.client.XmlControlChannel;
import anon.error.AnonServiceException;
import anon.error.ProtocolViolationException;
import anon.error.ServiceInterruptedException;
import anon.pay.xml.NotRecoverableXMLError;
import anon.pay.xml.XMLAccountInfo;
import anon.pay.xml.XMLAiLoginConfirmation;
import anon.pay.xml.XMLChallenge;
import anon.pay.xml.XMLEasyCC;
import anon.pay.xml.XMLErrorMessage;
import anon.pay.xml.XMLPayRequest;
import anon.pay.xml.XMLResponse;
import anon.pay.xml.XMLPriceCertificate;
import anon.util.XMLUtil;

import java.util.Date;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;
import java.util.Enumeration;
import anon.IServiceContainer;
import anon.infoservice.ListenerInterface;
import anon.infoservice.MixCascade;
import anon.infoservice.MixPosition;
import java.util.Hashtable;
import java.sql.Timestamp;

/**
 * This control channel is used for communication with the AI
 * (AccountingInstance or Abrechnungsinstanz in German) which lives in the first
 * mix. The AI sends a request when it wants a cost confirmation from us. This
 * thread waits for incoming requests and sends the requested confirmations to
 * the AI.
 *
 * @author Bastian Voigt, Tobias Bayer, Rolf Wendolsky
 * @version 1.0
 */
public class AIControlChannel extends XmlControlChannel
{
	public static final long MAX_PREPAID_INTERVAL = 4000000; // 4MB
	public static final long MIN_PREPAID_INTERVAL = 5000; // 500 kb
	public static final long AI_LOGIN_TIMEOUT = 120000; // 2 minutes
	
	private static final long MULTIPLE_LOGIN_BLOCK_TIME = 30 * 1000l;

	private static final int FIRST_MIX = 0;
	private static final String PREPAID_AMOUNT_IN_PAY_REQ_MIXVERSION = "00.08.42";
	
	public static final boolean REVERT_PRE_PREPAID = true;
	
	public static final Hashtable HASH_PREPAID_ON_SERVICES = new Hashtable(); 
	
  //codes for AI events that can be fired
  //private static final int EVENT_UNREAL = 1;

  private int m_aiLogin_timeout = AnonClient.DEFAULT_LOGIN_TIMEOUT;

  private XMLErrorMessage m_lastErrorMessage;

  private static long m_totalBytes = 0;

  private boolean m_bPrepaidReceived = false;
  private long m_prepaidBytes = 0;
  private boolean m_bMultiplexerClosed = false;

  private Vector m_aiListeners = new Vector();

  private PacketCounter m_packetCounter;
  private EmptyAccountPacketObserver m_packetCountEmptyObserver;
  private boolean m_bEmptyMessageSent = false;

  private MixCascade m_connectedCascade;
  
  private PayAccount m_payAccount;

  private XMLEasyCC m_initialCC;

  private final Vector m_aiLoginSyncObject = new Vector(1);

  /**
   * indicates that the prepaid Bytes are transmitted by a PayRequest 
   * instead of a challenge. (supported by mixes of version >= 08.42 )
   */
  private boolean m_prepaidAmountInPayRequest = false;
  
  private boolean m_bLastCCSignaled = false;
  
public AIControlChannel(Multiplexer a_multiplexer,
						  PacketCounter a_packetCounter, IServiceContainer a_serviceContainer,
						  MixCascade a_connectedCascade) {
    super(ChannelTable.CONTROL_CHANNEL_ID_PAY, a_multiplexer, a_serviceContainer,true);
    m_packetCounter = a_packetCounter;
	m_connectedCascade = a_connectedCascade;
	
	if (m_packetCounter == null)
	{
		throw new NullPointerException();
	}
	
	//m_synchronizedAILogin =  true;
	
	m_packetCountEmptyObserver = new EmptyAccountPacketObserver(m_connectedCascade.getConcatenatedPriceCertHashes());
	
	m_packetCounter.addObserver(m_packetCountEmptyObserver);
	
	String firstMixSoftwareVersion = 
		m_connectedCascade.getMixInfo(FIRST_MIX).getServiceSoftware().getVersion();
	
	
	/* check the first mix software version to determine if client 
	 * can already perform the new synchronized ai login procedure.
	 * (for mix version >= 00.07.20) and transmitting of prepaid bytes 
	 * in the PayRequest (for mix version >= 00.08.42). 
	 */
	
	if(firstMixSoftwareVersion!= null)
	{
		/* TODO does not work... Prepaid amount MUST come BEFORE process initial CC...
		 * For some weird reason this is not the case with the test mix...?
		m_prepaidAmountInPayRequest =
			firstMixSoftwareVersion.compareTo(PREPAID_AMOUNT_IN_PAY_REQ_MIXVERSION) >= 0;
			*/
	}
	
	LogHolder.log(LogLevel.INFO, LogType.PAY,
			"Mix "+m_connectedCascade.getMixInfo(FIRST_MIX).getName() + 
			(m_prepaidAmountInPayRequest ? " supports " : " does not support ")+
			"improved prepaid bytes negotiation.");
	
  }

  public void addAIListener(IAIEventListener a_aiListener) {
	 synchronized (m_aiListeners)
	 {
	    if (!m_aiListeners.contains(a_aiListener)) 
	    {
	      m_aiListeners.addElement(a_aiListener);
	    }
	 }
  }

  /**
   * proccessXMLMessage - this is called when a new request is coming in.
   *
   * @param docMsg
   *          Document
   */
  public void processXmlMessage(Document docMsg)
  {
	  Element elemRoot = docMsg.getDocumentElement();
	  String tagName = elemRoot.getTagName();
	  
	 
	  if (!tagName.equals(XMLErrorMessage.XML_ELEMENT_NAME) && 
			(m_payAccount == null || m_payAccount.getCurrentCredit() <= 0))
	  {
		  elemRoot = new XMLErrorMessage(XMLErrorMessage.ERR_ACCOUNT_EMPTY, 
				  m_payAccount , m_connectedCascade).toXmlElement(docMsg);
		  tagName = XMLErrorMessage.XML_ELEMENT_NAME;
	  }
	  
	  try
	  {
		  if (tagName.equals(XMLPayRequest.XML_ELEMENT_NAME))
		  {
			  XMLPayRequest theRequest = new XMLPayRequest(elemRoot);
			  processPayRequest(theRequest);
		  }
		  else if (tagName.equals(XMLAiLoginConfirmation.XML_ELEMENT_NAME))
		  {
			  XMLAiLoginConfirmation loginConfirm =
				  new XMLAiLoginConfirmation(elemRoot);
			  /*LogHolder.log(LogLevel.EXCEPTION, LogType.PAY,
						"Thread "+ Thread.currentThread().getName()+" getting AI Login Confirmation with code "
						+loginConfirm.getCode()+" and message "+loginConfirm.getMessage());*/

			  synchronized(m_aiLoginSyncObject)
			  {
				  if(loginConfirm.getCode()==XMLErrorMessage.ERR_OK)
				  {
					  m_aiLoginSyncObject.addElement(new Object());
				  }
				  m_aiLoginSyncObject.notifyAll();
			  }
		  }
		  else if (tagName.equals(XMLErrorMessage.XML_ELEMENT_NAME))
		  {
			  XMLErrorMessage error = new XMLErrorMessage(elemRoot, m_payAccount, m_connectedCascade);
			  LogHolder.log(LogLevel.EXCEPTION, LogType.PAY,
						"For account " + m_payAccount.getAccountNumber() + ", processing AI ErrorMessage "+error.getXmlErrorCode()+": "+error.getMessage());
			  
			  if (error.getXmlErrorCode() == XMLErrorMessage.ERR_MULTIPLE_LOGIN)
			  {
				  if (error.getService() != null && error.getService() instanceof MixCascade)
				  {
					  Vector vecListenerInterfaces = 
						  ((MixCascade)error.getService()).getListenerInterfaces();
					  for (int i = 0; vecListenerInterfaces != null && 
					  		i < vecListenerInterfaces.size(); i++)
					  {
						  ((ListenerInterface)vecListenerInterfaces.elementAt(i)).blockInterface(
								  MULTIPLE_LOGIN_BLOCK_TIME);
					  }
				  }
			  }
				 
			  /*
			  if (error.getXmlErrorCode() != XMLErrorMessage.ERR_BLOCKED
					  && m_payAccount.isBlocked())
			  {
				  LogHolder.log(LogLevel.ERR, LogType.PAY, 
						  "We should see a 'blocked' error message for account " + m_payAccount.getAccountNumber() + "!");
				  error = new XMLErrorMessage(error.getErrorCode(), error.getMessage(), 
						  error.getAccount(), error.getService());
			  }
			  */
			  
			  if (error.getXmlErrorCode() ==  XMLErrorMessage.ERR_ACCOUNT_EMPTY &&
				!checkAccountChanged())
			  {
				  // find an account that is not empty - if possible...
				  synchronized (PayAccountsFile.getInstance())
				  {
					  updateBalance(m_payAccount, false); // show that account is empty
					  PayAccount currentAccount = PayAccountsFile.getInstance().getChargedAccount(
									  m_connectedCascade.getPIID(), m_payAccount);
	
					  if (currentAccount != null)
					  {
						  PayAccountsFile.getInstance().setActiveAccount(currentAccount);
					  }
					  else
					  {
						  if (!PayAccountsFile.getInstance().isAIAccountErrorIgnored())
						  {
							  error = new NotRecoverableXMLError(error);
						  }
						  getServiceContainer().keepCurrentService(false); // reconnect to another cascade if possible
					  }
				  }
			  }
			  else if (error.getXmlErrorCode() ==  XMLErrorMessage.ERR_BLOCKED && 
					  !checkAccountChanged())
			  {
				  m_payAccount.setBlocked(true);
				  synchronized (PayAccountsFile.getInstance())
				  {
					  PayAccount currentAccount = 
						  PayAccountsFile.getInstance().getChargedAccount(
									  m_connectedCascade.getPIID());
	
					  if (currentAccount != null)
					  {
						  PayAccountsFile.getInstance().setActiveAccount(currentAccount);
					  }
					  else
					  {
						  if (!PayAccountsFile.getInstance().isAIAccountErrorIgnored())
						  {
							  error = new NotRecoverableXMLError(error);
						  }
						  getServiceContainer().keepCurrentService(false); // reconnect to another cascade if possible
					  }
				  }
			  }
			  else
			  {
				  getServiceContainer().keepCurrentService(false); // reconnect to another cascade if possible
			  }
			  m_lastErrorMessage = error;
			  if (error instanceof NotRecoverableXMLError)
			  {
				  error = ((NotRecoverableXMLError)error).getSource();
			  }
			  PayAccountsFile.getInstance().signalAccountError(error);
		  }
		  else if (tagName.equals(XMLChallenge.XML_ELEMENT_NAME))
		  {
			  processChallenge(new XMLChallenge(elemRoot));
		  }
		  else if (tagName.equals(XMLEasyCC.XML_ELEMENT_NAME))
		  {
			  processInitialCC(new XMLEasyCC(elemRoot));
		  }
		  else
		  {
			  LogHolder.log(LogLevel.WARNING, LogType.PAY,
							"Received unknown payment control channel message '" + tagName + "'");
		  }
	  }
	  catch (Exception ex)
	  {
		  LogHolder.log(LogLevel.EXCEPTION, LogType.PAY, ex);
		  getServiceContainer().keepCurrentService(false); // reconnect to another cascade if possible
		  PayAccountsFile.getInstance().signalAccountError(
			  new XMLErrorMessage(XMLErrorMessage.ERR_BAD_REQUEST,
								  ex.getClass().getName() + ": " + ex.getMessage(), 
								  m_payAccount, m_connectedCascade));
	  }
  }

  private synchronized void handlePrepaidBytesReceived(int prepaidBytes, PayAccount activeAccount)
  {
	  if (activeAccount == null)
	  {
		  throw new NullPointerException("Active Account must not be null!");
	  }
	  if ( prepaidBytes > 0 )	// ignore prepaid bytes smaller than zero
	  {
		 if (!m_bPrepaidReceived)
		 {
			 m_prepaidBytes = prepaidBytes;
			 
			 PreviousPrepdaidBytes ppBytes = getPreviousPrepaidBytes();
			 if (ppBytes != null && !ppBytes.m_dateTimeout.before(new Date()))
			 {
				 // get the difference
				 ppBytes.m_prepaidBytes = Math.max(0, ppBytes.m_prepaidBytes - m_prepaidBytes);
			 }
			 
			 m_bPrepaidReceived = true;
			 activeAccount.updateCurrentBytes(prepaidBytes * ( -1)); // subtract from transferred bytes
		 }
	  }
  }
  
  /**
   * processChallenge
   *
   * @param xMLChallenge
   *          XMLChallenge
   */
  private synchronized void processChallenge(XMLChallenge chal) throws Exception
  {
	  byte[] arbChal = chal.getChallengeForSigning();

	  LogHolder.log(LogLevel.NOTICE, LogType.PAY, "Received " + chal.getPrepaidBytes() + " prepaid bytes.");
	  
	  if (chal.getType() != null && !chal.getType().equals(XMLChallenge.TYPE_MIX))
	  {
		  throw new Exception("Challenge has wrong type: " + chal.getType());
	  }
	  
	  if (m_payAccount == null)
	  {
		  throw new Exception("Received Challenge from AI but ActiveAccount not set!");
	  }
	  
	  //Although it is deprecated to receive the prepaid bytes with the challenge
	  //they're still handled here for backward compatibility reasons.
	  if (!m_prepaidAmountInPayRequest)
	  {
		  handlePrepaidBytesReceived(chal.getPrepaidBytes(), m_payAccount);
	  }
	  /*if (!m_bPrepaidReceived && chal.getPrepaidBytes() > 0)
	  {
		  m_prepaidBytes = chal.getPrepaidBytes();
		 System.out.println("Prepaid bytes: "+m_prepaidBytes);
		  m_bPrepaidReceived = true;
		  // ignore prepaid bytes smaller than zero
		  acc.updateCurrentBytes(chal.getPrepaidBytes() * ( -1)); // remove transferred bytes
	  }*/

	  byte[] arbSig = ByteSignature.sign(arbChal, m_payAccount.getPrivateKey());
	  XMLResponse response = new XMLResponse(arbSig, null);
	  this.sendXmlMessage(XMLUtil.toXMLDocument(response));
  }

  private boolean checkAccountChanged()
  {
	  if (m_payAccount == null || m_payAccount != PayAccountsFile.getInstance().getActiveAccount())
	  {
		  synchronized (m_aiListeners)
		  {
			  LogHolder.log(LogLevel.WARNING, LogType.PAY, "Active account has changed while connected to paid service!");
			  for (int i = 0; i < m_aiListeners.size(); i++)
			  {
				  ((IAIEventListener)m_aiListeners.elementAt(i)).accountChanged(m_payAccount, m_connectedCascade);
			  }
		  }
		  
		  return true;
	  }
	  return false;
  }
  
  /**
   * process a XMLPayRequest message, which might request a XMLAccountCertificate,
   * request a XMLBalance, or contain a XMLEasyCC which the AI asks the JAP to sign
   *
   * @param request
   *          XMLPayRequest
   */
  private synchronized void processPayRequest(XMLPayRequest request) 
  { 
	  if (checkAccountChanged())
	  {
		  return;
	  }
	  
	//if requested, send account certificate
	if (request.isInitialCCRequest())
	{
		if (m_prepaidAmountInPayRequest)
		{
			handlePrepaidBytesReceived(request.getPrepaidBytes(), m_payAccount);
		}
		processInitialCC(request.getCC());
		return;
	}
	if (request.isAccountRequest())
	{
		try
		{
			sendAccountCert();
		}
		catch (AnonServiceException a_e)
		{
			LogHolder.log(LogLevel.ALERT, LogType.PAY, "Could not send account certificate!", a_e);
		}
	}

	//if sent, process cost confirmation    
      try 
      {
			 processCcToSign(request.getCC());
      }
      catch (Exception ex1) 
      {
        // the account stated by the AI does not exist or is not currently
        // active
        // @todo handle this exception
        LogHolder.log(LogLevel.ERR, LogType.PAY, ex1);
      }

  }

  private void updateBalance(final PayAccount currentAccount, final boolean a_bSynchronous)
  {
	  Runnable runUpdate;
	  if (currentAccount == null)
	  {
		  return;
	  }

	  runUpdate = new Runnable()
	  {
		  public void run()
		  {
			  try
			  {
				  if (a_bSynchronous)
				  {
					  /* set a short connection timeout as this update blocks connection to a Cascade */
					  currentAccount.fetchAccountInfo(false, 2000); 
				  }
				  else
				  {
					  currentAccount.fetchAccountInfo(false);
				  }
			  }
			  catch (Exception ex)
			  {
				  LogHolder.log(LogLevel.DEBUG, LogType.PAY, ex);
			  }
		  }
	  };

	  if (a_bSynchronous)
	  {
		  LogHolder.log(LogLevel.DEBUG, LogType.PAY, "Fetching new Balance from BI.");
		  runUpdate.run();
	  }
	  else
	  {
		  LogHolder.log(LogLevel.DEBUG, LogType.PAY, "Fetching new Balance from BI asynchronously.");
		  Thread t = new Thread(runUpdate);
		  t.setDaemon(true);
		  t.start();

	  }
  }

	/**
	 * processCcToSign: to be called by processPayRequest
	 * (only the initial, old CC is sent as a naked XMLEasyCC,
	 * new CCs to sign are sent inside a XMLPayRequest)
	 *
	 */
	private synchronized void processCcToSign(XMLEasyCC cc) throws Exception
	{
		if (cc == null || cc.getConcatenatedPriceCertHashes() == null || 
			!cc.getConcatenatedPriceCertHashes().equals(m_connectedCascade.getConcatenatedPriceCertHashes()))
		{
			return;
		}
		
		if (checkAccountChanged())
		{
			throw new Exception("Active account has changed!");
		}
		
		long transferedBytes;
		XMLEasyCC myLastCC;
		
		//check CC for proper account number
		if ( (m_payAccount == null) || (m_payAccount.getAccountNumber() != cc.getAccountNumber()))
		{
			throw new Exception("Received CC with wrong accountnumber");
		}

		/*
		if (!m_bInitialCCSent)
		{
			LogHolder.log(LogLevel.WARNING, LogType.PAY, "CC requested before initial CC was sent!");
		}*/


	  // long transferedBytes = currentAccount.updateCurrentBytes(m_packetCounter);
		m_payAccount.updateCurrentBytes(m_packetCounter);
	  //LogHolder.log(LogLevel.DEBUG, LogType.PAY, "AI requests to sign " + transferedBytes + " transferred bytes");

	//System.out.println("PC Hash looked: " + cc.getConcatenatedPriceCertHashes());
	  myLastCC = m_payAccount.getAccountInfo().getCC(cc.getConcatenatedPriceCertHashes());
	  //System.out.println("toSign: " + cc.getConcatenatedPriceCertHashes());
	  long confirmedBytes = 0;
	  if (myLastCC != null)
	  {
		  if (m_initialCC == null)
		  {
			  // this seems to be the first connection to the cascade
			  LogHolder.log(LogLevel.WARNING, LogType.PAY,
							"No initial CC available! The Mix might have lost its CC.");
			  if (m_prepaidBytes > 0)
			  {
				  // remove any prepaid bytes received before
				  m_payAccount.updateCurrentBytes(m_prepaidBytes);
			  }
		  }
		  else
		  {
			  confirmedBytes = myLastCC.getTransferredBytes();
			  LogHolder.log(LogLevel.DEBUG, LogType.PAY, "Transferred bytes of last CC: " + confirmedBytes);
		  }
	  }

	  //System.out.println("To confirm1: " + cc.getTransferredBytes() + " Transferred1: " + transferedBytes);


	  transferedBytes = m_payAccount.getCurrentBytes();  // maybe transferred bytes have changed above...
	  m_totalBytes = transferedBytes;

	  // check if bytes asked for in CC match bytes transferred
	  long newPrepaidBytes = //(cc.getTransferredBytes() - confirmedBytes) + m_prepaidBytes - transferedBytes;
		  cc.getTransferredBytes() - transferedBytes;
	  long diff = newPrepaidBytes - m_connectedCascade.getPrepaidInterval();

	  /*
	  if (transferedBytes < confirmedBytes)
	  {
		  LogHolder.log(LogLevel.ALERT, LogType.PAY, "Less transferred than confirmed bytes! "+
			  "Transferred: " + transferedBytes + " Confirmed: " + confirmedBytes);
	  }*/

	  if (diff > 0)
	  {
		  // this may happen with fast downloads; the bytes are counted in the Mix before JAP receives them
		  LogHolder.log(LogLevel.WARNING, LogType.PAY,
						"Illegal number of prepaid bytes for signing. Difference/Spent/CC/PrevCC: " +
						diff + "/" + transferedBytes + "/" + cc.getTransferredBytes() + "/" + confirmedBytes);

		  if (transferedBytes < 0)
		  {
			  LogHolder.log(LogLevel.WARNING, LogType.PAY, "The mix might have lost a CC. " +
							"Resetting transferred bytes to zero for now...");
			  m_payAccount.updateCurrentBytes(transferedBytes * (-1));
			  transferedBytes = m_payAccount.getCurrentBytes();
		  }
		  else if (cc.getTransferredBytes() < confirmedBytes)
		  {
			  LogHolder.log(LogLevel.WARNING, LogType.PAY,
							"Requested less than confirmed before! Maybe a CC did get lost!");
			  //this.fireAIEvent(EVENT_UNREAL, diff); // do not show this problem to the user...
		  }
	  }
	  cc.setTransferredBytes(transferedBytes + m_connectedCascade.getPrepaidInterval());


	  //get pricecerts and check against hashes in CC
	  //get price certs from connected cascade
	  /**
	   * We don't even bother to check the hashes in the CC,
	   * the JAP just fills in the ones he knows from the Cascade
	   */
	  if (cc.getTransferredBytes() > confirmedBytes)
	  {
		// cc.setTransferredBytes(-1l);
	  
		  if (!m_bLastCCSignaled)
		  {
			  long lTransferred = cc.getTransferredBytes();
			  long lRemaining = cc.getTransferredBytes();
			  long lTotalToZero;
			  
			  if (myLastCC != null)
			  {
				  lRemaining -= myLastCC.getTransferredBytes();
			  }
			  
			  if (m_payAccount.getCurrentCredit() < lRemaining)
			  {
				  lTotalToZero = m_payAccount.getCurrentCredit();
				  if (myLastCC != null)
				  {
					  lTotalToZero += myLastCC.getTransferredBytes();
				  }
				  
				  cc.setTransferredBytes(lTotalToZero);
				  cc.setLastCC(true);
				  m_bLastCCSignaled = true;
				  
				  LogHolder.log(LogLevel.WARNING, LogType.PAY, "For account " + m_payAccount.getAccountNumber() +
						  ", " + lRemaining +" credits were requested. Should provide " + lTransferred + 
						  ", but can only provide " +  lTotalToZero + ". Missing credits: " + (lTransferred - lTotalToZero));
			  }
		  }
		  cc.setPriceCerts(m_connectedCascade.getPriceCertificateHashes());
		  cc.setPIID(m_payAccount.getAccountCertificate().getPIID());
		  cc.setCascadeID(m_connectedCascade.getId());
		  cc.sign(m_payAccount.getPrivateKey());
		  if (m_payAccount.addCostConfirmation(cc, true) <= 0)
		  {
			  LogHolder.log(LogLevel.WARNING, LogType.PAY, "Added old cost confirmation!");
		  }
	  }
	  else if (myLastCC != null && m_initialCC != null)
	  {
		  // resend the last valid CC
		 // System.out.println("old CC: " + cc.getConcatenatedPriceCertHashes());
		  cc = myLastCC;
	  }
	  else
	  {
		   LogHolder.log(LogLevel.EMERG, LogType.PAY, "Creating zero CC!!");
		  // send a zero CC
		  cc.setTransferredBytes(0);
		  cc.setPriceCerts(m_connectedCascade.getPriceCertificateHashes());
		  cc.setPIID(m_payAccount.getAccountCertificate().getPIID());
		  cc.setCascadeID( m_connectedCascade.getId());
		  cc.sign(m_payAccount.getPrivateKey());
		  m_payAccount.addCostConfirmation(cc, true);
	  }
	  //System.out.println("Processed: " + cc.getTransferredBytes());

	  //System.out.println("To confirm2: " + cc.getTransferredBytes() + " Transferred2: " + transferedBytes);

	  if (m_initialCC == null)
	  {
		  // this seems to be the first connection to the cascade
		  LogHolder.log(LogLevel.NOTICE, LogType.PAY, 
				  "Seems to be the first connection to service. Setting initial CC to current CC...");
		  m_initialCC = cc;
	  }

	  //System.out.println(cc.getTransferredBytes());

	 if (!m_payAccount.isCharged(new Timestamp(System.currentTimeMillis())))
	 {
		 synchronized(m_aiLoginSyncObject)
		 {
			 if (!m_bMultiplexerClosed)
			 {
				 m_packetCountEmptyObserver.setEmpty();
			 }
		 }
	 }
	  sendXmlMessage(XMLUtil.toXMLDocument(cc));
	 
	}
	 
	 private final class EmptyAccountPacketObserver implements Observer
	 {
		 private String m_concatenatedPCHashes;
		 private boolean m_bEmpty = false;
		 
		private EmptyAccountPacketObserver(String a_concatenatedPCHashes)
		{
			m_concatenatedPCHashes = a_concatenatedPCHashes;
		}
		 
		public void setEmpty()
		{
			m_bEmpty = true;
		}
		
		 public void update(Observable o, Object arg)
		 {
			 new Thread(new Runnable()
			 {// Thread is necessary to prevent deadlock with anon.client.Multiplexer: notifyObservers(new PacketProcessedEvent
				 public void run()
				 {
			 synchronized (AIControlChannel.this)
			 {
				 synchronized (m_aiLoginSyncObject)
				 {
					 if (!m_bEmptyMessageSent)
					 {
						 try
						 {
							 long prepaidBytes;
							 long diff;
							 m_payAccount.updateCurrentBytes(m_packetCounter);
							 XMLEasyCC myLastCC = m_payAccount.getAccountInfo().getCC(m_concatenatedPCHashes);
							 prepaidBytes = myLastCC.getTransferredBytes() - m_payAccount.getCurrentBytes();
							 diff =  m_payAccount.getCurrentCreditCalculatedAlsoNegative();
							 if (diff < 0)
							 {
								 prepaidBytes += diff;
							 }
							 //System.out.println("Prepaid: " + prepaidBytes + " diff: " + diff + " current: " + account.getCurrentBytes() + " transferred: " + myLastCC.getTransferredBytes());
				
						
							 if (!m_bEmpty && prepaidBytes < 1000000)
							 {
								/*
								 *  We need to send a cost confirmation! Maybe we did not receive a
								 *  CC request from the mix?
								 */
								 
								 /* TODO patch the mix so that this gets possible... */
								 /*
								try 
								{
									XMLEasyCC cc = new XMLEasyCC(m_payAccount.getAccountNumber(), 
											myLastCC.getTransferredBytes() + Math.min(m_connectedCascade.getPrepaidInterval(), MAX_PREPAID_INTERVAL), 
											null, m_connectedCascade.getPriceCertificateHashes(), m_connectedCascade.getId(), m_payAccount.getPIID());
									 processCcToSign(cc);
								}
								catch (Exception ex1) 
								{
									LogHolder.log(LogLevel.ERR, LogType.PAY, ex1);
								}
								return;
								*/
							 }
							 
							 if (!m_bEmpty || prepaidBytes > 0)
							 {
								 return;
							 }
						 }
						 catch (Exception a_e)
						 {
							 LogHolder.log(LogLevel.EXCEPTION, LogType.PAY, a_e);
						 }
					 }
					 /*
					 if (PayAccountsFile.getInstance().getChargedAccount(m_connectedCascade.getPIID()) == null)
					 {
						 new Thread()
						 {
							 public void run()
							 {*/
								 PayAccountsFile.getInstance().signalAccountError(
											new XMLErrorMessage(XMLErrorMessage.ERR_ACCOUNT_EMPTY,
													m_payAccount, m_connectedCascade));
							 /*}
						 }.start();
					 }*/
					 
					 // close the current connection by notifying an empty account
					synchronized (m_aiListeners)
					 {
						 for (int i = 0; i < m_aiListeners.size(); i++)
						 {
							 ((IAIEventListener)m_aiListeners.elementAt(i)).accountEmpty(
									 m_payAccount, m_connectedCascade);
						 }
						 m_bEmptyMessageSent = true;
					 }
			 	}
			 }
				 }
			 }).start();
		 }
	 }


  public void sendAccountCert() throws AnonServiceException
  {

	  String message = null;
	  Vector priceCerts = m_connectedCascade.getPriceCertificates();
	  String mixID;
	  XMLPriceCertificate priceCert;
	  Timestamp now = new Timestamp(System.currentTimeMillis());
	  
	  if (m_payAccount != null)
	  {
		  throw new ProtocolViolationException(m_connectedCascade, "Account certificate already received!");
	  }
	  
	  synchronized (PayAccountsFile.getInstance())
	  {
		  m_payAccount = PayAccountsFile.getInstance().getActiveAccount();
	
		  if (m_payAccount == null || !m_payAccount.isCharged(now) || (m_payAccount.getBI() == null) ||
			  !m_payAccount.getBI().getId().equals(m_connectedCascade.getPIID()))
		  {
			  PayAccount currentAccount = null;
			  PayAccount openTransactionAccount = null;
			  if (m_payAccount != null && 
					  //activeAccount.getSpent() == 0)
					  m_payAccount.getCurrentSpent() == 0)// spent means account has been used
			  {
				  openTransactionAccount = PayAccountsFile.getInstance().getActiveAccount();
			  }
	
			  Vector accounts = PayAccountsFile.getInstance().getAccounts(m_connectedCascade.getPIID());
			  if (accounts.size() > 0)
			  {
				  for (int i = 0; i < accounts.size(); i++)
				  {
					  currentAccount = (PayAccount) accounts.elementAt(i);
					  if (currentAccount.isCharged(now))
					  {
						  break;
					  }
					  else if (openTransactionAccount == null && 
							  //currentAccount.getSpent() == 0 &&
							  (m_payAccount == null || m_payAccount.getCurrentSpent() == 0) &&
							  !currentAccount.hasExpired())
					  {
						  openTransactionAccount = currentAccount;
					  }
					  currentAccount = null;
				  }
				  if (currentAccount != null)
				  {
					  PayAccountsFile.getInstance().setActiveAccount(currentAccount);
				  }
				  else if (openTransactionAccount != null)
				  {
					  PayAccountsFile.getInstance().setActiveAccount(openTransactionAccount);
				  }
				  
				  PayAccount activeAccount = PayAccountsFile.getInstance().getActiveAccount();
				  
				  // check if the account is charged; if not, try to get the latest balance from each account
				  if ((activeAccount== null ||
					  !activeAccount.isCharged(now) || 
					  !activeAccount.isSamePaymentInstance(m_connectedCascade.getPIID())) &&
					  accounts.size() > 0)
				  {
					  LogHolder.log(LogLevel.WARNING, LogType.PAY,
									"No charged account is available for connecting. Trying to update balances...");
					  for (int i = 0; i < accounts.size(); i++)
					  {
						  currentAccount = (PayAccount) accounts.elementAt(i);
						  if (currentAccount.canDoMonthlyOverusage(now))
						  {
							  PayAccountsFile.getInstance().setActiveAccount(currentAccount);
							  break;
						  }
						  else if (currentAccount.getBalance() == null || currentAccount.shouldUpdateAccountInfo())
						  {
							  // update the account if the timestamp is quite old
							  //updateBalance(currentAccount, true);
							  updateBalance(currentAccount, false); // do not block!
							  if (currentAccount.isCharged(now))
							  {
								  PayAccountsFile.getInstance().setActiveAccount(currentAccount);
								  break;
							  }
						  }
					  }
				  }
			  }
		  }
		  
		  m_payAccount = PayAccountsFile.getInstance().getActiveAccount();
	  }
	  
	  PayAccountsFile.getInstance().signalAccountRequest(m_connectedCascade);
	  
	  
	  if (m_payAccount == null || m_payAccount.getCurrentCredit() <= 0)
	  {
		  getServiceContainer().keepCurrentService(false); // reconnect to another service if possible
		  throw new XMLErrorMessage(XMLErrorMessage.ERR_ACCOUNT_EMPTY,
				  "No active account is available! Cannot connect to paid cascade.",
					  m_payAccount , m_connectedCascade);
	  }

	  if (priceCerts.size() != m_connectedCascade.getNumberOfMixes())
	  {
		  message = "Not all Mixes in cascade " + m_connectedCascade.getId() + " have price certs! " +
		  "PriceCerts/MixIDs:" + priceCerts.size() + "/" + m_connectedCascade.getNumberOfMixes();
	  }
	  else
	  {
		  for (int i = 0; i < m_connectedCascade.getNumberOfMixes(); i++)
		  {
			  priceCert = ( (XMLPriceCertificate) priceCerts.elementAt(i));
			  mixID = m_connectedCascade.getMixId(i);
			  PaymentInstanceDBEntry pi = m_payAccount.getBI();
			  if (!priceCert.verify(pi))
			  {
				  message = "Price certificate of cascade " + m_connectedCascade.getId() + " for mix " +
					  mixID + " cannot be verified for payment instance " + pi + " (" + m_payAccount.getPIID() +")" + "!";
				  break;
			  }

			  if (!priceCert.getSubjectKeyIdentifier().equals(mixID))
			  {
				  message = "SKI in price certificate of cascade " + m_connectedCascade.getId() +
					  " differs from Mix ID! SKI:" + priceCert.getSubjectKeyIdentifier() + " MixID: " + mixID;
				  break;
			  }
		  }
	  }

	  if (message != null)
	  {
		  LogHolder.log(LogLevel.ERR, LogType.PAY, message);

		  getServiceContainer().keepCurrentService(false); // reconnect to another cascade if possible
		  PayAccountsFile.getInstance().signalAccountError(
				  new XMLErrorMessage(XMLErrorMessage.ERR_INVALID_PRICE_CERTS, message, 
						  m_payAccount, m_connectedCascade));


		  throw new AnonServiceException(m_connectedCascade, message);
	  }

	  m_payAccount.resetCurrentBytes();
	  sendXmlMessage(XMLUtil.toXMLDocument(m_payAccount.getAccountCertificate()));
	  /*
	   * new ai login procedure: wait until all messages are
	   * exchanged or until login is timed out
	   */
	  AnonServiceException error = null;
	  synchronized(m_aiLoginSyncObject)
	  {
		  /* we wait until the mix confirms a successful/unsuccessful login or
		   * the connection timed out. 
		   */
		  LogHolder.log(LogLevel.INFO, LogType.PAY, "Performing new synchronous AI login");
		  try {
			m_aiLoginSyncObject.wait(m_aiLogin_timeout);
		  }
		  catch (InterruptedException e) {
			/* This happens when a user pushes the anonymity off button before
			 * the synchronous login hasn't finished. Therefore just leave. false will be returned
			 * automatically.
			 */
			 error = new ServiceInterruptedException(m_connectedCascade);
		  }
		  //LogHolder.log(LogLevel.ALERT, LogType.PAY, m_aiLoginSyncObject);
		  if ( m_aiLoginSyncObject.size() <= 0) 
		  {
			  if (m_lastErrorMessage != null) // maybe we got a payment error?
			  {
				  error = m_lastErrorMessage;
			  }
			  else
			  {
				  error = new AnonServiceException(m_connectedCascade, "No login confirmation found!");
			  }
		  }
		  m_aiLoginSyncObject.removeAllElements();
	  }
	  if (error != null)
	  {
		throw error;  
	  }
  }


  public static long getBytes() {
    return m_totalBytes;
  }

  	private PreviousPrepdaidBytes getPreviousPrepaidBytes()
  	{
			return (PreviousPrepdaidBytes)HASH_PREPAID_ON_SERVICES.get(
					m_connectedCascade.getConcatenatedPriceCertHashes() + m_payAccount.getAccountNumber());
  	}
  
	/**
	 * processInitialCC: last step of connecting to a pay cascade:
	 * take last CC as sent by AI as base value for future CCs
	 * Also, send a CC for (bytes in last CC + prepay interval of cascade)
	 * to avoid triggering the cascade's hardlimit by starting to transfer bytes without prepaying
	 *
	 * @param a_cc XMLEasyCC: the last CC that the JAP sent to this Cascade, as returned from the AI
	 */
	private synchronized void processInitialCC(XMLEasyCC a_cc) 
	{
		  if (checkAccountChanged())
		  {
			  return;
		  }
		
		String msg = "AI has sent an INVALID last cost confirmation.";

		if (a_cc.verify(m_payAccount.getPublicKey()))
		{
			try
			{
				//compare number
				if (a_cc.getNrOfPriceCerts() != m_connectedCascade.getNrOfPriceCerts())
				{
					LogHolder.log(LogLevel.EXCEPTION, LogType.PAY,
								  "number of price certificates in cost confirmation does not match " +
								  "number of price certs in cascade");
					getServiceContainer().keepCurrentService(false); // reconnect to another cascade if possible
					PayAccountsFile.getInstance().signalAccountError(
						new XMLErrorMessage(XMLErrorMessage.ERR_INVALID_PRICE_CERTS,
											"AI sent CC will illegal number of price certs" +
											a_cc.getNrOfPriceCerts(), m_payAccount,
											m_connectedCascade));
					return;
				}

				/**
				 * Check if the CC contains the correct price certs
				 */
				//get hashes from CC
				Hashtable hashPriceCertHashesInCC = a_cc.getPriceCertHashes();
				Enumeration priceCertHashesInCascade = m_connectedCascade.getPriceCertificateHashes().keys();
				Hashtable inCascade = m_connectedCascade.getPriceCertificateHashes();
				String curCascadeHash;
				String curCcHash;
				MixPosition ski;
				int i = 0;
				while (priceCertHashesInCascade.hasMoreElements()) //enough to use one enum in condition, since we already checked for equal size
				{
					ski = ((MixPosition)priceCertHashesInCascade.nextElement());
					curCascadeHash = (String)inCascade.get(ski);
					curCcHash = (String)hashPriceCertHashesInCC.get(ski);

					if (curCcHash == null || !curCascadeHash.equals(curCcHash))
					{
						String message = "AI sent CC with illegal price cert hash for mix " +
							(ski.getPosition() + 1) + " (" + (i + 1) + ")" +  "!";
						if (curCcHash == null)
						{
							message += " Price certificate for this Mix was not found in CC!";
						}
						LogHolder.log(LogLevel.WARNING, LogType.PAY, message);
						getServiceContainer().keepCurrentService(false); // reconnect to another cascade if possible
						PayAccountsFile.getInstance().signalAccountError(
							new XMLErrorMessage(XMLErrorMessage.ERR_INVALID_PRICE_CERTS, message,
									m_payAccount, m_connectedCascade));
						return;
					}
					hashPriceCertHashesInCC.remove(ski); // do not count duplicate price certs
					i++;
				}
				
				if (m_connectedCascade.getConcatenatedPriceCertHashes() == null || 
					a_cc.getConcatenatedPriceCertHashes() == null ||
						!m_connectedCascade.getConcatenatedPriceCertHashes().equals(a_cc.getConcatenatedPriceCertHashes()))
				{
					PayAccountsFile.getInstance().signalAccountError(
							new XMLErrorMessage(XMLErrorMessage.ERR_INVALID_PRICE_CERTS, "Price certificate hashes for a CC for service " + 
									m_connectedCascade.getName() + " cannot be verified!",
									m_payAccount, m_connectedCascade));
						return;
				}
				
				//System.out.println("Inital1 CC: " + a_cc.getConcatenatedPriceCertHashes());

				LogHolder.log(LogLevel.DEBUG, LogType.PAY,
							  "AI has sent a valid last cost confirmation. Adding it to account.");
				//no need to verify the price certificates of the last CC, since they might have changed since then
				//System.out.println("PC Hash stored: " + a_cc.getConcatenatedPriceCertHashes());

				if (m_initialCC == null)
				{
					m_payAccount.updateCurrentBytes(a_cc.getTransferredBytes());
					m_initialCC = a_cc;
				}
				else
				{
					long diff = a_cc.getTransferredBytes() - m_initialCC.getTransferredBytes();
					LogHolder.log(LogLevel.WARNING, LogType.PAY,
								  "Updated initial CostConfirmation! Difference: " + diff);
					m_payAccount.updateCurrentBytes(diff);		
				}
				long confirmedbytes = a_cc.getTransferredBytes();

				long diff = m_payAccount.addCostConfirmation(a_cc, true);
				if (diff < 0)
				{
					//confirmedbytes = 0;
					m_payAccount.updateCurrentBytes(diff);
					LogHolder.log(LogLevel.WARNING, LogType.PAY, "Received old cost confirmation!");
				}
				else if (diff > 0)
				{
					LogHolder.log(LogLevel.WARNING, LogType.PAY, "Restored lost cost confirmation!");
				}
				//m_payAccount.updateCurrentBytes(diff);

				//System.out.println("Initial: " + a_cc.getTransferredBytes() + ":" + a_cc.getConcatenatedPriceCertHashes());
				//get Cascade's prepay interval
				

				long currentlyTransferedBytes = m_payAccount.getCurrentBytes(); //currentAccount.updateCurrentBytes(m_packetCounter);
				long bytesToPay =
					m_connectedCascade.getPrepaidInterval() - (confirmedbytes - currentlyTransferedBytes); 
				//NOTE: (confirmedbytes - currentlyTransferedBytes) in this case means: my prepaidBytes (on this cascade). 

				long oldBytes = a_cc.getTransferredBytes();
				XMLEasyCC newCC = new XMLEasyCC(a_cc);
				//send CC for up to <last CC + prepay interval> bytes
				if (bytesToPay > 0)
				{
					PreviousPrepdaidBytes ppBytes = getPreviousPrepaidBytes();
					if (ppBytes != null)
					{
						if (ppBytes.m_dateTimeout.before(new Date()))
						{
							HASH_PREPAID_ON_SERVICES.remove(
									m_connectedCascade.getConcatenatedPriceCertHashes() + m_payAccount.getAccountNumber());
							ppBytes = null;
						}
						else if (ppBytes.m_prepaidBytes + m_prepaidBytes + bytesToPay > MAX_PREPAID_INTERVAL)
						{
							PayAccountsFile.getInstance().signalAccountError(
									new XMLErrorMessage(XMLErrorMessage.ERR_BAD_REQUEST, "The service " + 
											m_connectedCascade.getName() + " requests too many prepaid bytes. From previous logins, we " +
											"have:" + ppBytes.m_prepaidBytes + " Now additionally: " + bytesToPay,
											m_payAccount, m_connectedCascade));
							return;
						}
					}
					
					
					if (m_payAccount.getCurrentCredit() < bytesToPay)
					{
						long originBytesToPay = bytesToPay;
						bytesToPay = m_payAccount.getCurrentCredit();
						newCC.setLastCC(true);
						m_bLastCCSignaled = true;
						  
						LogHolder.log(LogLevel.WARNING, LogType.PAY, "For account " + m_payAccount.getAccountNumber() +
							", " + originBytesToPay +" credits were requested. Should provide " + (confirmedbytes + originBytesToPay) + 
							", but can only provide " +  (confirmedbytes + bytesToPay) + ". Missing credits: " + (originBytesToPay - bytesToPay));
					}
					
					newCC.setTransferredBytes(confirmedbytes + bytesToPay);
				}
				else
				{
					newCC.setTransferredBytes(confirmedbytes);
				}
				m_prepaidBytes += bytesToPay;
				
				//System.out.println("bytestopay: " + bytesToPay);
				
				//newCC.setTransferredBytes(1000000000l);
				newCC.sign(m_payAccount.getPrivateKey());
				if (bytesToPay > 0 && m_payAccount.addCostConfirmation(newCC, true) <= 0)
				{
					LogHolder.log(LogLevel.WARNING, LogType.PAY, "Sending old cost confirmation! " +
								  "Diff (ShouldBe)/Old/New:" + bytesToPay + "/" + oldBytes + "/" +
								  newCC.getTransferredBytes());
				}

				//System.out.println("Initial: " + newCC.getTransferredBytes());

				// always send the message to tell the Mix the CC was received
				sendXmlMessage(XMLUtil.toXMLDocument(newCC));

				return;
			}
			catch (Exception e)
			{
				LogHolder.log(LogLevel.EXCEPTION, LogType.PAY, msg, e);
			}
		}
		else
		{
			LogHolder.log(LogLevel.ERR, LogType.PAY, msg);
		}

		// an error occurred; reconnect to another cascade if possible

		getServiceContainer().keepCurrentService(false);
		PayAccountsFile.getInstance().signalAccountError(
			new XMLErrorMessage(XMLErrorMessage.ERR_WRONG_DATA, msg,
					m_payAccount, m_connectedCascade));
	}
	
	protected void finalize() throws Throwable
	{
		m_packetCounter.deleteObserver(m_packetCountEmptyObserver);

		super.finalize(); //not necessary if extending Object.
	}

	public void multiplexerClosed()
	{
		synchronized(m_aiLoginSyncObject)
		{
			m_bMultiplexerClosed = true;
			m_packetCounter.deleteObserver(m_packetCountEmptyObserver);
			m_aiLoginSyncObject.notifyAll();
		}
		
		// protection against services that "steal" prepaid bytes upon each connection
		
		if (m_payAccount == null)
		{
			return;
		}
		
		XMLAccountInfo accountInfo = m_payAccount.getAccountInfo();
		
		if (accountInfo == null)
		{
			return;
		}
		
		XMLEasyCC myLastCC = accountInfo.getCC(m_connectedCascade.getConcatenatedPriceCertHashes());
		
		if (myLastCC == null)
		{
			return;
		}
		
		long lPrepaidBytes = myLastCC.getTransferredBytes() - m_payAccount.getCurrentBytes();
		 //+ (m_connectedCascade.getPrepaidInterval() - m_prepaidBytes);
		
		//System.out.println(lPrepaidBytes + ":" + m_prepaidBytes);
		if (lPrepaidBytes > 0)
		{
			PreviousPrepdaidBytes ppBytes = getPreviousPrepaidBytes();
			if (ppBytes != null)
			{
				if (ppBytes.m_dateTimeout.before(new Date()))
				{
					ppBytes.m_prepaidBytes = 0;
				}
			}
			else
			{
				ppBytes = new PreviousPrepdaidBytes();
			}
		
			ppBytes.m_prepaidBytes += lPrepaidBytes;
			ppBytes.m_dateTimeout = new Date(System.currentTimeMillis() + 1000 * 60 * 5);
			
			
			//System.out.println("closed:" + lPrepaidBytes);
			HASH_PREPAID_ON_SERVICES.put(
					m_connectedCascade.getConcatenatedPriceCertHashes() + m_payAccount.getAccountNumber(),
					ppBytes);
			
			int iLogLevel = LogLevel.NOTICE;
			
			if (ppBytes.m_prepaidBytes > m_connectedCascade.getPrepaidInterval())
			{
				iLogLevel = LogLevel.WARNING;
			}
			
			LogHolder.log(iLogLevel, LogType.MISC, "Stored " + ppBytes.m_prepaidBytes + 
					" prepaid bytes for account " + m_payAccount.getAccountNumber() + 
					" with current bytes " + m_payAccount.getCurrentBytes() +
					" while transferred " + myLastCC.getTransferredBytes() +
					" and reported prepaid " + m_prepaidBytes + 
					" on service " + m_connectedCascade.getName() + " (" +
					m_connectedCascade.getConcatenatedPriceCertHashes() + ")"+
					" with prepaid interval " + m_connectedCascade.getPrepaidInterval());
		}
		else
		{
			LogHolder.log(LogLevel.ALERT, LogType.MISC, "prepaid:" + m_prepaidBytes + " : " + lPrepaidBytes);
		}
	}

	/*public void setSynchronizedAILogin(boolean synchronizedAILogin)
	{
		synchronized(m_aiLoginSyncObject)
		{
			m_synchronizedAILogin = synchronizedAILogin;
		}
	}*/
	
	/* only for synchronized login */
	public void setAILoginTimeout(int aiLogin_timeout)
	{
		synchronized(m_aiLoginSyncObject)
		{
			m_aiLogin_timeout = aiLogin_timeout;
		}
	}
	
	public boolean isPrepaidAmountInPayRequest() 
	{
		return m_prepaidAmountInPayRequest;
	}

	public void setPrepaidAmountInPayRequest(boolean prepaidAmountInPayRequest) 
	{
		this.m_prepaidAmountInPayRequest = prepaidAmountInPayRequest;
	}
	
	private class PreviousPrepdaidBytes
	{
		private Date m_dateTimeout;
		private long m_prepaidBytes;
	}
}
