/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
  may be used to endorse or promote products derived from this software without specific
  prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.pay;

import java.security.SecureRandom;
import java.sql.Timestamp;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Observable;
import java.util.Vector;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import anon.crypto.AsymmetricCryptoKeyPair;
import anon.crypto.DSAKeyPair;
import anon.crypto.DSAKeyPool;
import anon.error.AccountEmptyException;
import anon.infoservice.Database;
import anon.infoservice.MixCascade;
import anon.pay.xml.XMLAccountCertificate;
import anon.pay.xml.XMLBalance;
import anon.pay.xml.XMLErrorMessage;
import anon.pay.xml.XMLGenericText;
import anon.pay.xml.XMLJapPublicKey;
import anon.pay.xml.XMLPassivePayment;
import anon.util.IMiscPasswordReader;
import anon.util.IXMLEncodable;
import anon.util.XMLParseException;
import anon.util.XMLUtil;
import anon.util.captcha.ICaptchaSender;
import anon.util.captcha.IImageEncodedCaptcha;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;


import anon.pay.xml.XMLAccountInfo;


/**
 * This class encapsulates a collection of accounts. One of the accounts in the collection
 * is always active, except when the collection is empty.
 *
 * GUI classes can register a IPaymentListener with this class to be informed about all
 * payment specific events.
 *
 * The class can be initialized from an
 * XML structure and can also save all internal information in an XML structure before
 * shutdown.
 * For saving the accounts information, the following XML structure is used:
 * <pre>
 * &lt;PayAccountsFile version="1.0"&gt;
 *    &lt;MainAccountNumber&gt;123465&lt;/MainAccountNumber&gt;
 *    &lt;Accounts&gt;
 *       &lt;Account version="1.0"&gt;
 * 		    &lt;AccountCertificate&gt;...&lt;/AccountCertificate&gt; // Kontozertiufkat von der BI unterschrieben
 * 		    &lt;RSAPrivateKey&gt;...&lt;/RSAPrivateKey&gt; //der geheime Schl?ssel zum Zugriff auf das Konto
 * 		    &lt;TransferCertificates&gt; //offenen Transaktionsummern
 * 			    ....
 * 		    &lt;/TransferCertifcates&gt;
 * 		    &lt;AccountInfo&gt;...&lt;/AccountInfo&gt; //Kontostand (siehe XMLAccountInfo)
 *       &lt;/Account&gt;
 *        .
 *        .
 *        .
 *    &lt;/Accounts&gt;
 * &lt;/PayAccountsFile&gt;
 * </pre>
 *
 * @author Bastian Voigt, Tobias Bayer
 * @version 1.0
 */
public class PayAccountsFile extends Observable implements IXMLEncodable, IBIConnectionListener, IMessageListener
{
	public static final String XML_ELEMENT_NAME = "PayAccounts";
	
	public static final String METHOD_COUPON = "Coupon";
	
	public static final String XML_ELEMENT_AFFILIATE = "Affiliate";
	public static final String XML_ELEMENT_AFFILIATES = "Affiliates";
	
	public static final Integer CHANGED_AUTO_UPDATE = new Integer(0);

	private static final String XML_ATTR_IGNORE_AI_ERRORS = "ignoreAIErrorMessages";
	private static final String XML_ATTR_ENABLE_BALANCE_AUTO_UPDATE = "autoUpdateBalance";
	private static final String XML_ATTR_IS_NEW_USER = "isNewUser";
	
	public static final String MSG_DO_PREMIUM_PAYMENT = 
		PayAccountsFile.class.getName() + ".doPremiumPayment";
	public static final String MSG_DO_PREMIUM_PAYMENT_ALTERNATIVE = 
		PayAccountsFile.class.getName() + ".doPremiumPaymentAlternative";
	public static final String MSG_GET_FREE_CODE = 
		PayAccountsFile.class.getName() + ".getFreeCode";
	public static final String MSG_PAY_CREATE_ACCOUNT = 
		PayAccountsFile.class.getName() + ".payCreateAccount";
	public static final String MSG_NO_PAYMENT_INSTANCE = PayAccountsFile.class.getName() + ".noPaymentInstance";
	public static final String MSG_CREATING_KEY_PAIR = PayAccountsFile.class.getName() + ".creatingKeyPair";
	public static final String MSG_CREATING_KEY_PAIR_ERROR = PayAccountsFile.class.getName() + ".creatingKeyPairError";
	public static final String MSG_UPDATING_ACCOUNT_DATA = PayAccountsFile.class.getName() + ".updatingAccountData";
	public static final String MSG_UPDATING_ACCOUNT_DATA_ERROR = PayAccountsFile.class.getName() + ".updatingAccountDataError";
	public static final String MSG_CREATING_ACCOUNT = PayAccountsFile.class.getName() + ".creatingAccount";
	public static final String MSG_CREATING_ACCOUNT_ERROR = PayAccountsFile.class.getName() + ".creatingAccountError";
	public static final String MSG_CREATING_ACCOUNT_ERROR_UNREACHABLE = PayAccountsFile.class.getName() + ".creatingAccountErrorUnreachable";
	public static final String MSG_CREATING_ACCOUNT_ERROR_FORBIDDEN = PayAccountsFile.class.getName() + ".creatingAccountErrorForbidden";
	public static final String MSG_SAVING_CONFIG_FILE = PayAccountsFile.class.getName() + ".savingConfigFile";
	public static final String MSG_SAVING_CONFIG_FILE_ERROR = PayAccountsFile.class.getName() + ".savingConfigFileError";
	public static final String MSG_VERIFYING_COUPON = PayAccountsFile.class.getName() + ".verifyingCoupon";
	public static final String MSG_ACTIVATING_COUPON = PayAccountsFile.class.getName() + ".activatingCoupon";
	public static final String MSG_ACTIVATING_COUPON_ERROR = PayAccountsFile.class.getName() + ".activatingCouponError";
	public static final String MSG_ACTIVATING_COUPON_NOT_ACCEPTED = PayAccountsFile.class.getName() + ".activatingCouponNotAccepted";
	public static final String MSG_ACTIVATING_COUPON_NOT_A_NEW_USER = PayAccountsFile.class.getName() + ".activatingCouponNotANewUser";
	public static final String MSG_ERROR_ALLOW_NON_ANONYMOUS_OR_CONNECT = PayAccountsFile.class.getName() + ".errorAllowNonAnonymousOrConnect";
	public static final String MSG_ERROR_ALLOW_NON_ANONYMOUS_OR_CONNECT_INFOSERVICE = 
		PayAccountsFile.class.getName() + ".errorAllowNonAnonymousOrConnectInfoService";
	
	
	public static final String MSG_ERROR_DELETION_ACCOUNT_BLOCKED = PayAccountsFile.class.getName() + ".errorDeletionAccountBlocked";
	

	private static boolean m_bIsInitialized = false;
	private boolean m_bIgnoreAIAccountErrorMessages = false;
	private boolean m_bEnableBalanceAutoUpdate = true;

	private Hashtable m_hashAffiliate = new Hashtable();

	/** contains a vector of PayAccount objects, one for each account */
	private Vector m_Accounts = new Vector();
	
	private boolean m_bIsNewUser = true;

	/** the active account */
	private PayAccount m_ActiveAccount = null;

	/** the one and only accountsfile */
	private static PayAccountsFile ms_AccountsFile = null;

	private Vector m_paymentListeners = new Vector();

	private Vector m_messageListeners = new Vector();

	private MyAccountListener m_MyAccountListener = new MyAccountListener();
	
	private DSAKeyPool m_keyPool;
	private static int ms_keyPoolSize = 2;
	
	private IAffiliateOptOut m_affiliateOptOut = new IAffiliateOptOut()
	{
		public boolean isAffiliateAllowed()
		{
			return true;
		}
	};

	/**
	 * At this time, the implementation supports only one single BI. In the future
	 * a feature should be added to have support for multiple BIs, so that the
	 * AccountCertificate also contains a BIName. The infoservice should then
	 * publish information about the known BIs and also which MixCascade works
	 * with which BI.
	 *
	 * However, at the moment there is only one static BI which is used for all
	 * cascades and all accounts. This is the reason why we have this field in
	 * the singleton class.
	 */
	//private BI m_theBI;

	// singleton!
	private PayAccountsFile()
	{
		m_keyPool = new DSAKeyPool(ms_keyPoolSize);
		m_keyPool.start();
	}
	
	public static interface IAffiliateOptOut
	{
		boolean isAffiliateAllowed();
	}

	/**
	 * returns the one and only accountsfile.
	 * Note: If {@link init(BI, Element)} was not yet called,
	 * you get an empty instance which is not really useful.
	 */
	public static PayAccountsFile getInstance()
	{
		if (ms_AccountsFile == null)
		{
			ms_AccountsFile = new PayAccountsFile();
		}
		return ms_AccountsFile;
	}
	
	public AsymmetricCryptoKeyPair createAccountKeyPair()
	{
		return m_keyPool.popKeyPair();
	}
	
	public AsymmetricCryptoKeyPair createAccountKeyPair(int a_keyLengthForTesting)
	{
		return DSAKeyPair.getInstance(new SecureRandom(), a_keyLengthForTesting, 60);
	}

	/**
	 * Defined if error messages from the AI should be ignored. The connection might be closed, but the
	 * listeners will not get informed about the problem.
	 * @param a_bIgnore boolean
	 */
	public void setIgnoreAIAccountError(boolean a_bIgnore)
	{
		m_bIgnoreAIAccountErrorMessages = a_bIgnore;
	}

	public boolean isNewUser()
	{
		return m_bIsNewUser;
	}
	
	/**
	 * Returns if account balances are automatically updated.
	 * @return if account balances are automatically updated
	 */
	public boolean isBalanceAutoUpdateEnabled()
	{
		return m_bEnableBalanceAutoUpdate;
	}

	public void setBalanceAutoUpdateEnabled(boolean a_bEnable)
	{
		synchronized (this)
		{
			if (m_bEnableBalanceAutoUpdate != a_bEnable)
			{
				m_bEnableBalanceAutoUpdate = a_bEnable;
				setChanged();
			}
			notifyObservers(CHANGED_AUTO_UPDATE);
		}
	}

	/**
	 * Returns if error messages from the AI should be ignored.
	 * @return boolean
	 */
	public boolean isAIAccountErrorIgnored()
	{
		return m_bIgnoreAIAccountErrorMessages;
	}

	public synchronized boolean importAccounts(Element elemAccountsFile, 
			IMiscPasswordReader a_passwordReader) throws XMLParseException, Exception
	{
		return importAccounts(elemAccountsFile, a_passwordReader, false);
	}
	
	
	public synchronized boolean importAccounts(Element elemAccountsFile, IMiscPasswordReader a_passwordReader,
			 boolean a_bOnlyWithCorrectPassword) throws XMLParseException,
		Exception
	{
		boolean bAccountAdded = false;
		
		XMLUtil.assertNodeName(elemAccountsFile, XML_ELEMENT_NAME);
		Element elemAccounts = (Element) XMLUtil.getFirstChildByName(elemAccountsFile, "Accounts");
		Node elemAccount = XMLUtil.getFirstChildByName(elemAccounts, PayAccount.XML_ELEMENT_NAME);
		PayAccount theAccount;
		while (elemAccount != null)
		{
			theAccount = new PayAccount((Element)elemAccount, a_passwordReader);

			if (a_bOnlyWithCorrectPassword && theAccount.getPrivateKey() == null)
			{
				continue;
			}
			
			try
			{
				addAccount(theAccount);
				bAccountAdded = true;
			}
			catch(AccountAlreadyExistingException a_e)
			{
				LogHolder.log(LogLevel.WARNING, LogType.PAY, "Account " + 
						theAccount.getAccountNumber() + " already existed in our configuration and was not added.");
			}
			
			//do NOT explicitly add PayAccountsFile as MessageListener - addAccount() already includes that, would duplicate all messages to have two Listeners
			//theAccount.addMessageListener(PayAccountsFile.getInstance());


			//load messages from the balances we already have (future messages are only displayed if different from existing ones)
			if (theAccount.getAccountInfo() != null)
			{
				XMLBalance existingBalance = theAccount.getAccountInfo().getBalance();
				if (existingBalance != null)
				{
					PayMessage existingMessage = existingBalance.getMessage();
					if (existingMessage != null && !existingMessage.getShortMessage().equals(""))
					{
						getInstance().messageReceived(existingMessage);
					}
				}
			}
			
			// Move to next sibling until we find the next element or until the list is empty.
			while ((elemAccount = XMLUtil.getNextSibling(elemAccount)) != null && !(elemAccount instanceof Element));
		}
		return bAccountAdded;
	}
	
	/**
	 * Performs the initialization.
	 * @param a_passwordReader  a password reader for encrypted account files; message: AccountNumber
	 * @return boolean succeeded?
	 */
	public static synchronized boolean init(Element elemAccountsFile, IMiscPasswordReader a_passwordReader,
							   	boolean a_bForceAIErrors, int a_keyPoolSize, 
							   	IAffiliateOptOut a_affiliateOptOut)
	{
		if (m_bIsInitialized)
		{
			return false;
		}
		
		if (a_affiliateOptOut != null)
		{
			getInstance().m_affiliateOptOut = a_affiliateOptOut;
		}
		
		if (a_keyPoolSize >= 0)
		{
			ms_keyPoolSize = a_keyPoolSize;
		}
		
		//ms_AccountsFile.m_theBI = theBI;
		if (elemAccountsFile != null && elemAccountsFile.getNodeName().equals(XML_ELEMENT_NAME))
		{
			if (a_bForceAIErrors)
			{
				getInstance().m_bIgnoreAIAccountErrorMessages = false;
			}
			else
			{
				/* TODO temporarily disallow loading even in beta version
				getInstance().m_bIgnoreAIAccountErrorMessages =
					XMLUtil.parseAttribute(elemAccountsFile, XML_ATTR_IGNORE_AI_ERRORS, false);
					*/
				getInstance().m_bIgnoreAIAccountErrorMessages = false;
			}
			getInstance().m_bEnableBalanceAutoUpdate =
				XMLUtil.parseAttribute(elemAccountsFile, XML_ATTR_ENABLE_BALANCE_AUTO_UPDATE, true);
			
			getInstance().m_bIsNewUser =
				XMLUtil.parseAttribute(elemAccountsFile, XML_ATTR_IS_NEW_USER, getInstance().m_bIsNewUser);
			
			Element elemActiveAccount = (Element) XMLUtil.getFirstChildByName(elemAccountsFile,
				"ActiveAccountNumber");
			
			
			NodeList nodesAffiliates = XMLUtil.getElementsByTagName(
					XMLUtil.getFirstChildByName(elemAccountsFile, XML_ELEMENT_AFFILIATES), 
					XML_ELEMENT_AFFILIATE);
			String tmpPiid, tmpAffiliate;
			if (nodesAffiliates != null)
			{
				for (int i = 0; i < nodesAffiliates.getLength(); i++)
				{
					tmpAffiliate = XMLUtil.parseValue(nodesAffiliates.item(i), null);
					tmpPiid = XMLUtil.parseAttribute(nodesAffiliates.item(i), "piid", null);
					if (tmpAffiliate != null && tmpAffiliate.length() > 0 && 
							tmpPiid != null && tmpPiid.length() > 0 && 
							!tmpAffiliate.endsWith("_null")) // fix
					{
						getInstance().m_hashAffiliate.put(tmpPiid, tmpAffiliate);
						getInstance().m_bIsNewUser = false;
					}
				}
			}
			
			long activeAccountNumber = Long.parseLong(XMLUtil.parseValue(elemActiveAccount, "0"));

			try
			{
				getInstance().importAccounts(elemAccountsFile, a_passwordReader);
			}
			catch (Exception a_e)
			{
				LogHolder.log(LogLevel.EXCEPTION, LogType.PAY, a_e);
				return false;
			}
			
			// find activeAccount
			if (activeAccountNumber > 0)
			{
				Enumeration e = getInstance().m_Accounts.elements();
				while (e.hasMoreElements())
				{
					PayAccount current = (PayAccount) e.nextElement();
					if (current.getAccountNumber() == activeAccountNumber)
					{
						try
						{
							getInstance().setActiveAccount(current);
						}
						catch (Exception ex)
						{
						}
						break;
					}
				}
			}
			
			getInstance().checkAffiliateToDelete();
		}
		getInstance();
		m_bIsInitialized = true;
		return true;
	}
	
	private synchronized void checkAffiliateToDelete()
	{
		Hashtable hashPIFromAccounts = new Hashtable();
		Hashtable hashAffiliateClone = (Hashtable)m_hashAffiliate.clone();
		Enumeration e = m_Accounts.elements();
		String tmpPiid;
		while (e.hasMoreElements())
		{
			PayAccount current = (PayAccount) e.nextElement();
			hashPIFromAccounts.put(current.getPIID(), current.getPIID());
		}
		
		e = m_hashAffiliate.keys();
		while (e.hasMoreElements())
		{
			tmpPiid = (String)e.nextElement();
			if (!hashPIFromAccounts.containsKey(tmpPiid))
			{
				hashAffiliateClone.remove(tmpPiid);
				//getInstance().m_hashAffiliate.put(tmpPiid, null);
			}
		}
		m_hashAffiliate = hashAffiliateClone;
	}

	/**
	 * Thrown if an account with same account number was already existing when trying to add it.
	 */
	public static class AccountAlreadyExistingException extends Exception
	{

		/**
		 * 
		 */
		private static final long serialVersionUID = 1L;
	}

	/**
	 * constructs the xml structure
	 *
	 * @return Element
	 */
	public Element toXmlElement(Document a_doc)
	{
		return this.toXmlElement(a_doc, null, false);
	}

	public Element toXmlElement(Document a_doc, String a_password, boolean a_bDoNotDeactivate)
	{
		try
			{
				Element elemAccount;
				Element elemAccountsFile = a_doc.createElement(XML_ELEMENT_NAME);
				Enumeration enumKeys;
				String strBIID;

				elemAccountsFile.setAttribute(XML_ATTR_VERSION, "1.0");
				XMLUtil.setAttribute(elemAccountsFile, XML_ATTR_IGNORE_AI_ERRORS,
									 m_bIgnoreAIAccountErrorMessages);
				XMLUtil.setAttribute(elemAccountsFile, XML_ATTR_ENABLE_BALANCE_AUTO_UPDATE,
									 m_bEnableBalanceAutoUpdate);
				XMLUtil.setAttribute(elemAccountsFile, XML_ATTR_IS_NEW_USER, m_bIsNewUser);

				Element elem = a_doc.createElement("ActiveAccountNumber");
				XMLUtil.setValue(elem, getActiveAccountNumber());
				elemAccountsFile.appendChild(elem);
				
				if (m_hashAffiliate.size() > 0)
				{
					elem = a_doc.createElement(XML_ELEMENT_AFFILIATES);
					
					enumKeys = m_hashAffiliate.keys();
					while (enumKeys.hasMoreElements())
					{
						strBIID = (String)enumKeys.nextElement();
						if (m_hashAffiliate.get(strBIID) == null ||
								m_hashAffiliate.get(strBIID).toString().endsWith("_null")) // should not happen...
						{
							continue;
						}
						
						elemAccount = a_doc.createElement(XML_ELEMENT_AFFILIATE);
						XMLUtil.setValue(elemAccount, (String)m_hashAffiliate.get(strBIID));
						XMLUtil.setAttribute(elemAccount, "piid", strBIID);
						elem.appendChild(elemAccount);
					}
					elemAccountsFile.appendChild(elem);
				}

				elem = a_doc.createElement("Accounts");
				elemAccountsFile.appendChild(elem);

				synchronized (this)
				{
					for (int i = 0; i < m_Accounts.size(); i++)
					{
						PayAccount account = (PayAccount) m_Accounts.elementAt(i);
						elemAccount = account.toXmlElement(a_doc, a_password, a_bDoNotDeactivate);
						elem.appendChild(elemAccount);
					}
				}
				return elemAccountsFile;
			}
			catch (Exception ex)
			{
				LogHolder.log(LogLevel.EXCEPTION, LogType.PAY,
							  "Exception while creating PayAccountsFile XML: " + ex);
				return null;
			}
	}

	public boolean hasActiveAccount()
	{
		return m_ActiveAccount != null;
	}

	public PayAccount getActiveAccount()
	{
		return m_ActiveAccount;
	}


	public synchronized void setActiveAccount(PayAccount a_account)
	{
		PayAccount account = null;
		
		if (a_account != null)
		{
			// look whether this account is in the database
			account = getAccount(a_account.getAccountNumber(), a_account.getPIID());
		}
		
		if (account != null && account.getPrivateKey() != null)
		{
			synchronized (m_paymentListeners)
			{
				m_ActiveAccount = account;
				Enumeration enumListeners = m_paymentListeners.elements();
				while (enumListeners.hasMoreElements())
				{
					( (IPaymentListener) enumListeners.nextElement()).accountActivated(m_ActiveAccount);
				}
			}
		}
		else if (account == null)
		{
			m_ActiveAccount = null;
			synchronized (m_paymentListeners)
			{
				Enumeration enumListeners = m_paymentListeners.elements();
				while (enumListeners.hasMoreElements())
				{
					( (IPaymentListener) enumListeners.nextElement()).accountActivated(m_ActiveAccount);
				}
			}
		}
	}

	public long getActiveAccountNumber()
	{
		PayAccount activeAccount = m_ActiveAccount;
		if (activeAccount != null)
		{
			return activeAccount.getAccountNumber();
		}
		else
		{
			return -1;
		}
	}

	/**
	 * Liefert PayAccount zur angegebenen Kontonummer.
	 *
	 * @param accountNumber Kontonummer
	 * @return {@link PayAccount} oder null, wenn kein Konto unter der angebenen
	 * Kontonummer vorhanden ist
	 */
	public synchronized PayAccount getAccount(long accountNumber, String a_piid)
	{
		PayAccount tmp = null;
		Enumeration enumer = m_Accounts.elements();
		while (enumer.hasMoreElements())
		{
			tmp = (PayAccount) enumer.nextElement();
			if (tmp.getAccountNumber() == accountNumber && 
					((a_piid == tmp.getPIID() || 
					(a_piid != null && tmp.getPIID() != null && a_piid.equals(tmp.getPIID())))))
			{
				break;
			}
			else
			{
				tmp = null;
			}
		}

		return tmp;
	}

	/**
	 * Deletes the account from the accountsfile and saves the file to disk.
	 * If the deleted account was the active account, the first remaining account
	 * will become the active account.
	 *
	 * @param accountNumber account number
	 */
	public synchronized void deleteAccount(PayAccount a_account)
	{
		if (a_account == null || a_account.isLocked() || a_account.isStatusUnknown())
		{
			return;
		}
		
		PayAccount accountToDelete = null;
		synchronized (this)
		{
			accountToDelete = getAccount(a_account.getAccountNumber(), a_account.getPIID());

			if (accountToDelete != null)
			{
				for (int i = 0; i < m_Accounts.size(); i++)
				{
					accountToDelete = (PayAccount) m_Accounts.elementAt(i);
					if (accountToDelete.getAccountNumber() == a_account.getAccountNumber())
					{
						m_Accounts.removeElementAt(i);
						break;
					}
				}
				//if deleted account had a message, remove it
				//because a deleted account will not be update any more -> JapNewView would not realize the message is gone
				if (accountToDelete.getBalance() != null)
				{
					PayMessage oldMessage = accountToDelete.getBalance().getMessage();
					if (oldMessage != null && !oldMessage.getShortMessage().equals("") )
					{
						fireMessageRemoved(oldMessage);
					}
				}

	            //get a new active account, if possible
				if (getActiveAccount() == accountToDelete)
				{
					if (m_Accounts.size() > 0)
					{
						setActiveAccount((PayAccount)m_Accounts.elementAt(0));
					}
					else
					{
						setActiveAccount(null);
					}
				}
			}
			
			checkAffiliateToDelete();
		}

		if (accountToDelete != null)
		{
			synchronized (m_paymentListeners)
			{
				Enumeration enumListeners = m_paymentListeners.elements();
				while (enumListeners.hasMoreElements())
				{
					( (IPaymentListener) enumListeners.nextElement()).accountRemoved(accountToDelete);
				}
			}
		}
	}


	public String getAffiliate(String a_strPIID, boolean a_bForceAffiliate)
	{
		if (a_strPIID == null)
		{
			return null;
		}
		
		if (!m_affiliateOptOut.isAffiliateAllowed() && !a_bForceAffiliate)
		{
			return null;
		}
		
		return (String)m_hashAffiliate.get(a_strPIID);
	}
	
	/**
	 * Returns an enumeration of all accounts
	 *
	 * @return Enumeration of {@link PayAccount}
	 */
	public Enumeration getAccounts()
	{
		return ((Vector)m_Accounts .clone()).elements();
	}

	/**
	* iterates over all accounts, and whenever there is a message it is sent to all MessageListeners
	* reason for existence: JapNewView needs to load all messages at start-up
	* (PayAccountsFile does already fire all message when it loads the accounts from the file, but that possibly happens before a MessageListener is
	* initialized and added as a listener to PayAccountsFile)
	*/
	public static void fireKnownMessages()
	{
		Enumeration allAccounts = getInstance().getAccounts();
		while (allAccounts.hasMoreElements() )
		{
			PayAccount theAccount = (PayAccount) allAccounts.nextElement();
			XMLAccountInfo info = theAccount.getAccountInfo();
			if (info != null)
			{
				XMLBalance existingBalance = theAccount.getAccountInfo().getBalance();
				PayMessage existingMessage = existingBalance.getMessage();
				if (existingMessage != null && !existingMessage.getShortMessage().equals(""))
				{
					ms_AccountsFile.fireMessageReceived(existingMessage);
				}
			}
		}
	}

	public synchronized PayAccount getAlternativeChargedAccount(String a_piid)
	{	
		return getChargedAccount(a_piid, getActiveAccount());
	}
	
	public PayAccount getAccountWaitingForTransaction(String a_piid)
	{
		Vector accounts = getAccounts(a_piid);
		PayAccount currentAccount = null;

		for (int i = 0; i < accounts.size(); i++)
		{
			currentAccount = (PayAccount) accounts.elementAt(i);
			if (currentAccount.isWaitingForTransaction())
			{
				break;
			}
			currentAccount = null;
		}
		
		return currentAccount;
	}
	
	public PayAccount getChargedAccount(String a_piid)
	{
		return getChargedAccount(a_piid, null);
	}
	
	public PayAccount getChargedAccount(String a_piid, PayAccount a_excludeAccount)
	{
		/*
		if (a_piid == null)
		{
			return null;
		}
		*/
		
		Vector accounts = getAccounts(a_piid);
		Timestamp now = new Timestamp(System.currentTimeMillis());
		PayAccount currentAccount = null;
		
		if (a_excludeAccount != null && !accounts.contains(a_excludeAccount))
		{
			// good, we can choose the first account we will find
			a_excludeAccount = null;
		}

		for (int i = 0; i < accounts.size(); i++)
		{
			currentAccount = (PayAccount) accounts.elementAt(i);
			if (!currentAccount.isBlocked() && currentAccount.isCharged(now) &&
				(a_excludeAccount == null || a_excludeAccount != currentAccount))
			{
				break;
			}
			currentAccount = null;
		}
		return currentAccount;
	}


	public Vector getAccounts(String a_piid)
	{
		Vector vecAccounts = new Vector();
		synchronized (m_Accounts)
		{
			Enumeration enumer = m_Accounts.elements();
			PayAccount currentAccount;
			PayAccount activeAccount = getActiveAccount();
	
			
			//if (a_piid != null && a_piid.trim().length() > 0)
			{
				addAccount(vecAccounts, a_piid, activeAccount);
				while (enumer.hasMoreElements())
				{
					currentAccount = (PayAccount) enumer.nextElement();
					if (currentAccount != activeAccount)
					{
						addAccount(vecAccounts, a_piid, currentAccount);
					}
				}
			}
		}

		return vecAccounts;
	}
	
	private void addAccount(Vector a_vecAccounts, String a_piid, PayAccount a_account)
	{
		PaymentInstanceDBEntry pi;
		String piid;
		
		if (a_account == null || a_vecAccounts == null)
		{
			return;
		}
		
		pi = a_account.getBI();
		if (pi == null)
		{
			// this may happen sometimes... but it should not!
			LogHolder.log(LogLevel.ERR, LogType.PAY,
						  "Payment instance for account nr. " +
						  a_account.getAccountNumber() +
						  " not found!");
			//continue;
			piid = a_account.getPIID();
		}
		else
		{
			piid = pi.getId();
		}
		
		if (piid == null)
		{
			LogHolder.log(LogLevel.ALERT, LogType.PAY,
					  "Payment instance for account nr. " +
					  a_account.getAccountNumber() +
					  " is null!");
			return;
		}
		
		if (a_piid == null || piid.equals(a_piid))
		{
			a_vecAccounts.addElement(a_account);
		}
	}

	/**
	 * Adds a new account
	 *
	 * @param account new account
	 * @throws Exception If the same account was already added
	 */
	public synchronized void addAccount(PayAccount newAccount) throws AccountAlreadyExistingException
	{
		PayAccount tmp;
		boolean activeChanged = false;

		Enumeration enumer = m_Accounts.elements();
		while (enumer.hasMoreElements())
		{
			tmp = (PayAccount) enumer.nextElement();
			if (tmp.getAccountNumber() == newAccount.getAccountNumber() &&
				(tmp.getPrivateKey() != null || newAccount.getPrivateKey() == null))
			{
				if (tmp.getPIID() == newAccount.getPIID())
				{
					throw new AccountAlreadyExistingException();
				}
				else if (tmp.getPIID() != null && newAccount.getPIID() != null &&
						tmp.getPIID().equals(newAccount.getPIID()))
				{
					throw new AccountAlreadyExistingException();
				}
			}
		}
		newAccount.addAccountListener(m_MyAccountListener);
		newAccount.addMessageListener(this);
		m_Accounts.addElement(newAccount);
		if (newAccount.isUsed())
		{
			m_bIsNewUser = false;
		}

		if (m_ActiveAccount == null && newAccount.getPrivateKey() != null)
		{
			m_ActiveAccount = newAccount;
			activeChanged = true;
		}
		
		if (getAffiliate(newAccount.getPIID(), true) == null && 
			newAccount.getAffiliate() != null)
		{
			m_hashAffiliate.put(newAccount.getPIID(), newAccount.getAffiliate());
		}


		// notify payment listeners
		synchronized (m_paymentListeners)
		{
			Enumeration enumListeners = m_paymentListeners.elements();
			IPaymentListener pl;
			while (enumListeners.hasMoreElements())
			{
				pl = (IPaymentListener) enumListeners.nextElement();
				pl.accountAdded(newAccount);
				if (activeChanged == true)
				{
					pl.accountActivated(newAccount);
				}
			}
			enumListeners = null;
			pl = null;
		}
	}

	/**
	 * getNumAccounts
	 *
	 * @return int
	 */
	public int getNumAccounts()
	{
		return m_Accounts.size();
	}

	/**
	 * getAccountAt
	 *
	 * @param rowIndex int
	 * @return PayAccount
	 */
	public synchronized PayAccount getAccountAt(int rowIndex)
	{
		return (PayAccount) m_Accounts.elementAt(rowIndex);
	}

	/**
	 * isInitialized
	 *
	 * @return boolean
	 */
	public boolean isInitialized()
	{
		return m_bIsInitialized;
	}

	public void addPaymentListener(IPaymentListener listener)
	{
		synchronized (m_paymentListeners)
		{
			if (listener != null)
			{
				m_paymentListeners.addElement(listener);
			}
		}
	}

	public void removePaymentListener(IPaymentListener a_listener)
	{
		synchronized (m_paymentListeners)
		{
			if (m_paymentListeners.contains(a_listener))
			{
				m_paymentListeners.removeElement(a_listener);
			}
		}
	}

	public void addMessageListener(IMessageListener listener)
	{
		synchronized (m_messageListeners)
		{
			if (listener != null)
			{
				m_messageListeners.addElement(listener);
			}
		}
	}

	private void fireMessageReceived(PayMessage message)
	{
		Enumeration enumListeners = ( (Vector) m_messageListeners.clone()).elements();
		while (enumListeners.hasMoreElements())
		{
			( (IMessageListener) enumListeners.nextElement()).messageReceived(message);
		}
	}

	private void fireMessageRemoved(PayMessage message)
	{
		Enumeration enumListeners = ( (Vector) m_messageListeners.clone()).elements();
		while (enumListeners.hasMoreElements())
		{
			( (IMessageListener) enumListeners.nextElement()).messageRemoved(message);
		}

	}


	/**
	 * getBI
	 *
	 * @return BI
	 */
	/*	public BI getBI()
	{
		return m_theBI;
	 }*/

	private boolean m_bGotPremiumAccount = false;
	
	/**
	 * Listens to changes
	 * inside the accounts and forwards the events to our paymentListeners
	 */
	private class MyAccountListener implements IAccountListener
	{
		/**
		 * accountChanged
		 *
		 * @param acc PayAccount
		 */
		public void accountChanged(PayAccount acc)
		{
			// fire event
			Enumeration enumListeners;
			//synchronized (m_paymentListeners) // deadly for jdk 1.1.8...
			{
				/*
				 *  Clone the vector and leave synchronisation block as otherwise there
				 * would be a deadlock with fireChangeEvent in PayAccount:
				 * PayAccount.m_accountListeners:Vector,
				 * PayAccountsFile.m_paymentListeners:Vector
				 */
				enumListeners = ((Vector)m_paymentListeners.clone()).elements();
			}

			if (acc != null)
			{
				String strAffiliate = acc.getAffiliate();
				if (strAffiliate != null && getAffiliate(acc.getPIID(), true) == null)
				{
					m_hashAffiliate.put(acc.getPIID(), strAffiliate);
				}
				
				while (enumListeners.hasMoreElements())
				{
					( (IPaymentListener) enumListeners.nextElement()).creditChanged(acc);
				}
				
				if (acc.isUsed())
				{
					m_bIsNewUser = false;
				}
			}
		}
	}

	public PayAccount createAccount(PaymentInstanceDBEntry a_bi, XMLGenericText a_terms) throws Exception
	{
		AsymmetricCryptoKeyPair keyPair = createAccountKeyPair();
		if (keyPair == null)
		{
			return null;
		}
		return createAccount(a_bi, keyPair, a_terms);
	}
	
	public static boolean isNewUserLetter(char a_startingLetter)
	{
		return a_startingLetter == 'X' || a_startingLetter == 'Y';
	}
	
	public static boolean isPromotionLetter(char a_startingLetter)
	{
		return a_startingLetter == 'P';
	}
	
	public boolean isNewUserAllowed(String a_code)
	{
		if (a_code == null)
		{
			return true;
		}
		if (isNewUserLetter(a_code.charAt(0)) && !isNewUser())
		{
			return false;
		}
		return true;
	}
	
	public boolean activateCouponCode(String a_code, PayAccount a_account, boolean a_bPreCheckOnly) throws Exception
	{	
		boolean bValid = true;
		
		if (a_account == null)
		{
			throw new NullPointerException("No account given!");
		}
		
		if ((a_code = PayAccount.checkCouponCode(a_code)) == null)
		{
			return false;
		}
		
		BIConnection piConn = null;
		if (a_bPreCheckOnly)
		{
			bValid = false;
			piConn = new BIConnection(a_account.getBI());
			try
			{
				piConn.connect();
				piConn.authenticate(a_account);
				bValid = piConn.checkCouponCode(a_code, a_account);
			}
			finally
			{
				piConn.disconnect();
			}
			
			if (!isNewUserAllowed(a_code))
			{
				return false;
			}
			
			// do not really use this code; we only checked it
			return bValid;
		}
		
		
		 XMLPassivePayment paymentToSend = new XMLPassivePayment(a_account.getPIID());
		 //if coupon was used, get its code and put it into an XMLPassivePayment
		
		 paymentToSend.addData(XMLPassivePayment.KEY_COUPONCODE, a_code);
		 paymentToSend.setPaymentName(METHOD_COUPON);
		 long accNum = a_account.getAccountNumber();
		 paymentToSend.addData(XMLPassivePayment.KEY_ACCOUNTNUMBER, new Long(accNum).toString());
		 
		 piConn = null;
		 try
		 {
			 piConn = new BIConnection(a_account.getBI());
			 piConn.connect();
			 piConn.authenticate(a_account);
			 if (!piConn.sendPassivePayment(paymentToSend, a_account))
			 {
				 bValid = false;
			 }
		 }
		 finally 
		 {
			 if (piConn != null)
			 {
				 piConn.disconnect();
			 }
		 }
		
		 if (!bValid)
		 {
			 throw new Exception("Coupon code was not accepted when charging at the payment instance!");
		 }
		
		return true;
	}
	
	
	/**
	 * Creates a new Account.
	 * Generates an RSA or DSA key pair and then registers a new account with the BI.
	 * This can take a while, so the user should be notified before calling this.
	 *
	 * At the moment, only DSA should be used, because RSA is not supported by the
	 * AI implementation
	 * @param a_keyPair RSA should not be used at the moment
	 *
	 */
	public PayAccount createAccount(PaymentInstanceDBEntry a_bi,
			AsymmetricCryptoKeyPair a_keyPair, XMLGenericText a_terms)
		throws Exception
	{
		return createAccount(a_bi, a_keyPair, a_terms, false);
	}
	
	public PayAccount createAccount(PaymentInstanceDBEntry a_bi,
									AsymmetricCryptoKeyPair a_keyPair, XMLGenericText a_terms,
									boolean a_bLock) 
		throws Exception
	{
		XMLJapPublicKey xmlKey = new XMLJapPublicKey(a_keyPair.getPublic());

		LogHolder.log(LogLevel.DEBUG, LogType.PAY,
					  "Attempting to create account at PI " + a_bi.getName());
		//System.out.println(a_bi.getId());
		BIConnection biConn = new BIConnection(a_bi);
		biConn.addConnectionListener(this);
		XMLAccountCertificate cert;
		try
		{
			biConn.connect();
			cert = biConn.registerNewAccount(xmlKey, a_keyPair.getPrivate());
		}
		finally
		{
			biConn.disconnect();
		}

		// add the new account to the accountsFile
		PayAccount newAccount = new PayAccount(cert, a_keyPair.getPrivate(), a_bi, a_terms, a_bLock);
		addAccount(newAccount);
		return newAccount;
	}

	/**
	 * signalAccountRequest
	 */
	public void signalAccountRequest(MixCascade a_connectedCascade)
		throws AccountEmptyException
	{
		synchronized (m_paymentListeners)
		{
			Enumeration enumListeners = m_paymentListeners.elements();
			while (enumListeners.hasMoreElements())
			{
				((IPaymentListener) enumListeners.nextElement()).accountCertRequested(a_connectedCascade);	
			}
		}
	}

	/**
	 * signalAccountError
	 *
	 * @param msg XMLErrorMessage
	 */
	public void signalAccountError(XMLErrorMessage msg)
	{
		synchronized (m_paymentListeners)
		{
			IPaymentListener currentListener = null;
			Enumeration enumListeners = m_paymentListeners.elements();
			while (enumListeners.hasMoreElements())
			{
				currentListener = (IPaymentListener) enumListeners.nextElement();
				currentListener.accountError(msg, m_bIgnoreAIAccountErrorMessages);
			}
		}
		if (msg.getXmlErrorCode() == XMLErrorMessage.ERR_BLOCKED)
		{
			PayAccount account = getAccount(msg.getAccountNumber(), msg.getPIID());
			if (account != null)
			{
				account.setBlocked(true);
				m_MyAccountListener.accountChanged(account);
			}
		}
	}

	public Vector getPaymentInstances(boolean a_bProductiveOnly)
	{
		Vector vecPIs = Database.getInstance(PaymentInstanceDBEntry.class).getEntryList();
		Vector retPIs;
		
		//if (a_bProductiveOnly)
		{
			retPIs = new Vector();
			for (int i = 0; i < vecPIs.size(); i++)
			{
				if (a_bProductiveOnly != ((PaymentInstanceDBEntry)vecPIs.elementAt(i)).isTest())
				{
					retPIs.addElement(vecPIs.elementAt(i));
				}
			}
		}
	/*	else
		{
			retPIs = vecPIs;
		}
		*/
		
		
		return retPIs;
	}


	public Vector getPaymentInstances()
	{
		return Database.getInstance(PaymentInstanceDBEntry.class).getEntryList();
	}

	public PaymentInstanceDBEntry getBI(String a_piID)
	{
		PaymentInstanceDBEntry theBI =
			(PaymentInstanceDBEntry)Database.getInstance(PaymentInstanceDBEntry.class).getEntryById(a_piID);
		return theBI;
	}

	/**
	 * This method is called whenever a captcha has been received from the
	 * Payment Instance.
	 * @param a_source Object
	 * @param a_captcha IImageEncodedCaptcha
	 */
	public void gotCaptcha(ICaptchaSender a_source, final IImageEncodedCaptcha a_captcha)
	{
		synchronized (m_paymentListeners)
		{
			Enumeration enumListeners = m_paymentListeners.elements();
			while (enumListeners.hasMoreElements())
			{
				( (IPaymentListener) enumListeners.nextElement()).gotCaptcha(a_source, a_captcha);
			}
		}
	}

	/**
	 * just passes through a received message to the PayAccountsFile's MessageListeners
	 */
	public void messageReceived(PayMessage message)
	{
		fireMessageReceived(message);
	}

	public void messageRemoved(PayMessage message)
	{
		fireMessageRemoved(message);
	}
	
	//the DSA keyPool is a daemon thread!
	protected void finalize()
	{
		if(m_keyPool != null)
		{
			m_keyPool.stop();
		}
	}
}
