/*
 Copyright (c) 2000 - 2004, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

  - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */

package anon.pay;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.Random;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import anon.crypto.JAPCertificate;
import anon.infoservice.Constants;
import anon.infoservice.ListenerInterface;
import anon.infoservice.ServiceSoftware;
import anon.util.JAPMessages;
import anon.util.XMLParseException;
import anon.util.XMLUtil;
import anon.crypto.MultiCertPath;
import anon.crypto.SignatureVerifier;
import anon.crypto.XMLSignature;
import anon.infoservice.AbstractDistributableCertifiedDatabaseEntry;
import java.util.Date;
import anon.crypto.IVerifyable;
import logging.LogHolder;
import logging.LogLevel;
import anon.crypto.SignatureCreator;
import logging.LogType;

/** Holds the information of a payment instance for storing in the InfoService.*/
public class PaymentInstanceDBEntry extends AbstractDistributableCertifiedDatabaseEntry
		implements IVerifyable
{
	public static final String XML_ELEMENT_NAME = "PaymentInstance";
	public static final String XML_ELEMENT_CONTAINER_NAME = "PaymentInstances";

	private static final String MSG_TEST_NAME =
		PaymentInstanceDBEntry.class.getName() + ".testInstanceAlternativeName";
	
	private static final String XML_ELEM_NAME = "Name";
	private static final String XML_ELEM_CERT = "Certificate";
	private static final String XML_ELEM_NET = "Network";

	private static final String XML_ELEM_WEBSHOP_URLS = "WebshopURLs";
	private static final String XML_ELEM_FREE_CODE_URLS = "FreeCodeURLs";
	private static final String XML_ELEM_WEBSHOP_URL = "URL";
	private static final String XML_ATTR_WEBSHOP_AFFILIATE_ARGUMENT = "affiliateArgument";
	private static final String XML_ATTR_WEBSHOP_ARGUMENT = "shopArgument";
	private static final String XML_ATTR_WEBSHOP_LANGUAGE_ARGUMENT = "language";
	private static final String XML_ATTR_WEBSHOP_TRANSACTION_ARGUMENT = "transactionArgument";
	private static final String XML_ATTR_WEBSHOP_RATE_ARGUMENT = "rateArgument";

	/**
	 * This is the ID of this payment instance.
	 */
	private String m_strPaymentInstanceId;
	
	private boolean m_bIsTest = false;

	/**
	 * Stores the XML representation of this PaymentInstanceDBEntry.
	 */
	private Element m_xmlDescription;

	private XMLSignature m_signature;
	private MultiCertPath m_certPath;
	
	private Hashtable m_hashWebshopURLs = new Hashtable();
	private Hashtable m_hashFreeCodeURLs = new Hashtable();
	
	private String m_affiliateArgument;
	private String m_argLanguage;
	private String m_shopArgument;
	private String m_argTransaction;
	private String m_argRate;

	/**
	 * Stores the time when this payment instance entry was created by the origin payment instance.
	 *  This value is used to determine the more recent
	 * payment instance entry, if two entries are compared (higher version number -> more recent entry).
	 */
	private long m_creationTimeStamp;

	private long m_serialNumber;

	private Vector m_listenerInterfaces;
	private String m_name;
	private String m_strOrganisation;

	public PaymentInstanceDBEntry(Element elemRoot) throws XMLParseException
	{
		this(elemRoot, 0);
	}

	/** Creates a PaymentInstanceDBEntry which represents a payment instance.*/
	public PaymentInstanceDBEntry(Element elemRoot, long a_expireTime) throws XMLParseException
	{
		super(a_expireTime == 0 ?
			  System.currentTimeMillis() + Constants.TIMEOUT_PAYMENT_INSTANCE : a_expireTime);
		XMLUtil.assertNotNull(elemRoot);
		String name;

		/* store the XML representation */
		m_xmlDescription = elemRoot;
		

		name = XMLUtil.parseValue(XMLUtil.getFirstChildByName(elemRoot, XML_ELEM_NAME), null);
		if (name == null) // || !name.equals(m_name))
		{
			throw new XMLParseException(XML_ELEM_NAME);
		}

		m_signature = SignatureVerifier.getInstance().getVerifiedXml(elemRoot,
			SignatureVerifier.DOCUMENT_CLASS_PAYMENT);
		if (m_signature != null)
		{
			m_certPath = m_signature.getMultiCertPath();
			if (m_certPath != null)
			{
				m_strOrganisation = m_certPath.getSubject().getOrganisation();
			}
		}

		/* get the ID */
		m_strPaymentInstanceId = elemRoot.getAttribute(XML_ATTR_ID);
		if(!checkId() ) 
		{
			throw new XMLParseException(elemRoot.getNodeName(),"Invalid Payment-Instance ID: " + m_strPaymentInstanceId );
		}
		
		m_name = XMLUtil.parseValue(XMLUtil.getFirstChildByName(elemRoot, XML_ELEM_NAME), "");
		checkName();

		/* get the creation timestamp */
		m_creationTimeStamp = XMLUtil.parseValue(XMLUtil.getFirstChildByName(elemRoot, XML_LAST_UPDATE), -1L);
		if (m_creationTimeStamp == -1)
		{
			throw new XMLParseException(XML_LAST_UPDATE);
		}

		m_serialNumber = XMLUtil.parseAttribute(elemRoot, XML_ATTR_SERIAL, m_creationTimeStamp);

		Node nodeURL = XMLUtil.getFirstChildByName(elemRoot, XML_ELEM_WEBSHOP_URLS);
		NodeList nodesURL = XMLUtil.getElementsByTagName(nodeURL, XML_ELEM_WEBSHOP_URL);
		
		m_affiliateArgument = XMLUtil.parseAttribute(nodeURL, 
				XML_ATTR_WEBSHOP_AFFILIATE_ARGUMENT, null);
		
		m_argLanguage = XMLUtil.parseAttribute(nodeURL, 
				XML_ATTR_WEBSHOP_LANGUAGE_ARGUMENT, null);
		
		m_shopArgument = XMLUtil.parseAttribute(nodeURL, 
				XML_ATTR_WEBSHOP_ARGUMENT, null);
		
		m_argRate = XMLUtil.parseAttribute(nodeURL, 
				XML_ATTR_WEBSHOP_RATE_ARGUMENT, null);
		
		m_argTransaction = XMLUtil.parseAttribute(nodeURL, 
				XML_ATTR_WEBSHOP_TRANSACTION_ARGUMENT, null);
		
		m_hashWebshopURLs = new Hashtable();
		for (int i = 0; nodesURL != null && i < nodesURL.getLength(); i++)
		{
			try
			{
				m_hashWebshopURLs.put(
						XMLUtil.parseAttribute(nodesURL.item(i), XML_ATTR_LANGUAGE, "en"), 
						new URL(XMLUtil.parseValue(nodesURL.item(i), null)));
			}
			catch (MalformedURLException e) 
			{
				LogHolder.log(LogLevel.NOTICE, LogType.PAY, e);
			}
		}
		if (m_hashWebshopURLs.size() == 0)
		{
			m_hashWebshopURLs = null;
		}
		
		nodeURL = XMLUtil.getFirstChildByName(elemRoot, XML_ELEM_FREE_CODE_URLS);
		nodesURL = XMLUtil.getElementsByTagName(nodeURL, XML_ELEM_WEBSHOP_URL);
		m_hashFreeCodeURLs = new Hashtable();
		for (int i = 0; nodesURL != null && i < nodesURL.getLength(); i++)
		{
			try
			{
				m_hashFreeCodeURLs.put(
						XMLUtil.parseAttribute(nodesURL.item(i), XML_ATTR_LANGUAGE, "en"), 
						new URL(XMLUtil.parseValue(nodesURL.item(i), null)));
			}
			catch (MalformedURLException e) 
			{
				LogHolder.log(LogLevel.NOTICE, LogType.PAY, e);
			}
		}
		if (m_hashFreeCodeURLs.size() == 0)
		{
			m_hashFreeCodeURLs = null;
		}

		Node listenerInterfacesNode = XMLUtil.getFirstChildByName(
			  XMLUtil.getFirstChildByName(elemRoot, XML_ELEM_NET), ListenerInterface.XML_ELEMENT_CONTAINER_NAME);
		XMLUtil.assertNotNull(listenerInterfacesNode);


		NodeList listenerInterfaceNodes =
			((Element)listenerInterfacesNode).getElementsByTagName(ListenerInterface.XML_ELEMENT_NAME);
		if (listenerInterfaceNodes.getLength() == 0)
		{
			throw (new XMLParseException(ListenerInterface.XML_ELEMENT_NAME));
		}
		m_listenerInterfaces = new Vector();
		for (int i = 0; i < listenerInterfaceNodes.getLength(); i++)
		{
			m_listenerInterfaces.addElement(new ListenerInterface((Element) (listenerInterfaceNodes.item(i))));
		}
	}


	public PaymentInstanceDBEntry(String a_id, String a_name,
								  JAPCertificate a_cert,
								  Enumeration a_listeners,
								  String software_version, long creationTime, 
								  long a_serialNumber, Hashtable a_hashURLs, Hashtable a_hashFreeCodeURLs,
								  String a_affiliateArgument, String a_argTransaction,
								  String a_argPlanName, String a_shopArgument, String a_argLanguage)
	{
		super(System.currentTimeMillis() + Constants.TIMEOUT_PAYMENT_INSTANCE);
		m_shopArgument = a_shopArgument;
		m_argLanguage = a_argLanguage;
		m_affiliateArgument = a_affiliateArgument;
		m_argTransaction = a_argTransaction;
		m_argRate = a_argPlanName;
		m_strPaymentInstanceId = a_id;
		m_creationTimeStamp = creationTime;
		m_serialNumber = a_serialNumber;
		m_name = a_name;
		
		if (a_hashURLs != null)
		{
			m_hashWebshopURLs = (Hashtable)a_hashURLs.clone();
			if (m_hashWebshopURLs.size() == 0)
			{
				m_hashWebshopURLs = null;
			}
		}
		
		if (a_hashFreeCodeURLs != null)
		{
			m_hashFreeCodeURLs = (Hashtable)a_hashFreeCodeURLs.clone();
			if (m_hashFreeCodeURLs.size() == 0)
			{
				m_hashFreeCodeURLs = null;
			}
		}
		

		Document doc = XMLUtil.createDocument();
		Element elemRoot = doc.createElement(XML_ELEMENT_NAME);
		doc.appendChild(elemRoot);
		XMLUtil.setAttribute(elemRoot, XML_ATTR_ID, m_strPaymentInstanceId);
		XMLUtil.setAttribute(elemRoot, XML_ATTR_SERIAL, m_serialNumber);

		Element elemName = doc.createElement(XML_ELEM_NAME);
		XMLUtil.setValue(elemName, m_name);

		elemRoot.appendChild(elemName);
		ServiceSoftware software = new ServiceSoftware(software_version);
		elemRoot.appendChild(software.toXmlElement(doc));
		Element elemNet = doc.createElement(XML_ELEM_NET);
		elemRoot.appendChild(elemNet);
		Element elemListeners = doc.createElement(ListenerInterface.XML_ELEMENT_CONTAINER_NAME);
		elemNet.appendChild(elemListeners);
		while (a_listeners.hasMoreElements())
		{
			ListenerInterface li = (ListenerInterface) a_listeners.nextElement();
			elemListeners.appendChild(li.toXmlElement(doc));
		}
		
		if (m_hashWebshopURLs != null)
		{
			Element elemShopUrls = doc.createElement(XML_ELEM_WEBSHOP_URLS);
			Enumeration enumUrls = m_hashWebshopURLs.keys();
			String strLang;
			Element elemShopUrl;
			
			if (m_affiliateArgument != null)
			{
				XMLUtil.setAttribute(elemShopUrls, 
						XML_ATTR_WEBSHOP_AFFILIATE_ARGUMENT, m_affiliateArgument);
			}
			
			if (m_shopArgument != null)
			{
				XMLUtil.setAttribute(elemShopUrls, 
						XML_ATTR_WEBSHOP_ARGUMENT, m_shopArgument);
			}
			
			if (m_argLanguage != null)
			{
				XMLUtil.setAttribute(elemShopUrls, 
						XML_ATTR_WEBSHOP_LANGUAGE_ARGUMENT, m_argLanguage);
			}
			
			if (m_argTransaction != null)
			{
				XMLUtil.setAttribute(elemShopUrls, 
						XML_ATTR_WEBSHOP_TRANSACTION_ARGUMENT, m_argTransaction);
			}
			
			if (m_argRate != null)
			{
				XMLUtil.setAttribute(elemShopUrls, 
						XML_ATTR_WEBSHOP_RATE_ARGUMENT, m_argRate);
			}
			
			while (enumUrls.hasMoreElements())
			{
				elemShopUrl = doc.createElement(XML_ELEM_WEBSHOP_URL);
				strLang = (String)enumUrls.nextElement();
				XMLUtil.setAttribute(elemShopUrl, XML_ATTR_LANGUAGE, strLang);
				XMLUtil.setValue(elemShopUrl, 
						((URL)m_hashWebshopURLs.get(strLang)).toString());
				elemShopUrls.appendChild(elemShopUrl);
			}
			elemRoot.appendChild(elemShopUrls);
		}
		
		
		if (m_hashFreeCodeURLs != null)
		{
			Element elemShopUrls = doc.createElement(XML_ELEM_FREE_CODE_URLS);
			Enumeration enumUrls = m_hashFreeCodeURLs.keys();
			String strLang;
			Element elemShopUrl;
			
			while (enumUrls.hasMoreElements())
			{
				elemShopUrl = doc.createElement(XML_ELEM_WEBSHOP_URL);
				strLang = (String)enumUrls.nextElement();
				XMLUtil.setAttribute(elemShopUrl, XML_ATTR_LANGUAGE, strLang);
				XMLUtil.setValue(elemShopUrl, 
						((URL)m_hashFreeCodeURLs.get(strLang)).toString());
				elemShopUrls.appendChild(elemShopUrl);
			}
			elemRoot.appendChild(elemShopUrls);
		}
		
		
		Element elemLastUpdate = doc.createElement(XML_LAST_UPDATE);
		XMLUtil.setValue(elemLastUpdate, m_creationTimeStamp);
		elemRoot.appendChild(elemLastUpdate);
	    if (a_cert != null)
		{
			Element elemCert = doc.createElement(XML_ELEM_CERT);
			elemRoot.appendChild(elemCert);
			elemCert.appendChild(a_cert.toXmlElement(doc));

			m_signature = SignatureCreator.getInstance().getSignedXml(
				SignatureVerifier.DOCUMENT_CLASS_PAYMENT, elemRoot);
			if (m_signature != null)
			{
				m_certPath = m_signature.getMultiCertPath();
			}

			if (m_certPath == null)
			{
				LogHolder.log(LogLevel.EXCEPTION, LogType.MISC, "Document could not be signed!");
			}
			m_strOrganisation = a_cert.getSubject().getOrganisation();
		}

		m_xmlDescription = elemRoot;
	}
	
	public URL getWebshopURL()
	{
		return getWebshopURL(0, null);
	}
	
	public URL getFreeCodeURL()
	{
		URL url = null;
		if (m_hashFreeCodeURLs != null && m_hashFreeCodeURLs.size() > 0)
		{
			url = (URL)m_hashFreeCodeURLs.get(
					JAPMessages.getLocale().getLanguage().toLowerCase());
			if (url == null)
			{
				url = (URL)m_hashFreeCodeURLs.get("en");
			}
		}
		return url;
	}
	
	public URL getWebshopURL(long a_lTransactionID, String a_rateID)
	{
		URL url = null;
		String strAffiliate;
		Hashtable hashArgs = new Hashtable();
		
		if (m_hashWebshopURLs != null && m_hashWebshopURLs.size() > 0)
		{
			url = (URL)m_hashWebshopURLs.get(
					JAPMessages.getLocale().getLanguage().toLowerCase());
			if (url == null)
			{
				url = (URL)m_hashWebshopURLs.get("en");
			}
			
			strAffiliate = PayAccountsFile.getInstance().getAffiliate(m_strPaymentInstanceId, false);
			if (m_affiliateArgument != null && strAffiliate != null)
			{
				int index = strAffiliate.indexOf("_");
				if (m_shopArgument != null && index > 0 && strAffiliate.length() > index + 1)
				{
					String strShop = strAffiliate.substring(0, index);
					strAffiliate = strAffiliate.substring(index + 1, strAffiliate.length());
					hashArgs.put(m_shopArgument, strShop);
				}
				
				hashArgs.put(m_affiliateArgument, strAffiliate);
			}
			
			
			
			if (m_argLanguage != null)
			{
				hashArgs.put(m_argLanguage, JAPMessages.getLocale().getLanguage());
			}
			
			if (m_argRate != null && a_rateID != null)
			{
				hashArgs.put(m_argRate, a_rateID);
			}
			
			if (m_argTransaction != null && a_lTransactionID > 0)
			{
				hashArgs.put(m_argTransaction, new Long(a_lTransactionID));
			}
			
			if (url != null && hashArgs.size() > 0)
			{
				String strUrl = url.toString();
				String strKey;
				
				Enumeration enumKeys = hashArgs.keys();
				while (enumKeys.hasMoreElements())
				{
					if (strUrl.indexOf("?") > 0)
					{
						strUrl += "&";
					}
					else
					{
						strUrl += "?";
					}
					
					strKey = (String)enumKeys.nextElement();
					strUrl += strKey + "=" + hashArgs.get(strKey).toString();
				}
				
				try 
				{
					url = new URL(strUrl);
				} 
				catch (MalformedURLException e) 
				{
					LogHolder.log(LogLevel.ALERT, LogType.PAY, e);
				}
			}
		}
		return url;
	}
	
	public boolean isPersistanceDeletionAllowed()
	{
		return XMLUtil.getStorageMode() == XMLUtil.STORAGE_MODE_AGRESSIVE;
	}
	
	public void deletePersistence()
	{
		if (isPersistanceDeletionAllowed())
		{
			// m_xmlDescription = null;
			m_signature = null;
		}
	}

	public boolean isVerified()
	{
		if (m_certPath != null)
		{
			return m_certPath.isVerified();
		}
		return false;
	}

	public boolean isValid()
	{
		if (m_certPath != null)
		{
			return m_certPath.isValid(new Date());
		}
		return false;
	}
	
	public XMLSignature getSignature()
	{
		return m_signature;
	}
	
	public MultiCertPath getCertPath()
	{
		return m_certPath;
	}

	public String toString()
	{
		return getName();
	}

	public String getOrganisation()
	{
		return m_strOrganisation;
	}
	
	public String getId()
	{
		return m_strPaymentInstanceId;
	}
	
	public boolean equals(Object a_paymentInstance)
	{
		PaymentInstanceDBEntry paymentInstance;
		
		if (!(a_paymentInstance instanceof PaymentInstanceDBEntry) || a_paymentInstance == null)
		{
			return false;
		}
		
		paymentInstance = (PaymentInstanceDBEntry)a_paymentInstance;
		
		if (paymentInstance.getId() == getId() || paymentInstance.getId().equals(getId()))
		{
			return true;
		}
		
		return false;
	}
	
	public int hashCode()
	{
		if (m_strPaymentInstanceId == null)
		{
			return 0;
		}
		return m_strPaymentInstanceId.hashCode();
	}
	

	public String getName()
	{
		return m_name;
	}

	/**
	 * Returns the listener interfaces of this PI in a random order.
	 * @return the listener interfaces of this PI in a random order
	 */
	public Enumeration getListenerInterfaces()
	{
		Random random = new Random();
		int currentIndex;
		Vector interfaces = (Vector)m_listenerInterfaces.clone();
		Vector interfacesReturned = new Vector();
		while (interfaces.size() > 0)
		{
			currentIndex = Math.abs(random.nextInt() % interfaces.size());
			/*
			if (((ListenerInterface)interfaces.elementAt(currentIndex)).getHost().equals("91.184.37.84"))
			{
				interfacesReturned.removeAllElements();
				interfacesReturned.addElement(interfaces.elementAt(currentIndex));
				System.out.println("use pi1 interface");
				break;
			}*/
			
			interfacesReturned.addElement(interfaces.elementAt(currentIndex));
			interfaces.removeElementAt(currentIndex);
		}

		return interfacesReturned.elements();
	}

	/**
	 * Returns the time when this payment instance entry was created by the origin payment instance.
	 *
	 * @return A version number which is used to determine the more recent payment instance entry, if two
	 *         entries are compared (higher version number -> more recent entry).
	 */
	public long getVersionNumber()
	{
		return m_serialNumber;
	}

	public long getLastUpdate()
	{
		return m_creationTimeStamp;
	}

	/**
	 * This returns the filename (InfoService command), where this PaymentInstanceDBEntry is posted at
	 * other InfoServices. It's always '/paymentinstance'.
	 *
	 * @return The filename where the information about this PaymentInstanceDBEntry is posted at other
	 *         InfoServices when this entry is forwarded.
	 */
	public String getPostFile()
	{
		return "/paymentinstance";
	}

	public Element getXmlStructure()
	{
		return m_xmlDescription;
	}

	public boolean isTest()
	{
		return m_bIsTest;
	}
	
	private void checkName()
	{
		if (m_name != null)
		{
			StringTokenizer tokenizer = new StringTokenizer(m_name);
			while (tokenizer.hasMoreElements())
			{
				if (tokenizer.nextToken().toLowerCase().equals("test"))
				{
					m_bIsTest = true;
					String strName = JAPMessages.getString(MSG_TEST_NAME);
					if (!strName.equals(MSG_TEST_NAME))
					{
						m_name = strName;
					}
					break;
				}
			}			
		}
	}

}
