package anon.pay;

import java.sql.Timestamp;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import anon.util.IXMLEncodable;
import anon.util.XMLParseException;
import anon.util.XMLUtil;

public class Transaction implements IXMLEncodable 
{
	public static final String XML_ELEMENT_NAME = "Transaction";
	
	private static final String XML_ELEM_ACCOUNT_NUMBER = "AccountNumber";
	private static final String XML_ELEM_AMOUNT_EURO_CENT = "AmountEuroCent";
	private static final String XML_ELEM_CREATION_TIME = "CreationTime";
	private static final String XML_ELEM_USED_TIME = "UsedTime";
	private static final String XML_ELEM_PAYMENT_METHOD = "PaymentMethod";
	private static final String XML_ELEM_RATE = "Rate";
	private static final String XML_ATTR_DONE = "done";
	
	public static final long T_EXPIRE = 1000 * 60 * 60 * 24 * 14l;
	
	private long m_lTransactionID;
	private String m_strPaymentMethod;
	private long m_lAccountNr;
	private int m_iAmountInEuroCent;
	private String m_rate, m_rateID;
	private boolean m_bDone;
	private Timestamp m_tCreation;
	private Timestamp m_tExpiration;
	private Timestamp m_tUsed;
	
	public Transaction(long a_lTransactionID, long a_lAccountNr,
			int a_iAmountInEuroCent, String a_rate, String a_rateID, 
			Timestamp a_tCreation, Timestamp a_tUsed, String a_strPaymentMethod, boolean a_bDone)
	{
		m_lTransactionID = a_lTransactionID;
		m_strPaymentMethod = a_strPaymentMethod;
		m_tUsed = a_tUsed;
		m_lAccountNr = a_lAccountNr;
		m_iAmountInEuroCent = a_iAmountInEuroCent;
		m_rate = a_rate;
		m_rateID = a_rateID;
		m_bDone = a_bDone;
		m_tCreation = a_tCreation;
		// two weeks 
		m_tExpiration = new Timestamp(m_tCreation.getTime() + T_EXPIRE); 
	}
	
	public Transaction(Element a_element) throws XMLParseException
	{
		XMLUtil.assertNodeName(a_element, XML_ELEMENT_NAME);
		
		m_lTransactionID = XMLUtil.parseAttribute(a_element, XML_ATTR_ID, 0l);
		if (m_lTransactionID <= 0)
		{
			throw new XMLParseException(XML_ELEMENT_NAME);
		}
		m_bDone = XMLUtil.parseAttribute(a_element, XML_ATTR_DONE, false);
		
		m_lAccountNr = XMLUtil.parseValue(
				XMLUtil.getFirstChildByName(a_element, XML_ELEM_ACCOUNT_NUMBER), 0l);
		
		if (m_lAccountNr <= 0)
		{
			throw new XMLParseException(XML_ELEM_ACCOUNT_NUMBER);
		}
		
		m_iAmountInEuroCent = XMLUtil.parseValue(
				XMLUtil.getFirstChildByName(a_element, XML_ELEM_AMOUNT_EURO_CENT), 0);
		
		String strTimestamp =
			XMLUtil.parseValue(
					XMLUtil.getFirstChildByName(a_element, XML_ELEM_CREATION_TIME), null);
		if (strTimestamp != null)
		{
			m_tCreation = Timestamp.valueOf(strTimestamp);
		}
		if (m_tCreation == null)
		{
			throw new XMLParseException(XML_ELEM_CREATION_TIME);
		}
		m_tExpiration = new Timestamp(m_tCreation.getTime() + T_EXPIRE);
		
		strTimestamp = XMLUtil.parseValue(
				XMLUtil.getFirstChildByName(a_element, XML_ELEM_USED_TIME), null);
		if (strTimestamp != null)
		{
			m_tUsed = Timestamp.valueOf(strTimestamp);
		}
		
		m_strPaymentMethod = XMLUtil.parseValue(
				XMLUtil.getFirstChildByName(a_element, XML_ELEM_PAYMENT_METHOD), null);
		
		
		m_rate = XMLUtil.parseValue(
				XMLUtil.getFirstChildByName(a_element, XML_ELEM_RATE), null);
		m_rateID = XMLUtil.parseAttribute(
				XMLUtil.getFirstChildByName(a_element, XML_ELEM_RATE), XML_ATTR_ID, null);
	}
	
	public Element toXmlElement(Document a_doc)
	{
		Element elemRoot = a_doc.createElement(XML_ELEMENT_NAME);
		Element elem;
		
		
		XMLUtil.setAttribute(elemRoot, XML_ATTR_ID, m_lTransactionID);
		XMLUtil.setAttribute(elemRoot, XML_ATTR_DONE, m_bDone);
		
		elem = a_doc.createElement(XML_ELEM_ACCOUNT_NUMBER);
		XMLUtil.setValue(elem, m_lAccountNr);
		elemRoot.appendChild(elem);
		
		elem = a_doc.createElement(XML_ELEM_AMOUNT_EURO_CENT);
		XMLUtil.setValue(elem, m_iAmountInEuroCent);
		elemRoot.appendChild(elem);
		
		elem = a_doc.createElement(XML_ELEM_CREATION_TIME);
		XMLUtil.setValue(elem, m_tCreation.toString());
		elemRoot.appendChild(elem);
		
		if (m_tUsed != null)
		{
			elem = a_doc.createElement(XML_ELEM_USED_TIME);
			XMLUtil.setValue(elem, m_tUsed.toString());
			elemRoot.appendChild(elem);
		}
		
		elem = a_doc.createElement(XML_ELEM_PAYMENT_METHOD);
		XMLUtil.setValue(elem, m_strPaymentMethod);
		elemRoot.appendChild(elem);;
		
		elem = a_doc.createElement(XML_ELEM_RATE);
		XMLUtil.setValue(elem, m_rate);
		XMLUtil.setAttribute(elem, XML_ATTR_ID, m_rateID);
		elemRoot.appendChild(elem);
		
		return elemRoot;
	}
	
	public String getPaymentMethod()
	{
		return m_strPaymentMethod;
	}
	
	public long getID()
	{
		return m_lTransactionID;
	}
	
	public long getAccountNumber()
	{
		return m_lAccountNr;
	}
	
	public int getAmountEuroCent()
	{
		return m_iAmountInEuroCent;
	}
	
	public String getRateName()
	{
		return m_rate;
	}
	
	public String getRateID()
	{
		return m_rateID;
	}

	public boolean isUsed()
	{
		return m_bDone;
	}
	
	public Timestamp getUsedTime()
	{
		return m_tUsed;
	}
	
	public Timestamp getCreationTime()
	{
		return m_tCreation;
	}
	
	/**
	 * Do not suggest any payments after this time. After about 4 week, all references to 
	 * this transaction will be automatically deleted in the PI.
	 * @return
	 */
	public boolean hasExpired()
	{
		return hasExpired(new Timestamp(System.currentTimeMillis()));
	}
	
	public boolean hasExpired(Timestamp a_tNow)
	{
		return !m_bDone && m_tExpiration.before(a_tNow);
	}
}
