/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
  may be used to endorse or promote products derived from this software without specific
  prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.pay.xml;

import java.sql.Timestamp;
import java.util.Enumeration;
import java.util.Hashtable;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;


import org.w3c.dom.Document;
import org.w3c.dom.Element;

import anon.pay.PayAccount;
import anon.util.IXMLEncodable;
import anon.util.XMLParseException;
import anon.util.XMLUtil;

/**
 * This class represents an XML AccountInfo structure
 * (XMLBalance plus cost confirmations for the account)
 *
 *
 * The following XML structure is used:
 * <pre>
 * &lt;AccountInfo&gt;
 *    &lt;Balance&gt;
 *       &lt;AccountNumber&gt; ..&lt;/AccountNumber&gt;
 *       &lt;Deposit&gt;...&lt;/Deposit&gt;
 *       &lt;Spent&gt;....&lt;/Spent&gt;
 *       &lt;Validtime&gt;...&lt;/Validtime&gt;
 *       &lt;Timestamp&gt;...&lt;/Timestamp&gt;
 *       &lt;Signature&gt; //Unterschrift der BI
 *          ...
 *       &lt;/Signature&gt;
 *    &lt;/Balance&gt;
 *    &lt;CostConfirmations&gt; //Kostenbest\u00E4tigungen, die
 *		                           von den einzelnen AI's
 *		                           abgerechnet wurden
 *       &lt;CC&gt;...&lt;/CC&gt;
 *       &lt;CC&gt;...&lt;/CC&gt;
 *    &lt;/CostConfirmations&gt;
 * &lt;/AccountInfo&gt;
 * </pre>
 *
 * @author Bastian Voigt
 */
public class XMLAccountInfo implements IXMLEncodable //extends XMLDocument
{
	//~ Instance fields ********************************************************

	public static final String XML_ELEMENT_NAME = "AccountInfo";
	
	public static final String XML_ELEMENT_NAME_COST_CONFIRMATIONS = "CostConfirmations";
	
	/** the balance certificate */
	private XMLBalance m_balance = null;
	
	private PayAccount m_callbackAccount;

	/**
	 * a collection of costconfirmations (one for each mixcascade
	 * that was used with this account)
	 */
	private Hashtable m_costConfirmations = new Hashtable();
	
	private Hashtable m_pastMonthlyVolumeBytes = new Hashtable();
	
	// defines - in local time - tha last update of the balance
	private Timestamp m_tLastBalanceUpdate = new Timestamp(0);

	//~ Constructors ***********************************************************

	public static class PastMonthsVolumeBytes implements IXMLEncodable
	{	
		public static final String XML_ELEMENT_NAME_PMB = "PastMonthsVolumeBytes";
		
		private long m_lBytes = 0;
		private String m_strPriceCertHashes = null;
		private Timestamp m_tUpdatedOn = null;
		
		public PastMonthsVolumeBytes(String a_strPriceCertHashes, long a_lBytes, Timestamp a_tUpdatedOn)
		{
			if (a_lBytes < 0)
			{
				throw new IllegalArgumentException("Past Bytes may not be < 0!");
			}
			if (a_strPriceCertHashes == null)
			{
				throw new IllegalArgumentException("Price cert hashed may not be null!");
			}
			
			m_strPriceCertHashes = a_strPriceCertHashes;
			m_lBytes = a_lBytes;
			
			m_tUpdatedOn = a_tUpdatedOn;
		}
		
		public PastMonthsVolumeBytes(Element a_elem) throws XMLParseException
		{
			XMLUtil.assertNodeName(a_elem, XML_ELEMENT_NAME_PMB);
			
			m_strPriceCertHashes = XMLUtil.parseAttribute(a_elem, "priceCertHashes", null);
			if (m_strPriceCertHashes == null)
			{
				throw new XMLParseException("priceCertHashes", XMLParseException.NODE_NULL_TAG);
			}
			
			m_tUpdatedOn = new Timestamp(XMLUtil.parseAttribute(a_elem, "updatedOn", 0l));
			m_lBytes = XMLUtil.parseValue(a_elem, 0l);
		}
		
		public long getBytes()
		{
			return m_lBytes;
		}
		
		public void setUpdatedOn(Timestamp a_tUpdatedOn)
		{
			m_tUpdatedOn = a_tUpdatedOn;
		}
		
		public String getConcatenatedPriceCertHashes()
		{
			return m_strPriceCertHashes;
		}
		
		public Timestamp getUpdatedOn()
		{
			return m_tUpdatedOn;
		}
		
		public Element toXmlElement(Document a_doc)
		{
			Element elem = a_doc.createElement(XML_ELEMENT_NAME_PMB);
			
			XMLUtil.setAttribute(elem, "priceCertHashes", m_strPriceCertHashes);
			XMLUtil.setAttribute(elem, "updatedOn", m_tUpdatedOn.getTime());
			XMLUtil.setValue(elem, m_lBytes);
			
			return elem;
		}
	}
	
	public XMLAccountInfo(XMLBalance bal)
	{
		m_balance = bal;
		if (m_balance != null)
		{
			m_tLastBalanceUpdate = new Timestamp(System.currentTimeMillis());
		}
	}

	/**
	 * Creates an AccountInfo object from a string.
	 */
	public XMLAccountInfo(String xml) throws Exception
	{
		Document doc = XMLUtil.toXMLDocument(xml);
		setValues(doc.getDocumentElement());
	}

	public XMLAccountInfo()
	{}

	/**
	 * Creates a Balance from  an existing XML docuemnt
	 *
	 * @param xml the node that represents the Balance
	 * @param verifier JAPSignature must be initialized and ready to verify XML (or null)
	 * @throws Exception on invalid xml format or invalid signature
	 */
	public XMLAccountInfo(Element xml) throws Exception
	{
		setValues(xml);
	}

	//~ Methods ****************************************************************
	public Element toXmlElement(Document a_doc)
	{
		Element elemRoot = a_doc.createElement(XML_ELEMENT_NAME);
		elemRoot.setAttribute("version", "1.1");
		Element elem;

		XMLUtil.setAttribute(elemRoot, "lastBalanceUpdate", m_tLastBalanceUpdate.toString());
		
		// add balance
		elem = m_balance.toXmlElement(a_doc);
		elemRoot.appendChild(elem);

		// add CCs
		Element elemCCs = a_doc.createElement("CostConfirmations");
		elemRoot.appendChild(elemCCs);
		synchronized (m_costConfirmations)
		{
			Enumeration enumer = m_costConfirmations.elements();
			XMLEasyCC cc;
			while (enumer.hasMoreElements())
			{
				cc = (XMLEasyCC) enumer.nextElement();
				elem = cc.toXmlElement(a_doc);
				elemCCs.appendChild(elem);
			}
		}
		
		// add past bytes
		elemCCs = a_doc.createElement("PastMonthsVolumeBytesContainer");
		elemRoot.appendChild(elemCCs);
		synchronized (m_pastMonthlyVolumeBytes)
		{
			Enumeration enumer = m_pastMonthlyVolumeBytes.elements();
			PastMonthsVolumeBytes pastBytes;
			while (enumer.hasMoreElements())
			{
				pastBytes = (PastMonthsVolumeBytes) enumer.nextElement();
				elem = pastBytes.toXmlElement(a_doc);
				elemCCs.appendChild(elem);
			}
		}
		
		
		
		return elemRoot;
	}

	/**
	 * Adds a cost confirmation xml structure to the accountinfo.
	 * Note: If a cost confirmation for the same AI is already present
	 * it will be overwritten. Only adds the CC if it has the same amount or
	 * more bytes than the previous one.
	 *
	 * @param xmlCC XMLEasyCC
	 * @return the difference in the number of bytes between the old and the
	 * new costconfirmation for the same AI
	 */
	public long addCC(XMLEasyCC a_cc, boolean a_bForcedAdd)
	{
		long oldBytes = 0;
		XMLEasyCC cc;
		synchronized (m_costConfirmations)
		{
			cc = (XMLEasyCC)m_costConfirmations.get(a_cc.getConcatenatedPriceCertHashes());			
			if (cc != null)
			{
				oldBytes = cc.getTransferredBytes();
			}
			if (a_bForcedAdd || a_cc.getTransferredBytes() >= oldBytes)
			{
				// do not add old cost confirmations
				m_costConfirmations.put(a_cc.getConcatenatedPriceCertHashes(), a_cc);
			}
		}
		return a_cc.getTransferredBytes() - oldBytes;
	}
	
	public void addPastMonthlyVolumeBytes(XMLAccountInfo a_info) throws Exception
	{
		synchronized (m_pastMonthlyVolumeBytes)
		{
			Enumeration enumBytes = a_info.m_pastMonthlyVolumeBytes.elements();
			PastMonthsVolumeBytes pastBytes;
			while (enumBytes.hasMoreElements())
			{
				pastBytes = (PastMonthsVolumeBytes)enumBytes.nextElement();
				// put new values here without checking the update time
				m_pastMonthlyVolumeBytes.put(pastBytes.getConcatenatedPriceCertHashes(), pastBytes);
			}
		}
	}
	
	public void addPastMonthlyVolumeBytes(PastMonthsVolumeBytes a_monthlyBytes)
	{
		Timestamp lastUpdate = new Timestamp(0);
		PastMonthsVolumeBytes cc;
		synchronized (m_pastMonthlyVolumeBytes)
		{
			cc = (PastMonthsVolumeBytes)m_pastMonthlyVolumeBytes.get(a_monthlyBytes.getConcatenatedPriceCertHashes());			
			if (cc != null)
			{
				lastUpdate = cc.getUpdatedOn();
			}
			if (a_monthlyBytes.getUpdatedOn() != null && 
					a_monthlyBytes.getUpdatedOn().getTime() >= lastUpdate.getTime())
			{
				m_pastMonthlyVolumeBytes.put(a_monthlyBytes.getConcatenatedPriceCertHashes(), a_monthlyBytes);
			}
		}
	}
	

	private void setValues(Element elemRoot) throws Exception
	{
		if (!elemRoot.getTagName().equals(XML_ELEMENT_NAME))
		{
			LogHolder.log(LogLevel.EXCEPTION, LogType.PAY, "invalid XML structure: "+XMLUtil.toString(elemRoot));
			throw new Exception("XMLAccountInfo wrong XML structure");
		}
		
		Element elemBalance = (Element) XMLUtil.getFirstChildByName(elemRoot, "Balance");
		m_balance = new XMLBalance(elemBalance);
		
		m_tLastBalanceUpdate = Timestamp.valueOf( 
			XMLUtil.parseAttribute(elemRoot, "lastBalanceUpdate", m_balance.getTimestamp().toString()));
		
		
		Element elemCCs = (Element) XMLUtil.getFirstChildByName(elemRoot, XML_ELEMENT_NAME_COST_CONFIRMATIONS);
		Element elemCC = (Element) XMLUtil.getFirstChildByName(elemCCs, XMLEasyCC.XML_ELEMENT_NAME);
		XMLEasyCC cc = null;
		while (elemCC != null)
		{
			cc = new XMLEasyCC(elemCC);
			m_costConfirmations.put(cc.getConcatenatedPriceCertHashes(), cc);
			elemCC = (Element) XMLUtil.getNextSiblingByName(elemCC, XMLEasyCC.XML_ELEMENT_NAME);
		}
		
		Element elemPastBytesContainer = (Element) XMLUtil.getFirstChildByName(elemRoot, "PastMonthsVolumeBytesContainer");
		Element elemPastBytes = (Element) XMLUtil.getFirstChildByName(elemPastBytesContainer, PastMonthsVolumeBytes.XML_ELEMENT_NAME_PMB);
		PastMonthsVolumeBytes pastbytes = null;
		while (elemPastBytes != null)
		{			
			pastbytes = new PastMonthsVolumeBytes(elemPastBytes);
			//System.out.println("Account " + m_balance.getAccountNumber() + ": " + pastbytes.getPriceCertHashes() + ":" + pastbytes.getBytes() + ":" + pastbytes.getUpdatedOn());
			m_pastMonthlyVolumeBytes.put(pastbytes.getConcatenatedPriceCertHashes(), pastbytes);
			elemPastBytes = (Element) XMLUtil.getNextSiblingByName(elemPastBytes, PastMonthsVolumeBytes.XML_ELEMENT_NAME_PMB);
		}		
	}

	public XMLBalance getBalance()
	{
		return m_balance;
	}

	/**
	 * getCC - returns the cost confirmation with the specified aiName
	 *
	 * @param string String
	 * @return XMLEasyCC
	 */
	public XMLEasyCC getCC(String priceCertHash)
	{
		return (XMLEasyCC)m_costConfirmations.get(priceCertHash);
	}
	
	public void setAccountCallback(PayAccount a_callbackAccount)
	{
		m_callbackAccount = a_callbackAccount;
	}
	
	public void checkMonthlyBytesUpdatedOn()
	{
		if (m_balance != null && m_balance.getVolumeBytesMonthly() > 0)
		{
			Timestamp tNow = new Timestamp(System.currentTimeMillis());

			PastMonthsVolumeBytes pastBytes;
			XMLEasyCC cc;
			long lPastBytes;
			PayAccount callbackAccount;
			
			synchronized (m_costConfirmations)
			{
				synchronized(m_pastMonthlyVolumeBytes)
				{
			
					if (!XMLBalance.isSameMonthlyPeriod(m_balance.getMonthlyBytesUpdatedOn(), tNow, m_balance.getStartDate(), true) 
							&& !XMLBalance.isSameMonthlyPeriod(m_tLastBalanceUpdate, tNow, m_balance.getStartDate(), true) &&
							!XMLBalance.isSameMonthlyPeriod(m_balance.getStartDate(), tNow, m_balance.getStartDate(), true))
					{
						Enumeration enumCCs = getCCs();
						while (enumCCs.hasMoreElements())
						{
							cc = (XMLEasyCC)enumCCs.nextElement();
							
							pastBytes = (PastMonthsVolumeBytes)m_pastMonthlyVolumeBytes.get(cc.getPriceCertHashes());
							
							
							if (pastBytes != null &&
								XMLBalance.isSameMonthlyPeriod(pastBytes.getUpdatedOn(), tNow, m_balance.getStartDate(), true))
							{
								// we have already updated the past bytes for this CC this month
								continue;
							}
							
							pastBytes = 
								new PastMonthsVolumeBytes(cc.getConcatenatedPriceCertHashes(),  cc.getTransferredBytes(), tNow);
							m_pastMonthlyVolumeBytes.put(pastBytes.getConcatenatedPriceCertHashes(), pastBytes);
						}
						callbackAccount = m_callbackAccount;
						if (callbackAccount != null)
						{
							callbackAccount.fireChangeEvent();
						}
					}
				}
			}
		}
	}

	/**
	 * Returns an enumeration of all the available cost confirmations for
	 * this account.
	 * @return Enumeration
	 */
	public Enumeration getCCs()
	{
		return ((Hashtable)m_costConfirmations.clone()).elements();
	}
	
	/**
	 * returns the overall sum of all the transfered bytes of all 
	 * CCs belonging to this account.
	 * @return
	 */
	public long getAllCCsTransferredBytes()
	{
		long sum = 0l;
		XMLEasyCC cc;
		PastMonthsVolumeBytes pastBytes;
		long lPastBytes;
		synchronized (m_costConfirmations)
		{
			synchronized (m_pastMonthlyVolumeBytes)
			{
				Enumeration e = m_costConfirmations.elements();
				while(e.hasMoreElements())
				{
					cc = (XMLEasyCC) e.nextElement();
					pastBytes = 
						(PastMonthsVolumeBytes)m_pastMonthlyVolumeBytes.get(cc.getConcatenatedPriceCertHashes());
					if (pastBytes == null)
					{
						lPastBytes = 0;
					}
					else
					{
						lPastBytes = pastBytes.getBytes();
					}
					sum += (cc).getTransferredBytes() - lPastBytes;
				}
			}
		}
		return sum;
	}
	
	/**
	 * setBalance
	 *
	 * @param b1 XMLBalance
	 */
	public void setBalance(XMLBalance b1)
	{
		m_balance = b1;
		if (m_balance != null)
		{
			m_tLastBalanceUpdate = new Timestamp(System.currentTimeMillis());
		}
	}
	
	public Timestamp getLastBalanceUpdateLocalTime()
	{
		return m_tLastBalanceUpdate;
	}

	/**
	 * XMLAccountInfo
	 *
	 * @param document Document
	 */
	public XMLAccountInfo(Document document) throws Exception
	{
		setValues(document.getDocumentElement());
		m_tLastBalanceUpdate = new Timestamp(System.currentTimeMillis());
	}
}
