package anon.pay.xml;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import anon.AnonServerDescription;
import anon.error.AnonServiceException;
import anon.pay.PayAccount;
import anon.util.IXMLEncodable;
import anon.util.XMLUtil;

/**
 * This class encapsulates an error or success message.
 * In order to be independent from the HTTP protocol on the higher layer,
 * this is now used instead of http errorcodes.
 *
 * @author Bastian Voigt
 */
public class XMLErrorMessage extends AnonServiceException implements IXMLEncodable
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	public static final int ERR_OK = 0;
	public static final int ERR_INTERNAL_SERVER_ERROR = 1;
	public static final int ERR_WRONG_FORMAT = 2;
	public static final int ERR_WRONG_DATA = 3;
	public static final int ERR_KEY_NOT_FOUND = 4;
	public static final int ERR_BAD_SIGNATURE = 5;
	public static final int ERR_BAD_REQUEST = 6;
	public static final int ERR_NO_ACCOUNTCERT = 7;
	public static final int ERR_NO_BALANCE = 8;
	public static final int ERR_NO_CONFIRMATION = 9;
	public static final int ERR_ACCOUNT_EMPTY = 10;
	public static final int ERR_CASCADE_LENGTH = 11;
	public static final int ERR_DATABASE_ERROR = 12;
	public static final int ERR_INSUFFICIENT_BALANCE = 13;
	public static final int ERR_NO_FLATRATE_OFFERED = 14;
	public static final int ERR_INVALID_CODE = 15;
	public static final int ERR_OUTDATED_CC = 16;
	public static final int ERR_INVALID_PRICE_CERTS = 17;
	public static final int ERR_MULTIPLE_LOGIN = 18;
	public static final int ERR_NO_RECORD_FOUND = 19;
	public static final int ERR_SUCCESS_BUT_WITH_ERRORS = 20;
	public static final int ERR_BLOCKED = 21;
	public static final int ERR_NOT_SYNCHRONIZED = 22;

	private int m_iErrorCode;
	private IXMLEncodable m_oMessageObject;
	
	private PayAccount m_account;

	/** default error descriptions */
	/* length of the array is checked, you need to have a number of strings here that
	   matches the number of constants defined for error codes !! */
	private static final String[] m_errStrings =
		{
		"Success",
		"Internal Server Error",
		"Wrong format",
		"Wrong Data",
		"Key not found",
		"Bad Signature",
		"Bad request",
		"No account certificate",
		"No balance",
		"No cost confirmation",
		"Account is empty",
		"Cascade too long",
		"Database error",
		"Insufficient balance",
		"No flatrate offered",
		"Invalid code",
		"outdated CC",
		"Invalid price certificates",
		"multiple login is not allowed",
		"no record found",
		"operation succeeded, but there were errors",
		"this account is blocked", //21
		"your clock in not synchronized",
	};

	public static final String XML_ELEMENT_NAME = "ErrorMessage";

	/**
	 * Parses an XMLErrorMessage object from DOM Document
	 *
	 * @param document Document
	 */
	public XMLErrorMessage(Document doc, PayAccount a_account, AnonServerDescription a_service) throws Exception
	{
		super(a_service, getMessage(Integer.parseInt(doc.getDocumentElement().getAttribute("code")),
				XMLUtil.parseValue(doc.getDocumentElement(), "")));
		setValues(doc.getDocumentElement(), a_account);
	}

	public XMLErrorMessage(Element element, PayAccount a_account, AnonServerDescription a_service
			) throws Exception
	{
		super(a_service, getMessage(Integer.parseInt(element.getAttribute("code")),  
				XMLUtil.parseValue(element, "")));
		setValues(element, a_account);
	}

	public XMLErrorMessage(int errorCode, String message)
	{
		this(errorCode, message, null, null);
	}
	
	public XMLErrorMessage(int errorCode)
	{
		this(errorCode, (PayAccount)null, (AnonServerDescription)null);
	}
	
	public XMLErrorMessage(int errorCode, PayAccount a_account)
	{
		this(errorCode, a_account, (AnonServerDescription)null);
	}

	/**
	 * Creates an errorMessage object. The errorcode should be one of the
	 * above ERR_* constants.
	 * @param errorCode int one of the above constants
	 * @param message String a human-readable description of the error
	 */
	public XMLErrorMessage(int errorCode, String message, PayAccount a_account, 
			AnonServerDescription a_service)
	{
		super(a_service, getMessage(errorCode, message));
		m_iErrorCode = errorCode;
		m_account = a_account;
	}

	public XMLErrorMessage(int errorCode, String message, IXMLEncodable messageObject,
			PayAccount a_account, 
			AnonServerDescription a_service)
	{
		super(a_service, getMessage(errorCode, message));
		m_iErrorCode = errorCode;
		m_oMessageObject = messageObject;
		m_account = a_account;
	}

	public XMLErrorMessage(int errorCode, AnonServerDescription a_service)
	{
		this(errorCode, null, null, a_service);
	}
	
	/**
	 * Uses a default description String
	 * @param errorCode int
	 */
	public XMLErrorMessage(int errorCode, PayAccount a_account, AnonServerDescription a_service)
	{
		this(errorCode, null, a_account, a_service);
	}
	
	private static String getMessage(int errorCode, String a_message)
	{
		if (a_message != null)
		{
			return a_message;
		}
		if (errorCode < 0 || errorCode > m_errStrings.length)
		{
			return "Unknown message"; //says "message", not "error", since it doesn't have to be an error
		}
		else
		{
			return m_errStrings[errorCode];
		}
	}

	public Element toXmlElement(Document a_doc)
	{
		Element elemRoot = a_doc.createElement(XML_ELEMENT_NAME);
		elemRoot.setAttribute("code", Integer.toString(m_iErrorCode));
		XMLUtil.setValue(elemRoot, getMessage());
		if (m_oMessageObject != null)
		{
			Element elemObjectRoot = a_doc.createElement("MessageObject");
			Element elemObject = m_oMessageObject.toXmlElement(a_doc);
			elemObjectRoot.appendChild(elemObject);
			elemRoot.appendChild(elemObjectRoot);
		}
		return elemRoot;
	}
	
	public long getAccountNumber()
	{
		if (m_account == null)
		{
			return 0;
		}
		return m_account.getAccountNumber();
	}
	
	public PayAccount getAccount()
	{
		return m_account;
	}

	public String getPIID()
	{
		if (m_account == null)
		{
			return null;
		}
		return m_account.getPIID();
	}

	public int getXmlErrorCode()
	{
		return m_iErrorCode;
	}

	/**
	 * getMessageObject: object corresponding to the message, may be null
	 *
	 * @return IXMLEncodable: exact type will depend on the message code, to be looked up via getMessageObjectType
	 */
	public IXMLEncodable getMessageObject()
	{
		return m_oMessageObject;
	}

	public void setMessageObject(IXMLEncodable a_messageObject)
	{
		m_oMessageObject = a_messageObject;
	}

	private void setValues(Element elemRoot, PayAccount a_account) throws Exception
	{
		if (! (elemRoot.getTagName().equals(XML_ELEMENT_NAME)))
		{
			throw new Exception("Format error: Root element wrong tagname");
		}
		m_iErrorCode = Integer.parseInt(elemRoot.getAttribute("code"));
		m_account = a_account;

		
		try
		{
			Node nodeFound;
			Node objectRoot = XMLUtil.getFirstChildByName(elemRoot, "MessageObject");
			
			if (objectRoot == null)
			{
				return;
			}
			
			nodeFound = XMLUtil.getFirstChildByName(objectRoot, XMLAccountInfo.XML_ELEMENT_NAME);
			if (nodeFound != null && nodeFound instanceof Element)
			{
				m_oMessageObject = new XMLAccountInfo((Element)nodeFound);
			}
			
			nodeFound = XMLUtil.getFirstChildByName(objectRoot, XMLEasyCC.XML_ELEMENT_NAME);
			if (nodeFound != null && nodeFound instanceof Element)
			{
				m_oMessageObject = new XMLEasyCC((Element)nodeFound);
			}
		}
		catch (Exception a_e)
		{
			// ignore, not needed
		}
		
		
	}
}
