/*
 Copyright (c) 2000 - 2004, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

  - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.platform;

import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

import java.io.File;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Properties;
import java.util.StringTokenizer;

import anon.util.RecursiveFileTool;

/**
 * This class is instantiated by AbstractOS if the current OS is Linux
 */
public final class LinuxOS extends AbstractOS
{
	private boolean m_bKDE = false;
	private int m_iKDEVersion = 0;
	private boolean m_bGnome = false;

	private final static String LINUX_VBOX_NETADAPTER = "vboxnet";

	public static final String[] BROWSERLIST =
		{
		"firefox", "iexplore", "explorer", "mozilla", "konqueror", "mozilla-firefox", "opera"
	};

	public LinuxOS() throws Exception
	{
		String osName = System.getProperty("os.name", "").toLowerCase();
		if (osName.toLowerCase().indexOf("linux") == -1)
		{
			throw new Exception("Operating system is not Linux");
		}

		Properties properties = new Properties();
		try
		{
			properties.load(Runtime.getRuntime().exec("env").getInputStream());
			m_bKDE = Boolean.valueOf(properties.getProperty("KDE_FULL_SESSION")).booleanValue();
			if (m_bKDE)
			{
				m_iKDEVersion = Integer.valueOf(properties.getProperty("KDE_SESSION_VERSION")).intValue();
			}
		}
		catch (Exception a_e)
		{
		}
		m_bGnome = properties.getProperty("GNOME_DESKTOP_SESSION_ID") != null;
		
		initEnv("env");
	}

	protected boolean openLink(String a_link)
	{
		if (a_link == null)
		{
			return false;
		}

		if (m_bKDE)
		{
			return execute("kfmclient exec " + a_link);
		}
		else if (m_bGnome)
		{
			return execute("gnome-open " + a_link);
		}
		return false;
	}
	
	public String getAppdataDefaultDirectory(String a_applicationName, boolean a_bCreate)
	{
		return null;
	}

	public String getConfigPath(String a_applicationName, boolean a_bCreate)
	{
		//Return path in user's home directory with hidden file (preceded by ".")
		return System.getProperty("user.home", "") + "/.";
	}
	
	
	public boolean isVirtualBoxInterface(Object a_networkInterface)
	{

		String strDisplayName = getInterfaceName(a_networkInterface);
	    return (strDisplayName != null) && 
			    (strDisplayName.indexOf(LINUX_VBOX_NETADAPTER) >= 0);		
	}
	
	/**
	 * @author G. Koppen
	 */
	public boolean copyAsRoot(File a_sourceFile, File a_targetDirectory, AbstractRetryCopyProcess a_checkRetry)
	{
		if (a_sourceFile == null || a_targetDirectory == null || !a_targetDirectory.isDirectory())
		{
			return false;
		}
		
		Process pr;
		BufferedReader stdInput;
		String KDEPath =  "";
		String s = null;
		String[] command = new String[3];
		StringTokenizer tokenizerKDEPath;
		String tokenCurrent;
		File kdesuExec;
		String execCmd = null;
		boolean bUsedXterm = false;
		String sourcePath = a_sourceFile.getPath();
		String targetPath = a_targetDirectory.getPath() + "/";

		a_checkRetry.incrementProgress();
		if (m_bKDE)
		{
			if (m_iKDEVersion == 4)
			{
				try
				{
					command[0] = "kde4-config";
					command[1] = "--path";
					command[2] = "exe";
					pr = Runtime.getRuntime().exec(command);
					stdInput = new BufferedReader(new InputStreamReader(pr.getInputStream()));
					while (!procDone(pr))
					{
						while ((s = stdInput.readLine()) != null)
						{
							KDEPath = KDEPath + s;
						}
					}
					LogHolder.log(LogLevel.NOTICE, LogType.MISC, "The KDE-Path on your system is: " + KDEPath);
					tokenizerKDEPath = new StringTokenizer(KDEPath, ":");
					while (tokenizerKDEPath.hasMoreTokens())
					{
						tokenCurrent = tokenizerKDEPath.nextToken();
						kdesuExec = new File(tokenCurrent + "kdesu");
						if (kdesuExec.isFile())
						{	
							execCmd = tokenCurrent + "kdesu";
							break;
						}
					}
				}
				catch (IOException e)
				{
					LogHolder.log(LogLevel.EXCEPTION, LogType.MISC, e);
				}
				if (execCmd != null)
				{
					execCmd = execCmd + " 'cp -r " + sourcePath + " " + targetPath + "'";
					executeShell(execCmd);
				}
				else
				{
					LogHolder.log(LogLevel.ALERT, LogType.MISC, "No path to kdesu found, aborting...");
					return false;
				}
			}
      			else
			{
				execCmd = "kdesu 'cp -r " + sourcePath + " " + targetPath + "'";
				executeShell(execCmd);
			}
		}
		else if (m_bGnome)
		{
			execCmd = "gksu 'cp -r " + sourcePath + " " + targetPath + "'";
			executeShell(execCmd);
		}
		else
		{
			execCmd = "xterm -e su -c 'cp -r " + sourcePath + " " + targetPath + "'";
			executeShell(execCmd);
			bUsedXterm = true;
		}

		File target = new File(targetPath + a_sourceFile.getName());
		if (RecursiveFileTool.equals(target, a_sourceFile, true))
		{
			while (a_checkRetry.incrementProgress());
			return true;
		}
		else if (bUsedXterm && a_checkRetry != null && a_checkRetry.checkRetry())
		{
			a_checkRetry.reset();
			return copyAsRoot(a_sourceFile, a_targetDirectory, a_checkRetry);
		}
		
		return false;
	}
	   	
	private void executeShell(String a_cmd) 
	{
		try
		{
			String[] command = new String[3];
		    command[0] = "sh";
		    command[1] = "-c";
		    command[2] = a_cmd;
		    Runtime rt = Runtime.getRuntime();
		    Process pr = rt.exec(command);
		    pr.waitFor();
		}
		catch (Exception e)
		{
			LogHolder.log(LogLevel.EXCEPTION, LogType.MISC, e);
		}
	}
	
	private boolean procDone(Process a_process)
	{
		try
		{
			int v = a_process.exitValue();
			return true;
		}
		catch (IllegalThreadStateException e)
		{
			return false;
		}
	}
}
