/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.proxy;

import java.io.InterruptedIOException;
import java.net.ConnectException;
import java.net.ServerSocket;
import java.net.SocketException;
import java.util.Enumeration;
import java.util.Observable;
import java.util.Observer;
import java.util.Vector;

import anon.AnonChannel;
import anon.AnonServerDescription;
import anon.AnonService;
import anon.AnonServiceEventListener;
import anon.AnonServiceFactory;
import anon.ErrorCodes;
import anon.client.AbstractAutoSwitchedMixCascadeContainer;
import anon.client.AnonClient;
import anon.client.BasicTrustModel;
import anon.client.DummyTrafficControlChannel;
import anon.client.ITrustModel;
import anon.client.TrustModel;
import anon.error.AccountEmptyException;
import anon.error.AnonServiceException;
import anon.error.ConnectionEstablishmentTimeoutException;
import anon.error.INotRecoverableException;
import anon.error.NotRecoverableException;
import anon.error.RecoverableExceptionContainer;
import anon.error.ServiceInterruptedException;
import anon.infoservice.AbstractMixCascadeContainer;
import anon.infoservice.IMutableProxyInterface;
import anon.infoservice.InfoServiceDBEntry;
import anon.infoservice.ListenerInterface;
import anon.infoservice.MixCascade;
import anon.infoservice.MixInfo;
import anon.mixminion.MixminionServiceDescription;
import anon.pay.PayAccountsFile;
import anon.pay.PaymentInstanceDBEntry;
import anon.pay.xml.NotRecoverableXMLError;
import anon.terms.TermsAndConditionConfirmation;
import anon.tor.TorAnonServerDescription;
import anon.transport.connection.IStreamConnection;
import anon.util.ExceptionVariable;
import anon.util.ObjectQueue;
import anon.util.SocketGuard;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

/**
 * This calls implements a proxy one can use for convenient access to the
 * provided anonymous communication primitives, like {@link AnonChannel}
 * objects. Below you find an example which creates a proxy which uses the AN.ON
 * mix cascade for anonymous web surfing.
 * <pre>
 * {@code AnonProxy theProxy = new AnonProxy(serverSocket, null);

 * theProxy.setMixCascade( new MixCascade(null, null, hostNameOfMixCascade,
 * portNumberOfMixCascade) );

 * theProxy.start(); }
 * </pre>
 */

final public class AnonProxy implements AnonServiceEventListener
	{
		public static final int UNLIMITED_REQUESTS = Integer.MAX_VALUE;
		public static final int MIN_REQUESTS = 5;

		public final static int E_MIX_PROTOCOL_NOT_SUPPORTED = ErrorCodes.E_PROTOCOL_NOT_SUPPORTED;

		public final static int E_SIGNATURE_CHECK_FIRSTMIX_FAILED = ErrorCodes.E_SIGNATURE_CHECK_FIRSTMIX_FAILED;

		public final static int E_SIGNATURE_CHECK_OTHERMIX_FAILED = ErrorCodes.E_SIGNATURE_CHECK_OTHERMIX_FAILED;

		private static final int RECONNECT_INTERVAL = 800;

		private int m_maxRequests = UNLIMITED_REQUESTS;

		private int m_socketGuardTimeout = 0;

		private AnonClient m_Anon;

		private AnonService m_Tor;

		private AnonService m_Mixminion;

		private Vector m_anonServiceListener;

		private Thread threadRunOne;
		private Thread threadRunTwo;
		private final Object SYNC_THREAD_RUN = new Object();

		private Thread m_tInit;

		private ServerSocket m_socketListener;
		private ServerSocket m_socketListenerTwo;
		private IMutableProxyInterface m_proxyInterface = new IMutableProxyInterface.DummyMutableProxyInterface();

		private IProxyListener m_ProxyListener;

		private volatile int m_numChannels = 0;

		private boolean m_bReconnecting = false;
		private boolean m_bConnecting = false;
		private final Object THREAD_SYNC = new Object();
		private final Object SHUTDOWN_SYNC = new Object();
		private boolean m_bShuttingDown = false;

		private ProxyCallbackHandler m_callbackHandler;
		private final Object SYNC_CALLBACK_HANDLER = new Object();
		private final HTTPProxyCallback m_httpProxyCallback = new HTTPProxyCallback();
		private DecompressionProxyCallback m_decompressionProxyCallback = null;

		private TermsAndConditionConfirmation termsConfirmation = null;

		/**
		 * Stores the MixCascade we are connected to.
		 */
		private AbstractMixCascadeContainer m_containerMixCascade = new DummyMixCascadeContainer();
		private Observer m_observer;
		private boolean m_bWeChanged = false; // not needed now because of
																					// multi-threading, but maybe in the
																					// future..

		/**
		 * Stores the Tor params.
		 */
		private TorAnonServerDescription m_currentTorParams;

		/**
		 * Stores the Mixminion params.
		 */
		private MixminionServiceDescription m_currentMixminionParams;

		/**
		 * Stores, whether we use a forwarded connection (already active, when
		 * AnonProxy is created) or not.
		 */
		private boolean m_forwardedConnection;

		/**
		 * Stores the maximum dummy traffic interval in milliseconds -> we need
		 * dummy traffic with at least that rate. If this value is -1, there is no
		 * need for dummy traffic on a forwarded connection on the server side. This
		 * value is only meaningful, if m_forwardedConnection is true.
		 */
		private int m_maxDummyTrafficInterval = DummyTrafficControlChannel.DT_MAX_INTERVAL_MS;

		/**
		 * Creates a new AnonProxy. This proxy uses as default only the AN.ON
		 * service. If you also want to use TOR and Mixminion you have to enable
		 * them by calling setTorParams() and setMixmininoParams().
		 *
		 * @see setTorParams()
		 * @see setMixminionParams()
		 *
		 * @param a_listener
		 *          A ServerSocket, where the AnonProxy listens for new requests
		 *          (e.g. from a web browser).
		 */
		public AnonProxy(ServerSocket a_listener, TermsAndConditionConfirmation termsConfirmation, int a_socketTimeout)
			{
				this(a_listener, null, termsConfirmation);
			}

		public AnonProxy(DirectProxy a_directProxy, IMutableProxyInterface a_proxyInterface,
				TermsAndConditionConfirmation termsConfirmation)
			{
				this(a_directProxy, null, a_proxyInterface, termsConfirmation);

			}

		/**
		 * Creates a new AnonProxy. This proxy uses as default only the AN.ON
		 * service. If you also want to use TOR and Mixminion you have to enable
		 * them by calling setTorParams() and setMixmininoParams().
		 *
		 * @see setTorParams()
		 * @see setMixminionParams()
		 *
		 * @param a_listener
		 *          A ServerSocket, where the AnonProxy listens for new requests
		 *          (e.g. from a web browser).
		 * @param a_proxyInterface
		 *          describes a proxy the AnonProxy should use to establish
		 *          connections to the anon servers (e.g. if you are behind some
		 *          firewall etc.)
		 */
		public AnonProxy(ServerSocket a_listener, IMutableProxyInterface a_proxyInterface,
				TermsAndConditionConfirmation termsConfirmation)
			{
				this(null, a_listener, a_proxyInterface, termsConfirmation);
			}

		private AnonProxy(DirectProxy a_directProxy, ServerSocket a_listener, IMutableProxyInterface a_proxyInterface,
				TermsAndConditionConfirmation termsConfirmation)
			{
				if ((a_directProxy == null || a_directProxy.getSocketListener() == null) && a_listener == null)
					{
						throw new IllegalArgumentException("Socket listener is null!");
					}

				if (a_directProxy != null)
					{
						if (a_listener == null)
							{
								m_socketListener = a_directProxy.getSocketListener();
								m_socketListenerTwo = a_directProxy.getSocketListenerTwo();
							}
					}

				if (m_socketListener == null)
					{
						m_socketListener = a_listener;
					}
				if (a_proxyInterface != null)
					{
						m_proxyInterface = a_proxyInterface;
					}
				// HTTP
				m_Anon = new AnonClient(a_directProxy);
				m_Anon.setProxy(m_proxyInterface);
				setDummyTraffic(DummyTrafficControlChannel.DT_DISABLE);
				m_forwardedConnection = false;
				m_anonServiceListener = new Vector();
				m_Anon.removeEventListeners();
				m_Anon.addEventListener(this);
				this.termsConfirmation = termsConfirmation;
				// SOCKS\uFFFD
			}

		/**
		 * Creates a new AnonProxy with an already active mix connection.
		 *
		 * @param a_directProxy
		 * 					A {@link DirectProxy}, from which to get the
		 * 					{@link ServerSocket} objects to listen for requests on.
		 * @param a_proxyConnection
		 *          An already open connection to a mix (but not initialized, like
		 *          keys exchanged, ...).
		 * @param a_currentService
		 * @param a_maxDummyTrafficInterval
		 *          The minimum dummy traffic rate the connection needs. The value
		 *          is the maximum dummy traffic interval in milliseconds. Any call
		 *          of setDummyTraffic(), will respect this maximum interval value
		 *          -> bigger values set with setDummyTraffic (especially -1) result
		 *          in that maximum dummy traffic interval value. If this value is
		 *          -1, there is no need for dummy traffic on that connection on the
		 *          server side.
		 * @param termsConfirmation
		 */
		public AnonProxy(DirectProxy a_directProxy, IStreamConnection a_proxyConnection, MixCascade a_currentService,
				int a_maxDummyTrafficInterval, TermsAndConditionConfirmation termsConfirmation)
			{
				if (a_directProxy == null)
					{
						throw new IllegalArgumentException("Socket listener is null!");
					}
				m_socketListener = a_directProxy.getSocketListener();
				m_socketListenerTwo = a_directProxy.getSocketListenerTwo();
				m_Anon = new AnonClient(a_directProxy, a_proxyConnection, a_currentService);
				m_forwardedConnection = true;
				m_maxDummyTrafficInterval = a_maxDummyTrafficInterval;
				setDummyTraffic(a_maxDummyTrafficInterval);
				m_anonServiceListener = new Vector();
				m_Anon.removeEventListeners();
				m_Anon.addEventListener(this);
				this.termsConfirmation = termsConfirmation;
			}

		private void enableProxyCallback(ProxyCallback callback)
			{
				synchronized (SYNC_CALLBACK_HANDLER)
					{
						if (callback == null)
							{
								return;
							}

						if (m_callbackHandler == null)
							{
								LogHolder.log(LogLevel.WARNING, LogType.NET,
										"No ProxyCallbackHandler activated: cannot process HTTP headers.");
								return;
							}
						m_callbackHandler.registerProxyCallback(callback);
					}
			}

		private void disableProxyCallback(ProxyCallback callback)
			{
				synchronized (SYNC_CALLBACK_HANDLER)
					{
						if (callback != null && m_callbackHandler != null)
							{
								m_callbackHandler.removeCallback(callback);
							}
					}
			}

		public void setHTTPHeaderProcessingEnabled(boolean enable, boolean b_force_proxy_callback)
			{
				synchronized (SYNC_CALLBACK_HANDLER)
					{
						// alternative...
						if (b_force_proxy_callback) {
							m_httpProxyCallback.blockHTTPListeners(!enable);

							if (m_callbackHandler == null) {
								m_callbackHandler = new ProxyCallbackHandler();
							}

							enableProxyCallback(m_httpProxyCallback);
						} else {
							if (enable)
							{
								if (m_callbackHandler == null)
									{
										m_callbackHandler = new ProxyCallbackHandler();
									}

								enableProxyCallback(m_httpProxyCallback);
							}
							else
							{
								disableProxyCallback(m_httpProxyCallback);
								m_callbackHandler = null;
							}
						}
					}
			}

		public synchronized void setHTTPDecompressionEnabled(boolean enable)
			{
				synchronized (SYNC_CALLBACK_HANDLER)
					{
						if (enable)
							{
								if (m_decompressionProxyCallback == null)
									{
										m_decompressionProxyCallback = new DecompressionProxyCallback();
									}
								enableProxyCallback(m_decompressionProxyCallback);
							}
						else
							{
								disableProxyCallback(m_decompressionProxyCallback);
								m_decompressionProxyCallback = null;
							}
					}
			}

		public synchronized void removeHTTPConnectionListener(AbstractHTTPConnectionListener listener)
			{
				if (listener == null)
					{
						return;
					}

				synchronized (SYNC_CALLBACK_HANDLER)
					{
						m_httpProxyCallback.removeHTTPConnectionListener(listener);
					}
			}

		/*
		 * TODO: this also enables the experimental ConnectionWatch has to be
		 * handled by another enable method
		 */
		public void addHTTPConnectionListener(AbstractHTTPConnectionListener listener)
			{
				synchronized (SYNC_CALLBACK_HANDLER)
					{
						m_httpProxyCallback.addHTTPConnectionListener(listener);
						enableProxyCallback(m_httpProxyCallback);
					}
			}

		/**
		 * Sets a new MixCascade.
		 *
		 * @param newMixCascade
		 *          The new MixCascade we are connected to.
		 */
		/*
		 * private void setMixCascade(AbstractMixCascadeContainer newMixCascade) {
		 * if (newMixCascade == null) { m_currentMixCascade = new
		 * DummyMixCascadeContainer(); } else { m_currentMixCascade = new
		 * EncapsulatedMixCascadeContainer(newMixCascade); } //
		 * m_AICom.setAnonServer(newMixCascade); }
		 */

		/** Returns the current Mix cascade */
		public MixCascade getMixCascade()
			{
				if (m_Anon.isConnected())
					{
						return m_Anon.getCurrentService();
					}
				else
					{
						return m_containerMixCascade.getCurrentCascade();
					}
			}

		/**
		 * Sets the parameter for TOR (anonymous SOCKS). If NULL TOR proxy is
		 * disabled.
		 *
		 * @param newTorParams
		 *          The new parameters for TOR.
		 * @see TorAnonServerDescription
		 */
		public void setTorParams(TorAnonServerDescription newTorParams)
			{
				m_currentTorParams = newTorParams;
			}

		public TorAnonServerDescription getTorParams()
			{
				return m_currentTorParams;
			}

		/**
		 * Sets the parameter for Mixminion (anonymous remailer). If NULL Mixminion
		 * proxy is disabled.
		 *
		 * @param newMixminionParams
		 *          The new parameters for Mixminion. If NULL the Mixminion proxy is
		 *          disabled.
		 * @see MixminionServiceDescription
		 */
		public void setMixminionParams(MixminionServiceDescription newMixminionParams)
			{
				m_currentMixminionParams = newMixminionParams;
			}

		public MixminionServiceDescription getMixminionParams()
			{
				return m_currentMixminionParams;
			}

		public void setMaxConcurrentRequests(int a_maxRequests)
			{
				if (a_maxRequests > MIN_REQUESTS)
					{
						m_maxRequests = a_maxRequests;
					}
			}

		public int getMaxConcurrentRequests()
			{
				return m_maxRequests;
			}

		/**
		 * Changes the dummy traffic interval on the connection to the server. This
		 * method respects dummy traffic restrictions on a forwarded connection. If
		 * there is a minimum dummy traffic rate needed by the server, the dummy
		 * traffic interval gets never bigger than that needed rate on a forwarded
		 * connection (especially a interval value of -1 is ignored).
		 *
		 * @param a_interval
		 *          The interval for dummy traffic on the connection to the server
		 *          in milliseconds.
		 */
		public void setDummyTraffic(int a_interval)
			{
				try
					{
						if ((!m_forwardedConnection) || (m_maxDummyTrafficInterval < 0)
								|| (a_interval == DummyTrafficControlChannel.DT_DISABLE))
							{
								/* no dummy traffic restrictions */
								m_Anon.setDummyTraffic(a_interval);
							}
						else
							{
								/* there are dummy traffic restrictions */
								if (a_interval >= 0)
									{
										/* take the smaller interval */
										m_Anon.setDummyTraffic(Math.min(a_interval, m_maxDummyTrafficInterval));
									}
								else
									{
										/*
										 * we need dummy traffic with a minimum rate -> can't
										 * disable dummy traffic
										 */
										m_Anon.setDummyTraffic(m_maxDummyTrafficInterval);
									}
							}
					}
				catch (ClassCastException a_e)
					{
						// this is a direct proxy!
					}
			}

		/**
		 * Changes the timeout used for blocking an faulty interface of a Cascade.
		 */
		public void setInterfaceBlockTimeout(long a_msTimeout)
			{
				m_Anon.setInterfaceBlockTimout(a_msTimeout);
			}

		public void setDebug(boolean bDebug)
		{
			m_Anon.setDebug(bDebug);

		}

		public void stop()
			{
				synchronized (SHUTDOWN_SYNC)
				// synchronized (THREAD_SYNC)
					{
						// if (getMixCascade.isPayment()) {
						// LogHolder.log(LogLevel.ALERT, LogType.NET, "DEBUG InfoService",
						// new Exception()); }

						m_tInit = null;

						synchronized (SYNC_THREAD_RUN)
							{
								if (threadRunOne == null && threadRunTwo == null)
									{
										disconnected();
										return;
									}
							}

						m_bShuttingDown = true;

						m_Anon.shutdown(true);
						if (m_Tor != null)
							{
								m_Tor.shutdown(true);
							}
						if (m_Mixminion != null)
							{
								m_Mixminion.shutdown(true);
							}

						while (threadRunOne.isAlive())
							{
								try
									{
										threadRunOne.interrupt();
										threadRunOne.join(500);
									}
								catch (InterruptedException e)
									{
									}
							}
						while (threadRunTwo != null && threadRunTwo.isAlive())
							{
								try
									{
										threadRunTwo.interrupt();
										threadRunTwo.join(500);
									}
								catch (InterruptedException e)
									{
									}
							}

						m_Tor = null;
						m_Mixminion = null;
						synchronized (SYNC_THREAD_RUN)
							{
								threadRunTwo = null;
								threadRunOne = null;
							}

						packetMixed(0);
						/*
						 * synchronized (m_eventListeners) { final Enumeration
						 * eventListenersList = m_eventListeners.elements(); Thread
						 * notificationThread = new Thread(new Runnable() { public void
						 * run() { while (eventListenersList.hasMoreElements()) { (
						 * (AnonServiceEventListener)
						 * (eventListenersList.nextElement())).packetMixed(0); } } },
						 * "AnonClient: Zero PacketMixed notification (after shutdown)");
						 * notificationThread.setDaemon(true); notificationThread.start(); }
						 */

						while (m_bReconnecting)
							{
								try
									{
										SHUTDOWN_SYNC.wait(100);
									}
								catch (InterruptedException e)
									{
										// ignore
									}
							}

						disconnected();

						m_bShuttingDown = false;

						TrustModel.getCurrentTrustModel().unblockInterfacesFromDatabase();
						ListenerInterface.unblockInterfacesFromDatabase(InfoServiceDBEntry.class);
					}

				/*
				 * synchronized (THREAD_SYNC) { THREAD_SYNC.notify(); }
				 */

			}

		/**
		 * This is a queue
		 */
		public class RoundRobinRequestQueue
			{
				private final Vector vecPriorityRequests = new Vector();

				public void addPriority(AnonProxyRequest a_request)
					{
						synchronized (vecPriorityRequests)
							{
								if (!vecPriorityRequests.contains(a_request))
									{
										vecPriorityRequests.addElement(a_request);
									}
							}
					}

				public void removePriority(AnonProxyRequest a_request)
					{
						synchronized (vecPriorityRequests)
							{
								vecPriorityRequests.removeElement(a_request);
							}
					}

				public boolean isSlowDownUploads()
					{
						return vecPriorityRequests.size() > 0;
					}

				public boolean isBlockUploads()
					{
						return m_Anon.isSendingControlMessage();
					}
			}

		/**
		 * This class receives {@link SocketGuard} objects via the
		 * {@link OpenSocketRequester#pushSocket(SocketGuard) pushSocket()} method
		 * and starts {@link AnonProxyRequest}s, respecting the max. number of
		 * requests that class can process requests.
		 * <p>
		 * This class can be used to process sockets, that want to send a request
		 * without having to worry about blocking or sending too many requests at
		 * once.
		 */
		private class OpenSocketRequester implements Runnable
			{
				/**
				 * A queue used for {@link SocketGuard} objects, wrapping sockets the
				 * client used to send requests over.
				 *
				 * They will
				 */
				private ObjectQueue m_socketQueue = new ObjectQueue();

				/**
				 * {@link AnonProxy} object used to send requests and receive responses
				 * to the destination of the initial request.
				 */
				private AnonProxy m_proxy;

				/**
				 * Used by the underlying {@link AnonProxyRequest} objects to wait, when
				 * the given {@link AnonProxy} needs to reconnect.
				 */
				private Object m_syncObject;

				/**
				 * Signals if the {@link OpenSocketRequester} has been or is being
				 * closed.
				 */
				private boolean m_bIsClosed = false;

				/**
				 * Instantiates an {@link OpenSocketRequester}.
				 * @param a_proxy {@link AnonProxy} to communicate over.
				 * @param a_syncObject
				 */
				public OpenSocketRequester(AnonProxy a_proxy, Object a_syncObject)
					{
						m_proxy = a_proxy;
						m_syncObject = a_syncObject;
					}

				/**
				 * Adds a {@link SocketGuard} to the internal queue and notifies threads
				 * waiting on the queue.
				 *
				 * @param clientSocket The {@link SocketGuard} object to add to the
				 * queue.
				 */
				public void pushSocket(SocketGuard clientSocket)
					{
						synchronized (m_socketQueue)
							{
								m_socketQueue.push(clientSocket);
								m_socketQueue.notify();
							}
					}

				/**
				 * Will stop the {@link #run()} method and signal Threads waiting on the
				 * internal socket queue.
				 */
				public void close()
					{
						m_bIsClosed = true;
						synchronized (m_socketQueue)
							{
								m_socketQueue.notify();
							}
					}

				/**
				 * {@link OpenSocketRequester}
				 */
				public void run()
					{
						// While the thread is not interrupted or the object closing
						while (!Thread.currentThread().isInterrupted() && !m_bIsClosed)
							{
								// If there are sockets available and the max request number has
							  // not been reached
								if (m_socketQueue.getSize() > 0 && AnonProxyRequest.getNrOfRequests() < m_maxRequests)
									{
										try
											{
												// start a request using a socket from the queue
												new AnonProxyRequest(m_proxy, (SocketGuard) m_socketQueue.pop(), m_syncObject,
														m_callbackHandler);
											}
										catch (Exception e)
											{
												LogHolder.log(LogLevel.ERR, LogType.NET, e);
											}
									}
								else
									{
										try
											{
												synchronized (m_socketQueue)
													{
														/*
														 * If there were too many requests, assume that
														 * sockets are available and only wait 100ms to try
														 * again.
														 */
														if (AnonProxyRequest.getNrOfRequests() >= m_maxRequests)
															{
																m_socketQueue.wait(100);
															}
														else
															{
																/*
																 * Assume that there were no sockets to be
																 * processed and wait until we get notified
																 * by the queue, that a new socket is available.
																 */
																m_socketQueue.wait();
															}
													}
											}
										catch (InterruptedException ex)
											{
												break;
											}
									}
							}
						LogHolder.log(LogLevel.INFO, LogType.NET, "Open socket thread stopped.");
					}
			}

		/**
		 * This object continuously accepts connections on the given
		 * {@link ServerSocket} and, assuming, that there is a request to be sent
		 * out, pushes resulting connected sockets to an {@link OpenSocketRequester}
		 * for further processing.
		 */
		private class RunnableProxy implements Runnable
			{
				private ServerSocket m_serverSocket;

				/**
				 * Creates a {@link RunnableProxy} object, that listens for requests on
				 * the given {@link ServerSocket}.
				 *
				 * @param a_serverSocket
				 */
				public RunnableProxy(ServerSocket a_serverSocket)
					{
						m_serverSocket = a_serverSocket;
					}

				/**
				 * See {@link RunnableProxy}
				 */
				public void run()
					{
						Thread requesterThread;
						OpenSocketRequester requester;
						int oldTimeOut = 0;
						LogHolder.log(LogLevel.DEBUG, LogType.NET, "AnonProxy is running as Thread");

						// try saving old time out to restore later, when shutting down
						try
							{
								oldTimeOut = m_serverSocket.getSoTimeout();
							}
						catch (Exception e)
							{
								if (m_bShuttingDown)
									{
										return;
									}
							}

						// try setting new timeout
						try
							{
								m_serverSocket.setSoTimeout(2000);
							}
						catch (Exception e1)
							{
								LogHolder.log(LogLevel.DEBUG, LogType.NET, "Could not set accept time out!", e1);
							}

						// restore initial timeout upon shutting down
						if (m_bShuttingDown)
							{
								try
									{
										m_serverSocket.setSoTimeout(oldTimeOut);
									}
								catch (Exception e4)
									{
									}

								return;
							}

						requester = new OpenSocketRequester(AnonProxy.this, THREAD_SYNC);
						requesterThread = new Thread(requester, requester.getClass().getName());
						requesterThread.start();

						try
							{
								while (!Thread.currentThread().isInterrupted() && !m_bShuttingDown)
									{
										if (!isConnected())
											{
												// this is necessary so that the DirectProxy may get the
												// requests meanwhile
												Thread.sleep(250);
												continue;
											}

										/*
										 * wait for a connection to the server socket, indicating,
										 * that a request is to be made over the proxy
										 */
										SocketGuard socket = null;
										try
											{
												socket = new SocketGuard(m_serverSocket.accept(), m_socketGuardTimeout);
												// socket = m_serverSocket.accept();
											}
										catch (InterruptedIOException e)
											{
												continue;
											}

										/*
										 * Give the socket to the requester to do its thing.
										 */
										try
											{
												socket.setSoTimeout(0); // ensure that socket is
																								// blocking!
												requester.pushSocket(socket);
											}
										catch (SocketException soex)
											{
												socket = null;
												LogHolder.log(LogLevel.ERR, LogType.NET, "Could not set non-Blocking mode for Channel-Socket!",
														soex);
												continue;
											}
									}
							}
						catch (Exception e)
							{
								LogHolder.log(LogLevel.ERR, LogType.NET, e);
							}
						try
							{
							// restore old timeout
								m_serverSocket.setSoTimeout(oldTimeOut);
							}
						catch (Exception e4)
							{
							}
						requesterThread.interrupt();
						requester.close();

						try
							{
								requesterThread.join();
							}
						catch (InterruptedException ex)
							{
							}

						LogHolder.log(LogLevel.WARNING, LogType.NET,
								"JAPAnonProxyServer stopped socket: " + m_serverSocket.toString());
					}
			}

		/**
		 * Creates a channel to the destination given to this proxy object, using
		 * the underlying {@link AnonService}. It can be used to write and receive
		 * data.
		 *
		 * @param type Type of channel. See {@link AnonChannel}
		 * @return An {@link AnonChannel} object.
		 * @throws ConnectException If channel creation fails.
		 */
		AnonChannel createChannel(int type) throws ConnectException
			{
				if (type == AnonChannel.SOCKS)
					{
						if (m_Tor != null)
							{
								return m_Tor.createChannel(AnonChannel.SOCKS);
							}
						else if (getMixCascade().isSocks5Supported())
							{
								return m_Anon.createChannel(AnonChannel.SOCKS);
							}
						else
							{
								LogHolder.log(LogLevel.ERR, LogType.NET, "Received SOCKS request, but no SOCKS server is available.");
							}
					}
				else if (type == AnonChannel.HTTP)
					{
						return m_Anon.createChannel(AnonChannel.HTTP);
					}
				else if (type == AnonChannel.VPN)
					{
						return m_Anon.createChannel(AnonChannel.VPN);
					}
				else if (type == AnonChannel.SMTP)
					{
						if (m_Mixminion != null)
							{
								return m_Mixminion.createChannel(AnonChannel.SMTP);
							}
					}
				return null;
			}

		// synchronized
		void reconnect()
			{
				MixCascade cascade;
				boolean bReconnect;

				synchronized (THREAD_SYNC)
					{
						if (m_Anon.isConnected() || m_bShuttingDown || Thread.currentThread().isInterrupted())
							{
								m_bConnecting = false;
								return;
							}

						if (!m_containerMixCascade.isReconnectedAutomatically())
							{
								m_bConnecting = false;
								stop();
								THREAD_SYNC.notifyAll(); // maybe setting has changed meanwhile
								return;
							}
						if (m_bReconnecting)
							{
								// there already is a reconnect thread
								return;
							}

						m_bReconnecting = true;
						m_bConnecting = true;

						while (threadRunOne != null && m_containerMixCascade.isReconnectedAutomatically() && !m_Anon.isConnected()
								&& !m_bShuttingDown && !Thread.currentThread().isInterrupted())
							{
								bReconnect = m_Anon.getCurrentService() == m_containerMixCascade.getCurrentCascade();

								m_bWeChanged = true;
								cascade = m_containerMixCascade.getNextCascade();
								m_bWeChanged = false;

								int loglevel = LogLevel.WARNING;
								try
									{
										if (cascade.getListenerInterface(0).getHost().equals("0.0.0.0"))
											{
												loglevel = LogLevel.INFO;
											}
									}
								catch (Exception a_e)
									{
										// ignore
									}
								LogHolder.log(loglevel, LogType.NET,
										"Try reconnect to AN.ON service. Connecting to " + cascade.getName() + "...");

								Thread tInit = null;
								final ExceptionVariable retVar = new ExceptionVariable(null);

								if (!cascade.equals(m_Anon.getCurrentService()))
									{
										fireCurrentServiceChanged(cascade);
									}
								tInit = startInitThread(retVar, cascade, true, bReconnect);

								joinInitThread(tInit, retVar);

								finishInitThread(cascade, retVar, tInit);

								if (retVar.get() != null)
									{
										LogHolder.log(LogLevel.NOTICE, LogType.NET, retVar.get());

										if (retVar.get() instanceof ServiceInterruptedException
												|| !m_containerMixCascade.isReconnectedAutomatically()
												|| (!m_containerMixCascade.isServiceAutoSwitched()
														&& retVar.get() instanceof AnonServiceException &&
														// these errors cannot be 'healed'
														retVar.get() instanceof INotRecoverableException && ((AnonServiceException) retVar.get())
														.getService() != AbstractAutoSwitchedMixCascadeContainer.INITIAL_DUMMY_SERVICE))
											{
												if (retVar.get() instanceof INotRecoverableException)
													{
														connectionError((AnonServiceException) retVar.get());
													}
												break;
											}

										try
											{
												THREAD_SYNC.wait(RECONNECT_INTERVAL);
											}
										catch (InterruptedException ex)
											{
												break;
											}
									}
							}

						synchronized (SHUTDOWN_SYNC)
							{
								m_bReconnecting = false;
								m_bConnecting = false;

								if (!m_bShuttingDown && (threadRunOne == null || !isConnected())
										&& !m_containerMixCascade.isReconnectedAutomatically())
									{
										stop();
										THREAD_SYNC.notifyAll(); // maybe setting has changed
																							// meanwhile
									}
								SHUTDOWN_SYNC.notify();
							}

						if (isConnected())
							{
								m_containerMixCascade.reset();
							}

						return;
					}
			}

		public void setProxyListener(IProxyListener l)
			{
				m_ProxyListener = l;
			}

		private boolean interruptInit(Thread a_tInit)
			{
				synchronized (THREAD_SYNC)
					{
							// synchronized (SHUTDOWN_SYNC)
							{
								if (a_tInit != null)
									{
										LogHolder.log(LogLevel.WARNING, LogType.NET, "Interrupting init...", new InterruptedException());
										while (a_tInit.isAlive())
											{
												a_tInit.interrupt();
												try
													{
														a_tInit.join(100);
													}
												catch (InterruptedException e1)
													{
														// ignore
													}
											}
										LogHolder.log(LogLevel.WARNING, LogType.NET, "Init was interrupted successfully!");
										return true;
									}
							}
					}
				return false;
			}

		private final Vector m_currentStartThreads = new Vector();

		public int countStartThreads()
			{
				return m_currentStartThreads.size();
			}

		public void start(AbstractMixCascadeContainer a_newMixCascade) throws AnonServiceException
			{
				synchronized (THREAD_SYNC)
					{
						m_currentStartThreads.addElement(Thread.currentThread());
						m_bConnecting = true;

						try
							{
								start_internal(a_newMixCascade);
								m_bConnecting = false;
								m_currentStartThreads.removeElement(Thread.currentThread());
							}
						catch (AnonServiceException a_e)
							{
								// m_bConnecting = false;
								m_currentStartThreads.removeElement(Thread.currentThread());
								throw a_e;
							}
						catch (RuntimeException a_e)
							{
								m_bConnecting = false;
								m_currentStartThreads.removeElement(Thread.currentThread());
								throw a_e;
							}
					}
			}

		private void joinInitThread(Thread tInit, ExceptionVariable retVar)
			{
				if (tInit == null)
					{
						return;
					}

				long lTimeout = System.currentTimeMillis() + AnonClient.getLoginTimeout();

				while (tInit.isAlive())
					{
						m_bConnecting = true; // necessary as reconnect thread may reset
																	// this...
						try
							{
								THREAD_SYNC.wait(100);
							}
						catch (InterruptedException e)
							{
								interruptInit(tInit);
							}
						if (m_tInit == null)
							{
								// the stop method was called meanwhile!
								interruptInit(tInit);
							}
						else if (lTimeout < System.currentTimeMillis())
							{
								// timed out
								interruptInit(tInit);
								retVar.set(new ConnectionEstablishmentTimeoutException(m_Anon.getCurrentService()));
							}
					}
				m_bConnecting = true; // necessary as reconnect thread may reset this...
			}

		private void finishInitThread(MixCascade currentCascade, ExceptionVariable retVar, Thread tInit)
			{
				synchronized (SHUTDOWN_SYNC)
					{
						if (m_tInit == null || m_bShuttingDown)
							{
								// the stop method was called meanwhile!
								if (m_Anon.isConnected())
									{
										m_Anon.shutdown(false);
										disconnected();
									}
								retVar.set(new ServiceInterruptedException(currentCascade));
							}
						if (m_tInit == tInit)
							{
								m_tInit = null;
							}
					}
			}

		private Thread startInitThread(final ExceptionVariable retVar, final MixCascade nextCascade,
				final boolean a_bKeepCurrentService, final boolean a_bReconnect)
			{
				synchronized (SHUTDOWN_SYNC)
					{
						if (m_bShuttingDown)
							{
								return null;
							}

						Thread tInit;
						m_tInit = new Thread("AN.ON Connection establishment (key exchange etc.)")
							{
								public void run()
									{
										if (nextCascade.isPayment()
												&& PayAccountsFile.getInstance().getChargedAccount(nextCascade.getPIID()) == null)
											{
												/*
												 * This implementation prevents free service users to
												 * massively connect to Premium services if free
												 * services are down.
												 */
												try
													{
														PayAccountsFile.getInstance().signalAccountRequest(nextCascade);
													}
												catch (AccountEmptyException e1)
													{
														retVar.set(e1);
														return;
													}
											}

										try
											{
												m_Anon.initialize(nextCascade, m_containerMixCascade, termsConfirmation, a_bReconnect);
												if (a_bKeepCurrentService)
													{
														m_containerMixCascade.keepCurrentService(true);
													}
											}
										catch (AnonServiceException a_e)
											{
												retVar.set(a_e);
											}
									}
							};
						m_tInit.start();
						tInit = m_tInit;
						return tInit;
					}
			}

		private void start_internal(AbstractMixCascadeContainer a_newMixCascade) throws AnonServiceException
			{
				boolean bConnectionError = false;
				final ExceptionVariable retVar = new ExceptionVariable(null);
				Thread tInit;
				MixCascade currentCascade;

				// synchronize, so no shutdown occurs, while we are busy
				synchronized (SHUTDOWN_SYNC)
					{
						if (a_newMixCascade == null)
							{
								a_newMixCascade = new DummyMixCascadeContainer();
							}
						else
							{
								a_newMixCascade = new EncapsulatedMixCascadeContainer(a_newMixCascade);
							}

						// delete observers of old mix container
						if (m_observer != null)
							{
								m_containerMixCascade.deleteObserver(m_observer);
							}
						m_containerMixCascade = a_newMixCascade;

						m_bWeChanged = true;

						final MixCascade nextCascade = m_containerMixCascade.getNextCascade();
						currentCascade = nextCascade;

						if (m_bShuttingDown)
							{
								throw new ServiceInterruptedException(currentCascade);
							}

						m_bWeChanged = false;
						m_observer = new Observer()
							{
								public void update(Observable a_observable, Object a_object)
									{
										if (a_object != null && a_object instanceof MixCascade)
											{
												synchronized (THREAD_SYNC)
													{
														if (!m_bWeChanged)
															{
																new Thread(new Runnable()
																	{
																		public void run()
																			{
																				synchronized (THREAD_SYNC)
																					{
																						// do not react on m_bReconnecting -
																						// this will lead to deadlocks if
																						// network is off!!
																						if (m_tInit != null || m_Anon.isConnected())
																							{
																								try
																									{
																										m_containerMixCascade.keepCurrentService(true);
																										start(m_containerMixCascade);
																									}
																								catch (AnonServiceException a_ex)
																									{
																										LogHolder.log(LogLevel.ERR, LogType.NET,
																												"Switching service failed!", a_ex);
																									}
																							}
																					}
																			}
																	}).start();
															}
													}
											}
									}
							};
						m_containerMixCascade.addObserver(m_observer);

						if (nextCascade != null)
							{
								if ((m_Anon.isConnected() || m_tInit != null) && m_Anon.getCurrentService().equals(nextCascade)
										&& m_containerMixCascade.getTrustModel().isTrusted(nextCascade)) // if
																																											// it
																																											// is
																																											// not
																																											// trusted,
																																											// we
																																											// will
																																											// check
																																											// if
																																											// this
																																											// has
																																											// changed
																																											// during
																																											// initialization
									{
										// We already have this cascade activated!
										return;
									}
							}
						else
							{
								retVar.set(new NotRecoverableException(null, "Could not get cascade to connect. Next cascade is null!",
										ErrorCodes.E_UNKNOWN));
								throw (AnonServiceException) retVar.get();
							}

						THREAD_SYNC.notifyAll();

						interruptInit(m_tInit);

						if (threadRunOne != null || threadRunTwo != null)
							{
								m_Anon.shutdown(false);

								if (threadRunTwo != null)
									{
										while (threadRunTwo.isAlive())
											{
												try
													{
														threadRunTwo.interrupt();
														threadRunTwo.join(500);
													}
												catch (InterruptedException e)
													{
													}
											}
										threadRunTwo = null;
									}
								if (threadRunOne != null)
									{
										while (threadRunOne.isAlive())
											{
												try
													{
														threadRunOne.interrupt();
														threadRunOne.join(500);
													}
												catch (InterruptedException e)
													{
													}
											}
										threadRunOne = null;
									}
							}
						else
							{
								m_Anon.shutdown(true); // reset transferred bytes
							}

						LogHolder.log(LogLevel.NOTICE, LogType.NET, "Connecting to AN.ON service " + nextCascade.getName() + "...");
						m_numChannels = 0;

						fireCurrentServiceChanged(nextCascade);

						tInit = startInitThread(retVar, nextCascade, false, false);
					} // synchronize(SHUTDOWN_SYNC) end

				joinInitThread(tInit, retVar);

				synchronized (SHUTDOWN_SYNC)
					{
						finishInitThread(currentCascade, retVar, tInit);

						synchronized (SYNC_THREAD_RUN)
							{
								if (threadRunOne != null || threadRunTwo != null)
									{
										// it seems that another start thread has finished meanwhile
										return;
									}
							}

						if (retVar.get() != null)
							{
								if (retVar.get() instanceof ServiceInterruptedException
										|| !m_containerMixCascade.isReconnectedAutomatically()
										|| (!m_containerMixCascade.isServiceAutoSwitched() && retVar.get() instanceof AnonServiceException
												&&
												// these errors cannot be 'healed'
												retVar.get() instanceof INotRecoverableException && ((AnonServiceException) retVar.get())
												.getService() != AbstractAutoSwitchedMixCascadeContainer.INITIAL_DUMMY_SERVICE))
									{
										connectionError((AnonServiceException) retVar.get());
										throw (AnonServiceException) retVar.get();
									}
								else
									{
										bConnectionError = true;
									}
							}
						else
							{
								m_containerMixCascade.keepCurrentService(true);
								m_containerMixCascade.reset();
							}
						LogHolder.log(LogLevel.INFO, LogType.NET, "AN.ON initialized");

						if (m_currentTorParams != null)
							{
								m_Tor = AnonServiceFactory.getAnonServiceInstance(AnonServiceFactory.SERVICE_TOR);
								m_Tor.setProxy(m_proxyInterface);
								try
									{
										m_Tor.initialize(m_currentTorParams, null, termsConfirmation, false);
										LogHolder.log(LogLevel.DEBUG, LogType.NET, "Tor initialized");
									}
								catch (AnonServiceException a_ex)
									{
										LogHolder.log(LogLevel.EXCEPTION, LogType.NET, a_ex);
									}
							}
						if (m_currentMixminionParams != null)
							{
								m_Mixminion = AnonServiceFactory.getAnonServiceInstance(AnonServiceFactory.SERVICE_MIXMINION);
								m_Mixminion.setProxy(m_proxyInterface);
								try
									{
										m_Mixminion.initialize(m_currentMixminionParams, null, termsConfirmation, false);
									}
								catch (AnonServiceException a_ex)
									{
										LogHolder.log(LogLevel.EXCEPTION, LogType.NET, a_ex);
									}
								LogHolder.log(LogLevel.DEBUG, LogType.NET, "Mixminion initialized");
							}

						synchronized (SYNC_THREAD_RUN)
							{
								if (m_socketListenerTwo != null)
									{
										threadRunTwo = new Thread(new RunnableProxy(m_socketListenerTwo), "JAP - AnonProxy 2nd");
										threadRunTwo.setDaemon(true);
										threadRunTwo.start();
									}

								threadRunOne = new Thread(new RunnableProxy(m_socketListener), "JAP - AnonProxy");
								threadRunOne.setDaemon(true);
								threadRunOne.start();
							}
					}

				if (bConnectionError)
					{
						if (!m_containerMixCascade.isReconnectedAutomatically()
								|| (!m_containerMixCascade.isServiceAutoSwitched() && retVar.get() instanceof AnonServiceException &&
								// these errors cannot be 'healed'
										retVar.get() instanceof INotRecoverableException && ((AnonServiceException) retVar.get())
										.getService() != AbstractAutoSwitchedMixCascadeContainer.INITIAL_DUMMY_SERVICE))
							{

							}
						else if (retVar.get() instanceof INotRecoverableException)
							{
								if (retVar.get() instanceof NotRecoverableXMLError)
									{
										retVar.set(((NotRecoverableXMLError) retVar.get()).getSource());
									}
								else
									{
										retVar.set(new RecoverableExceptionContainer((AnonServiceException) retVar.get()));
									}
							}

						connectionError((AnonServiceException) retVar.get());
						throw (AnonServiceException) retVar.get();
					}
			}

		/**
		 * Decreases the internal channel count and notifies any
		 * {@link IProxyListener} objects registered to this {@link AnonProxy}.
		 */
		protected synchronized void decNumChannels()
			{
				m_numChannels--;
				if (m_ProxyListener != null)
					{
						m_ProxyListener.channelsChanged(m_numChannels);
					}
			}

		/**
		 * Increases the internal channel count and notifies any
		 * {@link IProxyListener} objects registered to this {@link AnonProxy}.
		 */
		protected synchronized void incNumChannels()
			{
				m_numChannels++;
				if (m_ProxyListener != null)
					{
						m_ProxyListener.channelsChanged(m_numChannels);
					}
			}

		/**
		 * Lets a potential {@link IProxyListener} object know, how many bytes of
		 * which type of protocol have been sent.
		 *
		 * @param bytes
		 * @param protocolType
		 */
		protected synchronized void transferredBytes(long bytes, int protocolType)
			{
				if (m_ProxyListener != null)
					{
						m_ProxyListener.transferedBytes(bytes, protocolType);
					}
			}

		private void fireDisconnected()
			{
				synchronized (m_anonServiceListener)
					{
						Enumeration e = m_anonServiceListener.elements();
						while (e.hasMoreElements())
							{
								((AnonServiceEventListener) e.nextElement()).disconnected();
							}
					}
			}

		private void fireConnecting(AnonServerDescription a_serverDescription, boolean a_bIsReconnect)
			{
				synchronized (m_anonServiceListener)
					{
						Enumeration e = m_anonServiceListener.elements();
						while (e.hasMoreElements())
							{
								((AnonServiceEventListener) e.nextElement()).connecting(a_serverDescription, a_bIsReconnect);
							}
					}
			}

		private void fireCurrentServiceChanged(AnonServerDescription a_serverDescription)
			{
				synchronized (m_anonServiceListener)
					{
						Enumeration e = m_anonServiceListener.elements();
						while (e.hasMoreElements())
							{
								((AnonServiceEventListener) e.nextElement()).currentServiceChanged(a_serverDescription);
							}
					}
			}

		private void fireConnectionEstablished(AnonServerDescription a_serverDescription)
			{
				synchronized (m_anonServiceListener)
					{
						if (a_serverDescription instanceof MixCascade)
							{
								int iPremiumProbability = -1;
								boolean bCountRedirect = false;
								MixCascade cascade = (MixCascade) a_serverDescription;

								MixInfo lastMix = cascade.getMixInfo(cascade.getNumberOfMixes() - 1);
								if (lastMix != null && (iPremiumProbability = lastMix.getPremiumProbability()) >= 0)
									{
										// this is a test service
										bCountRedirect = true;
									}
								else
									{
										PaymentInstanceDBEntry pi = null;
										if (cascade.isPayment())
											{
												pi = cascade.getPaymentInstance();
											}
										if (pi == null || pi.isTest())
											{
												bCountRedirect = true;
											}
									}
								m_httpProxyCallback.resetRedirect(iPremiumProbability, bCountRedirect);
							}

						Enumeration e = m_anonServiceListener.elements();
						while (e.hasMoreElements())
							{
								((AnonServiceEventListener) e.nextElement()).connectionEstablished(a_serverDescription);
							}
					}
			}

		public void connecting(AnonServerDescription a_serverDescription, boolean a_bIsReconnect)
			{
				LogHolder.log(LogLevel.WARNING, LogType.NET, "AnonProxy is connecting to: " + a_serverDescription);
				fireConnecting(a_serverDescription, a_bIsReconnect);
			}

		public void currentServiceChanged(AnonServerDescription a_serverDescription)
			{
				LogHolder.log(LogLevel.ALERT, LogType.NET, "AnonProxy changed current service to '" + a_serverDescription
						+ "'.");
				fireCurrentServiceChanged(a_serverDescription);
			}

		public void connectionEstablished(AnonServerDescription a_serverDescription)
			{
				LogHolder.log(LogLevel.ALERT, LogType.NET, "AnonProxy received connectionEstablished to '"
						+ a_serverDescription + "'.");
				fireConnectionEstablished(a_serverDescription);
			}

		public void disconnected()
			{
				LogHolder.log(LogLevel.ALERT, LogType.NET, "AnonProxy was disconnected from service "
						+ m_containerMixCascade.getCurrentCascade().getName() + ".");
				fireDisconnected();
			}

		public void connectionError(AnonServiceException a_e)
			{
				LogHolder
						.log(
								(a_e != null && a_e.getService() == AbstractAutoSwitchedMixCascadeContainer.INITIAL_DUMMY_SERVICE ? LogLevel.DEBUG
										: (a_e != null && a_e instanceof ServiceInterruptedException ? LogLevel.WARNING : LogLevel.ERR)),
								LogType.NET, "AnonProxy received connectionError", a_e, LogHolder.CALLING_METHOD_ADD);

				synchronized (m_anonServiceListener)
					{
						Enumeration e = m_anonServiceListener.elements();
						while (e.hasMoreElements())
							{
								((AnonServiceEventListener) e.nextElement()).connectionError(a_e);
							}
					}

				new Thread(new Runnable()
					{
						public void run()
							{
								reconnect();
							}
					}, "Connection error reconnect thread").start();
			}

		public synchronized void addEventListener(AnonServiceEventListener l)
			{
				if (l == null)
					{
						LogHolder.log(LogLevel.WARNING, LogType.NET, "Tried to add NULL listener to AnonProxy.");
						return;
					}

				synchronized (m_anonServiceListener)
					{
						Enumeration e = m_anonServiceListener.elements();
						while (e.hasMoreElements())
							{
								if (l.equals(e.nextElement()))
									{
										return;
									}
							}
						m_anonServiceListener.addElement(l);
					}
			}

		public synchronized void removeEventListener(AnonServiceEventListener l)
			{
				m_anonServiceListener.removeElement(l);
			}

		public boolean isConnected()
			{
				return m_Anon.isConnected();
			}

		public boolean isConnecting()
			{
				return m_bReconnecting || m_bConnecting;
			}

		public void packetMixed(long a_totalBytes)
			{
				if (isConnected() || a_totalBytes == 0)
					{
						synchronized (m_anonServiceListener)
							{
								Enumeration e = m_anonServiceListener.elements();
								while (e.hasMoreElements())
									{
										((AnonServiceEventListener) e.nextElement()).packetMixed(a_totalBytes);
									}
							}
					}
			}

		public void dataChainErrorSignaled(AnonServiceException a_e)
			{
				LogHolder.log(LogLevel.ERR, LogType.NET, a_e);
				m_containerMixCascade.keepCurrentService(false);
				m_Anon.shutdown(false);
				synchronized (m_anonServiceListener)
					{
						Enumeration e = m_anonServiceListener.elements();
						while (e.hasMoreElements())
							{
								((AnonServiceEventListener) e.nextElement()).dataChainErrorSignaled(a_e);
							}
					}
				reconnect();
			}

		public void integrityErrorSignaled(AnonServiceException a_e)
			{
				synchronized (m_anonServiceListener)
					{
						Enumeration e = m_anonServiceListener.elements();
						while (e.hasMoreElements())
							{
								((AnonServiceEventListener) e.nextElement()).integrityErrorSignaled(a_e);
							}
					}
			}

		/**
		 * Returns {@code null} or false on all methods.
		 */
		private class DummyMixCascadeContainer extends AbstractMixCascadeContainer
			{
				public DummyMixCascadeContainer()
					{
					}

				public MixCascade getNextCascade()
					{
						return null;
					}

				public MixCascade getNextRandomCascade()
					{
						return null;
					}

				public MixCascade getCurrentCascade()
					{
						return null;
					}

				public void keepCurrentService(boolean a_bKeepCurrentCascade)
					{
					}

				public boolean isServiceAutoSwitched()
					{
						return false;
					}

				public boolean isReconnectedAutomatically()
					{
						return false;
					}

				public ITrustModel getTrustModel()
					{
						return new BasicTrustModel();
					}
			}

		/**
		 * Wraps an {@link AbstractMixCascadeContainer}. Only difference is in
		 * {@link #isReconnectedAutomatically()}.
		 */
		private class EncapsulatedMixCascadeContainer extends AbstractMixCascadeContainer
			{
				private AbstractMixCascadeContainer m_mixCascadeContainer;

				public void addObserver(Observer a_observer)
					{
						m_mixCascadeContainer.addObserver(a_observer);
					}

				public void deleteObserver(Observer a_observer)
					{
						m_mixCascadeContainer.deleteObserver(a_observer);
					}

				public void deleteObservers()
					{
						m_mixCascadeContainer.deleteObservers();
					}

				public EncapsulatedMixCascadeContainer(AbstractMixCascadeContainer a_mixCascadeContainer)
					{
						m_mixCascadeContainer = a_mixCascadeContainer;
					}

				public void reset()
					{
						m_mixCascadeContainer.reset();
					}

				public ITrustModel getTrustModel()
					{
						return m_mixCascadeContainer.getTrustModel();
					}

				public MixCascade getNextRandomCascade()
					{
						return m_mixCascadeContainer.getNextRandomCascade();
					}

				public MixCascade getNextCascade()
					{
						return m_mixCascadeContainer.getNextCascade();
					}

				public MixCascade getCurrentCascade()
					{
						return m_mixCascadeContainer.getCurrentCascade();
					}

				public void keepCurrentService(boolean a_bKeepCurrentCascade)
					{
						m_mixCascadeContainer.keepCurrentService(a_bKeepCurrentCascade);
					}

				public boolean isServiceAutoSwitched()
					{
						return m_mixCascadeContainer.isServiceAutoSwitched();
					}

				/**
				 * Returns true if the contained {@link MixCascade} reconnects
				 * automatically and does not use a forwarded connection.
				 *
				 *  @see AbstractMixCascadeContainer#isReconnectedAutomatically()
				 *  @see AnonProxy#m_forwardedConnection
				 */
				public boolean isReconnectedAutomatically()
					{
						/** @todo reconnect is not yet supported with forwarded connections */
						return !m_forwardedConnection && m_mixCascadeContainer.isReconnectedAutomatically();
					}
			}
	}
