/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.proxy;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.util.Hashtable;

import anon.AnonChannel;
import anon.NotConnectedToMixException;
import anon.TooMuchDataForPacketException;
import anon.util.SocketGuard;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;

/**
 * Reads the request over the given client socket and opens an
 * {@link AnonChannel}, corresponding to the necessary protocol, to the
 * destination using the given {@link AnonProxy} object.
 * <p>
 * This allows the request to be intercepted and processed, if necessary, before
 * sending it to the destination over the channel. Similarly the response
 * received over the channel to this object can be processed before relaying to
 * the client over the client socket.
 */
public final class AnonProxyRequest implements Runnable
	{
	  /**
	   * Number of requests currently being processed.
	   */
		private static int ms_nrOfRequests = 0;

		private static final long TIMEOUT_RECONNECT = 60000;
		private static final int CHUNK_SIZE = 1000;

		/**
		 * Continuing count of processed requests.
		 */
		private static int ms_currentRequest;

		/**
		 * The {@link AnonChannel} connecting to the destination of the request.
		 */
		private AnonChannel m_Channel;

		/**
		 * {@link InputStream} from {@link #m_Channel}.
		 * Used in the {@link Response} object to read the response to this request.
		 */
		private InputStream m_InChannel;

		/**
		 * {@link OutputStream} from {@link #m_Channel}.
		 * Used in the {@link AnonProxyRequest#run()} method to send the request to
		 * the destination, after processing.
		 */
		private OutputStream m_OutChannel;

		/**
		 * The socket connected to the Browser sending the request.
		 */
		private SocketGuard m_clientSocket;

		/**
		 * {@link InputStream} from {@link #m_clientSocket}.
		 * Used to read in the request.
		 */
		private InputStream m_InSocket;

		/**
		 * {@link OutputStream} from {@link #m_clientSocket}.
		 * Used to write the received response to the client, after processing.
		 */
		private OutputStream m_OutSocket;

		private Thread m_threadResponse;

		private Thread m_threadRequest;

		private Hashtable m_hashParsedDomains = new Hashtable();

		private boolean m_bShowBrowserWarning = false;

		private AnonProxy m_Proxy;

		private volatile boolean m_bRequestIsAlive;

		/**
		 * The protocol the request is using, gathered from the first received byte
		 * or lack thereof in the {@link #run()} method.
		 */
		private int m_iProtocol;

		/**
		 * Used when {@link #m_Proxy} needs to reconnect, blocking all other
		 * instances using the same sync object.
		 */
		private final Object m_syncObject;

		private ProxyCallbackHandler m_callbackHandler = null;

		/**
		 * Indicates that the JonDo has to perform the contentEncoding itself. May
		 * happen if, due to the header replacement, the JonDo specifies content
		 * encodings which the client does not support.
		 */
		private String[] contentEncodings;
		private boolean internalEncodingRequired = false;

		/**
		 * Creates an {@link AnonProxyRequest} object and starts it in a Thread,
		 * listening for a request to come through the given client socket.
		 *
		 * @see #run()
		 *
		 * @param proxy The proxy to communicate over.
		 * @param clientSocket The socket to communicate with the browser with.
		 * @param a_syncObject Used when the given proxy needs to reconnect to block
		 * all other requests, using that proxy. Implies a connection between the
		 * proxy and this sync object.
		 * @param callbackHandler Contains {@link ProxyCallback} objects, that have
		 * the option to process the request and response bytes, before they're sent
		 * out to the destination or client respectively.
		 *
		 * @throws IOException
		 */
		AnonProxyRequest(AnonProxy proxy, SocketGuard clientSocket, Object a_syncObject,
				ProxyCallbackHandler callbackHandler) throws IOException
			{
				m_Proxy = proxy;
				m_clientSocket = clientSocket;
				m_syncObject = a_syncObject;
				m_clientSocket.setSoTimeout(0); // just to ensure that threads will
				// stop - really no timeout ? We had 1000...
				m_InSocket = clientSocket.getInputStream();
				m_OutSocket = clientSocket.getOutputStream();
				m_threadRequest = new Thread(this, "JAP - AnonProxy Request " + Integer.toString(ms_currentRequest));
				ms_currentRequest++;
				m_callbackHandler = callbackHandler;

				m_threadRequest.setDaemon(true);
				m_threadRequest.start();
			}

		/**
		 * Get the anonymity distribution of the mix cascade used by the underlying
		 * proxy server.
		 *
		 * @see anon.infoservice.MixCascade#getDistribution()
		 *
		 * @return Distribution value from 1-6, 0 if unknown
		 */
		public int getAnonymityDistribution()
			{
				return m_Proxy.getMixCascade().getDistribution();
			}

		/**
		 * @see AnonProxyRequest#ms_nrOfRequests
		 * @return
		 */
		public static int getNrOfRequests()
			{
				return ms_nrOfRequests;
			}

		/**
		 * Reads first byte from the given client socket to recognize the kind of
		 * request. If there is nothing to read, SMTP is assumed.
		 * <p>
		 * After that a corresponding {@link AnonChannel} to the destination of the
		 * request is created to listen for a {@link Response} on.
		 * <p>
		 * Read the request in chunks and send the chunks to a {@link
		 * ProxyCallbackHandler} for intermittent processing, before sending it to
		 * the destination using the {@link AnonChannel} created earlier.
		 *
		 * @see Response
		 */
		public void run()
			{
				ms_nrOfRequests++;
				m_bRequestIsAlive = true;
				AnonChannel newChannel = null;
				boolean bIsSMTP = false;
				// Check for type
				int firstByte = 0;
				try
					{
						firstByte = m_InSocket.read();
					}
				catch (InterruptedIOException ex)
					{ // no request received so far - assume
						// SMTP, where we have to sent something
						// first
						try
							{
								bIsSMTP = true;
								newChannel = m_Proxy.createChannel(AnonChannel.SMTP);
								m_iProtocol = IProxyListener.PROTOCOL_OTHER;
								if (newChannel == null)
									{
										closeRequest();
										return;
									}
							}
						catch (Throwable to)
							{
								LogHolder.log(LogLevel.ERR, LogType.NET,
										"AnonProxyRequest - something was wrong with setting up a new SMTP channel -- Exception: " + to);
								closeRequest();
								return;
							}
					}
				catch (Throwable t)
					{
						closeRequest();
						return;
					}
				if (newChannel == null) // no SMTP - maybe HTTP or SOCKS
					{
						firstByte &= 0x00FF;
						for (; !Thread.currentThread().isInterrupted();)
							{
								try
									{
										newChannel = null;
										if (firstByte == 4 || firstByte == 5) // SOCKS
											{
												newChannel = m_Proxy.createChannel(AnonChannel.SOCKS);
												m_iProtocol = IProxyListener.PROTOCOL_OTHER;
											}
										else
											{
												newChannel = m_Proxy.createChannel(AnonChannel.HTTP);
												m_iProtocol = IProxyListener.PROTOCOL_WWW;
											}
										break;
									}
								catch (NotConnectedToMixException ec)
									{
									/*
									 * If the connection to the mix cascade was lost, try to
									 * reconnect.
									 */
										LogHolder.log(LogLevel.ERR, LogType.NET, "AnonProxyRequest - Connection to Mix lost");

										Thread timeoutThread = new Thread(new Runnable()
											{
												public void run()
													{
														m_Proxy.reconnect();
													}
											}, "Request reconnect thread");
										timeoutThread.start();
										// set timeout for this request
										long currentTime = System.currentTimeMillis();

										// wait some time for the reconnection to happen
										try
											{

												timeoutThread.join(TIMEOUT_RECONNECT);
											}
										catch (InterruptedException ex1)
											{
												Thread.currentThread().interrupt();
											}

										boolean success = true;
										synchronized (m_syncObject) // synchronize with connection
																								// thread
											{
												if (!m_Proxy.isConnected() && !Thread.currentThread().isInterrupted())
													{
														long remainder = currentTime + TIMEOUT_RECONNECT - System.currentTimeMillis();
														if (remainder > 0)
															{
																try
																	{
																		m_syncObject.wait(remainder);
																	}
																catch (InterruptedException ex2)
																	{
																		Thread.currentThread().interrupt();
																		break;
																	}
															}
													}

												if (!m_Proxy.isConnected())
													{
														success = false;
													}
											}

										if (!success)
											{
												// reconnect failed
												LogHolder.log(LogLevel.ERR, LogType.NET,
														"Requests terminated due to loss of connection to service!");
												closeRequest();
												return;
											}
									}
								catch (Exception e)
									{
										LogHolder.log(LogLevel.ERR, LogType.NET,
												"AnonProxyRequest - something was wrong with seting up a new channel. Exception: " + e);
										closeRequest();
										return;
									}
							}
						if (newChannel == null)
							{
								closeRequest();
								return;
							}
					}

				// channel was successfully created
				int len = 0;
				int aktPos = 0;
				if (!bIsSMTP) // only SOCKS and HTTP will read the first byte - but not
											// SMTP!
					{
						aktPos = 1;
					}

				byte[] buff = null;
				ProxyCallbackBuffer pcBuffer = null;
				try
					{
						m_Channel = newChannel;
						m_InChannel = newChannel.getInputStream();
						m_OutChannel = newChannel.getOutputStream();

						m_threadResponse = new Thread(new Response(), "JAP - AnonProxy Response for "
								+ Thread.currentThread().getName());
						m_threadResponse.start();

						buff = new byte[1900];
						buff[0] = (byte) firstByte;
					}
				catch (Throwable t)
					{
						closeRequest();
						return;
					}

				m_Proxy.incNumChannels();

				/*
				 * Read the request from the client socket in chunks and send it to the
				 * ProxyCallbackHandler if there.
				 */
				try
					{
						for (;;)
							{
								try
									{
									  //XXX Why len so many times?

										// determine the length to read minus what has been read
										len = Math.min(m_Channel.getOutputBlockSize(), 1900);
										len -= aktPos; // # of bytes to read

										// read up to len bytes
										len = m_InSocket.read(buff, aktPos, len);
										len += aktPos; // bytes read + offset
									}
								catch (InterruptedIOException ioe)
									{
										aktPos = aktPos + ioe.bytesTransferred;
										continue;
									}
								if (len < 0)
									{
										break;
									}
								try
									{
										if (m_callbackHandler != null && len > 0)
											{
												pcBuffer = new ProxyCallbackBuffer(buff, 0, len);
												try
													{
														m_callbackHandler.deliverUpstream(this, pcBuffer);
													}
												catch (ProxyCallbackDelayException e)
													{
													  // ProxyCallback needs more data to continue
														aktPos = 0;
														continue;
													}
												m_OutChannel.write(pcBuffer.getChunk(), 0, pcBuffer.getPayloadLength());
											}
										else
											{
												m_OutChannel.write(buff, 0, len);
											}
										/* everything was OK */
										aktPos = 0;
									}
								catch (TooMuchDataForPacketException e)
									{
										/*
										 * The implementation could not send all bytes. Reason:
										 * AnonChannel.getOutputBlockSize() is sometimes only a hint
										 * about the estimated size, the available size can be
										 * smaller (also some implementations need to know, whether
										 * there are more bytes available -> they request more bytes
										 * than they can send to find it out). Solution: Send the
										 * remaining bytes again in the next write-call.
										 */
										if (m_callbackHandler != null)
											{
												/*
												 * callback processing has most likely caused this
												 * exception and needs another handling.
												 */
												sendRemainingBytesRecursion(pcBuffer, e.getBytesSent(), m_OutChannel);
												// Postcondition: aktPos is always 0, because we sent
												// the whole chunk away;
												aktPos = 0;
											}
										else
											{
												byte[] tempBuff = new byte[buff.length - e.getBytesSent()];
												System.arraycopy(buff, e.getBytesSent(), tempBuff, 0, tempBuff.length);
												System.arraycopy(tempBuff, 0, buff, 0, tempBuff.length);
												aktPos = tempBuff.length;
											}
									}
								// LogHolder.log(LogLevel.DEBUG,LogType.NET,"Channel
								// "+Integer.toString(m_Channel.hashCode())+" Request Len:
								// "+re+" Read:
								// "+len);
								m_Proxy.transferredBytes(len - aktPos, m_iProtocol);
								Thread.yield();
							}
					}
				catch (IOException e)
					{
						LogHolder.log(LogLevel.DEBUG, LogType.NET, "Exception in AnonProxyRequest - upstream loop.", e);
					}
				catch (ProxyCallbackNotProcessableException cnpe)
					{
						try
							{
								m_OutSocket.write(cnpe.getErrorResponse());
							}
						catch (IOException e)
							{
							}
						LogHolder.log(LogLevel.ERR, LogType.NET, "chunk could not be processed. Terminating", cnpe);
					}
				closeRequest();
				m_Proxy.decNumChannels();
			}

		private static void sendRemainingBytesRecursion(ProxyCallbackBuffer pcBuffer, int sentBytes,
				OutputStream outputStream) throws IOException
			{
				byte[] tempBuff = new byte[pcBuffer.getPayloadLength() - sentBytes];
				System.arraycopy(pcBuffer.getChunk(), sentBytes, tempBuff, 0, tempBuff.length);
				System.arraycopy(tempBuff, 0, pcBuffer.getChunk(), 0, tempBuff.length);
				try
					{
						outputStream.write(tempBuff);
					}
				catch (TooMuchDataForPacketException e)
					{
						pcBuffer.setChunk(tempBuff);
						sendRemainingBytesRecursion(pcBuffer, e.getBytesSent(), outputStream);
					}
			}

		private synchronized void closeRequest()
			{
				if (m_bRequestIsAlive)
					{
						ms_nrOfRequests--;
						m_bRequestIsAlive = false;
					}
				try
					{
						if (m_Channel != null)
							{
								m_Channel.close();
							}
					}
				catch (Throwable t)
					{
					}
				try
					{
						m_InSocket.close();
					}
				catch (Throwable t)
					{
					}
				try
					{
						m_OutSocket.close();
					}
				catch (Throwable t)
					{
					}
				try
					{
						m_clientSocket.close();
					}
				catch (Throwable t)
					{
					}
				if (m_callbackHandler != null)
					{
						m_callbackHandler.closeRequest(this);
					}
			}

		protected void setHttpParsed(String a_strDomain)
			{
				if (a_strDomain != null && !isProxyKeepAliveEnabled())
					{
						m_hashParsedDomains.put(a_strDomain, a_strDomain);
					}
			}

		protected boolean isProxyKeepAliveEnabled()
			{
				return m_hashParsedDomains.size() > 0;
				/*
				 * return m_hashParsedDomains.size() > 1 || (a_strDomain != null &&
				 * m_hashParsedDomains.size() > 0 &&
				 * !m_hashParsedDomains.containsKey(a_strDomain));
				 */
			}

		public boolean isInternalEncodingRequired()
			{
				return internalEncodingRequired;
			}

		public void showBrowserWarning(boolean a_bShow)
			{
				m_bShowBrowserWarning = a_bShow;
			}

		public boolean isBrowserWarningShown()
			{
				return m_bShowBrowserWarning;
			}

		protected void setInternalEncodingRequired(boolean internalEncodingRequired)
			{
				this.internalEncodingRequired = internalEncodingRequired;
			}

		protected String[] getContentEncodings()
			{
				return contentEncodings;
			}

		protected void setContentEncodings(String[] contentEncodings)
			{
				this.contentEncodings = contentEncodings;
			}


		/**
		 * This object reads a response from an {@link AnonChannel} and sends it to
		 * the {@link ProxyCallbackHandler} for processing, before writing it to the
		 * client socket given in the constructor of the {@link AnonProxyRequest}
		 * owning this Response object.
		 *
		 */
		final class Response implements Runnable
			{
				Response()
					{
					}

				public void run()
					{
						int len = 0;
						byte[] buff = new byte[2900];

						try
							{
								ProxyCallbackBuffer pcBuffer = null;
								mainLoop: do
									{
										len = m_InChannel.read(buff, 0, CHUNK_SIZE);
										if (len < 0)
											{
												break;
											}
										int count = 0;
										for (;;)
											{
												try
													{
														// Are there ProxyCallback objects to handle first?
														if (m_callbackHandler != null && len > 0)
															{
																// strange looking but tests if a
																// a proxy callback correctly
																// processes chunk offsets
																// byte[] nchunk = new byte[len+10];
																// for (int i = 0; i < 10; i++)
																// {
																// nchunk[i] = (byte)'e';
																// }
																// System.arraycopy(buff, 0, nchunk, 10, len);
																// pcBuffer = new ProxyCallbackBuffer(nchunk,
																// 10, len+10);

																pcBuffer = new ProxyCallbackBuffer(buff, 0, len);
																try
																	{
																		m_callbackHandler.deliverDownstream(AnonProxyRequest.this, pcBuffer);
																	}
																catch (ProxyCallbackDelayException e)
																	{
																		continue mainLoop;
																	}

																// write out processed bytes
																m_OutSocket.write(pcBuffer.getChunk(), 0, pcBuffer.getPayloadLength());

																if (pcBuffer.getStatus() == ProxyCallback.STATUS_FINISHED)
																	{
																		break mainLoop;
																	}
																// belongs to the above mentioned offset test
																// byte[] nnchunk = new
																// byte[pcBuffer.getPayloadLength()-10];
																// System.arraycopy(pcBuffer.getChunk(), 10,
																// nnchunk, 0, pcBuffer.getPayloadLength()-10);
																// m_OutSocket.write(nnchunk, 0,
																// pcBuffer.getPayloadLength()-10);
															}
														else
															{
															  // no ProxyCallbacks just write out the response
																m_OutSocket.write(buff, 0, len);
															}

														m_OutSocket.flush();
														break;
													}
												catch (InterruptedIOException ioe)
													{
														LogHolder.log(LogLevel.EMERG, LogType.NET,
																"Should never be here: Timeout in sending to Browser!");
													}

												count++;
												if (count > 3)
													{
														// timed out 4 times, giving up
														throw new IOException("Could not send to Browser...");
													}
											}

										// ProxyCallback was not finished, but able to write out bytes
										m_Proxy.transferredBytes(len, m_iProtocol);
										Thread.yield();
									}
								while (len >= 0 && !m_Channel.isClosed());
							}
						catch (IOException e)
							{
								if (!m_Proxy.isConnected() && !m_Proxy.isConnecting())
									{
										LogHolder.log(LogLevel.ERR, LogType.NET, e);
									}
								else
									{
										LogHolder.log(LogLevel.INFO, LogType.NET, e);
									}
							}
						catch (ProxyCallbackNotProcessableException cnpe)
							{
								LogHolder.log(LogLevel.ERR, LogType.NET, cnpe);
								try
									{
										m_OutSocket.write(cnpe.getErrorResponse());
									}
								catch (IOException e)
									{
									}
							}
						try
							{
								m_clientSocket.close();
							}
						catch (IOException e)
							{
							}
						try
							{
								Thread.sleep(500);
							}
						catch (InterruptedException e)
							{
							}
						if (m_bRequestIsAlive)
							{
								m_threadRequest.interrupt();
							}
						buff = null;
					}
			}
	}
