/*
 Copyright (c) 2000, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
  may be used to endorse or promote products derived from this software without specific
  prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.proxy;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.PushbackInputStream;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.Observable;
import java.util.Observer;
import java.util.StringTokenizer;
import java.util.TimeZone;
import java.util.Hashtable;
import java.util.Vector;

import anon.infoservice.ProxyInterface;
import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;
import anon.infoservice.ImmutableProxyInterface;
import anon.infoservice.IMutableProxyInterface;
import anon.util.BooleanVariable;
import anon.util.JAPMessages;
import anon.util.SocketGuard;

/**
 * This class describes an HTTP proxy, when anonymity has been deactivated.
 * It is the pendant to the AnonProxy class.
 */
final public class DirectProxy
{
	private static final String GENERAL_RULE = "*";
	private static final int BUFFER_PUSHBACK = 2000;

	private static final String JONDOPROXY = "JonDoProxy";
	private static final String JONDOPROXY_ARG = JONDOPROXY.toLowerCase() + "=";

	private static final long TEMPORARY_REMEMBER_TIME_SECONDS = 5000l;
	/*
	private static final long TEMPORARY_REMEMBER_TIME <= 60000l;
	private static final long TEMPORARY_REMEMBER_TIME_NO_WARNING = 300000l;
	*/
	private static final long TEMPORARY_REMEMBER_TIME = 10000l;
	private static final long TEMPORARY_REMEMBER_TIME_NO_WARNING = 10000l;

	private AllowProxyConnectionCallback m_callback;
	private IMutableProxyInterface m_proxyInterface =
		new IMutableProxyInterface.DummyMutableProxyInterface();

	private ServerSocket m_socketListener;
	private ServerSocket m_socketListenerTwo;
	private final Object THREAD_SYNC = new Object();
	private boolean m_bInterrupted = false;

	private volatile Thread threadRunLoop;
	private BooleanVariable m_bIsRunningOne = new BooleanVariable(false);
	private volatile Thread threadRunLoopTwo;
	private BooleanVariable m_bIsRunningTwo = new BooleanVariable(false);

	private final Hashtable rememberedDomains = new Hashtable();
	private final Object SYNC_CALLBACK = new Object();

	private final Vector m_vecThreads = new Vector();

	public DirectProxy(ServerSocket a_socket, ServerSocket a_socketTwo, IMutableProxyInterface a_proxyInterface,
			AllowProxyConnectionCallback a_callback)
	{
		m_socketListener = a_socket;
		m_socketListenerTwo = a_socketTwo;
		if (a_proxyInterface != null)
		{
			m_proxyInterface = a_proxyInterface;
		}
		setAllowUnprotectedConnectionCallback(a_callback);
	}

	private synchronized void setAllowUnprotectedConnectionCallback(
			AllowProxyConnectionCallback a_callback)
	{
		synchronized (rememberedDomains)
		{
			if (m_callback != a_callback)
			{
				m_callback = a_callback;
				rememberedDomains.clear();
				if (m_callback != null)
				{
					m_callback.setRulesChanged(false);
					rememberedDomains.put(GENERAL_RULE, new RememberedRequestRight(GENERAL_RULE,
							true, System.currentTimeMillis() + TEMPORARY_REMEMBER_TIME_SECONDS));
				}
			}
		}
	}

	/**
	 * This class holds the URI, method (HTTP or "CONNECT"), port and a
	 * JonDoProxy argument, that were read from a Request received from the
	 * client.
	 */
	public static class RequestInfo
	{
		private String m_strURI;
		private String m_strMethod;
		private int m_port;
		private String m_jdpArg;

		protected RequestInfo(String a_strURI, String a_strMethod, int a_port, String a_jdpArg)
		{
			m_strURI = a_strURI;
			m_strMethod = a_strMethod;
			m_port = a_port;
			m_jdpArg = a_jdpArg;
		}

		public String getJonDoProxyArgument()
		{
			return m_jdpArg;
		}

		public String getHost()
		{
			return m_strURI;
		}

		public String getMethod()
		{
			return m_strMethod;
		}

		public int getPort()
		{
			return m_port;
		}
	}


	public static abstract class AllowProxyConnectionCallback implements Observer
	{
		public static final String RULES_CHANGED_OBSERVABLE_NOTIFIER = "RulesChanged";
		
		private boolean m_bRulesChanged = false;
		
		public static class Answer
		{
			private boolean m_bRemembered;
			private boolean m_bAllow;
			private boolean m_bTimeout;

			public Answer(boolean a_bAllow, boolean a_bRemembered, boolean a_bTimeout)
			{
				m_bAllow = a_bAllow;
				m_bRemembered = a_bRemembered;
				m_bTimeout = a_bTimeout;
			}

			public boolean isRemembered()
			{
				return m_bRemembered;
			}

			public boolean isTimedOut()
			{
				return m_bTimeout;
			}

			public boolean isAllowed()
			{
				return m_bAllow;
			}
		}

		public abstract boolean isConnecting();

		public boolean haveRulesChanged()
		{
			return m_bRulesChanged;
		}

		public URL getHTMLHelpPath()
		{
			return null;
		}

		private void setRulesChanged(boolean a_bChanged)
		{
			m_bRulesChanged = a_bChanged;
		}

		public void update(Observable o, Object arg)
		{
			if (arg != null && arg == RULES_CHANGED_OBSERVABLE_NOTIFIER)
			{
				m_bRulesChanged = true;
			}
		}

		public abstract boolean isNonAnonymousAccessForbidden();

		public abstract String getApplicationName();

		public abstract String getAllowNonAnonymousSettingName();

		public abstract boolean isAskedForAnyNonAnonymousRequest();

		public abstract Answer callback(RequestInfo a_requestInfo);
	}


	public boolean isConnected()
	{
		synchronized (THREAD_SYNC)
		{
			return threadRunLoop != null;
		}
	}

	public ServerSocket getSocketListener()
	{
		return m_socketListener;
	}

	public ServerSocket getSocketListenerTwo()
	{
		return m_socketListenerTwo;
	}

	public void reset()
	{
		AllowProxyConnectionCallback callback = m_callback;
		if (callback != null)
		{
			callback.setRulesChanged(false);
		}
		synchronized (rememberedDomains)
		{
			rememberedDomains.clear();
			rememberedDomains.put(GENERAL_RULE, new RememberedRequestRight(GENERAL_RULE,
					true, System.currentTimeMillis() + TEMPORARY_REMEMBER_TIME_SECONDS, false));
		}
	}

	public synchronized boolean start()
	{
		if (m_socketListener == null)
		{
			return false;
		}

		synchronized (THREAD_SYNC)
		{
			// exit if already running
			if (threadRunLoop != null && threadRunLoop.isAlive())
			{
				return true;
			}

			stop();

			// reset domains
			reset();

			m_bInterrupted = false;

			LogHolder.log(LogLevel.WARNING, LogType.NET, "Starting up direct proxy...");

			m_bIsRunningOne.set(true);
			threadRunLoop = new Thread(new DirectProxyRunnable(m_socketListener, m_bIsRunningOne), "JAP - Direct Proxy");
			threadRunLoop.setDaemon(true);
			threadRunLoop.start();

			if (m_socketListenerTwo != null)
			{
				m_bIsRunningTwo.set(true);
				threadRunLoopTwo = new Thread(new DirectProxyRunnable(m_socketListenerTwo, m_bIsRunningTwo), "JAP - Direct Proxy");
				threadRunLoopTwo.setDaemon(true);
				threadRunLoopTwo.start();
			}

			return true;
		}
	}

	/**
	 * Returns the stored {@link RememberedRequestRight} object for the URI
	 * described by the given {@link RequestInfo} object.
	 *
	 * @param requestInfo RequestInfo object describing a URI
	 * @return The associated RememberedRequestRight, the general rule if not
	 * found or null if the found Request timed out already.
	 */
	private RememberedRequestRight getRequestRight(RequestInfo requestInfo)
	{
		if (requestInfo == null)
		{
			return null;
		}
		synchronized (rememberedDomains)
		{
			RememberedRequestRight requestRight =
				(RememberedRequestRight)rememberedDomains.get(requestInfo.getHost());

			/*
			 * No rule for the URI in the request info was found. Check for a
			 * general rule affecting all URIs.
			 */
			if (requestRight == null)
			{
				requestRight = (RememberedRequestRight)rememberedDomains.get(GENERAL_RULE); // if a rule for all requests is set
			}

			/*
			 * If a rule was found, but it has timed out already remove that
			 * rule from the list and return null.
			 */
			if (requestRight != null && requestRight.isTimedOut())
			{
				rememberedDomains.remove(requestRight.getURI());
				requestRight = null;
			}
			return requestRight;
		}
	}

	/**
	 *
	 */
	private class ConnectionHandler extends Thread
	{
		private PushbackInputStream clientInputStream;
		private Socket m_clientSocket;

		public ConnectionHandler(Socket a_clientSocket)
		{
			m_clientSocket = a_clientSocket;
		}


		public synchronized void close()
		{
			try
			{
				SocketGuard.close(m_clientSocket);
			}
			catch (IOException e)
			{
				// ignore
			}


			try
			{
				if (clientInputStream != null)
				{
					clientInputStream.close();
				}
			}
			catch (IOException e)
			{
				// ignore
			}
		}


		/**
		 *
		 */
		protected void handleConnection()
		{
			RequestInfo requestInfo;
			RememberedRequestRight requestRight;

			// check if the client socket is set
			if (m_clientSocket == null)
			{
				return;
			}

			// try to get the input stream for the client socket
			try
			{
				clientInputStream = new PushbackInputStream(m_clientSocket.getInputStream(), BUFFER_PUSHBACK);
			}
			catch (IOException ex)
			{
				LogHolder.log(LogLevel.ERR, LogType.NET, "Could not handle HTTP connection!", ex);
				close();
				return;
			}

			// check if the client socket was closed due to error above
			synchronized (this)
			{
				if (SocketGuard.isClosed(m_clientSocket))
				{
					LogHolder.log(LogLevel.ERR, LogType.NET,
							"Could not handle HTTP connection! Socket is closed.");
					close();
					return;
				}
			}

			/*
			 * if no callback is set or the rules have previously changed
			 * discard all remembered domains
			 */
			synchronized (rememberedDomains)
			{
				if (m_callback == null || m_callback.haveRulesChanged())
				{
					rememberedDomains.clear();
					if (m_callback != null)
					{
						m_callback.setRulesChanged(false);
					}
				}

				requestInfo = DirectProxyConnection.getURI(clientInputStream, BUFFER_PUSHBACK);
				requestRight = getRequestRight(requestInfo);
			}

			if (requestInfo == null)
			{
				try
				{
					SocketGuard.close(m_clientSocket);
				}
				catch (IOException e)
				{
					LogHolder.log(LogLevel.ERR, LogType.NET, e);
				}

				try
				{
					clientInputStream.close();
				}
				catch (IOException e)
				{
					LogHolder.log(LogLevel.ERR, LogType.NET, e);
				}

				return;
			}

			if (requestInfo != null && requestRight == null)
			{
				// only one callback is allowed at a time...
				synchronized (SYNC_CALLBACK)
				{
					// maybe another thread/callback has set the right meanwhile...
					requestRight = getRequestRight(requestInfo);

					if (requestRight == null)
					{
						AllowProxyConnectionCallback.Answer answer;
						if (m_callback != null)
						{
							answer = m_callback.callback(requestInfo);
						}
						else
						{
							answer = new AllowProxyConnectionCallback.Answer(false, false, false);
						}

						if (answer != null)
						{
							long lTimeout;

							if (!answer.isTimedOut())
							{
								lTimeout = Long.MAX_VALUE;
							}
							else if (answer.isAllowed())
							{
								lTimeout = System.currentTimeMillis() + TEMPORARY_REMEMBER_TIME_NO_WARNING;
							}
							else
							{
								lTimeout = System.currentTimeMillis() + TEMPORARY_REMEMBER_TIME;
							}

							synchronized (rememberedDomains)
							{
								if (m_callback == null || m_callback.haveRulesChanged())
								{
									rememberedDomains.clear();
									if (m_callback != null)
									{
										m_callback.setRulesChanged(false);
									}
								}

								if (answer.isRemembered())
								{
									if (answer.isTimedOut() && m_callback != null &&
										m_callback.isAskedForAnyNonAnonymousRequest())
									{
										lTimeout =
											System.currentTimeMillis() + TEMPORARY_REMEMBER_TIME_SECONDS;
									}

									requestRight = new RememberedRequestRight(GENERAL_RULE,
											!answer.isAllowed(), lTimeout);
								}
								else
								{
									requestRight = new RememberedRequestRight(requestInfo.getHost(),
											!answer.isAllowed(), lTimeout);

								}
								rememberedDomains.put(requestRight.getURI(), requestRight);
							}
						}
					}
				}
			}

			if (requestRight != null && !requestRight.isWarningShown())
			{
				if (getProxyInterface() != null && getProxyInterface().isValid() &&
					getProxyInterface().getProtocol() == ProxyInterface.PROTOCOL_TYPE_HTTP)
				{
					new DirectConViaHTTPProxy(m_clientSocket, clientInputStream).run();
				}
				else
				{
					new DirectProxyConnection(m_clientSocket, clientInputStream, DirectProxy.this);
				}
			}
			else
			{
				new SendAnonWarning(m_clientSocket, clientInputStream,
						requestRight, DirectProxy.this, requestInfo).run();
			}
		}
	}

	/**
	 * This Runnable continuously accepts connections on the given
	 * {@link ServerSocket}.
	 */
	private class DirectProxyRunnable implements Runnable
	{
		private ServerSocket m_serverSocket;
		private BooleanVariable m_bRunning;

		/**
		 * Sets up a Runnable to accept connections on the given {@link
		 * ServerSocket}.
		 * @param a_serverSocket Socket to accept connections with.
		 * @param a_bRunning XXX unnecessary, because it's never read, only set?
		 */
		public DirectProxyRunnable(ServerSocket a_serverSocket, BooleanVariable a_bRunning)
		{
			m_serverSocket = a_serverSocket;
			m_bRunning = a_bRunning;
		}
		
	public void run()
	{
		try
		{
			m_serverSocket.setSoTimeout(2000);
		}
		catch (Exception e1)
		{
			LogHolder.log(LogLevel.DEBUG, LogType.NET, "Could not set accept time out!", e1);
		}

		while (!m_bInterrupted && !Thread.currentThread().isInterrupted())
		{
			Socket socket = null;
			try
			{
				socket = m_serverSocket.accept();
			}
			catch (InterruptedIOException e1)
			{
				Thread.yield();
				continue;
			}
			catch (SocketException e2)
			{
				LogHolder.log(LogLevel.ERR, LogType.NET, "Accept socket exception: " + e2);
				break;
			}
			catch (IOException a_e)
			{
				LogHolder.log(LogLevel.EXCEPTION, LogType.NET, "Socket could not accept!" + a_e);
				break;
			}

			/*
			 * If the Runnable was interrupted during accepting a connection
			 * close the socket and stop accepting.
			 */
			if (m_bInterrupted  || Thread.currentThread().isInterrupted())
			{
				try
				{
					SocketGuard.close(socket);
				}
				catch (IOException a_e)
				{
					LogHolder.log(LogLevel.ERR, LogType.NET, a_e);
				}

				break;
			}

			/*
			 * Set the timeout to 0ms, so I/O operations on this socket will
			 * block from now on.
			 *
			 * If this fails close the socket and start accepting again.
			 */
			try
			{
				socket.setSoTimeout(0);
			}
			catch (SocketException soex)
			{
				LogHolder.log(LogLevel.ERR, LogType.NET,
							  "Could not set socket to blocking mode! Exception: " + soex);
				try
				{
					SocketGuard.close(socket);
				}
				catch (IOException a_e)
				{
					LogHolder.log(LogLevel.ERR, LogType.NET, a_e);
				}
				socket = null;
				continue;
			}

			/*
			 * If the Runnable was interrupted during accepting a connection
			 * close the socket and stop accepting.
			 */
			if (m_bInterrupted  || Thread.currentThread().isInterrupted())
			{
				try
				{
					SocketGuard.close(socket);
				}
				catch (IOException a_e)
				{
					LogHolder.log(LogLevel.ERR, LogType.NET, a_e);
				}

				break;
			}

			/*
			 * Accepting and changing timeout successful.
			 */
			ConnectionHandler connectionHandler = new ConnectionHandler(socket)
			{
				public void run()
				{
					this.handleConnection();
					synchronized(m_vecThreads)
					{
						m_vecThreads.removeElement(Thread.currentThread());
						m_vecThreads.notifyAll();
					}
				}
			};
		
			m_vecThreads.addElement(connectionHandler);
			connectionHandler.start();
		}
		
		m_bRunning.set(false);
		LogHolder.log(LogLevel.NOTICE, LogType.NET, "Accept was stopped.");
		
		Vector vecThreads = (Vector)m_vecThreads.clone();
		
		if (vecThreads.size() > 0)
		{
			LogHolder.log(LogLevel.WARNING, LogType.NET, "Closing remaining threads...");
		}
		
		ConnectionHandler thread;
		
		while (vecThreads.size() > 0)
		{
			LogHolder.log(LogLevel.WARNING, LogType.NET, "Closing thread...");
			
			thread = (ConnectionHandler)vecThreads.lastElement();
			while (thread.isAlive())
			{
				LogHolder.log(LogLevel.WARNING, LogType.NET, "Closing last thread from a total of " + vecThreads.size());
				
				thread.interrupt();
				thread.close();
				
				Thread.yield();
				LogHolder.log(LogLevel.INFO, LogType.NET, "Before sync...");
				synchronized(m_vecThreads)
				{
					if (thread.isAlive())
					{
						try 
						{
							m_vecThreads.wait(250);
						} 
						catch (InterruptedException e)
						{
							// ignore
						}
					}
				}
			}
			vecThreads.removeElement(thread);
			LogHolder.log(LogLevel.WARNING, LogType.NET, "Thread closed. " + vecThreads.size() + " threads left.");
		}
		
		if (m_vecThreads.size() > 0)
		{
			LogHolder.log(LogLevel.ALERT, LogType.NET, 
					"Possible insecurity or memory leak: Direct Proxy Server has " + m_vecThreads.size() + " remaining threads!!");
		}
		
		
		LogHolder.log(LogLevel.INFO, LogType.NET, "Direct Proxy Server stopped.");
	}
	}
	
	public boolean allowDomain(URL a_url)
	{
		if (a_url == null)
		{
			return false;
		}
		RequestInfo info = parseDomain(a_url.toString(), true, null);
		
		
		if (info == null)
		{
			return false;
		}
		
		RememberedRequestRight requestRight = 
			new RememberedRequestRight(info.getHost(), 
				false, Long.MAX_VALUE);

		rememberedDomains.put(requestRight.getURI(), requestRight);
		return true;
	}

	private static ServerSocket closeThread(Thread a_thread, ServerSocket a_socket,
			BooleanVariable a_bRunning)
	{
		if (a_thread == null)
		{
			return a_socket;
		}
		
		int i = 0;
		while (a_thread.isAlive())
		{
			if (a_bRunning.isTrue())
			{
				a_thread.interrupt();
			}
			Thread.yield();
			try
			{
				a_thread.join(250);
			}
			catch (InterruptedException e)
			{
				//LogHolder.log(LogLevel.ERR, LogType.NET, "Direct Proxy Server could not be stopped!!!");
			}
			
			if (i > 10)
			{
				LogHolder.log(LogLevel.ALERT, LogType.NET, "Shutting down direct proxy (" +
						(a_bRunning.isTrue() ? "Running" : "")
						+ ") at " 
						+ a_socket.getInetAddress().getHostAddress() + ":" + a_socket.getLocalPort() +
						"...");
				/*
				try
				{
					threadRunLoop.stop();
				}
				catch (Exception e)
				{
				}
				*/
			}
			
			i++;
		}
		
		return a_socket;
	}
	
	public synchronized void stop()
	{
		synchronized (THREAD_SYNC)
		{
			if (threadRunLoop == null)
			{
				return;
			}
			
			LogHolder.log(LogLevel.WARNING, LogType.NET, "Shutting down direct proxy...");
			
			m_bInterrupted = true;
			
			m_socketListener = closeThread(threadRunLoop, m_socketListener, m_bIsRunningOne);
			m_socketListenerTwo = closeThread(threadRunLoopTwo, m_socketListenerTwo, m_bIsRunningTwo);
			
			threadRunLoop = null;
			threadRunLoopTwo = null;
		}
	}

	protected ImmutableProxyInterface getProxyInterface()
	{
		return m_proxyInterface.getProxyInterface(false).getProxyInterface();
	}

	/**
	 *  This class is used to inform the user that he tries to
	 *  send requests although anonymity mode is off.
	 */
	private static final class SendAnonWarning implements Runnable
	{
		private static final String MSG_BLOCKED = 
			DirectProxy.class.getName() + ".blocked";
		private static final String MSG_BLOCKED_ALL = 
			DirectProxy.class.getName() + ".blockedAll";
		private static final String MSG_BLOCKED_DOMAIN = 
			DirectProxy.class.getName() + ".blockedDomain";
		private static final String MSG_COUNTDOWN = 
			DirectProxy.class.getName() + ".countdown";	
		private static final String MSG_RELOAD = 
			DirectProxy.class.getName() + ".reload";	
		private static final String MSG_BLOCKED_PERMANENTLY = 
			DirectProxy.class.getName() + ".blockedPermanently";	
		private static final String MSG_ANON_MODE_OFF = 
			DirectProxy.class.getName() + ".htmlAnonModeOff";
		private static final String MSG_ANON_MODE_OFF_BUT_FORCED = 
			DirectProxy.class.getName() + ".htmlAnonModeOffButForced";
		private static final String MSG_CONNECTING = 
			DirectProxy.class.getName() + ".connecting";
		private static final String MSG_ANONYMITY_MODE = 
			DirectProxy.class.getName() + ".anonymityModeOff";
		private static final String MSG_WAIT_FOR_CONNECTION = 
			DirectProxy.class.getName() + ".waitForConnection";
		private static final String MSG_WAIT_FOR_CONNECTION_2 = 
			DirectProxy.class.getName() + ".waitForConnection2";
		private static final String MSG_HTML_ANON_MODE_SWITCH = 
			DirectProxy.class.getName() + ".htmlAnonModeSwitch";
		
		
		
		
		private Socket socket;
		private SimpleDateFormat dateFormatHTTP;
		private InputStream m_clientInputStream;
		private RememberedRequestRight m_requestRight;
		private DirectProxy m_directProxy;
		private RequestInfo m_info;

		public SendAnonWarning(Socket s, InputStream a_clientInputStream, 
				RememberedRequestRight a_requestRight, DirectProxy a_directProxy, RequestInfo a_info)
		{
			this.socket = s;
			m_info = a_info;
			m_directProxy = a_directProxy;
			m_requestRight = a_requestRight;
			m_clientInputStream = a_clientInputStream;
			dateFormatHTTP = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz", Locale.US);
			dateFormatHTTP.setTimeZone(TimeZone.getTimeZone("GMT"));
		}

		public void run()
		{
			long countDown;
			String[] addedMessage;
			String blockedMessage;
			try
			{
				// read something so that the browser realizes everything is OK
				if (m_clientInputStream != null)
				{
					m_clientInputStream.read();
				}
				else
				{
					socket.getInputStream().read();
				}
			}
			catch (IOException a_e)
			{
				// ignored
			}

			try
			{
				AllowProxyConnectionCallback callback = m_directProxy.m_callback;
				String applicationName = null;
				
				applicationName = (callback != null ? callback.getApplicationName() : "JonDo");
				if (applicationName == null)
				{
					applicationName = "JonDo";
				}
				
				String date = dateFormatHTTP.format(new Date());
				BufferedWriter toClient = new BufferedWriter(new OutputStreamWriter(socket.getOutputStream()));
				toClient.write("HTTP/1.1 202 Accepted\r\n");
				toClient.write("Content-type: text/html; charset=UTF-8\r\n");
				toClient.write("Expires: " + date + "\r\n");
				toClient.write("Date: " + date + "\r\n");
				toClient.write("Pragma: no-cache\r\n");
				toClient.write("Cache-Control: no-cache\r\n\r\n");
				toClient.write("<HTML>\n<HEAD>\n<TITLE>" + applicationName + "</TITLE>\n"+
						"<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />\n" +
						(m_info != null && m_info.getJonDoProxyArgument() != null ? 
						"<meta http-equiv=\"" + JONDOPROXY + "\" content=\"" + m_info.getJonDoProxyArgument() + "\" />\n" : "") + 
						"</HEAD>\n<BODY>\n");
				toClient.write("<PRE>" + date + "</PRE>\n");
				toClient.flush();
				
				
				String strAnonymityState;
				boolean bConnecting = (callback != null && callback.isConnecting());
					
				if (bConnecting)
				{
					strAnonymityState = JAPMessages.getString(MSG_CONNECTING).toUpperCase();
				}
				else
				{
					strAnonymityState = JAPMessages.getString(MSG_ANONYMITY_MODE) +
					" " + JAPMessages.getString("ngAnonOff").toUpperCase();
				}
				
				strAnonymityState = applicationName + ": " + strAnonymityState;
				strAnonymityState = "<CENTER><H1>" + strAnonymityState + "</H1>";
				
				if (m_requestRight == null || 
					callback == null || 
					callback.isNonAnonymousAccessForbidden())
				{	
					if (callback != null && callback.isConnecting())
					{
						if (callback.isNonAnonymousAccessForbidden() && 
							callback.getAllowNonAnonymousSettingName() != null)
						{
							strAnonymityState += JAPMessages.getString(
									MSG_ANON_MODE_OFF, new String[]{
											JAPMessages.getString(callback.getAllowNonAnonymousSettingName()),
											"<BR>" + JAPMessages.getString(MSG_WAIT_FOR_CONNECTION), JONDOPROXY});
						}
						else
						{
							strAnonymityState += JAPMessages.getString(MSG_WAIT_FOR_CONNECTION_2);
						}
					}
					else if (callback != null && callback.getAllowNonAnonymousSettingName() != null &&
							callback.isNonAnonymousAccessForbidden())
					{
						strAnonymityState += JAPMessages.getString(MSG_ANON_MODE_OFF_BUT_FORCED, new String[]{
								JAPMessages.getString("ngAnonOn")}) + "<br><br>" + 
								JAPMessages.getString(MSG_ANON_MODE_OFF, new String[]{
										JAPMessages.getString(callback.getAllowNonAnonymousSettingName()),
										"", JONDOPROXY});
					}
					else
					{
						strAnonymityState += JAPMessages.getString(MSG_ANON_MODE_OFF_BUT_FORCED, new String[]{
								JAPMessages.getString("ngAnonOn")});
					}
					
					byte[] byWrite = (strAnonymityState + "</CENTER>").getBytes("UTF-8");
					
					socket.getOutputStream().write(byWrite, 0, byWrite.length);
					
					//toClient.write(strAnonymityState + "</CENTER>");
				}
				else
				{
					countDown = m_requestRight.getCountDown();
					
					if (!bConnecting)
					{
						strAnonymityState += JAPMessages.getString(MSG_ANON_MODE_OFF_BUT_FORCED, new String[]{
								JAPMessages.getString("ngAnonOn")}) + "<BR><BR>";
					}
					
					if (!m_requestRight.isUserAction())
					{
						blockedMessage = "";
					}
					else if (m_requestRight.getURI().equals(GENERAL_RULE))
					{
						blockedMessage = JAPMessages.getString(MSG_BLOCKED_ALL) + "<BR>";
					}
					else
					{
						blockedMessage = JAPMessages.getString(MSG_BLOCKED_DOMAIN, 
								"<code>" + m_requestRight.getURI() + "</code>") + "<BR>";
					}
					
					if (countDown == Long.MAX_VALUE)
					{
						addedMessage = new String[]{
								(bConnecting ? 
										"<BR>" + JAPMessages.getString(MSG_WAIT_FOR_CONNECTION) : ""),
								(bConnecting ? JAPMessages.getString(MSG_HTML_ANON_MODE_SWITCH, JAPMessages.getString("ngAnonOff")) : JAPMessages.getString(MSG_BLOCKED_PERMANENTLY)) + 
								"<BR>" + JAPMessages.getString(MSG_RELOAD), JONDOPROXY};
					}
					/*
					else if (countDown / 1000 == 0)
					{
						addedMessage = new String[]{strAnonymityState,
								blockedMessage, 
								JAPMessages.getString(MSG_RELOAD)};
					}
					*/
					else 
					{
						addedMessage = new String[]{
								(bConnecting ? 
										"<BR>" + JAPMessages.getString(MSG_WAIT_FOR_CONNECTION) : ""),
								JAPMessages.getString(MSG_COUNTDOWN, 
										new String[]{"" + (countDown / 1000),
										(bConnecting ? JAPMessages.getString(MSG_HTML_ANON_MODE_SWITCH, JAPMessages.getString("ngAnonOff")) : JAPMessages.getString(MSG_BLOCKED_PERMANENTLY)),
											"<BR>" + JAPMessages.getString(MSG_RELOAD)}), JONDOPROXY};
					}
					byte[] byWrite = (strAnonymityState +
							blockedMessage +
							JAPMessages.getString(MSG_BLOCKED, addedMessage) +
							"</CENTER>").getBytes("UTF-8");
					
					socket.getOutputStream().write(byWrite, 0, byWrite.length);
				}
				
				/*
				if (callback != null && callback.getHTMLHelpPath() != null)
				{
					toClient.write("<BR><BR><CENTER>" +
							"<a href=\"" + "http://" + callback.getHTMLHelpPath().toString() + "\">" +
							//"<a href=\"" + "http://www.spiegel.de" + "\">" +
							"Get help here!</a>" + 
							"</CENTER>");
				}
				*/
				toClient.write("</BODY></HTML>\n");
				toClient.flush();
				toClient.close();
			}
			catch (Exception e)
			{
				LogHolder.log(LogLevel.EXCEPTION, LogType.NET, e);
			}
			finally
			{
				try
				{
					SocketGuard.close(socket);
				}
				catch (IOException a_e)
				{
					LogHolder.log(LogLevel.ERR, LogType.NET, a_e);
				}
			}
		}
	}

	private final class RememberedRequestRight
	{	
		private long m_timeRemembered;
		private boolean m_bWarn;
		private String m_URI;
		private boolean m_bUserAction;
		
		public RememberedRequestRight(String a_URI, boolean a_bWarn, long a_lTimeout)
		{
			this(a_URI, a_bWarn, a_lTimeout, true);
		}
		
		public RememberedRequestRight(String a_URI, boolean a_bWarn, long a_lTimeout,
				boolean a_bUserAction)
		{
			m_URI = a_URI;
			m_timeRemembered = a_lTimeout;
			m_bWarn = a_bWarn;
			m_bUserAction = a_bUserAction;
		}
		
		public boolean isUserAction()
		{
			return m_bUserAction;
		}
		
		public String getURI()
		{
			return m_URI;
		}
		
		public boolean isWarningShown()
		{
			return m_bWarn;
		}
		
		public long getCountDown()
		{
			if (m_timeRemembered == Long.MAX_VALUE)
			{
				return Long.MAX_VALUE;
			}
			
			long countDown = m_timeRemembered - System.currentTimeMillis();
			if (countDown < 0)
			{
				countDown = 0;
			}
			return countDown;
		}

		public boolean isTimedOut()
		{
			return m_timeRemembered < System.currentTimeMillis();
		}
	}

	/**
	 *  This class is used to transfer requests from a client socket to the
	 *  proxy defined by this class and responses by the proxy back to the
	 *  client.
	 */
	private final class DirectConViaHTTPProxy implements Runnable
	{
		private Socket m_clientSocket;
		private InputStream m_clientInputStream;

		/**
		 * Sets the client socket to send the response of the proxy to and the
		 * InputStream to read the client request from. If it is null, the
		 * InputStream of the client socket is used.
		 *
		 * @param a_clientSocket Socket the client writes to and reads from.
		 * @param a_clientInputStream The InputStream the client request will be
		 * coming from.
		 */
		public DirectConViaHTTPProxy(Socket a_clientSocket, InputStream a_clientInputStream)
		{
			m_clientSocket = a_clientSocket;
			m_clientInputStream = a_clientInputStream;
		}

		public void run()
		{
			try
			{
				/*
				 * Check if the InputStream can be used. Otherwise use the Input
				 * stream of the client socket.
				 */
				InputStream fromClient;
				if (m_clientInputStream != null)
				{
					fromClient = m_clientInputStream;
				}
				else
				{
					fromClient = m_clientSocket.getInputStream();
				}

				// create Socket to proxy server
				Socket serverSocket = new Socket(getProxyInterface().getHost(),
												 getProxyInterface().getPort());

				/*
				 * Prepare a Thread to send the response of the proxy back to
				 * the client, once it follows the request further down.
				 */
				DirectProxyResponse pr = new DirectProxyResponse(serverSocket.getInputStream(),
					m_clientSocket.getOutputStream());

				Thread responseThread = new Thread(pr, "JAP - DirectProxyResponse");
				responseThread.start();

				// Get an OutputStream to the proxy server.
				OutputStream toProxy = serverSocket.getOutputStream();

				// Transfer data client --> server

				/*
				 * If authorization is used for this proxy server, send the
				 * necessary information before relaying the client request.
				 */
				if (getProxyInterface().isAuthenticationUsed())
				{
					//we need to insert an authorization line...
					//read first line and after this insert the authorization
					String str = DirectProxyConnection.readLine(fromClient);
					str += "\r\n";
					toProxy.write(str.getBytes());

					str = getProxyInterface().getProxyAuthorizationHeaderAsString();
					toProxy.write(str.getBytes());
					toProxy.flush();
				}

				byte[] buff = new byte[1000];
				int len;

				/*
				 * Send the request from the client to the proxy server.
				 */
				while ( (len = fromClient.read(buff)) != -1)
				{
					if (len > 0)
					{
						toProxy.write(buff, 0, len);
						toProxy.flush();
					}
				}

				/*
				 * Wait for response thread to send the proxy response to the
				 * client.
				 */

				responseThread.join();

				/*
				 * Close all sockets and associated streams.
				 */
				toProxy.close();
				fromClient.close();
				serverSocket.close();
			}
			catch (IOException ioe)
			{
			}
			catch (Exception e)
			{
				LogHolder.log(LogLevel.EXCEPTION, LogType.NET, "JAPDirectConViaProxy: Exception: " + e);
			}
		}
	}
	
	protected static RequestInfo parseDomain(String strURI, boolean a_bSecondLevelOnly, String a_strMethod)
	{
		int port = 80;
		if (strURI != null && strURI.length() > 0)
		{
			String strJDPArg = null;
			int index = strURI.indexOf("//");
			if (index > 0 && strURI.length() > 2)
			{
				strURI = strURI.substring(index + 2, strURI.length());
			}
			
			index = strURI.indexOf("?");
			if (index > 0)
			{
				if (strURI.length() >= index + 3)
				{
					strJDPArg = strURI.substring(index + 1, strURI.length());
				}
				
				strURI = strURI.substring(0, index);
				
				if (strJDPArg != null)
				{
					index = strJDPArg.toLowerCase().indexOf(JONDOPROXY_ARG);
					if (index >= 0 && strJDPArg.length() >= index + JONDOPROXY_ARG.length() + 1)
					{
						strJDPArg = strJDPArg.substring(index + JONDOPROXY_ARG.length(), strJDPArg.length());
						index = strJDPArg.indexOf("&");
						if (index > 0)
						{
							strJDPArg = strJDPArg.substring(0, index);
						}
					}
					else
					{
						strJDPArg = null;
					}
				}
				else
				{
					strJDPArg = null;
				}
			}
			
			
			index = strURI.indexOf("/");
			if (index > 0)
			{
				strURI = strURI.substring(0, index);
			}
			// strip port
			index = strURI.lastIndexOf(":");
			if (index > 0 && strURI.length() > index + 1)
			{
				try
				{
					port = Integer.parseInt(strURI.substring(index + 1, strURI.length()));
				}
				catch (NumberFormatException a_e)
				{
					LogHolder.log(LogLevel.ERR, LogType.NET, "Could not parse port!", a_e);
				}
				strURI = strURI.substring(0, index);
			}
			
			
			
			while (strURI.endsWith("/"))
			{
				strURI = strURI.substring(0, strURI.length() - 1);
			}

			if (a_bSecondLevelOnly)
			{
				// test for IP
				index = strURI.lastIndexOf(".");
				if (index > 0 && strURI.length() > index + 1)
				{
					try
					{
						Integer.parseInt(strURI.substring(index + 1, strURI.length()));
						// this is an IP
					}
					catch (NumberFormatException a_e)
					{
						// this is a domain
						StringTokenizer tokenizer = new StringTokenizer(strURI, ".");
						while (tokenizer.countTokens() > 2)
						{
							tokenizer.nextToken();
						}
						strURI = tokenizer.nextToken() + "." + tokenizer.nextToken();
					}
				}
			}
			
			return new RequestInfo(strURI, a_strMethod, port, strJDPArg);
		}
		
		return null;	
	}
}
