package anon.transport.address;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;

/**
 * An Endpoint is the interface between an {@link IAddress} and its URN
 * representation.
 * <p>
 * It provides the mechanisms to parse a URN and allow structured access to it,
 * and to transform an {@link IAddress} back into a URN.
 *
 * @see Endpoint
 */
public class Endpoint {

	/**
	 * The transport identifier as extracted from the URN.
	 */
	protected String m_transportIdentifier;

	/**
	 * A map of the parameters, extracted from the URN. In the form of
	 * String:{@link AddressParameter}.
	 */
	protected Hashtable m_parameters;

	/**
	 * Returns the transport identifier of the URN.
	 */
	public String getTransportIdentifier() {
		return m_transportIdentifier;
	}

	/**
	 * Transforms a given {@link IAddress} into its URN representation. The
	 * transformation follows the following rules:
	 * 
	 * <pre>
	 *   EndpointURN ::= &quot;urn:endpoint:&quot;  Identifier ParameterGroup
	 *   Identifier == TransportIdentifier
	 *   ParameterGroup = Parameter | Parameter ParamterGroup
	 *   Parameter = &quot;:&quot;  ParamterName &quot;(&quot; ParamterValue  &quot;)&quot;
	 * </pre>
	 * 
	 * Identifier describes the transport identifier of the {@link IAddress}.
	 * ParameterName and ParameterValue describe the name and value of a
	 * parameter of the {@link IAddress}, all of which are inserted into the
	 * resulting URN.
	 * 
	 * @param a_address
	 *            The address to be transformed into a URN.
	 */
	public static String toURN(IAddress a_address) {
		String identifier = a_address.getTransportIdentifier();
		StringBuffer builder = new StringBuffer();
		builder.append("urn:endpoint:");
		builder.append(identifier);
		AddressParameter[] parameters = a_address.getAllParameters();
		for (int i = 0, length = parameters.length; i < length; i++) {
			builder.append(":");
			builder.append(parameters[i].getName());
			builder.append("(");
			builder.append(parameters[i].getValue());
			builder.append(")");
		}
		return builder.toString();
	}

	/**
	 * Tries to create a new instance of {@link Endpoint}, by extracting the
	 * transport identifier and parameters from the URN.
	 * 
	 * @param a_theURN
	 *            The URN to create the Endpoint object from.
	 * @throws MalformedURNException
	 *             If the URN is not in the expected format.
	 */
	public Endpoint(String a_theURN) throws MalformedURNException {
		StringTokenizer st=new StringTokenizer(a_theURN,":");
		String[] components = new String[st.countTokens()];
		int i=0;
		while(st.hasMoreElements())
			components[i++]=st.nextToken();
		if (components.length < 3)
			throw new MalformedURNException(
					"A valid Endpoint needs at least 3 Components");
		if (!components[0].equals("urn"))
			throw new MalformedURNException("URN must start with \"urn:\"");
		if (!components[1].equals("endpoint"))
			throw new MalformedURNException(
					"Can only handle Endpoint-Namespace. Is " + components[1]);
		m_transportIdentifier = components[2];

		// get Parameters
		m_parameters = new Hashtable();
		int length=components.length;
		for (i = 3; i < length; i++) {
			int parentheseIndex = components[i].indexOf("(");
			int decrementedLength = components[i].length() - 1;
			String name = components[i].substring(0, parentheseIndex);
			++parentheseIndex;
			String value = components[i].substring(parentheseIndex,
					decrementedLength);
			m_parameters.put(name, new AddressParameter(name, value));
		}
	}

	/**
	 * Returns the parameter with the given name.
	 *
	 * @param a_name The name of the parameter.
	 * 
	 * @return The value for the given name if it exists or null.
	 */
	public String getParameter(String a_name) {
		AddressParameter parameter = (AddressParameter) m_parameters.get(a_name);
		return parameter == null ? null : parameter.getValue();
	}

	/**
	 * Returns an array of all {@link AddressParameter} objects associated with
	 * this Endpoint.
	 * 
	 * @return An array of Key-Value pairs.
	 */
	public AddressParameter[] getAllParameters() {
		AddressParameter[] result = new AddressParameter[m_parameters.size()];
		Enumeration en=m_parameters.elements();
		int i=0;
		while(en.hasMoreElements())
			{
					result[i++]=(AddressParameter)en.nextElement();
			}
		return result;
	}

}
