package anon.transport.address;

import java.net.InetAddress;
import java.net.UnknownHostException;

/**
 * A {@link TcpIpAddress} identifies an {@link Endpoint} of a socket connection
 * by its unique pair of IP address and port number.
 */
public class TcpIpAddress implements IAddress {

	public static final String TRANSPORT_IDENTIFIER = "tcpip";

	private static final String IP_PARAMETER = "ip-address";
	private static final String PORT_PARAMETER = "port";

	/**
	 * The port number.
	 */
	protected int m_port;

	/**
	 * The IP address.
	 */
	protected InetAddress m_ipAddress;

	/**
	 * Creates a new {@link TcpIpAddress} based on the given host and port. The
	 * IP address is determined by a host lookup. If that fails, the IP address
	 * is null.
	 *
	 * @param a_host
	 *            The host name or a string of an IP address.
	 * @param a_port
	 *            The port number.
	 */
	public TcpIpAddress(String a_host, int a_port) {
		try {
			m_ipAddress = InetAddress.getByName(a_host);
		} catch (UnknownHostException e) {
			m_ipAddress = null;
		}

		m_port = a_port;
	}

	/**
	 * Creates a new {@link TcpIpAddress} based on the given {@link InetAddress}
	 * and port.
	 *
	 * @param a_ipAddress
	 * @param a_port The port number.
	 */
	public TcpIpAddress(InetAddress a_ipAddress, int a_port) {
		m_ipAddress = a_ipAddress;
		m_port = a_port;
	}

	/**
	 * Creates a new {@link TcpIpAddress}, where the values for port and IP
	 * address are determined from the given Endpoint.
	 *
	 * @throws AddressMappingException
	 */
	public TcpIpAddress(Endpoint a_baseAddress) throws AddressMappingException {
		String value = a_baseAddress.getParameter(IP_PARAMETER);
		if (value == null)
			throw new AddressMappingException("IP Parameter is missing");
		try {
			m_ipAddress = InetAddress.getByName(value);
		} catch (UnknownHostException e) {
			throw new AddressMappingException("IP-Address could not be parsed.");
		}
		value = a_baseAddress.getParameter(PORT_PARAMETER);
		if (value == null)
			throw new AddressMappingException("Port Parameter is missing");
		try {
			m_port = Integer.parseInt(value);
		} catch (NumberFormatException e) {
			throw new AddressMappingException("Port could not be parsed.");
		}
	}

	/**
	 * Returns the IP address.
	 */
	public InetAddress getIPAddress() {
		return m_ipAddress;
	}

	/**
	 * Returns the port.
	 */
	public int getPort() {
		return m_port;
	}

	/**
	 * Returns the current host name of the internal IP address. Might be
	 * different, than at initialization.
	 */
	public String getHostname() {
		return m_ipAddress.getHostName();
	}

	/**
	 * Returns an array of {@link AddressParameter} objects.
	 */
	public AddressParameter[] getAllParameters() {
		AddressParameter[] result = new AddressParameter[2];
		result[0] = new AddressParameter(IP_PARAMETER, m_ipAddress
				.getHostAddress());
		result[1] = new AddressParameter(PORT_PARAMETER, String.valueOf(m_port));
		return result;
	}
	/**
	 * Returns a string describing the transport protocol.
	 */
	public String getTransportIdentifier() {
		return TRANSPORT_IDENTIFIER;
	}
}
