/*
 Copyright (c) 2007, JonDos GmbH
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

  - Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.

  - Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

  - Neither the name of the JonDos GmbH nor the names of its contributors
 may be used to endorse or promote products derived from this software without specific
 prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
package anon.util;

/**
 * This is a generic FIFO object queue.
 *
 * @author Rolf Wendolsky
 */
public class ObjectQueue
	{
		/**
		 * Most recently added object. Last to go out.
		 */
		private QueueItem m_head = null;

		/**
		 * Least recently added object. First to go out.
		 */
		private QueueItem m_foot = null;

		/**
		 * Number of objects currently in the queue. If the same object has been
		 * pushed multiple times it will be counted multiple times.
		 */
		private int m_size = 0;

		/**
		 * Signals if the {@link #take()} method should stop blocking.
		 */
		private volatile boolean m_bClosed;

		/**
		 * Simple LinkedList-like list item used in the {@link ObjectQueue} class.
		 */
		private final class QueueItem
			{
				/**
				 * The actual object this {@link QueueItem} wraps.
				 */
				private Object m_object;

				/**
				 * A reference to the previous {@link QueueItem} in the
				 * {@link ObjectQueue}.
				 */
				private QueueItem m_previous = null;

				public QueueItem(Object a_object)
					{
						m_object = a_object;
						m_bClosed=false;
					}
			}

		/**
		 * Returns the number of objects currently in the queue. If the same object
		 * has been pushed multiple times it will be counted multiple times.
		 *
		 * @return Size of the queue.
		 */
		public int getSize()
			{
				return m_size;
			}

		/**
		 * Signals the closing of this queue, interrupting any running
		 * {@link #take()} method calls.
		 * <p>
		 * Note: This doesn't affect any other method calls and gets undone, when
		 * pushing an object into the queue.
		 */
		public void close()
			{
				m_bClosed=true;
			}

		/**
		 * Pushes an object to the head of the queue.
		 *
		 * @param a_object
		 */
		public synchronized void push(Object a_object)
			{
				QueueItem newItem = new QueueItem(a_object);
				m_size++;

				if (m_head == null)
					{
						// there is no item im the queue
						m_head = newItem;
						m_foot = newItem;
					}
				else
					{
						// there is more than one item in the queue
						m_head.m_previous = newItem;
						m_head = newItem;
					}
			}

		/**
		 * Returns the least recent object in the queue. If the queue is empty
		 * returns null.
		 *
		 * @return The tail of the queue, null if queue is empty.
		 */
		public synchronized Object pop()
			{
				Object object;
				if (m_head == null)
					{
						// there is not item in the queue
						return null;
					}
				else if (m_head == m_foot)
					{
						// there is only one object in the queue
						object = m_foot.m_object;
						m_head = null;
						m_foot = null;
					}
				else
					{
						// there is more than one object in the queue
						object = m_foot.m_object;
						m_foot = m_foot.m_previous;

					}
				m_size--;
				return object;
			}

		/**
		 * If the queue is empty this method blocks until a new item is available
		 * or the queue has been closed using {@link #close()}.
		 *
		 * @return The tail of the queue, null if the queue is closed.
		 */
		public Object take() /*throws InterruptedException*/
			{
				while(!m_bClosed)
					{
						Object o=pop();
						if(o!=null)
							return o;
						try
							{
								Thread.sleep(100);
							}
						catch (InterruptedException e)
							{
							}
					}
				return null;
			}

		/**
		 * Tries returning an object using {@link #pop()}. If no object was
		 * available it waits {@code a_msTimeout} milliseconds and calls
		 * {@link #pop()}, returning the result.
		 *
		 * @param a_msTimeout Amount of milliseconds to wait, if queue is empty.
		 * @return Object received from {@link #pop()}. null, if null was returned.
		 * @throws InterruptedException
		 */
		public Object poll(int a_msTimeout) throws InterruptedException
			{
				Object o=pop();
				if(o!=null)
					return o;
				Thread.sleep(a_msTimeout);
				return pop();
			}
		
		public synchronized boolean isEmpty()
		{
				return (m_size==0);
		}

	}
