/*
 Copyright (c) 2004, The JAP-Team
 All rights reserved.
 Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

 - Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation and/or
  other materials provided with the distribution.

 - Neither the name of the University of Technology Dresden, Germany nor the names of its contributors
   may be used to endorse or promote products derived from this software without specific
  prior written permission.


 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS
 OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE
 */
/*
 * based on tinySSL
 * (http://www.ritlabs.com/en/products/tinyweb/tinyssl.php)
 *
 */
package anon.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Method;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;


import logging.LogHolder;
import logging.LogLevel;
import logging.LogType;
import java.net.SocketException;
import java.util.Enumeration;
import java.util.Vector;

import anon.platform.AbstractOS;

/**
 * Wraps a {@link Socket} object and can close it after a given number of
 * milliseconds.
 */
public class SocketGuard
{
	private Thread m_threadCloseGuard = null;
	private Object SYNC_CLOSE = new Object();


	private Socket m_Socket;


	private InputStream m_istream;
	private OutputStream m_ostream;
	private String m_strCaller;

	/**
	 * Creates a {@link SocketGuard} for the given Socket, closing it after
	 * {@code a_forceCloseAfterMS} milliseconds, or not if 0.
	 *
	 * @param socket The socket to guard.
	 * @param a_forceCloseAfterMS The amount of time in milliseconds after which
	 * the socket is forcefully closed. Might happen earlier due to
	 * {@link #close()}. 0 to leave it open.
	 *
	 * @throws IOException
	 */
	public SocketGuard(Socket socket, final long a_forceCloseAfterMS) throws IOException
	{
		m_Socket = socket;
		m_strCaller = LogHolder.getCallingMethod(true);

		if (a_forceCloseAfterMS > 0 && isClosedSupported())
		{
			Runnable run = new Runnable()
			{
				public void run()
				{
					synchronized(SYNC_CLOSE)
					{
						if (m_threadCloseGuard != null)
						{
							LogHolder.log(LogLevel.DEBUG, LogType.NET,
									  "Waiting for socket for " + a_forceCloseAfterMS +
									  " milliseconds...");
							
							try
							{
								SYNC_CLOSE.wait(a_forceCloseAfterMS);
							}
							catch (InterruptedException ex)
							{
							}
							if (!isClosed(m_Socket))
							{
								LogHolder.log(LogLevel.ALERT, LogType.NET,
											  "Closing socket after " + a_forceCloseAfterMS +
											  " milliseconds! Called from: " + m_strCaller);

								try
								{
									close();
								}
								catch (IOException ex1)
								{
									LogHolder.log(LogLevel.ALERT, LogType.NET, ex1);
								}
								if (!isClosed(m_Socket))
								{
									LogHolder.log(LogLevel.ALERT, LogType.NET,
											  "Unable to close socket!! This is a serious problem. " +
											  "Called from: " + m_strCaller);
								}
							}
						}
					}
				}
			};
			m_threadCloseGuard = new Thread(run);
			m_threadCloseGuard.start();
		}

	
		m_istream = socket.getInputStream();
		m_ostream = socket.getOutputStream();

	}

	public Socket getSocket()
	{
		return m_Socket;
	}
	
	public InetAddress getInetAddress()
	{
		return m_Socket.getInetAddress();
	}

	public InputStream getInputStream()
	{
		return this.m_istream;
	}

	public OutputStream getOutputStream()
	{
		return this.m_ostream;
	}

	/**
	 * Closes the {@link Socket} this object holds and its {@link InputStream} and
	 * {@link OutputStream} .
	 *
	 * If an {@link IOException} occurred along the way, it gets thrown.
	 * @throws IOException
	 */
	public void close() throws IOException
	{
		IOException ex = null;

		// close the outputstream
		try
		{
			if (m_ostream != null)
			{
				m_ostream.close();
			}
		}
		catch (IOException a_e)
		{
			if (ex == null)
			{
				ex = a_e;
			}
		}

		// close the inputstream
		try
		{
			if (m_istream != null)
			{
				m_istream.close();
			}
		}
		catch (IOException a_e)
		{
			if (ex == null)
			{
				ex = a_e;
			}
		}

		// close the socket itself
		if (m_Socket != null)
		{
			m_Socket.close();
		}

		// closes the time out thread, if present
		if (m_threadCloseGuard != null)
		{
			synchronized(SYNC_CLOSE)
			{
				SYNC_CLOSE.notify();
				m_threadCloseGuard = null;
			}
		}

		// if an exception occurred along the way, throw it
		if (ex != null)
		{
			throw ex;
		}
	}

	/**
	 * Closes the given {@link SocketGuard} object.
	 *
	 * @see #close()
	 * @param a_socket
	 */
	public static void close(SocketGuard a_socket)
	{
		if (a_socket != null)
		{
			try
			{
				a_socket.close();
			}
			catch (IOException ex1)
			{
				LogHolder.log(LogLevel.ERR, LogType.NET, ex1);
			}
		}
	}


	/**
	 * Returns whether or not the internal Socket is closed or not.
	 * @return
	 */
	public boolean isClosed()
	{
		return isClosed(m_Socket);
	}


	public void setSoTimeout(int ms) throws SocketException
	{
		if (m_Socket != null)
		{
			m_Socket.setSoTimeout(ms);
		}
	}

	/**
	 * Returns whether or not the Socket implementation used by this JVM supports
	 * the {@code isClosed()} method.
	 *
	 * @return
	 */
	public static boolean isClosedSupported()
	{
		try 
		{
			Socket.class.getMethod("isClosed", new Class[]{});
			return true;
		} 
		catch (Exception e) 
		{
			return false;
		} 
	}
	
	public static void close(final Socket a_socket) throws IOException
	{
		if (a_socket == null)
		{
			return;
		}
		
		IReturnRunnable run = new IReturnRunnable()
		{
			IOException ex = null;
			
			public void run()
			{
				do
				{
					boolean bIsClosed = isClosed(a_socket);
					
					try
					{
						a_socket.getInputStream().close();
					}
					catch (Exception a_e)
					{
						// ignore
					}
					
					try
					{
						a_socket.getOutputStream().close();
					}
					catch (Exception a_e)
					{
						// ignore
					}
					
					try
					{
						a_socket.close();
						Thread.yield();
					}
					catch (IOException a_e)
					{
						ex = a_e;
						if (!bIsClosed && !isClosedSupported())
						{
							LogHolder.log(LogLevel.WARNING, LogType.NET, a_e);
						}
					}
				}
				while (isClosedSupported() && !isClosed(a_socket));
			}
			
			public Object getValue()
			{
				return ex;
			}
		};
		
		Thread tRun = new Thread(run);
		tRun.start();
		
		
		do 
		{
			try 
			{
				tRun.join();
			} 
			catch (InterruptedException e) 
			{
				// ignore, we definitely want to close this socket
			}
		}
		while (tRun.isAlive());

		
		if (run.getValue() != null && !isClosedSupported())
		{
			throw (IOException)run.getValue();
		}
	}
	
	/**
	 * 
	 * @param a_socket
	 * @return false if this method is not supported, or if the socket is closed; 
	 * does only return true if the socket is really closed
	 */
	public static boolean isClosed(Socket a_socket)
	{
		Socket socket = a_socket;
		if (socket != null)
		{
			try
			{
				return ( (Boolean) Socket.class.getMethod("isClosed", new Class[]
					{}).invoke(
						socket, new Object[]
						{})).booleanValue();
			}
			catch (Exception ex)
			{
				return false;
			}
			//return socket.isClosed();
		}
		return true;
	}

  /**
   * Creates a new Socket. A timeout can be set for this operation and also it is interruptible.
   * If there occurred an Exception while creating the Socket, this Exception is thrown. If the
   * creation process was interrupted from outside, an InterrupedException is thrown (but the
   * interrupted flag is not set on the thread). If a timeout was specified and reached, a normal
   * Exception is thrown.
   *
   * @param a_hostName The hostname of the target or null for the loopback interface
   *                   (see java.net.Socket).
   * @param a_port The port of the target.
   * @param a_createTimeout The timeout for creating the new Socket in milliseconds. A timeout of
   *                        0 means to wait forever (until Socket creation is ready).
   *
   * @return The created Socket.
   */
  public static Socket createSocket(final String a_hostName, final int a_port, long a_createTimeout) throws Exception {
    final Vector socketStore = new Vector();
    final Vector exceptionStore = new Vector();
    Thread createThread = new Thread(new Runnable() {
      /**
       * This is the implementation for the thread creating the new Socket.
       */
      public void run() {
        try {
          Socket serverConnection = new Socket(a_hostName, a_port);
          synchronized (socketStore) {
            if (Thread.interrupted()) {
              /* we are to late -> close the socket */
              serverConnection.close();
            }
            else {
              /* we are in time */
              socketStore.addElement(serverConnection);
            }
          }
        }
        catch (Exception e) {
          synchronized (exceptionStore) {
            exceptionStore.addElement(e);
          }
        }
      }
    });
    createThread.setDaemon(true);
    createThread.start();
    try {
      createThread.join(a_createTimeout);
    }
    catch (Exception e) {
      /* first try to interrupt the create thread, maybe later Java versions will stop the
       * socket operation then, also if the creation of the Socket will be successful, the
       * create thread will close the now useless Socket immediately
       */
      createThread.interrupt();
      synchronized (socketStore) {
        /* maybe it was to late and the Socket was already created -> close it */
        if (socketStore.size() > 0) {
          try {
            ((Socket)(socketStore.firstElement())).close();
          }
          catch (Exception e2) {
            /* do nothing */
          }
        }
      }
      /* now throw the InterruptedException */
      throw e;
    }
    /* if we reach this point, the socket create thread cam to the end or the time was up */
    Socket createdSocket = null;
    synchronized (socketStore) {
      synchronized (exceptionStore) {
        if (exceptionStore.size() > 0) {
          /* an exception occured while creating the socket -> throw it */
          throw ((Exception)(exceptionStore.firstElement()));
        }
        if (socketStore.size() > 0) {
          /* socket was created successfully */
          createdSocket = (Socket)(socketStore.firstElement());
        }
        else {
          /* no exception, no socket -> create thread not ready -> timeout occured -> throw an
           * exception
           */
          /* first try to interrupt the create thread, maybe later Java versions will stop the
           * socket operation then, also if the creation of the Socket will be successful, the
           * create thread will close the now useless Socket immediately
           */
          createThread.interrupt();
          throw (new Exception("Timeout occured while creating socket."));
        }
      }
    }
    return createdSocket;
  }

  public static ServerSocket createVirtualBoxServerSocket(int a_port, InetAddress a_excludeAddress)
  {
	  ServerSocket socketListener = null;
	  InetAddress address;
	  
	  // Try to identify any existing installation of VirtualBox
		Enumeration enumInterfaces;
		
		Class classNetworkInterface;
		Class classInet6Address;
		Method method;

			
		try 
		{
			// We need compatibility with Java 1.3...
			classNetworkInterface = Class.forName("java.net.NetworkInterface");
			classInet6Address = Class.forName("java.net.Inet6Address");
			method = classNetworkInterface.getMethod("getNetworkInterfaces", (Class[])null);
			enumInterfaces = (Enumeration)method.invoke(null, (Object[])null);
		} 
		catch (Exception e) 
		{
			LogHolder.log(LogLevel.ALERT, LogType.NET, "Could not get any network interfaces!", e);
			return null;
		}
		Object netInterface;
		Enumeration enumInetAddress;
		InetAddress ipv6Address = null;
		while (enumInterfaces != null && enumInterfaces.hasMoreElements() && socketListener == null)
		{
			netInterface = enumInterfaces.nextElement();
			if (AbstractOS.getInstance().isVirtualBoxInterface(netInterface))
			{
				try
				{
					method = classNetworkInterface.getMethod("getInetAddresses", (Class[])null);
					enumInetAddress = (Enumeration)method.invoke(netInterface, (Object[])null);
				}
				catch (Exception a_e)
				{
					LogHolder.log(LogLevel.ALERT, LogType.NET, 
							"Could not analyze network interfaces!", a_e);
					return null;
				}
				while (enumInetAddress.hasMoreElements())
				{
					address = (InetAddress)enumInetAddress.nextElement();
					
					if (classInet6Address.isInstance(address))
					{
						if (ipv6Address == null)
						{
							ipv6Address = address;
						}
						continue;
					}
					if (a_excludeAddress.equals(address))
					{
						// the user has manually given this address
						enumInterfaces = null;
						break;
					}
					LogHolder.log(LogLevel.NOTICE, LogType.NET, 
							"Try binding Listener on host: " + address);
					try
					{
						socketListener = new ServerSocket(a_port, 50, address);
						
						LogHolder.log(LogLevel.ALERT, LogType.NET, 
								"Listener was successfully bound to: " + address.getHostAddress() + ":" + a_port);
						break;
					}
					catch (IOException a_e)
					{
						LogHolder.log(LogLevel.WARNING, LogType.NET, 
								"Could not bind listener to host: " + address, a_e);
					}
				}
			}
		}
		if (socketListener == null && ipv6Address != null)
		{
			address = ipv6Address;
			try
			{
				socketListener = new ServerSocket(a_port, 50, address);
				
				LogHolder.log(LogLevel.ALERT, LogType.NET, 
						"Listener was successfully bound to: " + socketListener);
			}
			catch (IOException a_e)
			{
				LogHolder.log(LogLevel.WARNING, LogType.NET, 
						"Could not bind listener to host: " + address, a_e);
			}
		}
		return socketListener;
  	}
}
